
//#define TYPE3_CACHE

#define PDKDEBUG(X)
#define PDKDEBUG2(X)
#define PDKDEBUG3(X) 


///========================================================================
//
// QOutputDev.cc
//
// Copyright 1996 Derek B. Noonburg
// CopyRight 2002 Robert Griebl
// Copyright 2003 Philip Kohn
//
//========================================================================

#ifdef __GNUC__
#pragma implementation
#endif

#include "GString.h"
#include "Object.h"
#include "Stream.h"
#include "Link.h"
#include "GfxState.h"
#include "GfxFont.h"
#include "UnicodeMap.h"
#include "CharCodeToUnicode.h"
#include "FontFile.h"
#include "Error.h"

#include "qdocview.h"
#include "TextOutputDev.h"

#include "GList.h"
#include "QOutputDev.h"

#include <qfile.h>
#include <qtextstream.h>
#include <time.h>
#include <qfontdatabase.h>
#include "qqimage.h"

#define QT_NO_TRANSFORMATIONS


//------------------------------------------------------------------------
// Constants and macros
//------------------------------------------------------------------------


static inline QColor q_col ( const GfxRGB &rgb )
{
  //	return QColor ( lrint ( rgb. r * 255 ), lrint ( rgb. g * 255 ), lrint ( rgb. b * 255 ));
#if 0
  printf("color %f %f %f => %x %x %x\n",
	 (float)rgb.r, (float)rgb.g, (float)rgb.b,
	 getpix(rgb.r), getpix(rgb.g), getpix(rgb.b));
#endif
  return QColor(getpix(rgb.r), getpix(rgb.g), getpix(rgb.b));
}


//------------------------------------------------------------------------
// Font substitutions
//------------------------------------------------------------------------

struct QOutFontSubst {
	char * m_name;
	char * m_sname;
	bool   m_bold;
	bool   m_italic;
	QFont::StyleHint m_hint;
};

static QOutFontSubst qStdFonts [] = {
	{ "Helvetica",             "Helvetica", false, false, QFont::Helvetica },
	{ "Helvetica-Oblique",     "Helvetica", false, true,  QFont::Helvetica },
	{ "Helvetica-Bold",        "Helvetica", true,  false, QFont::Helvetica },
	{ "Helvetica-BoldOblique", "Helvetica", true,  true,  QFont::Helvetica },
	{ "Times-Roman",           "Times",     false, false, QFont::Times },
	{ "Times-Italic",          "Times",     false, true,  QFont::Times },
	{ "Times-Bold",            "Times",     true,  false, QFont::Times },
	{ "Times-BoldItalic",      "Times",     true,  true,  QFont::Times },
	{ "Courier",               "Courier",   false, false, QFont::Courier },
	{ "Courier-Oblique",       "Courier",   false, true,  QFont::Courier },
	{ "Courier-Bold",          "Courier",   true,  false, QFont::Courier },
	{ "Courier-BoldOblique",   "Courier",   true,  true,  QFont::Courier },

	{ "Symbol",                0,           false, false, QFont::AnyStyle },
	{ "Zapf-Dingbats",         0,           false, false, QFont::AnyStyle },
  
	{ 0,                       0,           false, false, QFont::AnyStyle }
};


char *fontTypeName(int type_num)
{
  char *type = "bad";
  switch(type_num) {
  case fontUnknownType:
    type = "Unknown";
    break;
  case fontType1:
    type = "T1";
    break;
  case fontType1C:
    type = "T1C";
    break;
  case fontType3:
    type = "T3";
    break;
  case fontTrueType:
    type = "TT";
    break;
  case fontCIDType0:
    type = "CID0";
    break;
  case fontCIDType0C:
    type = "CID0C";
    break;
  case fontCIDType2:
    type = "CID2";
    break;
  }
  return(type);
}

void
printFontDescription(GfxFont *gfxFont, fouble fsize)
{
  printf ("%s[%.2f]%s %s%s%s%s%s%s%s ",
	  ((gfxFont-> getName ( )) ? gfxFont-> getName ( )-> getCString ( ) : "" ),
	  (float)fsize,
	  fontTypeName(gfxFont->getType()),
	  gfxFont-> isBold ( )? " bold ":"",
	  gfxFont-> isItalic ( )? " ital":"",
	  gfxFont-> isFixedWidth ( )? " fix":"",
	  gfxFont-> isSerif ( ) ? " serif":"",
	  gfxFont-> isSymbolic ( ) ? " sym":"",
	  gfxFont-> isCIDFont ( )? " CID":"",
	  gfxFont-> getEmbeddedFontName ( ) ?  " embed":"");
  Ref ref;
  ref.num = -1;
  int fontID = gfxFont->getEmbeddedFontID(&ref);
  printf("{%s:%d} ", gfxFont->getEmbeddedFontName()?
	 gfxFont->getEmbeddedFontName()->getCString():"",
	 fontID);
}


void
QOutputDev::cacheSysFontDatabase()
{
  // this code causes repeated access to all the font files
  // which is not slow
  QFontDatabase fontDB;
  QStringList family_names;
  SysFontCacheEntry *font_entry;
  QFont f;

  family_names = fontDB.families( FALSE );
  for(int family_index = 0; family_index < family_names.count(); family_index++) {
    printf("FAMILY %s STYLES = %s  CHARSETS = %s\n",
	   family_names[family_index].latin1(),
	   fontDB.styles(family_names[family_index]).join(":").latin1(),
	   fontDB.charSets(family_names[family_index]).join(":").latin1());

    QStringList family_styles = fontDB.styles(family_names[family_index]);
    for(int style_index = 0; style_index < family_styles.count(); style_index++) {
      QValueList<int> family_sizes = fontDB.pointSizes( family_names[family_index],
							family_styles[style_index]);
      if (family_sizes.count() > 0) {
	for(int size_index = 0; size_index < family_sizes.count(); size_index++) { 
	  int size;
	  size = family_sizes[size_index];
#ifndef __i386__
	  // a hack, sizes on Z seem to be points*10
	  size /= 10;
#endif
	  f = QFont(family_names[family_index], size, 
		    fontDB.weight(family_names[family_index], family_styles[style_index]),
		    fontDB.italic(family_names[family_index], family_styles[style_index]));
#if 1
	  QFontMetrics fm(f);
#else
	  QPainter painter;
	  painter.setFont(f);
	  QFontMetrics fm = painter.fontMetrics();
#endif
	  
	  font_entry = new SysFontCacheEntry();
	  //	if (fm.width("l") == fm.width("m")) {
	  //	  // don't trust f.fixedPitch() or f.exactMatch(), they lie!!
	  //	  font_entry->fixed = true;
	  //	}
	  font_entry->font = f;
	  //	font_entry->name = family_names[family_index];
	  //	font_entry->size = family_sizes[size_index];
	  //	font_entry->pixel_size = f.pixelSize();
	  font_entry->avg_size = fm.width("etaoinr");
	  printf("FONT %s[%d] pixel size %d avg = %f bold %d italic %d\n",
		 family_names[family_index].latin1(),
		 family_sizes[size_index],
		 f.pixelSize(),
		 (float)font_entry->avg_size/ 7.0, f.bold(), f.italic());
	  m_sys_font_database.append(font_entry);
	}
      }
    }
  }
}


QFont QOutputDev::matchFont ( GfxFont *gfxFont, fp_t m11, fp_t m12, fp_t m21, fp_t m22 )
{
	static QDict<QOutFontSubst> stdfonts;

	fouble fsize = sqrt ( m21 * m21 + m22 * m22 );
	
	// this function needs to be rewritten:
	// use cache of fontDB stuff
	// is the avg_size stuff required?

	if (fsize <= 0.0) {
	  m21 *= fouble(1.0/36.0);
	  m22 *= fouble(1.0/36.0);
	  fsize = fouble(6.0) * sqrt( m21 * m21 + m22 * m22 );
	  if (fsize <= 0.0) {
	    printf("Bad font size %f %f = %f\n", (float)m21, (float)m22, (float)fsize);
	    fsize = 50.0;
	  }
	}

	if (qpdf_debug_level > 5 || fsize < 0.01) {
	  printf("matchFont matrix %f %f %f %f\n",
		 (float) m11, (float) m12,(float) m21, (float) m22);
	}
	m_good_font_match = true;

	// build dict for std. fonts on first invocation	
	if ( stdfonts. isEmpty ( )) {
		for ( QOutFontSubst *ptr = qStdFonts; ptr-> m_name; ptr++ ) {
			stdfonts. insert ( QString ( ptr-> m_name ), ptr );
		}	
	}

	int avg_size = 0;

	if (! gfxFont->isCIDFont()) {
	  avg_size = lrint((fouble)fsize * 
			   ( ((Gfx8BitFont*)gfxFont)->getWidth('e')
			     + ((Gfx8BitFont*)gfxFont)->getWidth('t')
			     + ((Gfx8BitFont*)gfxFont)->getWidth('a')
			     + ((Gfx8BitFont*)gfxFont)->getWidth('o')
			     + ((Gfx8BitFont*)gfxFont)->getWidth('i')
			     + ((Gfx8BitFont*)gfxFont)->getWidth('r')
			     + ((Gfx8BitFont*)gfxFont)->getWidth('n') ) );
	  if (qpdf_debug_level > 4) {
	    printf("8BitFont fsize=%f avg_size=%f\n", (float)fsize, (float)avg_size);
	  }
	} else {
	  // 0.55 too big, 0.50 too big, 0.45 sort of works
	  avg_size = lrint((fouble)fsize*7.0*0.35);
	  if (qpdf_debug_level > 0) {
	    printf("CIDFont fsize=%f avg_size=%f\n", (float)fsize, (float)avg_size);
	  }
	}

	QString fname ((gfxFont->getName() && gfxFont->getName()->getCString()) ? 
		       gfxFont-> getName ( )-> getCString ( ) : "<n/a>" );
	QFont f;
	// fast lookup for std. fonts
	QOutFontSubst *subst = stdfonts [fname];

	if ( subst ) {
	  if ( subst-> m_sname )
	    f. setFamily ( subst-> m_sname );
	  f. setStyleHint ( subst-> m_hint, (QFont::StyleStrategy) ( QFont::PreferOutline | QFont::PreferQuality ));	
	  f. setBold ( subst-> m_bold );
	  f. setItalic ( subst-> m_italic );
	} else {
	  QFont::StyleHint sty;
	  
	  if ( gfxFont-> isSerif ( ))
	    sty = QFont::Serif;
	  else if ( gfxFont-> isFixedWidth ( ))
	    sty = QFont::TypeWriter;
	  else
	    sty = QFont::Helvetica;
	  
	  f. setStyleHint ( sty, (QFont::StyleStrategy) ( QFont::PreferOutline | QFont::PreferQuality ));	
	  f. setBold ( gfxFont-> isBold ( ) > 0 );
	  f. setItalic ( gfxFont-> isItalic ( ) > 0 );
	  f. setFixedPitch ( gfxFont-> isFixedWidth ( ) > 0 );
	  
	  // common specifiers in font names 
	  if ( fname. contains ( "Oblique" ) || fname. contains ( "Italic" ))
	    f. setItalic ( true ); 
	  if ( fname. contains ( "Bold" ))
	    f. setWeight ( QFont::Bold );
	  if ( fname. contains ( "Demi" ))
	    f. setWeight ( QFont::DemiBold );
	  if ( fname. contains ( "Light" ))
	    f. setWeight ( QFont::Light );
	  if ( fname. contains ( "Black" ))
	    f. setWeight ( QFont::Black );
	}		
	// Treat x-sheared fonts as italic
	if (( m12 > -0.1 ) && ( m12 < 0.1 ) && ((( m21 > -5.0 ) && ( m21 < -0.1 )) || (( m21 > 0.1 ) && ( m21 < 5.0 )))) {
	  f. setItalic ( true ); 
	}	
	
	if (fsize < 2) {
	  if (qpdf_debug_level > 0) {
	    printf("Font size = %f (%f %f %f %f)\n", 
		   (float)fsize, (float)m11, (float)m12,
		   (float)m21, (float)m22);
	  }
	  fsize = 2; // type3 fonts misbehave sometimes  (???)
	}

	if (!m_reproportion_text) {
	  if (fsize > 128) {  // something in Qt seems to break
	    fsize = 128;
	  }
	  f.setPixelSize ( lrint(fsize) );
	  m_painter-> setFont ( f );

	} else {

	  // increase app startup speed, but pay the price one first adjust spacing
	  if (m_sys_font_database.count() == 0) {
	    cacheSysFontDatabase();
	  }

	  // this section needs to be rewritten
	  // it should use a cached array of font data from QFontDatabase

	  int best_avg_size;
	  int best_fit;
	  int best_font_index = -1;

	  best_avg_size = 100000;
	  best_fit = 100000;

	  // some fonts just have no char widths
	  if (avg_size < 1.0) {
	    // 0.6 was too big
	    // was 0.45 - worked, 0.45, 0.48 & 0.55 were a tiny bit too big
	    if (qpdf_debug_level > 0) {
	      printf("font avg_size = %f, fsize=%f, new avg_size=", (float)avg_size,
		     (float)fsize);
	    }
	    avg_size = 7.0 * 0.43 * fsize;
	    if (qpdf_debug_level > 0) {
	      printf("%f\n", (float)avg_size);
	    }
	  }

	  for(int retry=0; retry<=1; retry++) {
	    for(int font_index=0; font_index<m_sys_font_database.count(); font_index++) {
	      SysFontCacheEntry *font_entry;
	      
	      font_entry = m_sys_font_database[font_index];
	      if (retry == 0
		  && (font_entry->font.bold() != f.bold()
		      || font_entry->font.italic() != f.italic())) {
		continue;
	      }
	      
	      if (
		  // find closest match smaller than target = avg_size
		  (font_entry->avg_size <= avg_size
		   && QABS(font_entry->avg_size - avg_size) < best_fit)
		  ||
		  // if no smaller match, look for smallest larger match
		  (best_avg_size > avg_size
		   && font_entry->avg_size < best_avg_size)
		  ) {
		best_font_index = font_index;
		best_avg_size = font_entry->avg_size;
		best_fit = QABS(font_entry->avg_size - avg_size);
	      }
#if 0
	      printf("try %d avg_size %d font_index %d avg_size %d best_index %d best_avg_size %d best_fit %d\n",
		     retry, avg_size, font_index, font_entry->avg_size,
		     best_font_index, best_avg_size, best_fit);
#endif
	    }

	    if (best_avg_size <= avg_size) {
	      break;
	    }
	  }
	  if (best_avg_size > avg_size + 1) {
	    m_good_font_match = false;
	  }
	  //	  assert(best_font_index >= 0);
	  m_painter-> setFont ( m_sys_font_database[best_font_index]->font );
	}
	
	if (qpdf_debug_level > 0) {
	  //	  QFontMetrics fm(m_painter->fontMetrics());
	  printf("Font ");
	  printFontDescription(gfxFont, fsize);
	  printf ( "size %.2f got size %.2f %s%d/%d%s%s%s%s\n",
		   (float)(avg_size/7.0),
		   (float)(m_painter->fontMetrics().width("etaoinr")/7.0),
		   m_painter->font().family().ascii()? m_painter->font().family().ascii() : "", 
		   m_painter->font().pointSize(), 
		   m_painter->font().weight(), 
		   m_painter->font().bold()?" bold ":"", 
		   m_painter->font().italic()? " ital ":"", 
		   m_painter->font().underline()? " under ":"",
		   m_painter->font().strikeOut()? " strikeout ":"");
	}
	return f;
}



//------------------------------------------------------------------------
// QOutputDev
//------------------------------------------------------------------------


static char *CharSubstitutions[] = {
  "space"," ",
  "hyphen", "-",
  "emdash", "--",
  "endash", "-",
  "multiply", "*",
  "slash", "/",
  "acute","'",
  "breve","`",
  "brokenbar","|",
  "bullet","*",
  "circumflex","^",
  "copyright","@",
  "dagger","+",
  "daggerdbl","++",
  "divide","/",
  "dotlessi","i",
  "dieresis","\"",
  "ellipsis","...",
  "exclamdown","!",
  "fraction","/",
  "grave","`",
  "guillemotleft","<<",
  "guillemotright",">>",
  "guilsignglleft","<",
  "guilsignglright",">",
  "logicalnot","~",
  "minus","-",
  "mu","u",
  "onehalf","1/2",
  "onequarter","1/4",
  "periodcentered",".",
  "perthousand","%o",
  "plusminus","+/-",
  "quote", "'",
  "quoteright", "'",
  "quoteleft", "'",
  "quotedblleft", "\"",
  "quotedblright", "\"",
  "quotedblbase",",,",
  "quotesingle","'",
  "quotesinglbase",",",
  "threequarters","3/4",
  "thorn","p",
  "section","ss",
  "registered","@",
  "lessequal","<=",
  "greatequal",">=",
  "equal","=",
  0, 0
};

QOutputDev::QOutputDev ( QWidget *parent, const char *name, int flags ) 
{

	m_painter = 0;
	m_no_refresh = false;
	
	if (qpdf_debug_level > 6) {
	  globalParams->setPrintCommands(true);
	}

	// create text object
	m_text = new TextPage ( gFalse );

	for( char **ptr = CharSubstitutions; *ptr != NULL; ptr+=2) {
	  fontCharSubstitution.insert(*ptr, new QString(ptr[1]));
	}


	// set up the font cache and fonts
#ifdef RENDER_EMBEDDED_SYMBOL_FONTS	
	pdfFont = NULL;
	fontCache = new QOutputFontCache(this,
					 globalParams->getT1libControl(),
					 globalParams->getFreeTypeControl());
#endif
#ifdef TYPE3_CACHE
	nT3Fonts = 0;
	t3GlyphStack = NULL;
#endif
	// SHOULD BE AN OPTION
	m_reproportion_text = true;
	m_always_render_font = true;
	m_reproportion_and_render = false;

	if ((m_free_memory = getKernelFreeMemory()) == 0) {
	  m_free_memory = 2000000;
	}

	m_pixmap_ptr = NULL;
}

int
QOutputDev::getKernelFreeMemory()
{
  // free memory is used to determine how much to downsample
  // large images
  int total, used, memfree, shared, buffers, cached;
  QFile file( "/proc/meminfo" );
  memfree = 0;
  if ( file.open( IO_ReadOnly ) ) {
    QTextStream t( &file );
    QString dummy = t.readLine();	// title
    t >> dummy;
    t >> total;
    t >> used;
    t >> memfree;
    t >> shared;
    t >> buffers;
    t >> cached;
    m_free_memory = memfree;
    //    printf("dummy=%s total=%d used=%d freemem=%d shared=%d buffers=%d cached=%d\n", 
    //	   dummy.latin1(),total, used, memfree, shared, buffers, cached);
  }
  file.close();
  return(memfree);
}

QOutputDev::~QOutputDev ( ) 
{
	delete m_painter;
	delete m_text;
#ifdef RENDER_EMBEDDED_SYMBOL_FONTS	
	delete fontCache;
#endif
	if (m_pixmap_ptr && *m_pixmap_ptr)
	  delete *m_pixmap_ptr;
#ifdef TYPE3_CACHE
  for (int i = 0; i < nT3Fonts; ++i) {
    delete t3FontCache[i];
  }
#endif
}


void QOutputDev::startDoc ( XRef *_xref )
{
  xref = _xref;
#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
  fontCache->startDoc();
#endif
}



void QOutputDev::startPage ( int /*pageNum*/, GfxState *state ) 
{
  //  assert(m_pixmap_ptr);
  printf("allocate pixmap %d x %d\n",
	 lrint ( state-> getPageWidth ( )), lrint ( state-> getPageHeight ( )));

  if (*m_pixmap_ptr) {
    delete *m_pixmap_ptr;
  }
  *m_pixmap_ptr = new QPixmap ( QMAX(100,lrint ( state-> getPageWidth ( ))), 
				QMAX(100,lrint ( state-> getPageHeight ( ))) );
  if (*m_pixmap_ptr == NULL) {
    printf("Could not allocate pixmap for page size %d x %d\n",
	   lrint(state->getPageWidth()), lrint(state->getPageHeight()));
    return;
  }
  (*m_pixmap_ptr)->fill(white);

  if ((m_free_memory = getKernelFreeMemory()) == 0) {
    m_free_memory = 2000000;
  }

  if (m_painter) {
    delete m_painter;
  }
  m_painter = new QPainter(*m_pixmap_ptr);
  m_painter->setPen(black);
  m_painter->setBrush(black);

  m_text->startPage(state);

  m_last_x = m_last_y = 10000;

#ifdef TYPE3_CACHE
  for (int i = 0; i < nT3Fonts; ++i) {
    delete t3FontCache[i];
  }
  nT3Fonts = 0;
#endif

  m_found_render_font = false;
  m_type3_font = false;
  m_type3_msg = false;
  //  printf("QOutputDev::startPage\n");
}

void QOutputDev::endPage ( ) 
{
  //  printf("QOutputDev::endPage enter\n");
  processQtEvents();
  // should do some event processing inside coalesce too
  m_text->coalesce(false);  // physLayout
  processQtEvents();
  m_pixmap_ptr = NULL;
  delete m_painter;
  m_painter = 0;

#ifdef TYPE3_CACHE
  for (int i = 0; i < nT3Fonts; ++i) {
    delete t3FontCache[i];
  }
  nT3Fonts = 0;
#endif
  //  printf("QOutputDev::endPage exit\n");
}

void QOutputDev::drawLink ( Link *link, Catalog */*catalog*/ ) 
{
  fp_t x1, y1, x2, y2, w;
  LinkBorderStyle *borderStyle;
  GfxRGB rgb;

  link->getRect(&x1, &y1, &x2, &y2);
	
  borderStyle = link->getBorderStyle();
  if (borderStyle->getWidth() > 0) {
    borderStyle->getColor(&rgb.r, &rgb.g, &rgb.b);
    if ( w > 0 ) {
      int x, y, dx, dy;
      
      cvtUserToDev ( x1, y1, &x,  &y );
      cvtUserToDev ( x2, y2, &dx, &dy );
      
      QPen oldpen = m_painter-> pen ( );
      m_painter-> setPen ( QColor(rgb.r, rgb.g, rgb.b ) );
      if (borderStyle->getType() == linkBorderDashed) {
	m_painter-> setPen ( DashLine );
      }
      m_painter-> drawRect ( x, y, dx, dy );
      m_painter-> setPen ( oldpen );
    }
  }
}

void QOutputDev::saveState ( GfxState */*state*/ ) 
{
	QPDFDBG( printf ( "SAVE (CLIP=%d/%d)\n",  m_painter-> hasClipping ( ), !m_painter-> clipRegion ( ). isEmpty ( )));

	m_painter-> save ( );	
}

void QOutputDev::restoreState ( GfxState */*state*/ ) 
{
	m_painter-> restore ( );
	
//	m_painter-> setClipRegion ( QRect ( 0, 0, m_pixmap-> width ( ), m_pixmap-> height ( )));
//	m_painter-> setClipping ( false );
	QPDFDBG ( printf ( "RESTORE (CLIP=%d/%d)\n", m_painter-> hasClipping ( ), !m_painter-> clipRegion ( ). isEmpty ( )));
}

void QOutputDev::updateAll ( GfxState *state ) 
{
	updateLineAttrs ( state, gTrue );
//	updateFlatness ( state );
//	updateMiterLimit ( state );
	updateFillColor ( state );
	updateStrokeColor ( state );
	updateFont ( state );
}

void QOutputDev::updateCTM ( GfxState *state, fp_t /*m11*/, fp_t /*m12*/, fp_t /*m21*/, fp_t /*m22*/, fp_t /*m31*/, fp_t /*m32*/ ) 
{
	updateLineAttrs ( state, gTrue );
}

void QOutputDev::updateLineDash ( GfxState *state ) 
{
	updateLineAttrs ( state, gTrue );
}

void QOutputDev::updateFlatness ( GfxState */*state*/ ) 
{
	// not supported
	QPDFDBG( printf ( "updateFlatness not supported !\n" ));
}

void QOutputDev::updateLineJoin ( GfxState *state ) 
{
	updateLineAttrs ( state, gFalse );
}

void QOutputDev::updateLineCap ( GfxState *state ) 
{
	updateLineAttrs ( state, gFalse );
}

// unimplemented
void QOutputDev::updateMiterLimit ( GfxState */*state*/ ) 
{
	QPDFDBG( printf ( "updateMiterLimit not supported !\n" ));
}

void QOutputDev::updateLineWidth ( GfxState *state ) 
{
	updateLineAttrs ( state, gFalse );
}

void QOutputDev::updateLineAttrs ( GfxState *state, GBool updateDash ) 
{
	fp_t *dashPattern;
	int dashLength;
	fp_t dashStart;

	Qt::PenCapStyle  cap;
	Qt::PenJoinStyle join;
	int width;

	width = lrint ( state-> getTransformedLineWidth ( ));
	
	switch ( state-> getLineCap ( )) {
		case 0: cap = FlatCap; break;
		case 1: cap = RoundCap; break;
		case 2: cap = SquareCap; break;
		default:
			qWarning ( "Bad line cap style (%d)\n", state-> getLineCap ( ));
			cap = FlatCap;
			break;
	}
		
	switch (state->getLineJoin()) {
		case 0: join = MiterJoin; break;
		case 1: join = RoundJoin; break;
		case 2: join = BevelJoin; break;
		default:
			qWarning ( "Bad line join style (%d)\n", state->getLineJoin ( ));
			join = MiterJoin;
			break;
	}
	
	state-> getLineDash ( &dashPattern, &dashLength, &dashStart );

	QColor oldcol = m_painter-> pen ( ). color ( );
	GfxRGB rgb;

	state-> getStrokeRGB ( &rgb );
	oldcol = q_col ( rgb );

	m_painter-> setPen ( QPen ( oldcol, width, dashLength > 0 ? DashLine : SolidLine, cap, join ));

	if ( updateDash && ( dashLength > 0 )) {
		// Not supported by QT
/*	
		char dashList[20];
		if (dashLength > 20)
			dashLength = 20;
		for ( int i = 0; i < dashLength; ++i ) {
			dashList[i] = xoutRound(state->transformWidth(dashPattern[i]));
			if (dashList[i] == 0)
				dashList[i] = 1;
		}
		XSetDashes(display, strokeGC, xoutRound(dashStart), dashList, dashLength);
*/
	}
}

void QOutputDev::updateFillColor ( GfxState *state ) 
{
	GfxRGB rgb;
	state-> getFillRGB ( &rgb );

	m_painter-> setBrush ( q_col ( rgb ));

	if (qpdf_debug_level > 7) {
	  printf("updateFillColor %x %x %x (%f, %f, %f)\n", 
		 m_painter->brush().color().red(), m_painter->brush().color().green(), 
		 m_painter->brush().color().blue(),
		 (float)rgb.r, (float)rgb.g, (float)rgb.b);
	}

}

void QOutputDev::updateStrokeColor ( GfxState *state ) 
{
	GfxRGB rgb;
	state-> getStrokeRGB ( &rgb );
	
	QPen pen = m_painter-> pen ( );
	pen. setColor ( q_col ( rgb ));
	if (qpdf_debug_level > 7) {
	  printf("updateStrokeColor %x %x %x (%f, %f, %f)\n", 
		 pen.color().red(), pen.color().green(), pen.color().blue(),
		 (float)rgb.r, (float)rgb.g, (float)rgb.b);
	}
	m_painter-> setPen ( pen );
}

void QOutputDev::updateFont ( GfxState *state ) 
{
	fp_t m11, m12, m21, m22;
	GfxFont *gfxFont = state-> getFont ( );
	 
	if (! gfxFont) {
	  pdfFont = NULL;
	  return;
	}

	m_text-> updateFont ( state );
	m_text->getFontInfo(&m_min_word_space, &m_max_word_space, &m_max_char_space);
	if (qpdf_debug_level > 1) {
	  printf("adjust max_char_space = %f + %f * %f\n",
		 (float)m_max_char_space, 0.25, (float)m_min_word_space);
	}
	m_max_char_space += fouble(0.25) * m_min_word_space;
	m_max_word_space *= fouble(1.33);

	m_type3_font = (gfxFont->getType() == fontType3);
	if (m_type3_font) {
	  m_type3_msg = true;
	}
	if (m_type3_font) {
	  pdfFont = NULL;
	  return;
	}

	state-> getFontTransMat ( &m11, &m12, &m21, &m22 );
	m11 *= state-> getHorizScaling ( );
	m12 *= state-> getHorizScaling ( );
  
	// could use m_text->getFontSize() and save a sqrt
	matchFont ( gfxFont, m11, m12, m21, m22 );

	pdfFont = NULL;
#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
	// if (gfxFont->isSymbolic() || m_always_render_font) {
	if (m_always_render_font) {
	  //	if ( strstr(gfxFont->getName()->getCString(),"Math") != NULL
	  //	     || strstr(gfxFont->getName()->getCString(),"math") != NULL) {
	  pdfFont = fontCache->getFont(xref, gfxFont, m11, m12, m21, m22 );
	}
#endif

}

void QOutputDev::stroke ( GfxState *state ) 
{
	QPointArray points;
	QArray<int> lengths;

	// transform points
	int n = convertPath ( state, points, lengths );

	QPDFDBG( printf ( "DRAWING: %d POLYS\n", n ));
	
	// draw each subpath
	int j = 0;
	for ( int i = 0; i < n; i++ ) {
		int len = lengths [i];
	
		if ( len >= 2 ) {
			QPDFDBG( printf ( " - POLY %d: ", i ));
			QPDFDBG( for ( int ii = 0; ii < len; ii++ ))
				QPDFDBG( printf ( "(%d/%d) ", points [j+ii]. x ( ), points [j+ii]. y ( )));
			QPDFDBG( printf ( "\n" ));
			
			// for some reason, it doesn't like degenerate polygons
			// (two points that are equal)
			if (! (len == 2 
			       && points[j].x() == points[j+1].x()
			       && points[j].y() == points[j+1].y()) ) {
			  m_painter-> drawPolyline ( points, j, len );
			}
		}	
		j += len;
	}
	processQtEvents();
}

void QOutputDev::fill ( GfxState *state ) 
{
	doFill ( state, true );
}

void QOutputDev::eoFill ( GfxState *state ) 
{
	doFill ( state, false );
}

//
//  X doesn't color the pixels on the right-most and bottom-most
//  borders of a polygon.  This means that one-pixel-thick polygons
//  are not colored at all.  I think this is supposed to be a
//  feature, but I can't figure out why.  So after it fills a
//  polygon, it also draws lines around the border.  This is done
//  only for single-component polygons, since it's not very
//  compatible with the compound polygon kludge (see convertPath()).
//

//FIXME I've seen large negative numbers in points here
//  may be due to fixed.h stuff

//#define QPDFDBG(X) X

void QOutputDev::doFill ( GfxState *state, bool winding ) 
{
	QPointArray points;
	QArray<int> lengths;

	// transform points
	int n = convertPath ( state, points, lengths );


	QPDFDBG( printf ( "FILLING: %d POLYS\n", n ));
	
	QPen oldpen = m_painter-> pen ( );
	m_painter-> setPen ( QPen ( NoPen ));
	
	// draw each subpath
	int j = 0;
	for ( int i = 0; i < n; i++ ) {
		int len = lengths [i];
	
		if ( len >= 3 ) {
			QPDFDBG( printf ( " - POLY %d: ", i ));
			QPDFDBG( for ( int ii = 0; ii < len; ii++ ))
				QPDFDBG( printf ( "(%d/%d) ", points [j+ii]. x ( ), points [j+ii]. y ( )));
			QPDFDBG( printf ( "\n" ));
		
			m_painter-> drawPolygon ( points, winding, j, len );
		}	
		j += len;
	}
	m_painter-> setPen ( oldpen );

	processQtEvents();
}
//#define QPDFDBG(X) 

void QOutputDev::clip ( GfxState *state ) 
{
	doClip ( state, true );
}

void QOutputDev::eoClip ( GfxState *state ) 
{
	doClip ( state, false );
}


void QOutputDev::doClip ( GfxState *state, bool winding ) 
{
	QPointArray points;
	QArray<int> lengths;

	// transform points
	int n = convertPath ( state, points, lengths );

	QRegion region;
	
	QPDFDBG( printf ( "CLIPPING: %d POLYS\n", n ));
	

	if (n == 0) {
	  return;
	}

	// draw each subpath
	int j = 0;
	for ( int i = 0; i < n; i++ ) {
		int len = lengths [i];
		
		if ( len >= 3 ) {		
			QPointArray dummy;
			dummy. setRawData ( points. data ( ) + j, len );
		
			QPDFDBG( printf ( " - POLY %d: ", i ));
			QPDFDBG( for ( int ii = 0; ii < len; ii++ ) printf ( "(%d/%d) ", points [j+ii]. x ( ), points [j+ii]. y ( )));
			QPDFDBG( printf ( "\n" ));
		
			region |= QRegion ( dummy, winding );
		
			dummy. resetRawData ( points. data ( ) + j, len );
		}		
		j += len;
	}
	
	if ( m_painter-> hasClipping ( ))
		region &= m_painter-> clipRegion ( );

	m_painter-> setClipRegion ( region );
	m_painter-> setClipping ( true );
	
//	m_painter-> fillRect ( 0, 0, m_pixmap-> width ( ), m_pixmap-> height ( ), red );
//	m_painter-> drawText ( points [0]. x ( ) + 10, points [0]. y ( ) + 10, "Bla bla" );
	processQtEvents();
}



//
// Transform points in the path and convert curves to line segments.
// Builds a set of subpaths and returns the number of subpaths.
// If <fillHack> is set, close any unclosed subpaths and activate a
// kludge for polygon fills:  First, it divides up the subpaths into
// non-overlapping polygons by simply comparing bounding rectangles.
// Then it connects subaths within a single compound polygon to a single
// point so that X can fill the polygon (sort of).
//
int QOutputDev::convertPath ( GfxState *state, QPointArray &points, QArray<int> &lengths ) 
{
	GfxPath *path = state-> getPath ( );
	int n = path-> getNumSubpaths ( );
	int m = 0;
	lengths. resize ( n );

	// do each subpath
	for ( int i = 0; i < n; i++ ) {
		// transform the points
		lengths[m] = convertSubpath ( state, path-> getSubpath ( i ), points );
		if (lengths[m] != 0) {
		  m++;
		}
	}

	lengths.resize(m);
	return m;
}

//
// Transform points in a single subpath and convert curves to line
// segments.
//
int QOutputDev::convertSubpath ( GfxState *state, GfxSubpath *subpath, QPointArray &points )
{
	int oldcnt = points. count ( );
	fp_t x0, y0, x1, y1, x2, y2, x3, y3;

	int m = subpath-> getNumPoints ( );
	int i = 0;
	
	while ( i < m ) {
		if ( i >= 1 && subpath-> getCurve ( i )) {
		  //		  printf("subpath: ");
			state-> transform ( subpath-> getX ( i - 1 ), subpath-> getY ( i - 1 ), &x0, &y0 );
			//			printf("(%d,%d) ", lrint(x0), lrint(y0));
			if (x0 < 0 || y0 < 0 || x0 >= state->getPageWidth() || y0 >= state->getPageHeight()) goto skip;
			state-> transform ( subpath-> getX ( i ),     subpath-> getY ( i ),     &x1, &y1 );
			//			printf("(%d,%d) ", lrint(x1), lrint(y1));
			if (x1 < 0 || y1 < 0 || x1 >= state->getPageWidth() || y1 >= state->getPageHeight()) goto skip;
			state-> transform ( subpath-> getX ( i + 1 ), subpath-> getY ( i + 1 ), &x2, &y2 );
			//			printf("(%d,%d) ", lrint(x2), lrint(y2));
			if (x2 < 0 || y2 < 0 || x2 >= state->getPageWidth() || y2 >= state->getPageHeight()) goto skip;
			state-> transform ( subpath-> getX ( i + 2 ), subpath-> getY ( i + 2 ), &x3, &y3 );
			//			printf("(%d,%d) ", lrint(x3), lrint(y3));
			if (x3 < 0 || y3 < 0 || x3 >= state->getPageWidth() || y3 >= state->getPageHeight()) goto skip;
			
			QPointArray tmp;
			//			printf(" added\n");
			tmp. setPoints ( 4, lrint ( x0 ), lrint ( y0 ), lrint ( x1 ), lrint ( y1 ),
			                    lrint ( x2 ), lrint ( y2 ), lrint ( x3 ), lrint ( y3 ));

#if QT_VERSION < 300
			tmp = tmp. quadBezier ( );
			
			for ( uint loop = 0; loop < tmp. count ( ); loop++ ) {
				QPoint p = tmp. point ( loop );
				points. putPoints ( points. count ( ), 1, p. x ( ), p. y ( ));
			}
#else
			tmp = tmp. cubicBezier ( );
			
			points. putPoints ( points. count ( ), tmp. count ( ), tmp );			
#endif

			i += 3;
		} 
		else {
			state-> transform ( subpath-> getX ( i ), subpath-> getY ( i ), &x1, &y1 );
			//		printf("start subpaths (%d,%d)\n", lrint(x1), lrint(y1));
			if (x1 < 0 || y1 < 0 || x1 >= state->getPageWidth() || y1 >= state->getPageHeight()) goto skip;
			
			points. putPoints ( points. count ( ), 1, lrint ( x1 ), lrint ( y1 ));			
			++i;
		}
	}

	//	printf("return %d - %d\n", points.count(), oldcnt);
	return points. count ( ) - oldcnt;

 skip:
	//	printf("skipped path\n");
	points.resize(oldcnt);
	return 0;
}


class Debug_context {
  int size, last;
  char *ring;
  int post_size;
  int post_countdown;
public:
  Debug_context() {
    int i;
    size = 20;
    post_size = 20;
    ring = new char[size];
    last = 0;
    post_countdown = 0;
    for(i=0; i<size; i++) {
      ring[i] = '\0';
    }
  }
  void put(QString s) {
    int i;
    int n = s.length();
    for(i=0; i<n; i++) {
      put(s[i].latin1());
    }
  }

  void put(char c) {
    ring[last++] = c;
    if (last >= size) {
      last =0;
    }
    if (post_countdown > 0) {
      printf("%c", c);
    }
    if (--post_countdown == 0) {
      printf("\n");
    }
  }
  void print() {
    int i;
    printf("\n");
    for(i=last; i<size; i++) {
      if (ring[i] != '\0') {
	printf("%c", ring[i]);
      }
    }
    for(i=0; i<last; i++) {
      if (ring[i] != '\0') {
	printf("%c", ring[i]);
      }
    }
    post_countdown = post_size;
  }
} debug_context;


void
debug_font(int place, GfxFont *gfxFont, CharCode c, char *name)
{
  debug_context.print();
  printf("<<");
  printFontDescription(gfxFont, *gfxFont->getFontMatrix());
  printf("(0x%x)'%s'@%d>>\n",
	 c, name, place);
}



void QOutputDev::beginString ( GfxState *state, GString */*s*/ ) 
{
  m_text-> beginWord ( state , state->getCurX(), state->getCurY()); // PDK
  //  m_last_x = state->getCurX();
  //  m_last_y = state->getCurY();
  if (qpdf_debug_level > 0) debug_context.put('_');
}

void QOutputDev::endString ( GfxState */*state*/ ) 
{
  m_text-> endWord ( );
}


void 
QOutputDev::updateTextMat(GfxState *state)
{
  PDKDEBUG(printf("updateTextMat\n"));
}

void 
QOutputDev::updateCharSpace(GfxState *state)
{
  PDKDEBUG(printf("updateCharSpace\n"));
}

void 
QOutputDev::updateRender(GfxState *state)
{
}

void 
QOutputDev::updateRise(GfxState *state)
{
  PDKDEBUG(printf("updateRise\n"));
}

void 
QOutputDev::updateWordSpace(GfxState *state)
{
  PDKDEBUG(printf("updateWordSpace\n"));
}

void 
QOutputDev::updateHorizScaling(GfxState *state)
{
  PDKDEBUG(printf("updateHorizScaling\n"));
}

void 
QOutputDev::updateTextPos(GfxState *state)
{
  PDKDEBUG(printf("updateTextPos\n"));
}

void 
QOutputDev::updateTextShift(GfxState *state, fouble shift)
{
}

#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
void
QOutputDev::renderFont(GfxState *state, int render, 
		       fouble x1, fouble y1, fouble dx1, fouble dy1,
		       CharCode code, Unicode *u, int uLen)
{
  GfxRGB rgb;
    
  if (!(render & 1)) {
    state->getFillRGB(&rgb);
#if 0
      printf("fill color %f %f %f pos %f %f dpos %f %f code %d len %d\n",
	     (float)rgb.r, (float)rgb.g, (float)rgb.b,
	     (float)x1, (float)y1,
	     (float)dx1, (float)dy1,
	     code, uLen);
#endif      
      pdfFont->drawChar(state, *m_pixmap_ptr, &rgb, 
			x1, y1, dx1, dy1, code, u, uLen);
      m_found_render_font = true;
  }
    
  // stroke
  if ((render & 3) == 1 || (render & 3) == 2) {
    if (qpdf_debug_level > 0) {
      printf("stroke\n");
    }
    if (pdfFont->hasGetCharPath()) {
      if (qpdf_debug_level > 0) {
	printf("hasCharPath\n");
      }
    } else {
      // can't stroke the outline, so just fill it using the stroke
      // color
      state->getStrokeRGB(&rgb);
      pdfFont->drawChar(state, *m_pixmap_ptr, &rgb,
			x1, y1, dx1, dy1, code, u, uLen);
      m_found_render_font = true;
    }
  }
  processQtEvents();
  return;
}
  
#endif
  



void QTextOutputDev::beginString(GfxState *state, GString *s) {
  text->beginWord(state, state->getCurX(), state->getCurY());
}

void QTextOutputDev::endString(GfxState *state) {
  text->endWord();
}

void QTextOutputDev::updateFont(GfxState *state) {
  text->updateFont(state);
  text->getFontInfo(&m_min_word_space, &m_max_word_space, &m_max_char_space);
  if (qpdf_debug_level > 3) {
    printf("update font min word %f max word %f max char %f\n",
	   (float)m_min_word_space, (float)m_max_word_space, (float)m_max_char_space);
  }
  m_max_char_space = m_max_char_space + fouble(0.25) * m_min_word_space;
  m_max_word_space *= fouble(1.33);
  m_type3_font = (state->getFont() == NULL? false :
		  (state->getFont()->getType() == fontType3));
  if (m_type3_font) {
    m_type3_msg = true;
  }

  // need to guess if embedded font would be rendered
  // because: adjust_spacing would not be in effect even if the option is set
  // this changes the screen to word mapping
  m_embedded_font = false;
#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
  if (m_always_render_font && state->getFont()) {
    Ref ref;
    GfxFont *gfxFont = state->getFont();
    if (gfxFont->getEmbeddedFontID(&ref)) {
      switch (gfxFont->getType()) {
      case fontType1:
      case fontType1C:
#if HAVE_T1LIB_H
	if (t1libControl != fontRastNone) {
	  m_embedded_font = true;
	}
#endif
#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
	if (freetypeControl != fontRastNone) {
	  m_embedded_font = true;
	}
#endif
	break;
      case fontTrueType:
      case fontCIDType2:
#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
	if (freetypeControl != fontRastNone) {
	  m_embedded_font = true;
	}
#endif
#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
	if (freetypeControl != fontRastNone) {
	  m_embedded_font = true;
	}
#endif
	break;
      case fontCIDType0:
      case fontCIDType0C:
#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
	if (freetypeControl != fontRastNone) {
	  m_embedded_font = true;
	}
#endif
	break;
      }
    }
  }
#endif
}

void QTextOutputDev::startPage(int pageNum, GfxState *state) {
  //  printf("QTextOutputDev::startPage\n");
  text->startPage(state);
  m_last_x = m_last_y = 10000;
  m_type3_font = false;
  m_type3_msg = false;
  m_width = lrint( state-> getPageWidth ( ));
  m_height = lrint ( state-> getPageHeight ( ));

}

void QTextOutputDev::drawChar(GfxState *state, fouble x, fouble y,
			      fouble dx, fouble dy,
			      fouble originX, fouble originY,
			      CharCode code, Unicode *u, int uLen) {
  fp_t x1, y1, dx1, dy1;


  bool unknown_char = false;
  int render = state->getRender();

  int hack_dx1;
  fouble real_x1;
  fouble real_dx1;

  Unicode unicode[100];

  // check for invisible text -- this is used by Acrobat Capture
  if (( render & 3 ) == 3 ) {
    return;
  }

  if (m_type3_font) {
    return;
  }
  
  /*  transform into screen coord x1,y1 dx1,dy1 */

  x -= originX;
  y -= originY;
  state-> transform      ( x,  y,  &x1,  &y1 );
  state-> transformDelta ( dx, dy, &dx1, &dy1 );
  
  real_x1 = x1;
  real_dx1 = dx1;


  if (uLen == 0 && code == 0) {
    // some PDF files use CID 0, which is .notdef, so just ignore it
    return;
  }
  
  QString out;

  /*  hack 8 bit fonts for missing characters  */
  /*  (with more complete fonts on Zaurus, this wouldn't be needed) */

  if (uLen == 0 && code == 0x20) {
    out = " ";
  }
  
  /* This code is copied from QOutputDev::drawChar, 
     should be a function!!
  */
  if (state->getFont()->isCIDFont()) {
    /*   CID font, no games for now   */
    for(int i=0; i<uLen; i++) {
      /* okay, a few little games... */
      // TimesNewRomanPSMT
	if (u[i] == 0xfb01) {
	  out += "fi";
	} else if (u[i] == 0xfb02) {
	  out += "fl";
	} else if (u[i] == 0x2019 || u[i] == 0xffffffbc) {
	  out += "'";
	} else if (u[i] == 0x201c || u[i] == 0x201d) {
	  out += "\"";
	} else if (u[i] == 0x2014) {
	  out += "--";
	} else if (u[i] == 0x2013) {
	  out += "~";
	} else {
	  out += u[i];
	}
    }
  } else {
    for(int i=0; i<uLen; i++) {
      Unicode uu;
      uu = u[i];
      if (uu >= ' ' && uu <= '~') {
	out += uu;
      } else { // not printing ascii, and not CID
	Gfx8BitFont *font = (Gfx8BitFont*) state->getFont();
	char *cname = font->getCharName(code);
	if (cname != NULL) {
	  QString *s = QOutputDev::fontCharSubstitution[cname];
	  if (s != NULL) {
	    out += *s;
	    continue;
	  } 
	  
	  // no lookup substitute in table
	  QString sub;
	  if (cname[0] != '\0') {
	    // this often works: for 1 and 2 char letter names, take those letters
	    if (strlen(cname) < 3) {
	      sub = cname;
	    } else {
	      // real hacks
	      if (strcmp(cname,"C174") == 0  && code == 0xae) {
		sub = "fi";
	      } else if (strcmp(cname,"C175") == 0 && code == 0xaf) {
		sub = "fl";
	      } else if (strcmp(cname,"C177") == 0 
			 && (code == 0xb1 || code == 0x20)) {
		sub = "-";
	      } else if (strcmp(cname,"C128") == 0 
			 && code == 0x80) {
		sub = "ff";
	      } else {
		sub = uu;
	      }
	    }
	  } else {
	    /* no char code name to work with */
	    sub = QChar(uu);
	  }
	  out += sub;
	}
      }
    }
  }

  /*   adjust spacing hack */
  /* we do all this crap so that when going back to pdf mode
     from text mode, the screen coord are correct */
#if 0
  out_pre_adjust = out;

  if (m_reproportion_text && ! m_embedded_font) {
    // text hack: attempt to reproportion spacing for substituted font
    // 0.5 works most all the time, was 0.6
    if (state->getCharSpace() > 0.7) {
      int index;
    }
    
    if (QABS(state->getCurY() - m_last_y) > 0.08) {
      m_last_y = state->getCurY();
      m_last_x = x1;
      m_hack_x1 = x1;
    } else {
      if (x1 < m_last_x - m_max_word_space) {
	m_hack_x1 = x1;
      } else {
	if (x1 > m_last_x + m_max_char_space) {
	  if (x1 > m_last_x + m_max_word_space) {
	    m_hack_x1 = x1;
	  }
	}
      }
    }
    
    m_last_x = x1 + dx1;
    
    if (state->getCharSpace() != fouble(0.0)) {
      fouble dx2, dy2;
      // duplicate calculations in TextOutputDev.cc
      // ignoring word space for now
      state->textTransformDelta(state->getCharSpace() * state->getHorizScaling(), 0,
				&dx2, &dy2);
      m_last_x -= dx2;
    }
    
    hack_dx1 = dx1;  // fm.width(out);
    
    x1 = m_hack_x1;
    m_hack_x1 += hack_dx1;
    dx1 = hack_dx1;
  }
#endif
  
  /*  put it back into buffer for xpdf code */
  /*  (wouldn't need to do this if we could get pointer to unicode from QString) */
  uLen = out.length();
  for(int i=0; i<uLen; i++) {
    unicode[i] = out[i].unicode();
  }

  text-> addChar ( state, x, y, dx, dy, code, unicode, uLen,
		   real_x1, y1, real_dx1, dy1, x1, dx1);

#if 0
  fouble x1, y1, w1, h1;
  state-> transform      ( x,  y,  &x1,  &y1 );
  state-> transformDelta ( dx, dy, &w1, &h1 );
  text->addChar(state, x, y, dx, dy, c, u, uLen, 
		x1, y1, w1, h1, x1, w1);
#endif
}

//ARM_SPEEDUP
void QOutputDev::drawChar ( GfxState *state, fp_t x, fp_t y,
                           fp_t dx, fp_t dy, fp_t originX, fp_t originY,
                           CharCode code, Unicode *u, int uLen ) 
{

  fp_t x1, y1, dx1, dy1;


  bool unknown_char = false;
  int render = state->getRender();

  int hack_dx1;
  fouble real_x1;
  fouble real_dx1;

  Unicode unicode[100];

  // check for invisible text -- this is used by Acrobat Capture
  if (( render & 3 ) == 3 ) {
    return;
  }
  
  /*  transform into screen coord x1,y1 dx1,dy1 */

  x -= originX;
  y -= originY;
  state-> transform      ( x,  y,  &x1,  &y1 );
  state-> transformDelta ( dx, dy, &dx1, &dy1 );
  
  if (qpdf_debug_level > 5) {
    printf("x %f y %f origin %f %f dx %f dy %f x1 %f y1 %f dx1 %f dy1 %f\n",
	   (float) x, (float) y, (float)  originX, (float) originY,
	   (float) dx, (float) dy, (float) x1, (float) y1, (float) dx1,
	   (float) dy1);
  }

  real_x1 = x1;
  real_dx1 = dx1;


  if (uLen == 0 && code == 0) {
    // some PDF files use CID 0, which is .notdef, so just ignore it
    return;
  }
  
  QFontMetrics fm = m_painter-> fontMetrics ( );
  QString out;
  QString out_pre_adjust;

  /*  hack 8 bit fonts for missing characters  */
  /*  (with more complete fonts on Zaurus, this wouldn't be needed) */

  if (qpdf_debug_level > 20) {
    printf("adjust_space=%d pdfFont=%x CID=%d embedded=%s\n", m_reproportion_text,
	   pdfFont, state->getFont()->isCIDFont(), 
	   state->getFont()->getEmbeddedFontName()->getCString());
  }

#if 0
  if (! state->getFont()->isCIDFont() )
    //      && !state->getFont()->getEmbeddedFontName()) {
    Gfx8BitFont *font = (Gfx8BitFont*) state->getFont();
#endif

  if (uLen == 0 && code == 0x20) {
    if (qpdf_debug_level > 1) {
      printf("code = space\n");
    }
    out = " ";
  }
  
  if (state->getFont()->isCIDFont()) {
    /*   CID font, no games for now   */
    for(int i=0; i<uLen; i++) {
      /* okay, a few little games... */
      // TimesNewRomanPSMT
      if (u[i] == 0xfb01) {
	out += "fi";
      } else if (u[i] == 0xfb02) {
	out += "fl";
      } else if (u[i] == 0x2019 || u[i] == 0x2bc) {
	out += "'";
      } else if (u[i] == 0x201c || u[i] == 0x201d) {
	out += "\"";
      } else if (u[i] == 0x2014) {
	out += "--";
      } else if (u[i] == 0x2013) {
	out += "~";
      } else {
	out += u[i];
      }
      if (qpdf_debug_level > 0) {
	debug_context.put(u[i]);
	if (qpdf_debug_level > 1 && !fm.inFont(out[i])) {
	  debug_font(8,state->getFont(),u[i],"");
	}
      }
    }
  } else {
    for(int i=0; i<uLen; i++) {
      Unicode uu;
      uu = u[i];
      //      if (fm.inFont(uu)) {   // can't trust this, is says yes then draws a red box!
      if (uu >= ' ' && uu <= '~') {
	if(fm.inFont(uu)){
	  out += uu;
	} else { 
	  out += ' ';
	  if (qpdf_debug_level > 0) debug_font(4,state->getFont(),uu,"");
	  unknown_char = true;
	}
	if (qpdf_debug_level > 0) debug_context.put(uu);
	continue;
      }

      // not printing ascii, and not CID
      /* try to substitute char */
      Gfx8BitFont *font = (Gfx8BitFont*) state->getFont();
      char *cname = font->getCharName(code);
      if (cname != NULL) {
	QString *s = fontCharSubstitution[cname];
	if (s != NULL) {
	  out += *s;
	  if (qpdf_debug_level > 1) {
	    printf("substitution code %x cname %s => %s\n", code, cname, 
		   s->latin1());
	  }
	  continue;
	} 
	
	// no lookup substitute in table
	QString sub;
	if (cname[0] != '\0') {
	  // this often works: for 1 and 2 char letter names, take those letters
	  if (strlen(cname) < 3) {
	    sub = cname;
	  } else {
	    // real hacks
	    if (strcmp(cname,"C174") == 0  && code == 0xae) {
	      sub = "fi";
	    } else if (strcmp(cname,"C175") == 0 && code == 0xaf) {
	      sub = "fl";
	    } else if (strcmp(cname,"C177") == 0 
		       && (code == 0xb1 || code == 0x20)) {
	      sub = "-";
	    } else if (strcmp(cname,"C128") == 0 
		       && code == 0x80) {
	      sub = "ff";
	    } else {
	      if (fm.inFont(uu)) {
		sub = uu;
		if (qpdf_debug_level > 1) {
		  debug_font(1,state->getFont(),code,cname);
		}
	      } else {
		sub = '?';
		if (qpdf_debug_level > 0) debug_font(2,state->getFont(),code,cname);
		unknown_char = true;
	      }
	    }
	  }
	} else {
	  /* no char code name to work with */
	  QChar c = QChar(uu);
	  if (fm.inFont(c)) {
	    sub = c;
	    if (qpdf_debug_level > 1) {
	      debug_font(3,state->getFont(),code,cname);
	    }
	  } else {
	    if (qpdf_debug_level > 0) debug_font(4,state->getFont(),code,"");
	    unknown_char = true;
	    sub = '?';
	  }
	}
	if (qpdf_debug_level > 1) {
	  printf("subst char %x => %s => %s\n", code, cname, sub.latin1());
	}
	out += sub;
      }
    }
  }


  /*   adjust spacing hack */

  out_pre_adjust = out;

#if 0
    printf("x1=%f next_x1=%f last_x1=%f x1-last_x1=%f (%f %f) y1=%f code=%x out=%s\n",
	   (float)x1,(float)(x1+dx1),(float)m_last_x,(float)(x1 - m_last_x),
	   (float)m_min_word_space, (float)m_max_word_space,
	   (float)y1,code,out.latin1());
#endif
  
  
  // this works pretty well, but in files with embedded symbols and adjusted substituted fonts,
  // the symbols come out shifted from the text
  if (m_reproportion_text && (! pdfFont)) {

    // text hack: attempt to reproportion spacing for substituted font
    if (qpdf_debug_level > 3) {
      debug_context.print();
      printf("] %f ws %f-%f cs %f csp=%f wsp=%f hsca=%f\n",
	     (float)(x1 - m_last_x),
	     (float)m_min_word_space, (float)m_max_word_space,
	     (float)m_max_char_space,
	     (float)state->getCharSpace(), (float)state->getWordSpace(), 
	     (float)state->getHorizScaling());
    }
    // 0.5 works most all the time, was 0.6
    // 0.7 was the standard for a long time
    if (state->getCharSpace() > 0.7) {
      int index;
      while((index = out.find(' ')) >= 0) {
	out.remove(index,1);
      }
      if (qpdf_debug_level > 3) {
	printf("<insert space>charspace(%f)\n", 
	       (float)state->getCharSpace());
      }
      QString out2;
      for(index=0; index<out.length(); index++) {
	out2 = " " + out[index];
      }
      out = out2;
    }
    
    if (QABS(state->getCurY() - m_last_y) > 0.08) {
      if (qpdf_debug_level > 4) {
	printf("new line y=%f cury=%f lastY=%f dy=%f x1=%f y1=%f\n", (float)y,
	       (float)state->getCurY(),
	       (float)m_last_y,
	       (float)(QABS(state->getCurY()-m_last_y)), (float)x1, (float)y1);
      }
      m_last_y = state->getCurY();
      m_last_x = x1;
      m_hack_x1 = x1;
    } else {
      if (x1 < m_last_x - m_max_word_space) {
	if (qpdf_debug_level > 3) {
	  printf("(backtab %f < %f - maxspace %f (crit max_space=%f))\n",
		 (float)x1, (float)m_last_x, (float)m_max_word_space,
		 (float)(-x1 + m_last_x));
	}
	m_hack_x1 = x1;
      } else {
	if (x1 > m_last_x + m_max_char_space) {
	  //	if (x1 > m_last_x + m_min_word_space) {
	  if (x1 > m_last_x + m_max_word_space) {
	    if (qpdf_debug_level > 3) {
	      printf("(tab %f > %f + maxspace %f (crit max_space=%f))\n",
		     (float)x1, (float)m_last_x, (float)m_max_word_space,
		     (float)(x1 - m_last_x));
	    }
	    m_hack_x1 = x1;
	  } else {
	    if (out[0] != ' ' && out[out.length()-1] != ' ') {
	      if (qpdf_debug_level > 3) {
		printf("(SP+ %f > %f + minspace %f (crit min_space=%f))\n", 
		       (float)x1, (float)m_last_x, (float)m_max_char_space,
		       (float)(x1-m_last_x));
	      }
	      out = " " + out;
	    } else {
	      if (qpdf_debug_level > 4) {
		printf("(SP- %f < %f + %f (min_space=%f))\n",
		       (float)x1, (float)m_last_x, (float)m_max_char_space,
		       (float)(x1-m_last_x));
	      }
	    }
	  }
	} else {
	  if (qpdf_debug_level > 4) {
	    printf("(SP- %f<%f)", (float)(x1-m_last_x),(float)m_max_char_space);
	  }
	}
      }
    }
    
    m_last_x = x1 + dx1;
    
    if (state->getCharSpace() != fouble(0.0)) {
      fouble dx2, dy2;
      // duplicate calculations in TextOutputDev.cc
      // ignoring word space for now
      state->textTransformDelta(state->getCharSpace() * state->getHorizScaling(), 0,
				&dx2, &dy2);
      m_last_x -= dx2;
    }
    
    if (pdfFont) {
      hack_dx1 = dx1;  // need to fix this to account for spacing
    } else {
      hack_dx1 = fm.width(out);
    }
    
    x1 = m_hack_x1;
    m_hack_x1 += hack_dx1;
    dx1 = hack_dx1;
  }
  
  /*  put it back into buffer for xpdf code */
  /*  (wouldn't need to do this if we could get pointer to unicode from QString) */
  uLen = out_pre_adjust.length();
  for(int i=0; i<uLen; i++) {
    unicode[i] = out_pre_adjust[i].unicode();
  }
  
  m_text-> addChar ( state, x, y, dx, dy, code, 
		     unicode, uLen,
		     real_x1,  y1, real_dx1, dy1, x1, dx1);


#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
  if (pdfFont) {
#if 0
    if (m_reproportion_text 
	&& QABS(state->getCurY() - m_last_y) < 0.1
	&& x1 > m_last_x
	&& x1 - m_last_x < m_max_space) {
      x1 = m_hack_x1;
      m_hack_x1 += lrint(dx1) + 1;
    }
#endif
    renderFont(state, render, x1, y1, dx1, dy1, code, u, uLen);
    return;
  }
#endif


#ifndef QT_NO_TRANSFORMATIONS
  QWMatrix oldmat;
  
  // not needed unless transform?
  fp_t m11, m12, m21, m22;
  
  state-> getFontTransMat ( &m11, &m12, &m21, &m22 );
  m11 *= state-> getHorizScaling ( );
  m12 *= state-> getHorizScaling ( );
  //
  
  bool dorot = (( m12 < -0.1 ) || ( m12 > 0.1 )) && (( m21 < -0.1 ) || ( m21 > 0.1 ));		
  
  if ( dorot ) {	
    oldmat = m_painter-> worldMatrix ( );
    
    std::cerr << std::endl << "ROTATED: " << (double)m11 << ", " << (double)m12 << ", " << (double)m21 << ", " << (double)m22 << " / SIZE: " << (double)fsize << " / TEXT: " << str. local8Bit ( ) << std::endl << std::endl;
    
    QWMatrix mat ( lrint ( m11 / fsize ), lrint ( m12 / fsize ), -lrint ( m21 / fsize ), -lrint ( m22 / fsize ), lrint ( x1 ), lrint ( y1 ));
    
    m_painter-> setWorldMatrix ( mat );
    
    x1 = 0;
    y1 = 0;			
  }
#endif
  
  /*     set color    */

  QPen oldpen = m_painter-> pen ( );
  
  if (qpdf_debug_level > 6) {
    printf("'%s'(%5d,%5d) render %x color %x %x %x\n", 
	   out.latin1(), lrint(x), lrint(y),
	   state->getRender(),
	   m_painter->brush().color().red(), 
	   m_painter->brush().color().green(), m_painter->brush().color().blue());
  }

  if (!( state-> getRender ( ) & 1 )) {
    QPen fillpen = oldpen;
    
    fillpen. setColor ( m_painter-> brush ( ). color ( ));
    m_painter-> setPen ( fillpen );
    
  }	
  
#if 0
  if (unknown_char) {
    QPen fillpen = oldpen;
    fillpen. setColor ( Qt::red );
    m_painter-> setPen ( fillpen );
  }
#endif

  /*    draw characters using available fonts   */

  if (m_good_font_match) {
    m_painter-> drawText ( lrint ( x1 ), lrint ( y1 ), out );
  } else {
    m_painter-> fillRect ( lrint ( x1 ), lrint ( y1 ), 
			   lrint ( QMAX( fp_t(1), dx1 )), 
			   lrint ( QMAX( ((fouble)4.0), dy1 )), 
			   QBrush(Qt::gray, Qt::Dense1Pattern) );
  }
  
  /*    reset graphics context    */

  m_painter-> setPen ( oldpen );
  
#ifndef QT_NO_TRANSFORMATIONS
  if ( dorot )
    m_painter-> setWorldMatrix ( oldmat );
#endif	
  
  //	QPDFDBG( printf("%s (%f,%f=>%d,%d)\n", str.local8Bit().data(), (float)x1, (float)y1, lrint(x1), lrint(y1)); );
  //	QPDFDBG( printf ( "DRAW TEXT: \"%s\" at (%ld/%ld)\n", str. local8Bit ( ). data ( ), lrint ( x1 ), lrint ( y1 )));
  
  static int process_rate = 0;

  if (process_rate++ > 10) {
    process_rate = 0;
    processQtEvents();
  }
}

//ARM_SPEEDUP, SPEEDUP
void QOutputDev::drawImageMask ( GfxState *state, Object */*ref*/, Stream *str, int width, int height, GBool invert, GBool inlineImg ) 
{
  int xscale = 1;
  int yscale = 1;
  int sheight = height;
  int swidth = width;
  
  // get CTM, check for singular matrix
  fp_t *ctm = state-> getCTM ( );


  //  printf("x=%f y=%f w=%f h=%f\n", (float)ctm[4], (float)ctm[5],
  //	 (float)ctm[0], (float)ctm[3]);


  if ( fabs ( ctm [0] * ctm [3] - ctm [1] * ctm [2] ) < 0.000001 ) {
    printf("singular CTM is drawImageMask\n");
    if ( inlineImg ) {
      str-> reset ( );
      int j = height * (( width + 7 ) / 8 );
      for ( int i = 0; i < j; i++ )
	str->getChar();
      
      str->close();
    }
    return;
  }

  int ox = lrint ( ctm [4] );
  int oy = lrint ( ctm [5] );
  
  int ow = lrint ( ctm [0] );
  int oh = lrint ( ctm [3] );
		
  if (oh == 0 || ow == 0) {
    //    printf("drawImageMask oh=%d ow=%d\n", oh, ow);
    return;
  }

  if ( ow < 0 ) {
    ox += ow;
    ow = -ow;
  }
  if ( oh < 0 ) {
    oy += oh;
    oh = -oh;
  }

  // prescale by leaving out lines and pixels to give it a chance to fit in memory
  sheight = height;
  swidth = width;
  if (width > 130 || height > 130) {
    // maybe should be 2.5 or 3, given 2 is just at nyquist limit
    // todo => check available memory, then use 3 if possible
    const int factor = 2;
    if (width >= ow*factor) {
      xscale = width/(ow*factor);
      swidth = width/xscale;
    }
    if (height >= oh*factor) {
      yscale = height/(oh*factor);
      sheight = height/yscale;
    }
    
    // check to see if the image is still too large for limited memory on the Zaurus
    
#if 0
    printf("%d * %d * 8 + %d * %d * 4 = %d < %d - %d (kmem %d)\n",
	   sheight, swidth, ow, oh, 
	   sheight * swidth * 8 + ow * oh * 4, 
	   m_free_memory, free_memory_reserve, getKernelFreeMemory() );

#endif

    const int image_mem_multiplier = 4;
    const int image_mem_constant = 10000;
    const int image_mem_cutoff = 60000;

    int mem = (sheight * swidth + ow * oh ) * image_mem_multiplier 
      + image_mem_constant;
	   
    if (mem > image_mem_cutoff || mem > m_free_memory) {
      int freemem = getKernelFreeMemory();
      if (freemem == 0) 
	freemem = 200000;
      //      while(mem > freemem && (sheight > oh || swidth > ow)) {
      while(mem > freemem && (sheight > oh && swidth > ow)) {

	//    while (mem > 700000
	//	   || (mem > 200000 && (swidth > ow || sheight > oh))) {
	
	//    while (sheight * swidth > 200000
	//	   && (swidth > ow || sheight > oh)) {

	// maybe should use ratios and reduce dim with largest ratio
	if (sheight - oh > swidth - ow) {
	  yscale++;
	  sheight = height/yscale;
	} else {
	  xscale++;
	  swidth = width/xscale;
	}
	mem = (sheight * swidth + ow * oh) * image_mem_multiplier + image_mem_constant;

	printf("mem=(%d*%d + %d*%d)*%d + %d = %d > freemem=%d\n", 
	       sheight, swidth, ow, oh, 
	       image_mem_multiplier, image_mem_constant,
	       mem, freemem);
      }
    }

    printf("mask input %d x %d output %d x %d scale %d x %d = %d x %d\n",
	   width,height,ow,oh,xscale,yscale,swidth,sheight);
  }

  QQImage img ( swidth, sheight, 32 );	
	
  if (img.isNull()) {
    printf("can not allocate QQImage(%d,%d)\n", swidth, sheight);
    if ( inlineImg ) {
      str-> reset ( );
      int j = height * (( width + 7 ) / 8 );
      for ( int i = 0; i < j; i++ )
	str->getChar();
      
      str->close();
    }
    return;
  }


  GfxRGB rgb;	
  state-> getFillRGB ( &rgb );
  //  uint val = ( lrint ( rgb. r * 255 ) & 0xff ) << 16 
  //    | ( lrint ( rgb. g * 255 ) & 0xff ) << 8 | ( lrint ( rgb. b * 255 ) & 0xff );
  uint val = (getpix(rgb.r)<<16) | (getpix(rgb.b)<<8) | getpix(rgb.b);
  
  img. setAlphaBuffer ( true );
  
  QPDFDBG( printf ( "IMAGE MASK (%dx%d)\n", width, height ));
  
  // initialize the image stream
  ImageStream *imgStr = new ImageStream ( str, width, 1, 1 );
  imgStr-> reset ( );
  
  uchar **scanlines = img. jumpTable ( );
  
  if ( ctm [3] > 0 )
    scanlines += ( sheight - 1 );
  
  int realLines = 0;
  for ( int y = 0; y < sheight; y++) {
    QRgb *scanline = (QRgb *) *scanlines;
    Guchar *streamLine;
    
    if ( ctm [0] < 0 )
      scanline += ( swidth - 1 );
    
    streamLine = imgStr->getLine();
    realLines++;
    
    int xx;
    for ( int x = 0, xx = 0; x < swidth; x++, xx += xscale ) {
      Guchar alpha;
      
      alpha = streamLine[xx];
      
      if ( invert )
	alpha ^= 1;
      
      *scanline = ( alpha == 0 ) ? 0xff000000 | val : val;
      
      scanline += (ctm[0] < 0) ? -1 : 1;
    }
    
    scanlines += (ctm[3] > 0) ? -1 : 1;
    for(int yskip=1; yskip < yscale; yskip++) {
      if (realLines < height) {
	imgStr->getLine();
	realLines++;
      }
    }

    processQtEvents();
  }
  
  while(realLines < height) {
    imgStr->getLine();
    realLines++;
  }
  
#ifndef QT_NO_TRANSFORMATIONS			
  QWMatrix mat ( ctm [0] / swidth, ctm [1], ctm [2], ctm [3] / sheight, ctm [4], ctm [5] );
  
  std::cerr << "MATRIX T=" << mat. dx ( ) << "/" << mat. dy ( ) << std::endl 
	    << " - M=" << mat. m11 ( ) << "/" << mat. m12 ( ) << "/" << mat. m21 ( ) << "/" << mat. m22 ( ) << std::endl;
  
  QWMatrix oldmat = m_painter-> worldMatrix ( );
  m_painter-> setWorldMatrix ( mat, true );
  
#ifdef QWS
  QPixmap pm;
  pm. convertFromImage ( img );
  m_painter-> drawPixmap ( 0, 0, pm );
#else
  m_painter-> drawImage ( QPoint ( 0, 0 ), img );
#endif
  
  m_painter-> setWorldMatrix ( oldmat );
  
#else
  if (( ctm [1] < -0.1 ) || ( ctm [1] > 0.1 ) || ( ctm [2] < -0.1 ) || ( ctm [2] > 0.1 )) {
    printf (  "### ROTATED / SHEARED / ETC -- CANNOT DISPLAY THIS IMAGE ctm[1]=%f ctm[2]=%f\n", (float)ctm[1], (float)ctm[2] );
  }
  else {
    
    QPDFDBG( printf ( "DRAWING IMAGE MASKED: %d/%d - %dx%d\n", ox, oy, ow, oh ));
    
    //		putImageCache(img);
    processQtEvents();
#if 0
    printf("before (%d,%d)=>(%d,%d) depth=%d lines=%x line1=%x alpha=%d colors=%d\n",
	   img.width(), img.height(), ow, oh,
	   img.depth(),
	   img.jumpTable(),
	   img.scanLine(0), img.hasAlphaBuffer(), img.numColors());
#endif
    img = img. smoothScale ( ow, oh );
#if 0
    printf("after (%d,%d)=>(%d,%d) depth=%d lines=%x line1=%x alpha=%d colors=%d\n",
	   img.width(), img.height(), ow, oh,
	   img.depth(),
	   img.jumpTable(),
	   img.scanLine(0), img.hasAlphaBuffer(), img.numColors());
#endif
    processQtEvents();

    m_painter-> drawImage ( ox, oy, img );
  }
  
#endif
  
  delete imgStr;			       
  processQtEvents();
  return;
}


//ARM_SPEEDUP

// should use jpeg library for jpeg streams and let it resize the
//   image more efficiently (need to give Stream object a hint)
// should also average scaled pixels rather than sampling

void QOutputDev::drawImage(GfxState *state, Object */*ref*/, Stream *str, int width, int height, GfxImageColorMap *colorMap, int *maskColors, GBool inlineImg ) 
{
  int xscale = 1;
  int yscale = 1;
  int sheight = height;
  int swidth = width;
  
  // get CTM, check for singular matrix
  fp_t *ctm = state-> getCTM ( );
  
  if ( fabs ( ctm [0] * ctm [3] - ctm [1] * ctm [2] ) < 0.000001 ) {
    printf("singular CTM is drawImage\n");
    if ( inlineImg ) {
      str-> reset ( );
      int j = height * (( width + 7 ) / 8 );
      for ( int i = 0; i < j; i++ )
	str->getChar();
      
      str->close();
    }
    return;
  }
  
  int ox = lrint ( ctm [4] );
  int oy = lrint ( ctm [5] );
  
  int ow = lrint ( ctm [0] );
  int oh = lrint ( ctm [3] );
  
  if (oh == 0 || ow == 0) {
    //    printf("drawImageMask oh=%d ow=%d\n", oh, ow);
    return;
  }
  
  if ( ow < 0 ) {
    ox += ow;
    ow = -ow;
  }
  if ( oh < 0 ) {
    oy += oh;
    oh = -oh;
  }
  
  // prescale by leaving out lines and pixels to give it a chance to fit in memory
  sheight = height;
  swidth = width;
  
  
  if (width > 130 || height > 130) {
    // maybe should be 2.5 or 3, given 2 is just at nyquist limit
    // todo => check available memory, then use 3 if possible
    const int factor = 2;
    if (width >= ow*factor) {
      xscale = width/(ow*factor);
      swidth = width/xscale;
    }
    if (height >= oh*factor) {
      yscale = height/(oh*factor);
      sheight = height/yscale;
    }
    
    // check to see if the image is still too large for limited memory on the Zaurus
    
#if 0
    printf("%d * %d * 8 + %d * %d * 4 = %d < %d - %d (kmem %d)\n",
	   sheight, swidth, ow, oh, 
	   sheight * swidth * 8 + ow * oh * 4, 
	   m_free_memory, free_memory_reserve, getKernelFreeMemory() );

#endif

    const int image_mem_multiplier = 4;
    const int image_mem_constant = 10000;
    const int image_mem_cutoff = 60000;

    int mem = (sheight * swidth + ow * oh ) * image_mem_multiplier 
      + image_mem_constant;
	   
    if (mem > image_mem_cutoff || mem > m_free_memory) {
      int freemem = getKernelFreeMemory();
      if (freemem == 0) 
	freemem = 200000;
      //      while(mem > freemem && (sheight > oh || swidth > ow)) {
      while(mem > freemem && (sheight > oh && swidth > ow)) {

	//    while (mem > 700000
	//	   || (mem > 200000 && (swidth > ow || sheight > oh))) {
	
	//    while (sheight * swidth > 200000
	//	   && (swidth > ow || sheight > oh)) {

	// maybe should use ratios and reduce dim with largest ratio
	if (sheight - oh > swidth - ow) {
	  yscale++;
	  sheight = height/yscale;
	} else {
	  xscale++;
	  swidth = width/xscale;
	}
	mem = (sheight * swidth + ow * oh) * image_mem_multiplier + image_mem_constant;

	printf("mem=(%d*%d + %d*%d)*%d + %d = %d > freemem=%d\n", 
	       sheight, swidth, ow, oh, 
	       image_mem_multiplier, image_mem_constant,
	       mem, freemem);
      }
    }
    printf("input %d x %d output %d x %d scale %d x %d = %d x %d\n",
	   width,height,ow,oh,xscale,yscale,swidth,sheight);
  }
  
  QQImage img ( swidth, sheight, 32 );	
  
  if (img.isNull()) {
    printf("can not allocate QQImage(%d,%d)\n", swidth, sheight);
    if ( inlineImg ) {
      str-> reset ( );
      int j = height * (( width + 7 ) / 8 );
      for ( int i = 0; i < j; i++ )
	str->getChar();
      
      str->close();
    }
    return;
  }
  
  int nComps, nVals, nBits;
  
  // image parameters
  nComps = colorMap->getNumPixelComps ( );
  nVals = width * nComps;
  nBits = colorMap-> getBits ( );
  
#if 0
  printf("drawImage %d x %d comps %d/%d bits %d mask %x inline %d\n",
	 width, height, nComps, gfxColorMaxComps, 
	 nBits, maskColors, inlineImg);
#endif
  
  
  if ( maskColors )
    img. setAlphaBuffer ( true );
  
  QPDFDBG( printf ( "IMAGE (%dx%d)\n", width, height ));
  
  // initialize the image stream
  ImageStream *imgStr = new ImageStream ( str, width, nComps, nBits );
  imgStr-> reset ( );
  
  uchar **scanlines = img. jumpTable ( );
  
  if ( ctm [3] > 0 )
    scanlines += ( sheight - 1 );
  
  int realLines = 0;
  for ( int y = 0; y < sheight; y++) {
    QRgb *scanline = (QRgb *) *scanlines;
    Guchar *streamLine;
    
    if ( ctm [0] < 0 )
      scanline += ( swidth - 1 );
    
    streamLine = imgStr->getLine();
    realLines++;
    
    int xinc = xscale * nComps;
    int xx = 0;
    for ( int x = 0; x < swidth; x++, xx += xinc) {
      GfxRGB rgb;
      
      colorMap-> getRGB ( &streamLine[xx], &rgb );
      
      uint val = (getpix(rgb.r) << 16) | (getpix(rgb.g) << 8) | getpix(rgb.b);
      
      if ( maskColors ) {
	for ( int k = 0; k < nComps; ++k ) {
	  if (( streamLine[xx+k] < maskColors [2 * k] ) 
	      ||  streamLine[xx+k] > maskColors [2 * k] ) { 
	    val |= 0xff000000;
	    break;
	  }
	}
      }
      
      *scanline = val;
      
      scanline += (ctm[0] < 0) ? -1 : 1;
    }
    scanlines += (ctm[3] > 0) ? -1 : 1;
    for(int yskip=1; yskip < yscale; yskip++) {
      if (realLines < height) {
	imgStr->getLine();
	realLines++;
      }
    }
    
    processQtEvents();
  }
  while(realLines < height) {
    imgStr->getLine();
    realLines++;
  }
  
#ifndef QT_NO_TRANSFORMATIONS			
  QWMatrix mat ( ctm [0] / swidth, ctm [1], ctm [2], ctm [3] / sheight, ctm [4], ctm [5] );
  
  std::cerr << "MATRIX T=" << mat. dx ( ) << "/" << mat. dy ( ) << std::endl 
	    << " - M=" << mat. m11 ( ) << "/" << mat. m12 ( ) << "/" << mat. m21 ( ) << "/" << mat. m22 ( ) << std::endl;
  
  QWMatrix oldmat = m_painter-> worldMatrix ( );
  m_painter-> setWorldMatrix ( mat, true );
  
#ifdef QWS
  QPixmap pm;
  pm. convertFromImage ( img );
  m_painter-> drawPixmap ( 0, 0, pm );
#else	
  m_painter-> drawImage ( QPoint ( 0, 0 ), img );	
#endif
  
  m_painter-> setWorldMatrix ( oldmat );
  
#else // QT_NO_TRANSFORMATIONS
  
  if (( ctm [1] < -0.1 ) || ( ctm [1] > 0.1 ) || ( ctm [2] < -0.1 ) || ( ctm [2] > 0.1 )) {
    QPDFDBG( printf ( "### ROTATED / SHEARED / ETC -- CANNOT DISPLAY THIS IMAGE ctm[1]=%f ctm[2]=%f\n", (float)ctm[1], (float)ctm[2] ));
  }
  else {
    QPDFDBG( printf ( "DRAWING IMAGE: %d/%d - %dx%d\n", ox, oy, ow, oh ));
    
    //		putImageCache(img);
    img = img. smoothScale ( ow, oh );		
    processQtEvents();
    m_painter-> drawImage ( ox, oy, img );
  }
  
#endif
  
  
  delete imgStr;
  processQtEvents();
  return;
  
}


#ifdef OLD_VERSION

//ARM_SPEEDUP, SPEEDUP
void QOutputDev::drawImageMask ( GfxState *state, Object */*ref*/, Stream *str, int width, int height, GBool invert, GBool inlineImg ) 
{
  
  // get CTM, check for singular matrix
  fp_t *ctm = state-> getCTM ( );
  
  if ( fabs ( ctm [0] * ctm [3] - ctm [1] * ctm [2] ) < 0.000001 ) {
    qWarning ( "Singular CTM in drawImage\n" );
    
    if ( inlineImg ) {
      str-> reset ( );
      int j = height * (( width + 7 ) / 8 );
      for ( int i = 0; i < j; i++ )
	str->getChar();
      
      str->close();
    }
    return;
  }
  
  GfxRGB rgb;	
  state-> getFillRGB ( &rgb );
  uint val = (getpix(rgb.r)<<16) | (getpix(rgb.b)<<8) | getpix(rgb.b);
  
  QQImage img ( width, height, 32 );	
  img. setAlphaBuffer ( true );
  
  QPDFDBG( printf ( "IMAGE MASK (%dx%d)\n", width, height ));
  
  // initialize the image stream
  ImageStream *imgStr = new ImageStream ( str, width, 1, 1 );
  imgStr-> reset ( );
  
  uchar **scanlines = img. jumpTable ( );
  
  if ( ctm [3] > 0 )
    scanlines += ( height - 1 );
  
  for ( int y = 0; y < height; y++ ) {
    QRgb *scanline = (QRgb *) *scanlines;
    
    if ( ctm [0] < 0 )
      scanline += ( width - 1 );
    
    for ( int x = 0; x < width; x++ ) {
      Guchar alpha;
      
      imgStr-> getPixel ( &alpha );
      
      if ( invert )
	alpha ^= 1;
      
      *scanline = ( alpha == 0 ) ? 0xff000000 | val : val;
      
      ctm [0] < 0 ? scanline-- : scanline++;
    }
    ctm [3] > 0 ? scanlines-- : scanlines++;
    
    processQtEvents();
  }
  
#ifndef QT_NO_TRANSFORMATIONS			
  QWMatrix mat ( ctm [0] / width, ctm [1], ctm [2], ctm [3] / height, ctm [4], ctm [5] );
  
  std::cerr << "MATRIX T=" << mat. dx ( ) << "/" << mat. dy ( ) << std::endl 
	    << " - M=" << mat. m11 ( ) << "/" << mat. m12 ( ) << "/" << mat. m21 ( ) << "/" << mat. m22 ( ) << std::endl;
  
  QWMatrix oldmat = m_painter-> worldMatrix ( );
  m_painter-> setWorldMatrix ( mat, true );
  
#ifdef QWS
  QPixmap pm;
  pm. convertFromImage ( img );
  m_painter-> drawPixmap ( 0, 0, pm );
#else
  m_painter-> drawImage ( QPoint ( 0, 0 ), img );
#endif
  
  m_painter-> setWorldMatrix ( oldmat );
  
#else
  if (( ctm [1] < -0.1 ) || ( ctm [1] > 0.1 ) || ( ctm [2] < -0.1 ) || ( ctm [2] > 0.1 )) {
    printf (  "### ROTATED / SHEARED / ETC -- CANNOT DISPLAY THIS IMAGE  ctm[1]=%f ctm[2]=%f\n" , (float)ctm[1], (float)ctm[2]);
  }
  else {
    int x = lrint ( ctm [4] );
    int y = lrint ( ctm [5] );
    
    int w = lrint ( ctm [0] );
    int h = lrint ( ctm [3] );
    
    if ( w < 0 ) {
      x += w;
      w = -w;
    }
    if ( h < 0 ) {
      y += h;
      h = -h;
    }
    
    QPDFDBG( printf ( "DRAWING IMAGE MASKED: %d/%d - %dx%d\n", x, y, w, h ));
    
    //		putImageCache(img);
    img = img. smoothScale ( w, h );
    processQtEvents();
    m_painter-> drawImage ( x, y, img );
  }
  
#endif
  
  delete imgStr;			       
  processQtEvents();
}


//ARM_SPEEDUP
void QOutputDev::drawImage(GfxState *state, Object */*ref*/, Stream *str, int width, int height, GfxImageColorMap *colorMap, int *maskColors, GBool inlineImg ) 
{
  int nComps, nVals, nBits;
  
  // image parameters
  nComps = colorMap->getNumPixelComps ( );
  nVals = width * nComps;
  nBits = colorMap-> getBits ( );
  
#if 0
  printf("drawImage %d x %d comps %d/%d bits %d mask %x inline %d\n",
	 width, height, nComps, gfxColorMaxComps, 
	 nBits, maskColors, inlineImg);
#endif
  
  // get CTM, check for singular matrix
  fp_t *ctm = state-> getCTM ( );
  
  if ( fabs ( ctm [0] * ctm [3] - ctm [1] * ctm [2] ) < 0.000001 ) {
    qWarning ( "Singular CTM in drawImage\n" );
    
    if ( inlineImg ) {
      str-> reset ( );
      int j = height * (( nVals * nBits + 7 ) / 8 );
      for ( int i = 0; i < j; i++ )
	str->getChar();
      
      str->close();
    }
    return;
  }
  
  QQImage img ( width, height, 32 );
  
  if ( maskColors )
    img. setAlphaBuffer ( true );
  
  QPDFDBG( printf ( "IMAGE (%dx%d)\n", width, height ));
  
  // initialize the image stream
  ImageStream *imgStr = new ImageStream ( str, width, nComps, nBits );
  imgStr-> reset ( );
  
  Guchar pixBuf [gfxColorMaxComps];
  GfxRGB rgb;
  
  
  uchar **scanlines = img. jumpTable ( );
  
  if ( ctm [3] > 0 )
    scanlines += ( height - 1 );
  
  for ( int y = 0; y < height; y++ ) {
    QRgb *scanline = (QRgb *) *scanlines;
    
    if ( ctm [0] < 0 )
      scanline += ( width - 1 );
    
    for ( int x = 0; x < width; x++ ) {
      imgStr-> getPixel ( pixBuf );
      colorMap-> getRGB ( pixBuf, &rgb );
      
      //			uint val = ( lrint ( rgb. r * 255 ) & 0xff ) << 16 
      //			  | ( lrint ( rgb. g * 255 ) & 0xff ) << 8 
      //			  | ( lrint ( rgb. b * 255 ) & 0xff );
      uint val = (getpix(rgb.r)<<16) | (getpix(rgb.g)<<8) | getpix(rgb.b);
      
      if ( maskColors ) {
	for ( int k = 0; k < nComps; ++k ) {
	  if (( pixBuf [k] < maskColors [2 * k] ) || ( pixBuf [k] > maskColors [2 * k] )) { 
	    val |= 0xff000000;
	    break;
	  }
	}
      }
      *scanline = val;
      
      ctm [0] < 0 ? scanline-- : scanline++;
    }
    ctm [3] > 0 ? scanlines-- : scanlines++;
    
    processQtEvents();
  }
  
  
#ifndef QT_NO_TRANSFORMATIONS			
  QWMatrix mat ( ctm [0] / width, ctm [1], ctm [2], ctm [3] / height, ctm [4], ctm [5] );
  
  std::cerr << "MATRIX T=" << mat. dx ( ) << "/" << mat. dy ( ) << std::endl 
	    << " - M=" << mat. m11 ( ) << "/" << mat. m12 ( ) << "/" << mat. m21 ( ) << "/" << mat. m22 ( ) << std::endl;
  
  QWMatrix oldmat = m_painter-> worldMatrix ( );
  m_painter-> setWorldMatrix ( mat, true );
  
#ifdef QWS
  QPixmap pm;
  pm. convertFromImage ( img );
  m_painter-> drawPixmap ( 0, 0, pm );
#else	
  m_painter-> drawImage ( QPoint ( 0, 0 ), img );	
#endif
  
  m_painter-> setWorldMatrix ( oldmat );
  
#else // QT_NO_TRANSFORMATIONS
  
  if (( ctm [1] < -0.1 ) || ( ctm [1] > 0.1 ) || ( ctm [2] < -0.1 ) || ( ctm [2] > 0.1 )) {
    QPDFDBG( printf ( "### ROTATED / SHEARED / ETC -- CANNOT DISPLAY THIS IMAGE ctm[1]=%f ctm[2]=%f\n", (float)ctm[1], (float)ctm[2] ));
  }
  else {
    int x = lrint ( ctm [4] );
    int y = lrint ( ctm [5] );
    
    int w = lrint ( ctm [0] );
    int h = lrint ( ctm [3] );
    
    if ( w < 0 ) {
      x += w;
      w = -w;
    }
    if ( h < 0 ) {
      y += h;
      h = -h;
    }
    
    QPDFDBG( printf ( "DRAWING IMAGE: %d/%d - %dx%d\n", x, y, w, h ));
    
    //    printf("image %d x %d\n", width, height);
    //		putImageCache(img);
    img = img. smoothScale ( w, h );		
    processQtEvents();
    m_painter-> drawImage ( x, y, img );
  }
  
#endif
  
  
  delete imgStr;
  processQtEvents();
}

#endif



void QOutputDev::setPixmapPtr( QPixmap **pixmap )
{
  m_pixmap_ptr = pixmap;
}

void QOutputDev::makeTextMap( QDocText *textmap )
{
  m_text->makeTextMap(textmap);
  if (m_type3_msg) {
    textmap->appendText("\n<<Some text missing due to type 3 fonts>>\n");
  }
}

void
QOutputDev::processQtEvents()
{
  if (! m_no_refresh) {
    qApp-> processEvents ( );
  }
}


QDict<QString> QOutputDev::fontCharSubstitution;


#ifdef TYPE3_CACHE


T3FontCache::T3FontCache(Ref *fontIDA, fouble m11A, fouble m12A,
			 fouble m21A, fouble m22A,
			 int glyphXA, int glyphYA, int glyphWA, int glyphHA) {
  int i;

  //  printf("T3FontCache(%x) x=%d y=%d w=%d h=%d\n", this, glyphXA, glyphYA, glyphWA, glyphHA);
  fontID = *fontIDA;
  m11 = m11A;
  m12 = m12A;
  m21 = m21A;
  m22 = m22A;
  glyphX = glyphXA;
  glyphY = glyphYA;
  glyphW = glyphWA;
  glyphH = glyphHA;
  glyphSize = glyphW * glyphH;
  cacheAssoc = 8;
  if (glyphSize <= 256) {
    cacheSets = 8;
  } else if (glyphSize <= 512) {
    cacheSets = 4;
  } else if (glyphSize <= 1024) {
    cacheSets = 2;
  } else {
    cacheSets = 1;
  }
  //  printf("T3FontCache(%x) data sets %d assoc %d size %d = %d x %d\n", this,
  //	 cacheSets, cacheAssoc, glyphSize, glyphW, glyphH);
  cacheData = (Guchar *)gmalloc(cacheSets * cacheAssoc * glyphSize);
  //  cacheDataEnd = cacheData + cacheSets * cacheAssoc * glyphSize;
  cacheTags = (T3FontCacheTag *)gmalloc(cacheSets * cacheAssoc *
					sizeof(T3FontCacheTag));
  for (i = 0; i < cacheSets * cacheAssoc; ++i) {
    cacheTags[i].mru = i & (cacheAssoc - 1);
  }
  //  printf("alloc T3FontCache(%x) cacheData = %x  cacheTags = %x\n", this, cacheData, cacheTags);

  pixmap = new QPixmap(glyphW, glyphH, 32);
  painter = new QPainter(pixmap);
  //  printf("ALLOC T3FontCache %x pixmap %x painter %x\n", this, pixmap, painter);
  painter->setPen(Qt::black);
  painter->setBrush(Qt::black);
}


T3FontCache::~T3FontCache() {
  //  printf("FREE T3FontCache %x pixmap %x painter %x\n", this, pixmap, painter);
  gfree(cacheData);
  gfree(cacheTags);
  // must be done in this order, or bad things happen
  delete painter;
  delete pixmap;
}

// SL-5x00, SL-B500, iPaq have 320x240x16 FB
#define PIXEL_PTR_LANDSCAPE(x,y) ((pixmap_base) + ((x) * (pixmap_bpl)) + (((pixmapH-1)-(y)) * (pixmap_depth)) )
// SL-A300 have 240x320x16 FB, SL-C7x0 have 480x640x16 FB
#define PIXEL_PTR_PORTRAIT(x,y)  ((pixmap_base) + ((y) * (pixmap_bpl)) + ((x) * (pixmap_depth)) )
#define PIXEL_PTR(x,y) (isFBLandScape ? (PIXEL_PTR_LANDSCAPE(x, y)) : (PIXEL_PTR_PORTRAIT(x, y)))

GBool QOutputDev::beginType3Char(GfxState *state,
				 CharCode code, Unicode *u, int uLen) {
  Ref *fontID;
  fouble *ctm, *bbox;
  GfxRGB color;
  T3FontCache *t3Font;
  T3GlyphStack *t3gs;
  fouble x1, y1, xMin, yMin, xMax, yMax, xt, yt;
  int i, j;
  GfxFont *gfxFont;

  //  printf("beginType3Char(%x)\n", state);
  //  if (needFontUpdate) {
    updateFont(state);
    //  }
  if (!(gfxFont = state->getFont())) {
    return gFalse;
  }
  fontID = gfxFont->getID();
  ctm = state->getCTM();
  //  printf("beginType3Char ctm x %f %f %f  y %f %f %f\n",
  //	 (float)ctm[0],(float)ctm[1],(float)ctm[2],(float)ctm[3],(float)ctm[4],(float)ctm[5]);
  state->transform(fouble(0.0), fouble(0.0), &xt, &yt);

  // is it the first (MRU) font in the cache?
  if (!(nT3Fonts > 0 &&
	t3FontCache[0]->matches(fontID, ctm[0], ctm[1], ctm[2], ctm[3]))) {

    // is the font elsewhere in the cache?
    for (i = 1; i < nT3Fonts; ++i) {
      if (t3FontCache[i]->matches(fontID, ctm[0], ctm[1], ctm[2], ctm[3])) {
	t3Font = t3FontCache[i];
	for (j = i; j > 0; --j) {
	  t3FontCache[j] = t3FontCache[j - 1];
	}
	t3FontCache[0] = t3Font;
	break;
      }
    }
    if (i >= nT3Fonts) {

      // create new entry in the font cache
      if (nT3Fonts == xOutT3FontCacheSize) {
	delete t3FontCache[nT3Fonts - 1];
	--nT3Fonts;
      }
      for (j = nT3Fonts; j > 0; --j) {
	t3FontCache[j] = t3FontCache[j - 1];
      }
      ++nT3Fonts;
      bbox = gfxFont->getFontBBox();
      //      printf("bb %f %f %f %f xt %f yt %f\n", (float)bbox[0], (float)bbox[1], (float)bbox[2], (float)bbox[3], (float)xt, (float)yt);
      if (bbox[0] == 0 && bbox[1] == 0 && bbox[2] == 0 && bbox[3] == 0) {
	// broken bounding box -- just take a guess
	xMin = xt - 5;
	xMax = xMin + 30;
	yMax = yt + 15;
	yMin = yMax - 45;
      } else {
	state->transform(bbox[0], bbox[1], &x1, &y1);
	xMin = xMax = x1;
	yMin = yMax = y1;
	state->transform(bbox[0], bbox[3], &x1, &y1);
	if (x1 < xMin) {
	  xMin = x1;
	} else if (x1 > xMax) {
	  xMax = x1;
	}
	if (y1 < yMin) {
	  yMin = y1;
	} else if (y1 > yMax) {
	  yMax = y1;
	}
	state->transform(bbox[2], bbox[1], &x1, &y1);
	if (x1 < xMin) {
	  xMin = x1;
	} else if (x1 > xMax) {
	  xMax = x1;
	}
	if (y1 < yMin) {
	  yMin = y1;
	} else if (y1 > yMax) {
	  yMax = y1;
	}
	state->transform(bbox[2], bbox[3], &x1, &y1);
	if (x1 < xMin) {
	  xMin = x1;
	} else if (x1 > xMax) {
	  xMax = x1;
	}
	if (y1 < yMin) {
	  yMin = y1;
	} else if (y1 > yMax) {
	  yMax = y1;
	}
      }
      //      printf("xMin %f xMax %f yMin %f yMax %f\n", (float)xMin, (float)xMax, (float)yMin, (float)yMax);
      t3FontCache[0] = new T3FontCache(fontID, ctm[0], ctm[1], ctm[2], ctm[3],
	                               (int)floor(xMin - xt),
				       (int)floor(yMin - yt),
				       (int)ceil(xMax) - (int)floor(xMin) + 3,
				       (int)ceil(yMax) - (int)floor(yMin) + 3);
    }
  }
  t3Font = t3FontCache[0];

  fouble w1, h1;

  // is the glyph in the cache?
  i = (code & (t3Font->cacheSets - 1)) * t3Font->cacheAssoc;
  for (j = 0; j < t3Font->cacheAssoc; ++j) {
    if ((t3Font->cacheTags[i+j].mru & 0x8000) &&
	t3Font->cacheTags[i+j].code == code) {
      state->getFillRGB(&color);

      state-> transformDelta ( t3Font->cacheTags[i+j].wx, t3Font->cacheTags[i+j].wy,
			       &w1, &h1 );

#if TRY_TO_GET_TYPE3_TEXT
      m_text->addChar(state, 0, 0,
		      t3Font->cacheTags[i+j].wx, t3Font->cacheTags[i+j].wy,
		      code, u, uLen, xt, yt, w1, h1, xt, w1);
#endif
      //      printf("cacheData %x set %d assoc %d size %d\n", t3Font->cacheData, i / t3Font->cacheAssoc, j, t3Font->glyphSize);
      drawType3Glyph(t3Font, &t3Font->cacheTags[i+j],
		     t3Font->cacheData + (i+j) * t3Font->glyphSize,
		     xt, yt, &color);
      return gTrue;
    }
  }

  // push a new Type 3 glyph record
  t3gs = new T3GlyphStack();
  t3gs->next = t3GlyphStack;
  t3GlyphStack = t3gs;
  t3GlyphStack->cacheable = gFalse;
  t3GlyphStack->code = code;
  t3GlyphStack->cache = t3Font;
  t3GlyphStack->cacheIdx = i;
  t3GlyphStack->x = xt;
  t3GlyphStack->y = yt;
  t3GlyphStack->u = u;
  t3GlyphStack->uLen = uLen;
  
#if 0
  // this doesn't begin to cover all the many possible screen orientations (and flips!)
  // (it should work for the hardware I know about)
  bool isFBLandScape = (qt_screen->transformOrientation() != 0);
  QPixmap *pixmap = *m_pixmap_ptr;
  unsigned char *pixmap_base = pixmap->scanLine(0);
  const int pixmap_bpl = pixmap->bytesPerLine();
  const int pixmap_depth = pixmap->depth()/8;
  const int w = pixmap->width();
  const int h = pixmap->height();
  int x0 = xoutRound(t3GlyphStack->x + t3GlyphStack->cache->glyphX);
  int y0 = xoutRound(t3GlyphStack->y + t3GlyphStack->cache->glyphY);
  int gw = t3GlyphStack->cache->glyphW;
  int gh = t3GlyphStack->cache->glyphH;
  for (int y = 0; y < gh; ++y) {
    for (int x = 0; x < gw; ++x) {
      if (x0+x < 0 || y0 + y < 0) {
	continue;
      }
      *(unsigned long*)PIXEL_PTR(x0+x,y0+y) = 0xffffffff;
    }
  }
#endif

  return gFalse;
}



void QOutputDev::endType3Char(GfxState *state) {
  Guchar *p;
  int x, y;
  Gulong pixel;
  fouble alpha;
  T3GlyphStack *t3gs;
  fouble *ctm;


  //  printf("endType3Char(%x)\n", state);
  if (t3GlyphStack->cacheable) {

    //    image = t3GlyphStack->cache->image;
    //    XGetSubImage(display, pixmap, 0, 0,
    //		 t3GlyphStack->cache->glyphW, t3GlyphStack->cache->glyphH,
    //		 (1 << depth) - 1, ZPixmap, image, 0, 0);
    p = t3GlyphStack->cacheData;
    
    //    int x0 = xoutRound(t3GlyphStack->x + t3GlyphStack->cache->glyphX);
    //    int y0 = xoutRound(t3GlyphStack->y + t3GlyphStack->cache->glyphY);
    //    ctm = state->getCTM();
    //    int x0 = t3GlyphStack->cache->glyphX + ctm[4];
    //    int y0 = t3GlyphStack->cache->glyphY + ctm[5];
    int gw = t3GlyphStack->cache->glyphW;
    int gh = t3GlyphStack->cache->glyphH;

    //    printf("grab glyph cacheData(%x) x=%f %f %f y=%f %f  %f  w=%d h=%d\n", p,
    //    	   (float)t3GlyphStack->x, (float)t3GlyphStack->cache->glyphX, (float)ctm[4],
    //    	   (float)t3GlyphStack->y, (float)t3GlyphStack->cache->glyphY, (float)ctm[5],
    //    	   w, h);

    // this doesn't begin to cover all the many possible screen orientations (and flips!)
    // (it should work for the hardware I know about)
    bool isFBLandScape = (qt_screen->transformOrientation() != 0);
    QPixmap *pixmap = *m_pixmap_ptr;
    unsigned char *pixmap_base = pixmap->scanLine(0);
    const int pixmap_bpl = pixmap->bytesPerLine();
    const int pixmap_depth = pixmap->depth()/8;
    const int w = pixmap->width();
    const int h = pixmap->height();
    
    unsigned char *pix_ptr = pixmap_base;

    if (isFBLandScape) {
      unsigned char *pp = pixmap_base + pixmap_bpl;
      for (y = 0; y < gh; ++y) {
	pp -= 4;
	pix_ptr = pp;
	for (x = 0; x < gw; ++x) {
	  pixel = (*pix_ptr & 0xff);
	  pix_ptr += pixmap_bpl;
	  if (pixel <= (int)(0xff * 0.2)) {
	    *p++ = 4;
	  } else if (pixel <= (int)(0xff * 0.4)) {
	    *p++ = 3;
	  } else if (pixel <= (int)(0xff * 0.6)) {
	    *p++ = 2;
	  } else if (pixel <= (int)(0xff * 0.8)) {
	    *p++ = 1;
	  } else {
	    *p++ = 0;
	  }
	}
      }
    } else {
      for (y = 0; y < gh; ++y) {
	for (x = 0; x < gw; ++x) {
	  pixel = (*pix_ptr & 0xff);
	  pix_ptr += 4;
	  if (pixel <= (int)(0xff * 0.2)) {
	    *p++ = 4;
	  } else if (pixel <= (int)(0xff * 0.4)) {
	    *p++ = 3;
	  } else if (pixel <= (int)(0xff * 0.6)) {
	    *p++ = 2;
	  } else if (pixel <= (int)(0xff * 0.8)) {
	    *p++ = 1;
	  } else {
	    *p++ = 0;
	  }
	}
      }
    }

#if 0
    for (y = 0; y < gh; ++y) {
      for (x = 0; x < gw; ++x) {
	pixel = *(unsigned short*)PIXEL_PTR(x,y);   
	if (pixmap_depth == 2) {
	  alpha = ((fouble)(pixel >> 11)) / ((fouble)(0x1f));
	} else {
	  alpha = ((fouble)(pixel >> 8)) / ((fouble)(0xff));
	}
	//	printf("x=%d y=%d pixel=%x  alpha=%f\n", x0+x,y0+y, pixel, (float)alpha);
	//	  alpha = (fouble)((pixel >> rShift) & rMul) / (fouble)rMul;
	if (alpha <= 0.2) {
	  *p++ = 4;
	} else if (alpha <= 0.4) {
	  *p++ = 3;
	} else if (alpha <= 0.6) {
	  *p++ = 2;
	} else if (alpha <= 0.8) {
	  *p++ = 1;
	} else {
	  *p++ = 0;
	}
      }
    }
#endif

    //    pixmapW = t3GlyphStack->origPixmapW;
    //    pixmapH = t3GlyphStack->origPixmapH;

    *m_pixmap_ptr = t3GlyphStack->origPixmap;   // IS THIS NEEDED? (YES!)
    m_no_refresh = false;
    //    printf("reset painter %x isActive %d to origPainter %x isActive %d\n", m_painter, m_painter->isActive(),
    //	   t3GlyphStack->origPainter, t3GlyphStack->origPainter->isActive());
    m_painter = t3GlyphStack->origPainter;

    //    strokeGC = t3GlyphStack->origStrokeGC;
    //    fillGC = t3GlyphStack->origFillGC;
    //    clipRegion = t3GlyphStack->origClipRegion;

    //    if (p > t3GlyphStack->cache->cacheDataEnd) {
    //      printf("** out of range in cache %x p=%x\n", t3GlyphStack->cache->cacheDataEnd, p);
    //    }

    drawType3Glyph(t3GlyphStack->cache,
		   t3GlyphStack->cacheTag, t3GlyphStack->cacheData,
		   t3GlyphStack->x, t3GlyphStack->y, &t3GlyphStack->color);
    // the CTM must be restored here in order for TextPage::addChar to
    // work correctly
    ctm = state->getCTM();
    state->setCTM(ctm[0], ctm[1], ctm[2], ctm[3],
		  t3GlyphStack->origCTM4, t3GlyphStack->origCTM5);
  }

  fouble x1, y1, w1, h1;
  state-> transform      ( fouble(0.0), fouble(0.0),  &x1,  &y1 );
  state-> transformDelta ( t3GlyphStack->wx, t3GlyphStack->wy,
			   &w1, &h1 );

  // this usually results in garbage in the textmap for all the type3 PDF's I've looked at!
  // should fix it to either not bother with m_text (remove this call)
  // or count up non-printing characters and at some limit, replace text with "sorry" message
#if TRY_TO_GET_TYPE3_TEXT
  m_text->addChar(state, 0, 0, 
		  t3GlyphStack->wx, t3GlyphStack->wy,
		  t3GlyphStack->code, t3GlyphStack->u, t3GlyphStack->uLen,
		  x1, y1, w1, h1, x1, w1);
#endif

  t3gs = t3GlyphStack;
  t3GlyphStack = t3gs->next;
  delete t3gs;
}

void QOutputDev::drawType3Glyph(T3FontCache *t3Font,
				T3FontCacheTag *tag, Guchar *data,
				fouble x, fouble y, GfxRGB *color) {
  // this doesn't begin to cover all the many possible screen orientations (and flips!)
  // (it should work for the hardware I know about)
  bool isFBLandScape = (qt_screen->transformOrientation() != 0);
  QPixmap *pixmap = *m_pixmap_ptr;
  unsigned char *pixmap_base = pixmap->scanLine(0);
  const int pixmap_bpl = pixmap->bytesPerLine();
  const int pixmap_depth = pixmap->depth()/8;
  const int pixmapW = pixmap->width();
  const int pixmapH = pixmap->height();

  //  printf("drawType3Char(%x,%x)\n", t3Font, tag);
  //  QRgb xcolor;
  GfxRGB bg, rgb;
  Gulong map[5];
  char pixel;
  Guchar *p;
  int x0, y0, w0, h0;
  int ix, iy;

  // compute: (x0,y0) = position in destination pixmap
  //          (x1,y1) = position in the XImage
  //          (w0,h0) = size of XImage transfer
  x0 = xoutRound(x + t3Font->glyphX);
  y0 = xoutRound(y + t3Font->glyphY);
  w0 = t3Font->glyphW;
  h0 = t3Font->glyphH;
  // this is not clipping, but in the rare case
  // that a letter is partly outside the page, just nudge it
  // back on the page.  a small price to pay for a faster inner loop.
  if (x0 < 0) {
    x0 = 0;
  }
  if (x0 + w0 > pixmapW) {
    x0 = pixmapW - w0;
    if (x0 < 0) {
      return;
    }
  }
  if (y0 < 0) {
    y0 = 0;
  }
  if (y0 + h0 > pixmapH) {
    y0 = pixmapH - h0;
    if (y0 < 0) {
      return;
    }
  }

  //  XGetSubImage(display, pixmap, x0, y0, w0, h0,
  //	       (1 << depth) - 1, ZPixmap, image, x1, y1);
  //  xcolor = pixmap->pixel(x0 + t3Font->glyphW / 2, y0 + t3Font->glyphH / 2);

  int r, g, b;
  int dr, dg, db;
  // instead of interpolating, for efficiency, a delta step
  // is added repeatedly.  this does have rounding issues.
  if (pixmap_depth == 2) {
    int p565 = *(unsigned short*)PIXEL_PTR( (x0 + w0/2), (y0 + h0/2));
    r = p565 & 0x1f;
    dr = ((int)(color->r * 0x1f) - r)/4;
    p565 >>= 5;
    g = p565 & 0x3f;
    dg = ((int)(color->g * 0x3f) - g)/4;
    p565 >>= 6;
    b = p565 & 0x1f;
    db = ((int)(color->b * 0x1f) - b)/4;
    r += dr; g += dg; b += db;
    map[1] = (((r << 6) |  g) << 5) | b;
    r += dr; g += dg; b += db;
    map[2] = (((r << 6) |  g) << 5) | b;
    r += dr; g += dg; b += db;
    map[3] = (((r << 6) |  g) << 5) | b;
    r += dr; g += dg; b += db;
    map[4] = (((r << 6) |  g) << 5) | b;
  } else {
    unsigned char *pp = PIXEL_PTR(x0 + w0/2, y0 + h0/2);
    r = *pp++;
    dr = ((int)(color->r * 0xff) - r)/4;
    g = *pp++;
    dg = ((int)(color->g * 0xff) - g)/4;
    b = *pp++;
    db = ((int)(color->b * 0xff) - b)/4;
    r += dr; g += dg; b += db;
    map[1] = (((r << 8) |  g) << 8) | b;
    r += dr; g += dg; b += db;
    map[2] = (((r << 8) |  g) << 8) | b;
    r += dr; g += dg; b += db;
    map[3] = (((r << 8) |  g) << 8) | b;
    r += dr; g += dg; b += db;
    map[4] = (((r << 8) |  g) << 8) | b;
  }
    

#if 0
  if (pixmap_depth==2){
    int p565 = *(unsigned short*)PIXEL_PTR( (x0 + w0/2), (y0 + h0/2));
    bg.r = (fouble)(p565 & 0x1f)/(fouble)(0x1f);
    p565 >>= 5;
    bg.g = (fouble)(p565 & 0x3f)/(fouble)(0x3f);
    p565 >>= 6;
    bg.b = (fouble)(p565 & 0x1f)/(fouble)(0x1f);
  } else {
    unsigned char *pp = PIXEL_PTR(x0 + w0/2, y0 + h0/2);
    bg.r = (fouble)(*pp++)/(fouble)0xff; 
    bg.g = (fouble)(*pp++)/(fouble)0xff; 
    bg.b = (fouble)(*pp++)/(fouble)0xff;
  }

#define ENCODE_COLOR(x) ((pixmap_depth==2)?( ( ( ((int)((x.r)*(fouble)0x1f) << 6) | \
 ((int)((x.g)*(fouble)0x3f)) ) << 5 ) | ((int)((x.b)*(fouble)0x1f)) ) \
 : ( ( ( ((int)((x.r)*(fouble)0xff) << 8) | \
 ((int)((x.g)*(fouble)0xff)) ) << 8 ) | ((int)((x.b)*(fouble)0xff)) ) )

  rgb.r = 0.25 * (color->r + 3 * bg.r);
  rgb.g = 0.25 * (color->g + 3 * bg.g);
  rgb.b = 0.25 * (color->b + 3 * bg.b);
  map[1] = ENCODE_COLOR(rgb);
  //  printf("map[1] = %x  rgb = %f %f %f\n", map[1], (float)rgb.r, (float)rgb.g, (float)rgb.b);
  rgb.r = 0.5 * (color->r + bg.r);
  rgb.g = 0.5 * (color->g + bg.g);
  rgb.b = 0.5 * (color->b + bg.b);
  //  printf("map[2] = %x  rgb = %f %f %f\n", map[2], (float)rgb.r, (float)rgb.g, (float)rgb.b);
  map[2] = ENCODE_COLOR(rgb);
  rgb.r = 0.25 * (3 * color->r + bg.r);
  rgb.g = 0.25 * (3 * color->g + bg.g);
  rgb.b = 0.25 * (3 * color->b + bg.b);
  //  printf("map[3] = %x  rgb = %f %f %f\n", map[3], (float)rgb.r, (float)rgb.g, (float)rgb.b);
  map[3] = ENCODE_COLOR(rgb);
  map[4] = ENCODE_COLOR((*color));
#endif

  //  printf("color %f %f %f  bg %f %f %f  map %x %x %x %x\n",
  //	 (float)color->r, (float)color->g, (float)color->b, 
  //	 (float)bg.r, (float)bg.g, (float)bg.b, 
  //	 map[1], map[2], map[3], map[4]);
  p = data;


  if (pixmap_depth==2){
    int dx, dy;
    unsigned short *pp;
    if (isFBLandScape) {
      pp = (unsigned short*) PIXEL_PTR_LANDSCAPE(x0, y0);
      dx = (pixmap_bpl/2);
      dy = -w0 * dx - 1;
    } else {
      pp = (unsigned short*) PIXEL_PTR_PORTRAIT(x0, y0);
      dx = 1;
      dy = pixmap_bpl/2 - w0;
    }
    for (int yy = 0; yy < h0; ++yy) {
      for (int xx = 0; xx < w0; ++xx) {
	pixel = *p++;
	if (pixel > 0) {
	  *pp = (unsigned short)map[pixel];
	}
	pp += dx;
      }
      pp += dy;
    }
  } else {//32bpp
    const int dx = (isFBLandScape ? (pixmap_bpl/pixmap_depth) : 1);
    const int dy = (isFBLandScape ? (- w0 * dx -1) : (pixmap_bpl/pixmap_depth - w0));
    unsigned long* pp = (unsigned long*)PIXEL_PTR(x0, y0);
    for (int yy = 0; yy < h0; ++yy) {
      for (int xx = 0; xx < w0; ++xx) {
	pixel = *p++;
	//	printf("pixel=%d map=%x\n", pixel, map[pixel]);
	if (pixel > 0) {
	    *pp = (unsigned long)map[pixel]  ;
	}
	pp += dx;
      }
      pp += dy;
    }
  }

    //  for (iy = 0; iy < t3Font->glyphH; ++iy) {
    //    for (ix = 0; ix < t3Font->glyphW; ++ix) {
    //      pixel = *p++;
    //      if (pixel > 0) {
    //	XPutPixel(image, ix, iy, map[pixel]);
    //      }
    //    }
  //  XPutImage(display, pixmap, fillGC, image, x1, y1, x0, y0, w0, h0);
}

void QOutputDev::type3D0(GfxState *state, fouble wx, fouble wy) {
  //  printf("type3D0(%x)\n", state);
  t3GlyphStack->wx = wx;
  t3GlyphStack->wy = wy;
}

void QOutputDev::type3D1(GfxState *state, fouble wx, fouble wy,
			 fouble llx, fouble lly, fouble urx, fouble ury) {
  GfxColor fgColor;
  QRect rect;
  fouble *ctm;
  T3FontCache *t3Font;
  fouble xt, yt, xMin, xMax, yMin, yMax, x1, y1;
  int i, j;

  //  printf("type3D1(%x)\n", state);

  t3Font = t3GlyphStack->cache;
  t3GlyphStack->wx = wx;
  t3GlyphStack->wy = wy;

  // check for a valid bbox
  state->transform(0, 0, &xt, &yt);
  state->transform(llx, lly, &x1, &y1);
  xMin = xMax = x1;
  yMin = yMax = y1;
  state->transform(llx, ury, &x1, &y1);
  if (x1 < xMin) {
    xMin = x1;
  } else if (x1 > xMax) {
    xMax = x1;
  }
  if (y1 < yMin) {
    yMin = y1;
  } else if (y1 > yMax) {
    yMax = y1;
  }
  state->transform(urx, lly, &x1, &y1);
  if (x1 < xMin) {
    xMin = x1;
  } else if (x1 > xMax) {
    xMax = x1;
  }
  if (y1 < yMin) {
    yMin = y1;
  } else if (y1 > yMax) {
    yMax = y1;
  }
  state->transform(urx, ury, &x1, &y1);
  if (x1 < xMin) {
    xMin = x1;
  } else if (x1 > xMax) {
    xMax = x1;
  }
  if (y1 < yMin) {
    yMin = y1;
  } else if (y1 > yMax) {
    yMax = y1;
  }
  if (xMin - xt < t3Font->glyphX ||
      yMin - yt < t3Font->glyphY ||
      xMax - xt > t3Font->glyphX + t3Font->glyphW ||
      yMax - yt > t3Font->glyphY + t3Font->glyphH) {
    error(-1, "Bad bounding box in Type 3 glyph");
    return;
  }

  // allocate a cache entry
  t3GlyphStack->cacheable = gTrue;
  i = t3GlyphStack->cacheIdx;
  for (j = 0; j < t3Font->cacheAssoc; ++j) {
    if ((t3Font->cacheTags[i+j].mru & 0x7fff) == t3Font->cacheAssoc - 1) {
      t3Font->cacheTags[i+j].mru = 0x8000;
      t3Font->cacheTags[i+j].code = t3GlyphStack->code;
      t3GlyphStack->cacheTag = &t3Font->cacheTags[i+j];
      t3GlyphStack->cacheData = t3Font->cacheData + (i+j) * t3Font->glyphSize;
    } else {
      ++t3Font->cacheTags[i+j].mru;
    }
  }
  t3GlyphStack->cacheTag->wx = wx;
  t3GlyphStack->cacheTag->wy = wy;

  // prepare to rasterize the glyph
  //~ do we need to handle both fill and stroke color?
  // prepare to rasterize the glyph
  //~ do we need to handle both fill and stroke color?
  state->getFillRGB(&t3GlyphStack->color);
  //  if (reverseVideo) {
  //    t3GlyphStack->color.r = 1 - t3GlyphStack->color.r;
  //    t3GlyphStack->color.g = 1 - t3GlyphStack->color.g;
  //    t3GlyphStack->color.b = 1 - t3GlyphStack->color.b;
  //  }
  //  fgColor.c[0] = reverseVideo ? 1 : 0;
  fgColor.c[0] = 0;
  state->setFillColorSpace(new GfxDeviceGrayColorSpace());
  state->setFillColor(&fgColor);
  state->setStrokeColorSpace(new GfxDeviceGrayColorSpace());
  state->setStrokeColor(&fgColor);

  // IS THIS NEEDED? maybe only need to change painter?
  processQtEvents();
  t3GlyphStack->origPixmap = *m_pixmap_ptr;
  m_no_refresh = true;

  //  printf("save painter %x isActive %d into origPainter %x, cache painter %x isActive %d\n", m_painter, m_painter->isActive(),
  //  	 t3GlyphStack->origPainter, t3GlyphStack->cache->painter, t3GlyphStack->cache->painter->isActive());
  t3GlyphStack->origPainter = m_painter;
  *m_pixmap_ptr = t3GlyphStack->cache->pixmap;
  m_painter = t3GlyphStack->cache->painter;
  (*m_pixmap_ptr)->fill(Qt::white);

  //  t3GlyphStack->origPixmapW = pixmapW;
  //  t3GlyphStack->origPixmapH = pixmapH;
  //  t3GlyphStack->origPixmap = pixmap;
  //  t3GlyphStack->origStrokeGC = strokeGC;
  //  t3GlyphStack->origFillGC = fillGC;
  //  t3GlyphStack->origClipRegion = clipRegion;
  //  pixmapW = t3GlyphStack->cache->glyphW;
  //  pixmapH = t3GlyphStack->cache->glyphH;
  //  pixmap = t3GlyphStack->cache->pixmap;
  //  gcValues.foreground = BlackPixel(display, screenNum);
  //  gcValues.background = WhitePixel(display, screenNum);
  //  gcValues.line_width = 0;
  //  gcValues.line_style = LineSolid;
  //  strokeGC = XCreateGC(display, pixmap,
  //		       GCForeground | GCBackground | GCLineWidth | GCLineStyle,
  //		       &gcValues);
  updateLineAttrs(state, gTrue);
  //  gcValues.foreground = WhitePixel(display, screenNum);
  //  fillGC = XCreateGC(display, pixmap,
  //		     GCForeground | GCBackground | GCLineWidth | GCLineStyle,
  //		     &gcValues);
  //  XFillRectangle(display, pixmap, fillGC, 0, 0, pixmapW, pixmapH);
  //  XSetForeground(display, fillGC, BlackPixel(display, screenNum));
  //  clipRegion = XCreateRegion();
  //  rect.x = rect.y = 0;
  //  rect.width = pixmapW;
  //  rect.height = pixmapH;
  //  XUnionRectWithRegion(&rect, clipRegion, clipRegion);
  //  XSetRegion(display, strokeGC, clipRegion);
  //  XSetRegion(display, fillGC, clipRegion);
  ctm = state->getCTM();
  t3GlyphStack->origCTM4 = ctm[4];
  t3GlyphStack->origCTM5 = ctm[5];
  //  printf("setCTM %f %f => %f %f\n", (float)ctm[4], (float)ctm[5],
  //		(float)t3GlyphStack->cache->glyphX, (float)t3GlyphStack->cache->glyphY);
  state->setCTM(ctm[0], ctm[1], ctm[2], ctm[3], 
		-t3GlyphStack->cache->glyphX, -t3GlyphStack->cache->glyphY);
}
#endif


#ifdef RENDER_EMBEDDED_SYMBOL_FONTS

extern GBool openTempFile ( GString **name, FILE **f, char *mode, char *ext );

#if HAVE_T1LIB_H
#include "T1Font.h"
#endif
#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
#include "FTFont.h"
#endif
#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
#include "TTFont.h"
#endif

//------------------------------------------------------------------------
// QOutputFont
//------------------------------------------------------------------------

QOutputFont::QOutputFont(Ref *idA, fouble m11OrigA, fouble m12OrigA,
			 fouble m21OrigA, fouble m22OrigA,
			 fouble m11A, fouble m12A, fouble m21A, fouble m22A,
			 QOutputDev *xOutA) {
  id = *idA;
  xOut = xOutA;
  m11Orig = m11OrigA;
  m12Orig = m12OrigA;
  m21Orig = m21OrigA;
  m22Orig = m22OrigA;
  m11 = m11A;
  m12 = m12A;
  m21 = m21A;
  m22 = m22A;
}

QOutputFont::~QOutputFont() {
}

void QOutputFont::getCharPath(GfxState *state,
			      CharCode c, Unicode *u, int ulen) {
}

#if HAVE_T1LIB_H
//------------------------------------------------------------------------
// QOutputT1Font
//------------------------------------------------------------------------

QOutputT1Font::QOutputT1Font(Ref *idA, T1FontFile *fontFileA,
			     fouble m11OrigA, fouble m12OrigA,
			     fouble m21OrigA, fouble m22OrigA,
			     fouble m11A, fouble m12A,
			     fouble m21A, fouble m22A,
			     QOutputDev *xOutA):
  QOutputFont(idA, m11OrigA, m12OrigA, m21OrigA, m22OrigA,
	      m11A, m12A, m21A, m22A, xOutA)
{
  fouble matrix[4];

  fontFile = fontFileA;

  // create the transformed instance
  matrix[0] = m11;
  matrix[1] = -m12;
  matrix[2] = m21;
  matrix[3] = -m22;
  font = new T1Font(fontFile, matrix);
}

QOutputT1Font::~QOutputT1Font() {
  if (font) {
    delete font;
  }
}

GBool QOutputT1Font::isOk() {
  return font != NULL;
}

void QOutputT1Font::updateGC(QPainter *gc) {
}

void QOutputT1Font::drawChar(GfxState *state, QPixmap *pixmap,
			     GfxRGB *rgb,
			     fouble x, fouble y, fouble dx, fouble dy,
			     CharCode c, Unicode *u, int uLen) {
  font->drawChar(pixmap, xoutRound(x), xoutRound(y),
		 (int)(rgb->r * 65535), (int)(rgb->g * 65535),
		 (int)(rgb->b * 65535), c, u[0]);
}

void QOutputT1Font::getCharPath(GfxState *state,
				CharCode c, Unicode *u, int uLen) {
  font->getCharPath(c, u[0], state);
}
#endif // HAVE_T1LIB_H

#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
//------------------------------------------------------------------------
// QOutputFTFont
//------------------------------------------------------------------------

QOutputFTFont::QOutputFTFont(Ref *idA, FTFontFile *fontFileA,
			     fouble m11OrigA, fouble m12OrigA,
			     fouble m21OrigA, fouble m22OrigA,
			     fouble m11A, fouble m12A,
			     fouble m21A, fouble m22A,
			     QOutputDev *xOutA):
  QOutputFont(idA, m11OrigA, m12OrigA, m21OrigA, m22OrigA,
	      m11A, m12A, m21A, m22A, xOutA)
{
  fouble matrix[4];

  fontFile = fontFileA;

  // create the transformed instance
  matrix[0] = m11;
  matrix[1] = -m12;
  matrix[2] = m21;
  matrix[3] = -m22;
  font = new FTFont(fontFile, matrix);
}

QOutputFTFont::~QOutputFTFont() {
  if (font) {
    delete font;
  }
}

GBool QOutputFTFont::isOk() {
  return font != NULL;
}

void QOutputFTFont::updateGC(QPainter* gc) {
}

void QOutputFTFont::drawChar(GfxState *state,
			     QPixmap *pixmap,
			     GfxRGB *rgb,
			     fouble x, fouble y, fouble dx, fouble dy,
			     CharCode c, Unicode *u, int uLen) {
  //??  font->drawChar(pixmap, w, h, gc, xoutRound(x), xoutRound(y),
  //  printf("font=%x\n", font);
  font->drawChar(pixmap, xoutRound(x), xoutRound(y),
		 (int)(rgb->r * 65535), (int)(rgb->g * 65535),
		 (int)(rgb->b * 65535), c, uLen > 0 ? u[0] : 0);
}

void QOutputFTFont::getCharPath(GfxState *state,
				CharCode c, Unicode *u, int uLen) {
  font->getCharPath(c, u[0], state);
}
#endif // FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)

#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
//------------------------------------------------------------------------
// QOutputTTFont
//------------------------------------------------------------------------

QOutputTTFont::QOutputTTFont(Ref *idA, TTFontFile *fontFileA,
			     fouble m11OrigA, fouble m12OrigA,
			     fouble m21OrigA, fouble m22OrigA,
			     fouble m11A, fouble m12A,
			     fouble m21A, fouble m22A,
			     QOutputDev *xOutA):
  QOutputFont(idA, m11OrigA, m12OrigA, m21OrigA, m22OrigA,
	      m11A, m12A, m21A, m22A, xOutA)
{
  fouble matrix[4];

  fontFile = fontFileA;

  // create the transformed instance
  matrix[0] = m11;
  matrix[1] = -m12;
  matrix[2] = m21;
  matrix[3] = -m22;
  font = new TTFont(fontFile, matrix);
}

QOutputTTFont::~QOutputTTFont() {
  if (font) {
    delete font;
  }
}

GBool QOutputTTFont::isOk() {
  return font != NULL;
}

void QOutputTTFont::updateGC(QPainter* gc) {
}

void QOutputTTFont::drawChar(GfxState *state, Pixmap pixmap, int w, int h,
			     GC gc, GfxRGB *rgb,
			     fouble x, fouble y, fouble dx, fouble dy,
			     CharCode c, Unicode *u, int uLen) {
  font->drawChar(pixmap, w, h, gc, xoutRound(x), xoutRound(y),
		 (int)(rgb->r * 65535), (int)(rgb->g * 65535),
		 (int)(rgb->b * 65535), c, u[0]);
}
#endif // !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)

//------------------------------------------------------------------------
// QOutputFontCache
//------------------------------------------------------------------------

#if HAVE_T1LIB_H
QOutputT1FontFile::~QOutputT1FontFile() {
  delete fontFile;
  if (tmpFileName) {
    unlink(tmpFileName->getCString());
    delete tmpFileName;
  }
}
#endif

#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
QOutputFTFontFile::~QOutputFTFontFile() {
  delete fontFile;
  if (tmpFileName) {
    unlink(tmpFileName->getCString());
    delete tmpFileName;
  }
}
#endif

#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
QOutputTTFontFile::~QOutputTTFontFile() {
  delete fontFile;
  if (tmpFileName) {
    unlink(tmpFileName->getCString());
    delete tmpFileName;
  }
}
#endif

QOutputFontCache::QOutputFontCache(QOutputDev *xOutA,
				   FontRastControl t1libControlA,
				   FontRastControl freetypeControlA) {
  xOut = xOutA;

#if HAVE_T1LIB_H
  t1Engine = NULL;
  t1libControl = t1libControlA;
#endif

#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
  ftEngine = NULL;
#endif
#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
  ttEngine = NULL;
#endif
#if HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H
  freetypeControl = freetypeControlA;
#endif

  clear();
}

QOutputFontCache::~QOutputFontCache() {
  delFonts();
}

void QOutputFontCache::startDoc() {
  delFonts();
  clear();

#if HAVE_T1LIB_H
  if (t1libControl != fontRastNone) {
    t1Engine = new T1FontEngine(t1libControl == fontRastAALow ||
				t1libControl == fontRastAAHigh,
				t1libControl == fontRastAAHigh);
    if (t1Engine->isOk()) {
      // this doesn't really do anything since it is all
      // hardwared for Zaurus in SFont.cc
      if (QPixmap::defaultDepth() == 24) {
	t1Engine->useTrueColor(1, 0, 1, 8, 1, 16);
      } else {
	t1Engine->useTrueColor(1, 0, 1, 5, 1, 11);
      }
      //      if (trueColor) {
      //	t1Engine->useTrueColor(rMul, rShift, gMul, gShift, bMul, bShift);
      //      t1Engine->useTrueColor(rMul, rShift, gMul, gShift, bMul, bShift);
      //      } else {
      //	t1Engine->useColorCube(colors, numColors);
      //      }
    } else {
      delete t1Engine;
      t1Engine = NULL;
    }
  }
#endif // HAVE_T1LIB_H

#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
  if (freetypeControl != fontRastNone) {
    ftEngine = new FTFontEngine(freetypeControl == fontRastAALow ||
				freetypeControl == fontRastAAHigh);
    if (ftEngine->isOk()) {
      //      ftEngine->useTrueColor(rMul, rShift, gMul, gShift, bMul, bShift);
      if (QPixmap::defaultDepth() == 24) {
	  ftEngine->useTrueColor(1, 0, 1, 8, 1, 16);
	} else {
	  ftEngine->useTrueColor(1, 0, 1, 5, 1, 11);
	}
      }
    } else {
      delete ftEngine;
      ftEngine = NULL;
    }
//  }
#endif // FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)

#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
  if (freetypeControl != fontRastNone) {
    ttEngine = new TTFontEngine(display, visual, depth, colormap,
				freetypeControl == fontRastAALow ||
				  freetypeControl == fontRastAAHigh);
    if (ttEngine->isOk()) {
      if (trueColor) {
	ttEngine->useTrueColor(rMul, rShift, gMul, gShift, bMul, bShift);
      } else {
	ttEngine->useColorCube(colors, numColors);
      }
    } else {
      delete ttEngine;
      ttEngine = NULL;
    }
  }
#endif // !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
}

void QOutputFontCache::delFonts() {
  int i;

  for (i = 0; i < nFonts; ++i) {
    delete fonts[i];
  }

#if HAVE_T1LIB_H
  // delete Type 1 font files
  deleteGList(t1FontFiles, QOutputT1FontFile);
  if (t1Engine) {
    delete t1Engine;
  }
#endif

#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
  // delete FreeType font files
  deleteGList(ftFontFiles, QOutputFTFontFile);
  if (ftEngine) {
    delete ftEngine;
  }
#endif

#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
  // delete TrueType fonts
  deleteGList(ttFontFiles, QOutputTTFontFile);
  if (ttEngine) {
    delete ttEngine;
  }
#endif
}

void QOutputFontCache::clear() {
  int i;

  for (i = 0; i < xOutFontCacheSize; ++i) {
    fonts[i] = NULL;
  }
  nFonts = 0;

#if HAVE_T1LIB_H
  // clear Type 1 font files
  t1FontFiles = new GList();
#endif

#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
  // clear FreeType font cache
  ftFontFiles = new GList();
#endif

#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
  // clear TrueType font cache
  ttFontFiles = new GList();
#endif
}

QOutputFont *QOutputFontCache::getFont(XRef *xref, GfxFont *gfxFont,
				       fouble m11, fouble m12,
				       fouble m21, fouble m22) {
  QOutputFont *font;
  DisplayFontParam *dfp;
  GString *substName;
  fouble m11New, m12New, m21New, m22New;
  fouble w1, w2, v;
  fouble *fm;
  char *name;
  int index;
  int code;
  int i, j;

  // is it the most recently used font?
  if (nFonts > 0 && fonts[0]->matches(gfxFont->getID(), m11, m12, m21, m22)) {
    if (qpdf_debug_level > 1) {
      printf("got font from cache\n");
    }
    return fonts[0];
  }

  // is it in the cache?
  for (i = 1; i < nFonts; ++i) {
    if (fonts[i]->matches(gfxFont->getID(), m11, m12, m21, m22)) {
      font = fonts[i];
      for (j = i; j > 0; --j) {
	fonts[j] = fonts[j-1];
      }
      fonts[0] = font;
      if (qpdf_debug_level > 1) {
	printf("got font from cache\n");
      }
      return font;
    }
  }

  // try for a cached FontFile, an embedded font, or an external font
  // file
  font = NULL;
  switch (gfxFont->getType()) {
  case fontType1:
  case fontType1C:
#if HAVE_T1LIB_H
    if (t1libControl != fontRastNone) {
      font = tryGetT1Font(xref, gfxFont, m11, m12, m21, m22);
      if (qpdf_debug_level > 0) {
	if (qpdf_debug_level > 1 || font) {
	  printf("%s T1Lib font\n", font? "GOT" : "didn't get");
	}
      }
    }
#endif
#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
    if (!font) {
      if (freetypeControl != fontRastNone) {
	font = tryGetFTFont(xref, gfxFont, m11, m12, m21, m22);
	if (qpdf_debug_level > 0) {
	  if (qpdf_debug_level > 1 || font) {
	    printf("%s FreeType T1 font\n", font? "GOT" : "didn't get");
	  }
	}
      }
    }
#endif
    break;
  case fontTrueType:
  case fontCIDType2:
#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
    if (freetypeControl != fontRastNone) {
      font = tryGetFTFont(xref, gfxFont, m11, m12, m21, m22);
      if (qpdf_debug_level > 0) {
	if (qpdf_debug_level > 1 || font) {
	  printf("%s FreeType TT/T2 font\n", font? "GOT" : "didn't get");
	}
      }
    }
#endif
#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
    if (freetypeControl != fontRastNone) {
      font = tryGetTTFont(xref, gfxFont, m11, m12, m21, m22);
      if (qpdf_debug_level > 0) {
	if (qpdf_debug_level > 1 || font) {
	  printf("%s FreeType TT/T2 font\n", font? "GOT" : "didn't get");
	}
      }
    }
#endif
    break;
  case fontCIDType0:
  case fontCIDType0C:
#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
    if (freetypeControl != fontRastNone) {
      font = tryGetFTFont(xref, gfxFont, m11, m12, m21, m22);
      if (qpdf_debug_level > 0) {
	if (qpdf_debug_level > 1 || font) {
	  printf("%s FreeType CIDType0 font\n", font? "GOT" : "didn't get");
	}
      }
    }
#endif
    break;
  default:
    break;
  }

  // check for error
  if (!font) {
    // This will happen if the user specifies a bogus font in the
    // config file (a non-existent font file or a font that requires a
    // rasterizer that is disabled or wasn't built in), or if a CID
    // font has no associated font in the config file.
    if (gfxFont->isCIDFont()) {
      error(-1, "Couldn't find a font for the '%s' character collection",
	    ((GfxCIDFont *)gfxFont)->getCollection()->getCString());
    } else {
      error(-1, "Couldn't find a font for '%s'",
	    gfxFont->getName() ?
	        gfxFont->getName()->getCString() : "[unnamed]");
    }
    return NULL;
  }

  // insert font in cache
  if (nFonts == xOutFontCacheSize) {
    --nFonts;
    delete fonts[nFonts];
  }
  for (j = nFonts; j > 0; --j) {
    fonts[j] = fonts[j-1];
  }
  fonts[0] = font;
  ++nFonts;

  return font;
}

#if 0
QOutputFont *QOutputFontCache::tryGetFont(XRef *xref, DisplayFontParam *dfp,
					  GfxFont *gfxFont,
					  fouble m11Orig, fouble m12Orig,
					  fouble m21Orig, fouble m22Orig,
					  fouble m11, fouble m12,
					  fouble m21, fouble m22,
					  GBool subst) {
  QOutputFont *font;

  font = NULL;

  // create the new font
  switch (dfp->kind) {

  case displayFontT1:
#if HAVE_T1LIB_H
    if (t1libControl != fontRastNone && !gfxFont->isCIDFont()) {
      font = tryGetT1FontFromFile(xref, dfp->t1.fileName, gFalse, gfxFont,
				  m11Orig, m12Orig, m21Orig, m22Orig,
				  m11, m12, m21, m22, subst);
    }
#endif
#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
    if (!font) {
      if (freetypeControl != fontRastNone) {
	font = tryGetFTFontFromFile(xref, dfp->t1.fileName, gFalse, gfxFont,
				    m11Orig, m12Orig, m21Orig, m22Orig,
				    m11, m12, m21, m22, gFalse, subst);
      }
    }
#endif
#if !((FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)) || defined(HAVE_T1LIB_H))
    error(-1, "Config file specifies a Type 1 font,");
    error(-1, "but xpdf was not built with t1lib or FreeType2 support");
#endif
    break;

  case displayFontTT:
#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
    if (freetypeControl != fontRastNone) {
      font = tryGetFTFontFromFile(xref, dfp->tt.fileName, gFalse, gfxFont,
				  m11Orig, m12Orig, m21Orig, m22Orig,
				  m11, m12, m21, m22, gFalse, subst);
    }
#endif
#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
    if (freetypeControl != fontRastNone) {
      font = tryGetTTFontFromFile(xref, dfp->tt.fileName, gFalse, gfxFont,
				  m11Orig, m12Orig, m21Orig, m22Orig,
				  m11, m12, m21, m22, subst);
    }
#endif
#if !(HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
    error(-1, "Config file specifies a TrueType font,");
    error(-1, "but xpdf was not built with FreeType support");
    dfp = NULL;
#endif
    break;
  }

  return font;
}
#endif

#if HAVE_T1LIB_H
static void outputToFile(void *stream, char *data, int len) {
  static int process_events = 0;
  if (process_events++ > 100) {
    process_events = 0;
    qApp-> processEvents ( );
  }
  fwrite(data, 1, len, (FILE *)stream);
}

#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>

QOutputFont *QOutputFontCache::tryGetT1Font(XRef *xref,
					    GfxFont *gfxFont,
					    fouble m11, fouble m12,
					    fouble m21, fouble m22) {
  Ref *id;
  QOutputT1FontFile *xFontFile;
  QOutputFont *font;
  Ref embRef;
  GString *fileName;
  FILE *f;
  char *fontBuf;
  int fontLen;
  Type1CFontFile *ff;
  Object refObj, strObj;
  int c;
  int i;

  // check the already available font files
  id = gfxFont->getID();
  for (i = 0; i < t1FontFiles->getLength(); ++i) {
    xFontFile = (QOutputT1FontFile *)t1FontFiles->get(i);
    if (xFontFile->num == id->num && xFontFile->gen == id->gen &&
	!xFontFile->subst) {
      font = new QOutputT1Font(id, xFontFile->fontFile,
			       m11, m12, m21, m22,
			       m11, m12, m21, m22, xOut);
      if (!font->isOk()) {
	delete font;
	return NULL;
      }
      return font;
    }
  }

  // check for an embedded font
  if (gfxFont->getEmbeddedFontID(&embRef)) {
    if (qpdf_debug_level > 0) {
      printf("create embedded T1 Font\n");
    }
    // create the font file
    fileName = NULL;
    if (!openTempFile(&fileName, &f, "wb", NULL)) {
      error(-1, "Couldn't create temporary Type 1 font file");
      return NULL;
    }
    if (gfxFont->getType() == fontType1C) {
      if (!(fontBuf = gfxFont->readEmbFontFile(xref, &fontLen))) {
	fclose(f);
	return NULL;
      }
      ff = new Type1CFontFile(fontBuf, fontLen);
      if (!ff->isOk()) {
	delete ff;
	gfree(fontBuf);
	return NULL;
      }
      ff->convertToType1(outputToFile, f);
      delete ff;
      gfree(fontBuf);
    } else { // fontType1
      refObj.initRef(embRef.num, embRef.gen);
      refObj.fetch(xref, &strObj);
      refObj.free();
      strObj.streamReset();
      while ((c = strObj.streamGetChar()) != EOF) {
	fputc(c, f);
      }
      strObj.streamClose();
      strObj.free();
    }
    fclose(f);

    // create the Font
    font = tryGetT1FontFromFile(xref, fileName, gTrue, gfxFont,
				m11, m12, m21, m22,
				m11, m12, m21, m22, gFalse);

    // on systems with Unix hard link semantics, this will remove the
    // last link to the temp file
#if 0
    if (strncmp(fileName->getCString(),"/dev/ram",8) == 0) {
      printf("clearing ramdisk %s\n", fileName->getCString());
      int fd = open(fileName->getCString(),O_RDWR);
      if (fd > 0) {
	int ret = ioctl(fd, _IO(0x12,97));
	printf("returned %d\n", ret);
      }
      close(fd);
    } 
#endif
    unlink(fileName->getCString());

    delete fileName;
    
  // check for an external font file
  } else if ((fileName = gfxFont->getExtFontFile())) {
    font = tryGetT1FontFromFile(xref, fileName, gFalse, gfxFont,
				m11, m12, m21, m22,
				m11, m12, m21, m22, gFalse);

  } else {
    font = NULL;
  }
  return font;
}

QOutputFont *QOutputFontCache::tryGetT1FontFromFile(XRef *xref,
						    GString *fileName,
						    GBool deleteFile,
						    GfxFont *gfxFont,
						    fouble m11Orig,
						    fouble m12Orig,
						    fouble m21Orig,
						    fouble m22Orig,
						    fouble m11, fouble m12,
						    fouble m21, fouble m22,
						    GBool subst) {
  Ref *id;
  T1FontFile *fontFile;
  QOutputFont *font;

  // create the t1lib font file
  fontFile = new T1FontFile(t1Engine, fileName->getCString(),
			    ((Gfx8BitFont *)gfxFont)->getEncoding(),
			    gfxFont->getFontBBox());
  if (!fontFile->isOk()) {
    error(-1, "Couldn't create t1lib font from '%s'",
	  fileName->getCString());
    delete fontFile;
    if (deleteFile) {
      unlink(fileName->getCString());
    }
    return NULL;
  }

  // add to list
  id = gfxFont->getID();
  t1FontFiles->append(new QOutputT1FontFile(id->num, id->gen,
					    subst, fontFile,
					    deleteFile ? fileName->copy()
					               : (GString *)NULL));

  // create the Font
  font = new QOutputT1Font(gfxFont->getID(), fontFile,
			   m11Orig, m12Orig, m21Orig, m22Orig,
			   m11, m12, m21, m22, xOut);
  if (!font->isOk()) {
    delete font;
    return NULL;
  }
  return font;
}
#endif // HAVE_T1LIB_H

#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
QOutputFont *QOutputFontCache::tryGetFTFont(XRef *xref,
					    GfxFont *gfxFont,
                                            fouble m11, fouble m12,
					    fouble m21, fouble m22) {
  Ref *id;
  QOutputFTFontFile *xFontFile;
  QOutputFont *font;
  Ref embRef;
  GString *fileName;
  FILE *f;
#if 1 //~ need this until FT can handle fonts with missing tables
  char *fontBuf;
  int fontLen;
  TrueTypeFontFile *ff;
#endif
  Object refObj, strObj;
  int c;
  int i;

  // check the already available font files
  id = gfxFont->getID();
  for (i = 0; i < ftFontFiles->getLength(); ++i) {
    xFontFile = (QOutputFTFontFile *)ftFontFiles->get(i);
    if (xFontFile->num == id->num && xFontFile->gen == id->gen &&
	!xFontFile->subst) {
      font = new QOutputFTFont(id, xFontFile->fontFile,
			       m11, m12, m21, m22,
			       m11, m12, m21, m22, xOut);
      if (qpdf_debug_level > 0) {
	printf("use existing font file %s got %x\n",
	       xFontFile->tmpFileName, font->isOk());
      }
      if (!font->isOk()) {
	delete font;
	return NULL;
      }
      return font;
    }
  }

  // check for an embedded font
  if (gfxFont->getEmbeddedFontID(&embRef)) {
    if (qpdf_debug_level > 0) {
      printf("create embedded FT font\n");
    }
    // create the font file
    fileName = NULL;
    if (!openTempFile(&fileName, &f, "wb", NULL)) {
      error(-1, "Couldn't create temporary TrueType font file");
      return NULL;
    }
#if 1 //~ need this until FT can handle fonts with missing tables
    if (gfxFont->getType() == fontTrueType ||
	gfxFont->getType() == fontCIDType2) {
      if (!(fontBuf = gfxFont->readEmbFontFile(xref, &fontLen))) {
	fclose(f);
	if (qpdf_debug_level > 0) {
	  printf("readEmbFontFile failed\n");
	}
	return NULL;
      }
      ff = new TrueTypeFontFile(fontBuf, fontLen);
      ff->writeTTF(f);
      delete ff;
      gfree(fontBuf);
    } else {
      refObj.initRef(embRef.num, embRef.gen);
      refObj.fetch(xref, &strObj);
      refObj.free();
      strObj.streamReset();
      while ((c = strObj.streamGetChar()) != EOF) {
	fputc(c, f);
      }
      strObj.streamClose();
      strObj.free();
    }
#else
    refObj.initRef(embRef.num, embRef.gen);
    refObj.fetch(xref, &strObj);
    refObj.free();
    strObj.streamReset();
    while ((c = strObj.streamGetChar()) != EOF) {
      fputc(c, f);
    }
    strObj.streamClose();
    strObj.free();
#endif
    fclose(f);

    // create the Font
    font = tryGetFTFontFromFile(xref, fileName, gTrue, gfxFont,
				m11, m12, m21, m22,
				m11, m12, m21, m22, gTrue, gFalse);

    // on systems with Unix hard link semantics, this will remove the
    // last link to the temp file
    unlink(fileName->getCString());

    delete fileName;

  // check for an external font file
  } else if ((fileName = gfxFont->getExtFontFile())) {
    if (qpdf_debug_level > 0) {
      printf("try for external file\n");
    }
    font = tryGetFTFontFromFile(xref, fileName, gFalse, gfxFont,
				m11, m12, m21, m22,
				m11, m12, m21, m22, gFalse, gFalse);

  } else {
    if (qpdf_debug_level > 1) {
      printf("no emb or ext font\n");
    }
    font = NULL;
  }

  return font;
}

QOutputFont *QOutputFontCache::tryGetFTFontFromFile(XRef *xref,
						    GString *fileName,
						    GBool deleteFile,
						    GfxFont *gfxFont,
						    fouble m11Orig,
						    fouble m12Orig,
						    fouble m21Orig,
						    fouble m22Orig,
						    fouble m11, fouble m12,
						    fouble m21, fouble m22,
						    GBool embedded,
						    GBool subst) {
  Ref *id;
  FTFontFile *fontFile;
  QOutputFont *font;

  // create the FreeType font file
  //  printf("create font from %s\n", fileName->getCString());
  if (gfxFont->isCIDFont()) {
    if (gfxFont->getType() == fontCIDType2) {
      fontFile = new FTFontFile(ftEngine, fileName->getCString(),
				((GfxCIDFont *)gfxFont)->getCIDToGID(),
				((GfxCIDFont *)gfxFont)->getCIDToGIDLen(),
				embedded);
    } else { // fontCIDType0, fontCIDType0C
      fontFile = new FTFontFile(ftEngine, fileName->getCString(), embedded);
    }
  } else {
    char *buf;
    int len;
    FILE *f;
    TrueTypeFontFile *ff;
    Gushort *codeToGID;

    if (!(f = fopen(fileName->getCString(), "rb"))) {
      return NULL;
    }
    fseek(f, 0, SEEK_END);
    len = (int)ftell(f);
    fseek(f, 0, SEEK_SET);
    buf = (char *)gmalloc(len);
    if ((int)fread(buf, 1, len, f) != len) {
      gfree(buf);
      fclose(f);
      return NULL;
    }
    fclose(f);
    ff = new TrueTypeFontFile(buf, len);
    codeToGID = ((Gfx8BitFont *)gfxFont)->getCodeToGIDMap(ff);
    fontFile = new FTFontFile(ftEngine, fileName->getCString(),
			      ((Gfx8BitFont *)gfxFont)->getEncoding(),
			      codeToGID);
    gfree(codeToGID);
    delete ff;
    gfree(buf);
  }
  if (!fontFile->isOk()) {
    error(-1, "Couldn't create FreeType font from '%s'",
	  fileName->getCString());
    delete fontFile;
    if (deleteFile) {
      unlink(fileName->getCString());
    }
    return NULL;
  }
  // add to list
  id = gfxFont->getID();
  ftFontFiles->append(new QOutputFTFontFile(id->num, id->gen,
					    subst, fontFile,
					    deleteFile ? fileName->copy()
					               : (GString *)NULL));

  // create the Font
  font = new QOutputFTFont(gfxFont->getID(), fontFile,
			   m11Orig, m12Orig, m21Orig, m22Orig,
			   m11, m12, m21, m22, xOut);
  if (!font->isOk()) {
    if (qpdf_debug_level > 0) {
      printf("failed to create font\n");
    }
    delete font;
    return NULL;
  }
  return font;
}
#endif // FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)

#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
QOutputFont *QOutputFontCache::tryGetTTFont(XRef *xref,
					    GfxFont *gfxFont,
					    fouble m11, fouble m12,
					    fouble m21, fouble m22) {
  Ref *id;
  QOutputTTFontFile *xFontFile;
  QOutputFont *font;
  Ref embRef;
  GString *fileName;
  FILE *f;
  Object refObj, strObj;
  int c;
  int i;

  // check the already available font files
  id = gfxFont->getID();
  xFontFile = NULL;
  for (i = 0; i < ttFontFiles->getLength(); ++i) {
    xFontFile = (QOutputTTFontFile *)ttFontFiles->get(i);
    if (xFontFile->num == id->num && xFontFile->gen == id->gen &&
	!xFontFile->subst) {
      font = new QOutputTTFont(id, xFontFile->fontFile,
			       m11, m12, m21, m22,
			       m11, m12, m21, m22, xOut);


      if (qpdf_debug_level > 0) {
	printf("use existing font file %s got %x\n",
	       xFontFile->tmpFileName, font->isOk());
      }
      if (!font->isOk()) {
	delete font;
	return NULL;
      }
      return font;
    }
  }

  // check for an embedded font
  if (gfxFont->getEmbeddedFontID(&embRef)) {
    if (qpdf_debug_level > 0) {
      printf("created embedded TT font\n");
    }
    // create the font file
    fileName = NULL;
    if (!openTempFile(&fileName, &f, "wb", NULL)) {
      error(-1, "Couldn't create temporary TrueType font file");
      return NULL;
    }
    refObj.initRef(embRef.num, embRef.gen);
    refObj.fetch(xref, &strObj);
    refObj.free();
    strObj.streamReset();
    while ((c = strObj.streamGetChar()) != EOF) {
      fputc(c, f);
    }
    strObj.streamClose();
    strObj.free();
    fclose(f);

    // create the Font
    font = tryGetTTFontFromFile(xref, fileName, gTrue, gfxFont,
				m11, m12, m21, m22,
				m11, m12, m21, m22, gFalse);

    // on systems with Unix hard link semantics, this will remove the
    // last link to the temp file
    unlink(fileName->getCString());

    delete fileName;

  } else if ((fileName = gfxFont->getExtFontFile())) {
    font = tryGetTTFontFromFile(xref, fileName, gFalse, gfxFont,
				m11, m12, m21, m22,
				m11, m12, m21, m22, gFalse);

  } else {
    font = NULL;
  }

  return font;
}

QOutputFont *QOutputFontCache::tryGetTTFontFromFile(XRef *xref,
						    GString *fileName,
						    GBool deleteFile,
						    GfxFont *gfxFont,
						    fouble m11Orig,
						    fouble m12Orig,
						    fouble m21Orig,
						    fouble m22Orig,
						    fouble m11, fouble m12,
						    fouble m21, fouble m22,
						    GBool subst) {
  Ref *id;
  TTFontFile *fontFile;
  QOutputFont *font;

  // create the FreeType font file
  if (gfxFont->isCIDFont()) {
    // fontCIDType2
    fontFile = new TTFontFile(ttEngine, fileName->getCString(),
			      ((GfxCIDFont *)gfxFont)->getCIDToGID(),
			      ((GfxCIDFont *)gfxFont)->getCIDToGIDLen());
  } else {
    fontFile = new TTFontFile(ttEngine, fileName->getCString(),
			      ((Gfx8BitFont *)gfxFont)->getEncoding(),
			      ((Gfx8BitFont *)gfxFont)->getHasEncoding());
  }
  if (!fontFile->isOk()) {
    error(-1, "Couldn't create FreeType font from '%s'",
	  fileName->getCString());
    delete fontFile;
    if (deleteFile) {
      unlink(fileName->getCString());
    }
    return NULL;
  }

  // add to list
  id = gfxFont->getID();
  ttFontFiles->append(new QOutputTTFontFile(id->num, id->gen,
					    subst, fontFile,
					    deleteFile ? fileName->copy()
					               : (GString *)NULL));

  // create the Font
  font = new QOutputTTFont(gfxFont->getID(), fontFile,
			   m11Orig, m12Orig, m21Orig, m22Orig,
			   m11, m12, m21, m22, xOut);
  if (!font->isOk()) {
    delete font;
    return NULL;
  }
  return font;
}
#endif // !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)

#endif
