
//========================================================================
//
// QOutputDev.h
//
// Copyright 1996 Derek B. Noonburg
// Copyright 2003 Philip Kohn
//
//========================================================================

#define TYPE3_CACHE 1

#ifndef QOUTPUTDEV_H
#define QOUTPUTDEV_H

#ifdef __GNUC__
#pragma interface
#endif


#include "aconf.h"


class Object;

#include "config.h"

#include "CharTypes.h"
#include "GlobalParams.h"
#include "OutputDev.h"
#include "xpdf/GfxState.h"  // for GfxRGB only

class GString;
class GList;
struct GfxRGB;
class GfxFont;
class GfxSubpath;
class TextPage;
class QOutputFont;
class QOutputFontCache;
class Link;
class Catalog;
class DisplayFontParam;
class UnicodeMap;
class CharCodeToUnicode;


//class QPainter;
//class QPixmap;
//class QPointArray;


typedef fouble fp_t;

//------------------------------------------------------------------------
// Constants
//------------------------------------------------------------------------

const int free_memory_reserve = 400000;

//------------------------------------------------------------------------
// Misc types
//------------------------------------------------------------------------

//------------------------------------------------------------------------
// T3FontCache
//------------------------------------------------------------------------


#ifdef TYPE3_CACHE

struct T3FontCacheTag {
  Gushort code;
  Gushort mru;			// valid bit (0x8000) and MRU index
  fouble wx, wy;		// untransformed glyph metrics
};

class T3FontCache {
public:

  T3FontCache(Ref *fontID, fouble m11A, fouble m12A,
	      fouble m21A, fouble m22A,
	      int glyphXA, int glyphYA, int glyphWA, int glyphHA);
  ~T3FontCache();
  GBool matches(Ref *idA, fouble m11A, fouble m12A,
		fouble m21A, fouble m22A)
    { return fontID.num == idA->num && fontID.gen == idA->gen &&
	     m11 == m11A && m12 == m12A && m21 == m21A && m22 == m22A; }

  Ref fontID;			// PDF font ID
  fouble m11, m12, m21, m22;	// transform matrix
  int glyphX, glyphY;		// pixel offset of glyph pixmaps
  int glyphW, glyphH;		// size of glyph pixmaps, in pixels
  int glyphSize;		// size of glyph pixmaps, in bytes
  int cacheSets;		// number of sets in cache
  int cacheAssoc;		// cache associativity (glyphs per set)
  Guchar *cacheData;		// glyph pixmap cache
  Guchar *cacheDataEnd;
  T3FontCacheTag *cacheTags;	// cache tags, i.e., char codes
  QPixmap *pixmap;
  QPainter *painter;
};

struct T3GlyphStack {
  GBool cacheable;
  Gushort code;
  T3FontCache *cache;
  int cacheIdx;
  T3FontCacheTag *cacheTag;
  Guchar *cacheData;
  fouble x, y;
  Unicode *u;
  int uLen;
  GfxRGB color;
  // int origPixmapW, origPixmapH;
  QPixmap *origPixmap;
  QPainter *origPainter;
  //  GC origStrokeGC;
  //  GC origFillGC;
  //  QRegion origClipRegion;
  fouble origCTM4, origCTM5;
  fouble wx, wy;		// untransformed glyph metrics
  T3GlyphStack *next;
};
#endif




//------------------------------------------------------------------------
// QOutputDev
//------------------------------------------------------------------------

class QOutputDev : public OutputDev, public QObject {

public:

	// Constructor.
	QOutputDev( QWidget *parent, const char *name, int flags );

	// Destructor.
	virtual ~QOutputDev();

	//---- get info about output device

	// Does this device use upside-down coordinates?
	// (Upside-down means (0,0) is the top left corner of the page.)
	virtual GBool upsideDown() { return gTrue; }

	// Does this device use drawChar() or drawString()?
	virtual GBool useDrawChar() { return gTrue; }

	// Does this device use beginType3Char/endType3Char?  Otherwise,
	// text in Type 3 fonts will be drawn with drawChar/drawString.
	virtual GBool interpretType3Chars() { return gTrue; }
       
#ifdef TYPE3_CACHE
	virtual GBool beginType3Char(GfxState *state,
				     CharCode code, Unicode *u, int uLen);

	virtual void endType3Char(GfxState *state);
	//----- Type 3 font operators
	virtual void type3D0(GfxState *state, fouble wx, fouble wy);
	virtual void type3D1(GfxState *state, fouble wx, fouble wy,
			     fouble llx, fouble lly, fouble urx, fouble ury);
#endif

	// Does this device need non-text content?
	virtual GBool needNonText() { return gFalse; }

	//----- initialization and control

	// Start a page.
	virtual void startPage(int pageNum, GfxState *state);

	// End a page.
	virtual void endPage();

	//----- link borders
	virtual void drawLink(Link *link, Catalog *catalog);

	//----- save/restore graphics state
	virtual void saveState(GfxState *state);
	virtual void restoreState(GfxState *state);

	//----- update graphics state
	virtual void updateAll(GfxState *state);
	virtual void updateCTM(GfxState *state, fp_t m11, fp_t m12,
			 fp_t m21, fp_t m22, fp_t m31, fp_t m32);
	virtual void updateLineDash(GfxState *state);
	virtual void updateFlatness(GfxState *state);
	virtual void updateLineJoin(GfxState *state);
	virtual void updateLineCap(GfxState *state);
	virtual void updateMiterLimit(GfxState *state);
	virtual void updateLineWidth(GfxState *state);
	virtual void updateFillColor(GfxState *state);
	virtual void updateStrokeColor(GfxState *state);

	//----- update text state
	virtual void updateFont(GfxState *state);
	virtual void updateTextMat(GfxState *state);
	virtual void updateCharSpace(GfxState *state);
	virtual void updateRender(GfxState *state);
	virtual void updateRise(GfxState *state);
	virtual void updateWordSpace(GfxState *state);
	virtual void updateHorizScaling(GfxState *state);
	virtual void updateTextPos(GfxState *state);
	virtual void updateTextShift(GfxState *state, fouble shift);
	
	//----- path painting
	virtual void stroke(GfxState *state);
	virtual void fill(GfxState *state);
	virtual void eoFill(GfxState *state);

	//----- path clipping
	virtual void clip(GfxState *state);
	virtual void eoClip(GfxState *state);

	//----- text drawing
	virtual void beginString(GfxState *state, GString *s);
	virtual void endString(GfxState *state);

	virtual void drawChar(GfxState *state, fp_t x, fp_t y,
	                      fp_t dx, fp_t dy,
	                      fp_t originX, fp_t originY,
	                      CharCode code, Unicode *u, int uLen);

	//----- image drawing
	virtual void drawImageMask(GfxState *state, Object *ref, Stream *str,
	                          int width, int height, GBool invert,
	                          GBool inlineImg);
	virtual void drawImage(GfxState *state, Object *ref, Stream *str,
	                       int width, int height, GfxImageColorMap *colorMap,
	                       int *maskColors, GBool inlineImg);

#if 0
	// Find a string.  If <top> is true, starts looking at <l>,<t>;
	// otherwise starts looking at top of page.  If <bottom> is true,
	// stops looking at <l+w-1>,<t+h-1>; otherwise stops looking at bottom
	// of page.  If found, sets the text bounding rectange and returns
	// true; otherwise returns false.
	GBool findText ( Unicode *s, int len, GBool top, GBool bottom, int *xMin, int *yMin, int *xMax, int *yMax );
	bool findText ( const QString &str, int &l, int &t, int &w, int &h, bool top = 0, bool bottom = 0 );
	bool findText ( const QString &str, QRect &r, bool top = 0, bool bottom = 0 );
	// Get the text which is inside the specified rectangle.
	QString getText ( int left, int top, int width, int height );
	QString getText ( const QRect &r );

	// Get text after reflow (approx. in reading order)
	QString getTextFlows ( int left, int top, int width, int height );
	QString getTextFlows();
#endif
	
	//----- special QT access

	void startDoc(XRef *xref);
	void setPixmapPtr( QPixmap **pixmap );
	void makeTextMap(QDocText *textmap);
	void processQtEvents();

	bool m_reproportion_text;
	bool m_reproportion_and_render;
	bool m_always_render_font;
	bool m_found_render_font;

	int getKernelFreeMemory();

	static QDict<QString> fontCharSubstitution;

protected:
	void renderFont(GfxState *state, int render, 
			fouble x1, fouble y1, fouble dx1, fouble dy1,
			CharCode code, Unicode *u, int uLen);


	QPixmap **m_pixmap_ptr;
	QPainter *m_painter;
	bool m_no_refresh;
	TextPage *m_text;		// text from the current page

private:
	// hack to reproportion font mismatches
	int m_hack_x1;
	int m_hack_dx1;
	fouble m_last_x, m_last_y;
	fouble m_min_word_space, m_max_word_space, m_max_char_space;
	bool m_good_font_match;

	void cacheSysFontDatabase();
	struct SysFontCacheEntry {
	  QFont font;
	  int avg_size;
	};
	QValueList<SysFontCacheEntry*> m_sys_font_database;

	bool m_type3_font;
	bool m_type3_msg;

	int m_free_memory;

	QFont matchFont ( GfxFont *, fp_t m11, fp_t m12, fp_t m21, fp_t m22 );

	void updateLineAttrs ( GfxState *state, GBool updateDash );
	void doFill ( GfxState *state, bool winding );
	void doClip ( GfxState *state, bool winding );
	int convertPath ( GfxState *state, QPointArray &points, QArray<int> &lengths );
	int convertSubpath ( GfxState *state, GfxSubpath *subpath, QPointArray &points );


#ifdef TYPE3_CACHE
	void drawType3Glyph(T3FontCache *t3Font,
			    T3FontCacheTag *tag, Guchar *data,
			    fouble x, fouble y, GfxRGB *color);

	T3FontCache *			// Type 3 font cache
	  t3FontCache[xOutT3FontCacheSize];
	int nT3Fonts;			// number of valid entries in t3FontCache
	T3GlyphStack *t3GlyphStack;	// Type 3 glyph context stack
	//	int pixmapW, pixmapH;
#endif
	

#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
	QOutputFontCache *fontCache;
#endif
	QOutputFont *pdfFont;
	XRef *xref;


};



#define xoutRound(x) ((int)(x + 0.5))

#ifdef RENDER_EMBEDDED_SYMBOL_FONTS


#if HAVE_T1LIB_H
class T1FontEngine;
class T1FontFile;
class T1Font;
#endif

#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
class FTFontEngine;
class FTFontFile;
class FTFont;
#endif

#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
class TTFontEngine;
class TTFontFile;
class TTFont;
#endif

//------------------------------------------------------------------------
// QOutputFont
//------------------------------------------------------------------------

class QOutputFont {
public:

  QOutputFont(Ref *idA, fouble m11OrigA, fouble m12OrigA,
	      fouble m21OrigA, fouble m22OrigA,
	      fouble m11A, fouble m12A, fouble m21A, fouble m22A,
	      QOutputDev *xOutA);

  virtual ~QOutputFont();

  // Does this font match the ID and transform?
  GBool matches(Ref *idA, fouble m11OrigA, fouble m12OrigA,
		fouble m21OrigA, fouble m22OrigA)
    { return id.num == idA->num && id.gen == idA->gen &&
	     m11Orig == m11OrigA && m12Orig == m12OrigA &&
	     m21Orig == m21OrigA && m22Orig == m22OrigA; }

  // Was font created successfully?
  virtual GBool isOk() = 0;

  // Update <gc> with this font.
  virtual void updateGC(QPainter* gc) = 0;

  // Draw character <c>/<u> at <x>,<y> (in device space).
  virtual void drawChar(GfxState *state, QPixmap *pixmap,
			GfxRGB *rgb,
			fouble x, fouble y, fouble dx, fouble dy,
			CharCode c, Unicode *u, int uLen) = 0;

  // Returns true if this QOutputFont subclass provides the
  // getCharPath function.
  virtual GBool hasGetCharPath() { return gFalse; }

  // Add the character outline for <c>/<u> to the current path.
  virtual void getCharPath(GfxState *state,
			   CharCode c, Unicode *u, int ulen);

protected:

  Ref id;			// font ID
  fouble m11Orig, m12Orig,	// original transform matrix
         m21Orig, m22Orig;
  fouble m11, m12, m21, m22;	// actual transform matrix (possibly
				//   modified for font substitution)
  QOutputDev *xOut;
};

#if HAVE_T1LIB_H
//------------------------------------------------------------------------
// QOutputT1Font
//------------------------------------------------------------------------

class QOutputT1Font: public QOutputFont {
public:

  QOutputT1Font(Ref *idA, T1FontFile *fontFileA,
		fouble m11OrigA, fouble m12OrigA,
		fouble m21OrigA, fouble m22OrigA,
		fouble m11A, fouble m12A,
		fouble m21A, fouble m22A,
		QOutputDev *xOutA);

  virtual ~QOutputT1Font();

  // Was font created successfully?
  virtual GBool isOk();

  // Update <gc> with this font.
  virtual void updateGC(QPainter* gc);

  // Draw character <c>/<u> at <x>,<y>.
  virtual void drawChar(GfxState *state, QPixmap *pixmap,
			GfxRGB *rgb,
			fouble x, fouble y, fouble dx, fouble dy,
			CharCode c, Unicode *u, int uLen);

  // Returns true if this QOutputFont subclass provides the
  // getCharPath function.
  virtual GBool hasGetCharPath() { return gTrue; }

  // Add the character outline for <c>/<u> to the current path.
  virtual void getCharPath(GfxState *state,
			   CharCode c, Unicode *u, int ulen);

private:

  T1FontFile *fontFile;
  T1Font *font;
};
#endif // HAVE_T1LIB_H

#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
//------------------------------------------------------------------------
// QOutputFTFont
//------------------------------------------------------------------------

class QOutputFTFont: public QOutputFont {
public:

  QOutputFTFont(Ref *idA, FTFontFile *fontFileA,
		fouble m11OrigA, fouble m12OrigA,
		fouble m21OrigA, fouble m22OrigA,
		fouble m11A, fouble m12A,
		fouble m21A, fouble m22A,
		QOutputDev *xOutA);

  virtual ~QOutputFTFont();

  // Was font created successfully?
  virtual GBool isOk();

  // Update <gc> with this font.
  virtual void updateGC(QPainter* gc);

  // Draw character <c>/<u> at <x>,<y>.
  virtual void drawChar(GfxState *state, QPixmap *pixmap,
			GfxRGB *rgb,
			fouble x, fouble y, fouble dx, fouble dy,
			CharCode c, Unicode *u, int uLen);

  // Returns true if this QOutputFont subclass provides the
  // getCharPath function.
  virtual GBool hasGetCharPath() { return gTrue; }

  // Add the character outline for <c>/<u> to the current path.
  virtual void getCharPath(GfxState *state,
			   CharCode c, Unicode *u, int ulen);

private:

  FTFontFile *fontFile;
  FTFont *font;
};
#endif // FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)

#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
//------------------------------------------------------------------------
// QOutputTTFont
//------------------------------------------------------------------------

class QOutputTTFont: public QOutputFont {
public:

  QOutputTTFont(Ref *idA, TTFontFile *fontFileA,
		fouble m11OrigA, fouble m12OrigA,
		fouble m21OrigA, fouble m22OrigA,
		fouble m11A, fouble m12A,
		fouble m21A, fouble m22A,
		Display *displayA, QOutputDev *xOutA);

  virtual ~QOutputTTFont();

  // Was font created successfully?
  virtual GBool isOk();

  // Update <gc> with this font.
  virtual void updateGC(QPainter* gc);

  // Draw character <c>/<u> at <x>,<y>.
  virtual void drawChar(GfxState *state, QPixmap *pixmap,
			GfxRGB *rgb,
			fouble x, fouble y, fouble dx, fouble dy,
			CharCode c, Unicode *u, int uLen);

private:

  TTFontFile *fontFile;
  TTFont *font;
};
#endif // !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)

#if HAVE_T1LIB_H
class QOutputT1FontFile {
public:
  QOutputT1FontFile(int numA, int genA, GBool substA,
		    T1FontFile *fontFileA, GString *tmpFileNameA)
    { num = numA; gen = genA; subst = substA;
      fontFile = fontFileA; tmpFileName = tmpFileNameA; }
  ~QOutputT1FontFile();
  int num, gen;
  GBool subst;
  T1FontFile *fontFile;
  GString *tmpFileName;
};
#endif

#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
class QOutputFTFontFile {
public:
  QOutputFTFontFile(int numA, int genA, GBool substA,
		    FTFontFile *fontFileA, GString *tmpFileNameA)
    { num = numA; gen = genA; subst = substA;
      fontFile = fontFileA; tmpFileName = tmpFileNameA; }
  ~QOutputFTFontFile();
  int num, gen;
  GBool subst;
  FTFontFile *fontFile;
  GString *tmpFileName;
};
#endif

#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
class QOutputTTFontFile {
public:
  QOutputTTFontFile(int numA, int genA, GBool substA,
		    TTFontFile *fontFileA, GString *tmpFileNameA)
    { num = numA; gen = genA; subst = substA;
      fontFile = fontFileA; tmpFileName = tmpFileNameA; }
  ~QOutputTTFontFile();
  int num, gen;
  GBool subst;
  TTFontFile *fontFile;
  GString *tmpFileName;
};
#endif

class QOutputFontCache {
public:

  // Constructor.
  QOutputFontCache(QOutputDev *xOutA,
		   FontRastControl t1libControlA,
		   FontRastControl freetypeControlA);

  // Destructor.
  ~QOutputFontCache();

  // Initialize (or re-initialize) the font cache for a new document.
  void startDoc();

  // Get a font.  This creates a new font if necessary.
  QOutputFont *getFont(XRef *xref, GfxFont *gfxFont, fouble m11, fouble m12,
		       fouble m21, fouble m22);

private:

  void delFonts();
  void clear();
#if 0
  QOutputFont *tryGetFont(XRef *xref, DisplayFontParam *dfp, GfxFont *gfxFont,
			  fouble m11Orig, fouble m12Orig,
			  fouble m21Orig, fouble m22Orig,
			  fouble m11, fouble m12, fouble m21, fouble m22,
			  GBool subst);
#endif
#if HAVE_T1LIB_H
  QOutputFont *tryGetT1Font(XRef *xref, GfxFont *gfxFont,
			    fouble m11, fouble m12, fouble m21, fouble m22);
  QOutputFont *tryGetT1FontFromFile(XRef *xref, GString *fileName,
				    GBool deleteFile, GfxFont *gfxFont,
				    fouble m11Orig, fouble m12Orig,
				    fouble m21Orig, fouble m22Orig,
				    fouble m11, fouble m12,
				    fouble m21, fouble m22, GBool subst);
#endif
#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
  QOutputFont *tryGetFTFont(XRef *xref, GfxFont *gfxFont,
			    fouble m11, fouble m12, fouble m21, fouble m22);
  QOutputFont *tryGetFTFontFromFile(XRef *xref, GString *fileName,
				    GBool deleteFile, GfxFont *gfxFont,
				    fouble m11Orig, fouble m12Orig,
				    fouble m21Orig, fouble m22Orig,
				    fouble m11, fouble m12,
				    fouble m21, fouble m22,
				    GBool embedded, GBool subst);
#endif
#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
  QOutputFont *tryGetTTFont(XRef *xref, GfxFont *gfxFont,
			    fouble m11, fouble m12, fouble m21, fouble m22);
  QOutputFont *tryGetTTFontFromFile(XRef *xref, GString *fileName,
				    GBool deleteFile, GfxFont *gfxFont,
				    fouble m11Orig, fouble m12Orig,
				    fouble m21Orig, fouble m22Orig,
				    fouble m11, fouble m12,
				    fouble m21, fouble m22, GBool subst);
#endif

  QOutputDev *xOut;

  QOutputFont *
    fonts[xOutFontCacheSize];
  int nFonts;

#if HAVE_T1LIB_H
  FontRastControl t1libControl;	// t1lib settings
  T1FontEngine *t1Engine;	// Type 1 font engine
  GList *t1FontFiles;		// list of Type 1 font files
				//   [QOutputT1FontFile]
#endif

#if HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H
  FontRastControl		// FreeType settings
    freetypeControl;
#endif
#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
  FTFontEngine *ftEngine;	// FreeType font engine
  GList *ftFontFiles;		// list of FreeType font files
				//   [QOutputFTFontFile]
#endif
#if !FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
  TTFontEngine *ttEngine;	// TrueType font engine
  GList *ttFontFiles;		// list of TrueType font files
				//   [QOutputTTFontFile]
#endif
};
#endif


class QTextOutputDev : public TextOutputDev {
 public:
  QTextOutputDev(char *fileName, GBool physLayoutA,
		GBool rawOrderA, GBool append)
    : TextOutputDev(fileName, physLayoutA, rawOrderA, append) {
    m_reproportion_text = false;
    m_always_render_font = false;
    t1libControl =  globalParams->getT1libControl();
    freetypeControl = globalParams->getFreeTypeControl();
  }

  FontRastControl t1libControl;
  FontRastControl freetypeControl;

  // hack to reproportion font mismatches
  bool m_reproportion_text;
  bool m_always_render_font;
  int m_hack_x1;
  int m_hack_dx1;
  fouble m_last_x, m_last_y;
  fouble m_min_word_space, m_max_word_space, m_max_char_space;
  bool m_type3_font;
  bool m_type3_msg;
  bool m_embedded_font;

  int m_width, m_height;

  //----- text drawing
  virtual void beginString(GfxState *state, GString *s);
  virtual void endString(GfxState *state);
  virtual void drawChar(GfxState *state, fouble x, fouble y,
			fouble dx, fouble dy,
			fouble originX, fouble originY,
			CharCode c, Unicode *u, int uLen);
  virtual void updateFont(GfxState *state);
  virtual void startPage(int pageNum, GfxState *state);

};


#endif

