#define INCLUDE_MENUITEM_DEF

/*

To compile this you will have to change the files:
  qwidget.h    -  make the event function public
  qobject.h    -  make the activate_xxx functions public

*/

#include "qpdf.h"
#ifdef AUTOTEST
//#define MEM_LEAKS_EVENT_COUNT 50000

#include "qobjectlist.h"
#include "qmenudata.h"
#include "qwidgetlist.h"
#include "qwsevent_qws.h"
#include "qpe/global.h"
#include "qpe/applnk.h"
#include <fcntl.h>

/*
 Just so you know how much this blows, I tried all of these, and more!
      // qApp->event(me);
      // QApplication::postEvent(qApp->desktop(), me);
      //      m_stack->notify(this,me);
      //      qApp->desktop()->event(me);
      // qApp->notify(qApp->desktop(),me);
      //      qApp->postEvent( qApp->desktop(), me);
 Before realizing that there is no simple way (that I can find) to just
 send an event to the desktop and have it go to the correct place as if the
 mouse had moved, or key had been pressed!
*/

void
mouseEvent(QMouseEvent *me)
{
  if (qApp->activeWindow() == NULL) {
    return;
  }
  int winid = qApp->activeWindow()->winId();
  QWSMouseEvent qme;
  qme.simpleData.window = winid;
  qme.simpleData.x_root = me->x();
  qme.simpleData.y_root = me->y();
  qme.simpleData.state = me->state();
  qme.simpleData.time = QTime::currentTime().msec()
    + QTime::currentTime().second() * 1000
    + QTime::currentTime().minute() * 60 * 1000;

  if (me->x() >= 220 && me->y() <= 14) {
    delete me;
    return;
  }
  printf("mouseEvent win %d pos %d %d state %d time %d\n",
	 winid, me->x(), me->y(), me->state(), qme.simpleData.time);
  qApp->qwsProcessEvent(&qme);
  delete me;
}

void
keyboardEvent(QKeyEvent *ke)
{
  int winid = qApp->focusWidget()->winId();
  QWSKeyEvent qke;
  if (winid == 0) {
    return;
  }
  qke.simpleData.window = winid;
  qke.simpleData.unicode = 0;
  qke.simpleData.keycode = ke->key();
  qke.simpleData.modifiers = 0; //ke->state();
  qke.simpleData.is_press = (ke->type() == QEvent::KeyPress);
  qke.simpleData.is_auto_repeat = 0;
  qApp->focusWidget()->event(ke);
  //  QApplication::sendEvent(qApp->focusWidget(), ke);
  printf("keyboardEvent win %d key %x press %d\n",
	 winid, ke->key(), qke.simpleData.is_press);
  //qApp->qwsProcessEvent(&qke);
  delete ke;
}

void
QPDFWindow::autotester()
{
  static QTimer timer;
  int fh = open("/dev/random",0);
  if (fh > 0) {
    int rand;
    read(fh,&rand,sizeof(int));
    srand(rand);
    printf("seed %x\n", rand);
  } else {
    QTime midnight(0, 0, 0);
    srand(midnight.secsTo(QTime::currentTime()) );
  }
  connect(&timer, SIGNAL(timeout()), 
	  this, SLOT(autotester_timer_event()));
  timer.start(120, false);
}

static int random_range(int low, int high)
{
  return((rand() % (high - low + 1)) + low);
}

static int random_walk(int start, int maxstep, int low, int high)
{
  start += random_range(-maxstep, maxstep);
  if (start < low) {
    start = low;
  }
  if (start > high) {
    start = high;
  }
  return(start);
}

void
QPDFWindow::autotester_timer_event()
{
  
  static bool mouse_down = false;
  static bool mouse_held_down = false;
  static QPoint mouse_pos;
  static int openfile = 1;
  QMouseEvent *me;
  QKeyEvent *ke;
  int key;
  QMenuItem *mi;
  QPopupMenu *pm;
  QMenuBar *mb;
  QWidget *widget;
  QWidgetList *wlist;
  static DocLnkSet files;
  static DocLnk *doclnk;
  int xx;
  QWidget *widgets[5000];
  int num_widgets;

#ifdef MEM_LEAKS_EVENT_COUNT
  static int num_events = MEM_LEAKS_EVENT_COUNT;
  if (num_events-- < 0) {
    printf("(%d)\n", num_events);
    if (num_events < -30) {
      qApp->quit();
    }
    if (num_events < -100) {
      exit(0);
    }
    return;
  }
  printf("(%d)\n", num_events);
#endif

  if ((rand() % 500) == 0) {
    QTime midnight(0, 0, 0);
    srand(midnight.secsTo(QTime::currentTime()) );
  }

  //  if (((rand() % 20) == 0) && m_pdfview->getPath() == "" && !isBusy()) {
  if ((rand() % 90) == 0) {
    if (files.children().count() == 0) {
      Global::findDocuments(&files, "application/pdf");
    }
    printf("** open a file (of %d files)\n", files.children().count());
    if (files.children().count() != 0) {
      doclnk = ((QList<DocLnk>&)files.children()).at(rand() % files.children().count());
      openFile(doclnk->file());
      return;
    }
    //    openFile(doclnk);
  }

  if (mouse_down && !mouse_held_down) {
    if ((rand() % 2) == 0) {
      me = new QMouseEvent(QEvent::MouseButtonRelease, mouse_pos, 0, 0);
      printf("Mouse Click %d %d\n", mouse_pos.x(), mouse_pos.y());
      mouseEvent(me);
      mouse_down = false;
    } else {
      mouse_held_down = true;
    }
  }

  if ( (xx = (rand() % (isBusy()? 20 : 3))) != 0 ) {
     return;
  }

  if (mouse_down) {
    // I don't look at stuff happening while mouse is down (e.g. keyboard, menu, etc.)
    if ((rand() % 4) == 0) {
      me = new QMouseEvent(QEvent::MouseButtonRelease, mouse_pos, 0, 0);
      printf("Mouse Release %d %d\n", mouse_pos.x(), mouse_pos.y());
      mouseEvent(me);
      mouse_down = false;
      mouse_held_down = false;
    } else {
      mouse_pos.setX(random_walk(mouse_pos.x(), 150,
					  0, qApp->desktop()->width()-1));
      mouse_pos.setY(random_walk(mouse_pos.y(), 150,
					  0, qApp->desktop()->height()-1));
      me = new QMouseEvent(QEvent::MouseMove, mouse_pos, LeftButton, LeftButton);
      printf("Mouse me=%x Move %d %d\n", me, mouse_pos.x(), mouse_pos.y());
      mouseEvent(me);
    }
    return;
  }
  switch(rand() % 10) {
  case 0:
    mouse_pos.setX(random_walk(mouse_pos.x(), 70,
			       0, qApp->desktop()->width()-1));
    mouse_pos.setY(random_walk(mouse_pos.y(), 70,
			       0, qApp->desktop()->height()-1));
    me = new QMouseEvent(QEvent::MouseButtonPress, mouse_pos, LeftButton, LeftButton);
    printf("Mouse Press %d %d\n", mouse_pos.x(), mouse_pos.y());
    mouseEvent(me);
    mouse_down = true;
    break;

  case 1:
  case 2:
  case 3:
    
    wlist = qApp->allWidgets();
    num_widgets = 0;
    for(int i=0; i<wlist->count(); i++) {
      widget = wlist->at(i);
      if (!widget->isEnabled()) {
	continue;
      }
      if (strcmp(widget->className(),"QWidget") == 0
	&& strcmp(widget->name(),"qt_viewport") == 0) {
	continue;
      }
      if (strcmp(widget->className(),"QListBox") == 0
	&& strcmp(widget->name(),"in-combo") == 0) {
	continue;
      }
      if (widget->isFocusEnabled()) {
	widgets[num_widgets++] = widget;
      }
    }
    delete wlist;
    if (num_widgets == 0) {
      break;
    }
    widget = widgets[rand() % num_widgets];
    printf("%d widgets with focus enabled\n", num_widgets);
    static int backup = 0;
    if (backup > 0) {
      key = Key_Backspace;
      backup--;
    } else {  
      if ((rand() % 30) == 0) {
	backup = rand() % 50 + 1;
      }
      switch(rand() % 10) {
      case 0:
	key = random_range(Key_Escape, Key_Help);
	break;
	
      case 1:
	key = Key_Backspace;
	break;
	
      case 2:
      case 3:
      case 4:
      case 5:
	key = Key_Return;
	break;
	
      case 6:
      case 7:
	key = random_range(Key_0, Key_9);
	break;
	
      default:
	key = random_range(Key_Space, Key_AsciiTilde);
	break;
      }
    }
    
    printf("=== Keypress 0x%x (%c) to %s::%s\n", key, isalpha(key)? key : 'X',
	   widget->className(), widget->name());
    if (strcmp(widget->className(),"QPopupMenu") == 0
	&& strcmp(widget->name(),"unnamed") == 0) {
      widget->dumpObjectInfo();
    }
    if (backup > 0) {
      key = Key_Backspace;
      while(backup-- > 0) {
	ke = new QKeyEvent(QEvent::KeyPress, key, (key < 0xff)?key:0, 0, 
			   (key < 0xff)?QString(QChar(key)):QString());
	widget->event(ke);
	delete ke;
	ke = new QKeyEvent(QEvent::KeyRelease, key, key < 0xff?key:0, 0, 
			   (key < 0xff)?QString(QChar(key)):QString());
	widget->event(ke);
	delete ke;
      }
    } else {
      ke = new QKeyEvent(QEvent::KeyPress, key, (key < 0xff)?key:0, 0, 
			 (key < 0xff)?QString(QChar(key)):QString());
      widget->event(ke);
      delete ke;
      ke = new QKeyEvent(QEvent::KeyRelease, key, key < 0xff?key:0, 0, 
			 (key < 0xff)?QString(QChar(key)):QString());
      widget->event(ke);
      delete ke;
    }
    break;

  default:

    wlist = qApp->allWidgets();
    num_widgets = 0;
    for(int i=0; i<wlist->count(); i++) {
      widget = wlist->at(i);
      if (!widget->isEnabled()) {
	continue;
      }
      if (widget == m_pm_categories && (rand()&15) != 0) {
	continue;
      }
      if (widget->inherits("QPopupMenu") 
	  || widget->receivers("clicked()")
	  || widget->receivers("pressed()")) {
	if (qApp->loopLevel() > 1) {
	  if (widget->inherits("QButton")) {
	    if (((QButton*)widget)->text().find("OK") >= 0
		|| ((QButton*)widget)->text().find("Ok") >= 0) {
	      widgets[num_widgets++] = widget;
	    }
	  }
	} else {
	  widgets[num_widgets++] = widget;
	}
      }
    }
    delete wlist;
    if (num_widgets == 0) {
      break;
    }
    widget = widgets[rand() % num_widgets];

    if (widget->inherits("QPopupMenu")) {
      pm = (QPopupMenu*)widget;
      if (pm->count() <= 0)
	break;
      for(int ii=0; ii<30; ii++) {
	int id = pm->idAt(rand() % pm->count());
	if (!pm->isItemEnabled(id)) 
	  continue;
	mi = pm->findItem(id);
	if (mi->text().find("Backward") == 0) {
	  static int back = 0;
	  printf("*************** Backwards = %d  enabled=%d\n", back++, pm->isItemEnabled(id));
	}
	if (mi->text().find("Forward") == 0) {
	  static int forw = 0;
	  printf("*************** Forwards = %d enabled=%d\n", forw++, pm->isItemEnabled(id));
	}
	if (!mi->isEnabled()) {
	  printf("thought it was enabled!\n");
	  continue;
	}
	if (! mi->signal()) {
	  if (pm->receivers("activated(int)")) {
	    if (mi->text().find("All") == 0 
		|| mi->text().find("Unfiled") == 0
		|| mi->text().find("Personal") == 0
		|| mi->text().find("Business") == 0) {
	      continue;
	    }
	    printf("=== ACTIVATE %s %s\n", pm->name(), mi->text().latin1());
	    pm->activate_signal("activated(int)", id);
	    break;
	  }
	  printf("no signal\n");
	  continue;
	}
	if (mi->signal()->isBlocked()) {
	  printf("is blocked\n");
	  continue;
	}
	printf("=== Menu item %s %s (%s)\n", pm->name(), mi->text().latin1(),
	       mi->whatsThis().latin1());
	if (mi->text().find("Open") == 0) {
	  if ((rand() % 3) != 0) {
	    continue;
	  }
	  printf("OPEN FILE\n");
	  if (files.children().count() == 0) {
	    Global::findDocuments(&files, "application/pdf");
	  }
	  doclnk = ((QList<DocLnk>&)files.children()).at(rand() % files.children().count());
	  openFile(doclnk->file());
	  //	  openFile(doclnk);
	  return;
	}
	mi->signal()->activate();
	break;
      }
    } else {
      //      printf("WIDGET --- ");
      //      widget->dumpObjectInfo();
      if (widget->signalsBlocked()) {
	printf("=== ***blocked signals %s\n", widget->name());
      } else {
	// could go and connect, then disconnect a slot here
	if (widget->receivers(SIGNAL(pressed()))) {
	  printf("=== Pressed %s::%s\n", widget->className(),
		 widget->name());
	  widget->activate_signal("pressed()");
	  break;
	} 
	if (widget->receivers(SIGNAL(clicked()))) {
	  printf("=== Clicked %s::%s\n", widget->className(),
		 widget->name());
	  widget->activate_signal("clicked()");
	  break;
	} 
      }
    }
  }
}
      

#endif
