// Copyright 2003, Philip Kohn

#include "aconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <iostream>
#include <assert.h>

#include <qgfx_qws.h>
#include <qpixmap.h>
#include <qimage.h>
#include <qpainter.h>
#include <qdict.h>
#include <qtimer.h>
#include <qlabel.h>
#include <qscrollview.h>
#include <qmessagebox.h>
#include <qapplication.h>
#include <qfileinfo.h>

#include "qdocview.h"


QDocState::QDocState()
{
  clear();
}

void
QDocState::clear()
{
  page = 0;
  numPages = 0;
  page_offset = 0;
  isRotated = false;
  zoom = 0.8;
  // should be an OPTION
#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
  render_text = true;
  //  render_text = false;
#else
  render_text = false;
#endif
  reproportion_and_render = false;
  reproportion_text = true;
  //  reproportion_text = false; 
  auto_sense_render_text = false;
  path = "";  // bug in findRev comparing null strings...
  x = y = 0;
  doMove = false;
  mode = PDF_VIEW;
  width = height = -1;
  char_index = -1;
}

void
QDocState::set(QDocState &s)
{
  path = s.path;
  zoom = s.zoom;
  page = s.page;
  render_text = s.render_text;
  reproportion_text = s.reproportion_text;
  mode = s.mode;
  doMove = s.doMove;
  x = s.x; y = s.y;
  char_index = s.char_index;
}

void
QDocState::home()
{
  x = y = -1;
  doMove = true;
}

void
QDocState::pos(int xx, int yy)
{
  x = xx; y = yy;
  doMove = true;
}

QString
QDocState::filename()
{
  QFileInfo fi(path);
  return(fi.fileName());
}

QString
QDocState::basename()
{
  QFileInfo fi(path);
  return(fi.baseName());
}

void
QDocState::save(QString &str)
{
  str += QString::fromLatin1("<bookmark page=\"%1\" zoom=\"%2\" x=\"%3\" y=\"%4\" rotate=\"%5\" file=\"%6\" adjust=\"%7\" render=\"%8\" pageoffset=\"%9\"")
    .arg(page).arg(zoom).arg(x).arg(y).arg(isRotated).arg(path).arg(reproportion_text).arg(render_text).arg(page_offset) 
    + QString::fromLatin1(" char_index=\"%1\" mode=\"%2\">\n").arg(char_index).arg(mode);
}

int
QDocState::load(const QString &str, int index, bool loadPath)
{
  int nextIndex;
  XmlParser parser;
  nextIndex = parser.parse(str, index);
  if (!parser.checkHead("bookmark")) {
    //    printf("no bookmark header\n");
    return(-1);
  }
  if (loadPath && !parser.get("file").isEmpty()) {
    path = parser.get("file");
  }
  if (!parser.get("page").isEmpty()) {
    page = parser.get("page").toInt();
  }
  if (!parser.get("zoom").isEmpty()) {
    zoom = parser.get("zoom").toFloat();
  }
  if (!parser.get("x").isEmpty()) {
    x = parser.get("x").toInt();
  }
  if (!parser.get("y").isEmpty()) {
    y = parser.get("y").toInt();
  }
  if (!parser.get("char_index").isEmpty()) {
    char_index = parser.get("char_index").toInt();
  }
  if (!parser.get("rotate").isEmpty()) {
    isRotated = parser.get("rotate").toInt();
  }
  if (!parser.get("adjust").isEmpty()) {
    reproportion_text = parser.get("adjust").toInt();
  }
  if (!parser.get("render").isEmpty()) {
    render_text = parser.get("render").toInt();
  }
  if (!parser.get("pageoffset").isEmpty()) {
    page_offset = parser.get("pageoffset").toInt();
  }
  if (!parser.get("mode").isEmpty()) {
    mode = (ViewMode)parser.get("mode").toInt();
  }

  doMove = true;
  //  printf("**loaded state page = %d\n", page);
  return(nextIndex);
}


bool
QDocState::isSamePixmap(QDocState &state)
{
#if 0
  printf("isSamePixmap(%s==%s  %d==%d  %f==%f  %d==%d  %d==%d  %d==%d) %d\n",
	 path.latin1(), state.path.latin1(),
	 page, state.page,
	 (float)zoom, (float)state.zoom,
	 reproportion_and_render, state.reproportion_and_render,
	 render_text,state.render_text,
	 reproportion_text, state.reproportion_text,
	 page == state.page
	 && path == state.path
	 && zoom == state.zoom
#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
	 && reproportion_and_render == state.reproportion_and_render
	 && render_text == state.render_text
#endif
	 && reproportion_text == state.reproportion_text);
#endif

  return(page == state.page
	 && path == state.path
	 && zoom == state.zoom
#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
	 && reproportion_and_render == state.reproportion_and_render
	 && render_text == state.render_text
#endif
	 && reproportion_text == state.reproportion_text);
}

bool
QDocState::isSamePage(QDocState &state)
{
  return(page == state.page
	 && path == state.path);
}

