#ifndef _QDOCTEXT_H
#define _QDOCTEXT_H

#include "aconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <iostream>
#include <assert.h>

#include <qgfx_qws.h>
#include <qpixmap.h>
#include <qimage.h>
#include <qpainter.h>
#include <qdict.h>
#include <qtimer.h>
#include <qlabel.h>
#include <qscrollview.h>
#include <qmessagebox.h>
#include <qapplication.h>


class QDocTextWord {
 public:
  QDocTextWord(QRect pdf_box, QRect screen_box, int firstChar, int lastChar);
  ~QDocTextWord() {}
  bool contains(QPoint);
  bool containsPDF(QPoint);
  bool contains(int charIndex);

  int firstCharIndex, lastCharIndex;
  QRect pdf_box; // pdf box
  QRect box; // screen box (after hack, same as pdf box if no hack)
  //  QString debug_word_text;
};

class QDocText  {
 public:
  QDocText();
  virtual ~QDocText();
  void clear();
  int numWords();
  int numChars();
  int findWord(QPoint point, bool closest = true, bool lower_right = false);  // screen coord (hacked)
  int findWordPDF(QPoint point, bool closest = true); // PDF coord (unhacked)
  int findWord(QDocState state, bool closest = true);
  int findWord(QString string);
  bool findString(const QString &str, int *firstChar, int *lastChar);
  QString getText(int firstWord = 0, int lastWord = -1);

  bool isTextMapValid(QDocState &state) { return(isSamePixmap(state)); }
  bool isTextValid(QDocState &state) { return(isSamePage(state)); }

  bool isSamePage(QDocState &state) {
    return(m_state.isSamePage(state));
  }
  bool isSamePixmap(QDocState &state) {
    return(m_state.isSamePixmap(state));
  }
  void setState(QDocState *state) { m_state = *state; }
  void setText(QString str);
  
  int mapWordToChar(int wordNum, int *firstCharIndex = NULL, int *lastCharIndex = NULL);
  int mapCharToWord(int charNum);
  QRect wordBox(int word);  // word box on screen (with hack)
  QRect wordBoxPDF(int word); // word box in original (unhacked) PDF
  QRegion wordRegion(int firstWord, int lastWord);
  //  QString wordString(int firstWord, int lastWord);

  QDocState m_state;

  void addWord(int x1, int y1, int x2, int y2, 
	       int xs1, int xs2,
	       int first_char, int last_char);

  void appendText(QString msg);

 private:
  QString m_text;
  QList<QDocTextWord> m_words;
};
#endif
