// Copyright 2003, Philip Kohn

#ifdef __GNUC__
#pragma implementation
#endif

#include "qdocview.h"


extern QScreen *qt_screen;


QDocView::~QDocView()
{
  if (m_textmap)
    delete m_textmap;
  if (m_rotated_pixmap)
    delete m_rotated_pixmap;
  if (m_pixmap)
    delete m_pixmap;
}

bool
QDocView::textSelectionMode()
{
  return(m_doing_text_selection);
}

void
QDocView::setTextSelectionMode(bool m)
{
  if (! isTextMapValid() || m_textmap->numWords() <= 0) {
    m = false;
  }
  m_doing_text_selection = m;
  emit updateMenuItems();
}

void
QDocView::getTextSelection(int *from, int *to)
{
  *from = m_text_selection_start;
  *to = m_text_selection_end;
}

bool
QDocView::hasSelection()
{
  if (m_text_selection_end >= m_text_selection_start
      && m_text_selection_start >= 0
      && m_text_selection_end >= 0
      && isTextMapValid()) {
    return(true);
  }
  return(false);
}

void
QDocView::clearSelection()
{
  setSelection(QRegion ( ), false);
  setTextSelectionMode(false);
  m_text_selection_end = -2;
  m_text_selection_start = -1;
}

void
QDocView::setTextSelection(int from_word, int to_word, bool scroll)
{
  if (!isTextMapValid() && !getTextMap(m_state.page)) {
    return;
  }
  if (from_word < 0 || to_word < 0) {
    if (hasSelection()) {
      setSelection(getTextRegion(m_text_selection_start,
				 m_text_selection_end), scroll);
    }
    return;
  }
  if (from_word >= m_textmap->numWords()) {
    from_word = m_textmap->numWords() - 1;
  }
  if (to_word >= m_textmap->numWords()) {
    to_word = m_textmap->numWords() - 1;
  }
  m_text_selection_start = from_word;
  m_text_selection_end = to_word;
  //  printf("setTextSelection %d to %d\n", from_word, to_word);

  //  setSelection(getTextRegion(from_word, to_word), scroll);
  setSelection(getTextRegion(from_word, to_word), false);
  QPoint to_word_point = mapWordToPix(to_word);
  ensureVisible(to_word_point.x(), to_word_point.y());
}

QRegion
QDocView::getTextRegion(int from_word, int to_word)
{
  if (!isTextMapValid() && !getTextMap(m_state.page)) {
    return QRegion();
  }
  QRegion region;
  int i;
  if (to_word < 0 || to_word >= m_textmap->numWords()) {
    to_word = m_textmap->numWords() - 1;
  }
  if (from_word < 0 || from_word >= m_textmap->numWords()) {
    from_word = m_textmap->numWords() - 1;
  }
  if (from_word > to_word) {
    i = from_word;
    from_word = to_word;
    to_word = i;
  }
  for(i = from_word; i <= to_word; i++) {

    int x,y,w,h;
    QRect r = m_textmap->wordBox(i);
    r.rect(&x,&y,&w,&h);
    region |= rotate(m_textmap->wordBox(i));
  }
  return(region);
}
  
QString
QDocView::getText(int from_word, int to_word, int page)
{
  if (page < 0 || page == m_state.page) {
    return(m_textmap->getText(from_word, to_word));
  }
  if (!getTextMap(page)) {
    printf("failed to get textmap\n");
    return(QString::fromLatin1(""));
  }
  return(m_textmap->getText(from_word, to_word));
#if 0
  QDocText textmap;
  if (page < 0 || page == m_state.page) {
    return(m_textmap->getText(from_word, to_word));
  }
  if (!getTextMap(page, &textmap)) {
    printf("failed to get textmap\n");
    return(QString::fromLatin1(""));
  }
  return(textmap.getText(from_word, to_word));
#endif
}

QString
QDocView::selectedText()
{
  if (hasSelection()) {
    return(getText(m_text_selection_start, m_text_selection_end));
  } else {
    return QString();
  }
}

void 
QDocView::setPixPosition(QPoint p)
{
  setContentsPos(p.x(), p.y());
}

void 
QDocView::setWordPosition(int w)
{
  setPixPosition(mapWordToPix(w));
}

void 
QDocView::setCharPosition(int c)
{
  setPixPosition(mapWordToPix(mapCharToWord(c)));
}

void 
QDocView::setCharIndex(int c)
{
  m_state.char_index = c;
}

int 
QDocView::mapPixToWord(QPoint pix, bool closest, bool lower_right) 
{
  return(m_textmap->findWord(pix, closest, lower_right));
}

QPoint
QDocView::mapWordToPix(int word, bool lastChar)
{
  if (lastChar) {
    return(rotate(m_textmap->wordBox(word).topRight()));
  } else {
    return(rotate(m_textmap->wordBox(word).topLeft()));
  }
}

int 
QDocView::mapWordToChar(int w, bool lastChar)
{
  int first, last;
  m_textmap->mapWordToChar(w, &first, &last);
  return lastChar? last : first;
}

int 
QDocView::mapCharToWord(int w)
{
  if (!isTextValid(m_target_state)) {
    return(-1);
  }
  return(m_textmap->mapCharToWord(w));
}

QPoint
QDocView::absolutePosition(QPoint p)
{
  return(m_notes.absolutePosition(p));
}

bool
QDocView::findText ( const QString &str, int *from_char, int *to_char, int *found_page,
		     bool reverse, bool multipage )
{
  //  printf("findText(from=%d, to=%d)\n", *from_char, *to_char);
  if (! multipage) {
    if (!findTextInternal(str, from_char, to_char, state().page, state().page, found_page)) {
      return false;
    }
  } else {
    if (m_inside_read_page) {
      return false;
    }
    emit saveState();
    if (reverse) {
      *from_char = -1; *to_char = 0;
    } else {
      *from_char = 0; *to_char = -1;
    }
    emit setBusy ( true );
    if (reverse) {
      if (state().page <= 1 
	  || !findTextInternal(str, from_char, to_char, state().page-1, 1, found_page)) {
	emit setBusy(false);
	return false;
      }
    } else {
      if (state().page >= state().numPages
	  || !findTextInternal(str, from_char, to_char, state().page+1, state().numPages, 
			       found_page)) {
	emit setBusy(false);
	return false;
      }
    }
    emit setBusy(false);
  }
  return true;
}

bool
QDocView::isValid()
{
  if (!currentState()->isValid()) {
      printf("currentState mode %d not valid\n", m_target_state.mode);
  }
  //  if (!m_target_state.isValid()) {
  //    printf("m_target_state not valid\n");
  //  }
  //  return(m_state.isValid() && m_target_state.isValid());
  return(currentState()->isValid());
}

bool 
QDocView::findTextInternal(QString str, int *from_char, int *to_char, 
			   int first_page, int last_page,
			   int *found_page)
{
  bool reverse = false;
  int page;
  int len = str. length ( );
  QDocText tmp_textmap;
  QDocText *textmap;


  //  printf("findTextInternal (%s) char %d to %d, page %d to %d\n", str.latin1(), *from_char, *to_char, first_page, last_page);

  if (!isValid()
      || m_inside_read_page) {
    return false;
  }

  if (first_page < 0) {
    first_page = m_state.numPages;
  }
  if (last_page < 0) {
    last_page = m_state.numPages;
  }
  if (first_page > last_page) {
    reverse = true;
  }

  //  assert(first_page >= 1 && first_page <= m_state.numPages);
  //  assert(last_page >= 1 && last_page <= m_state.numPages);

  page = first_page;
  for(;;) {
    // interrupt with other changes (page# or mode)
    if (m_inside_read_page) {
      // changing state from UI
      //      printf("find: inside read page, exit\n");
      return(false);
    }
    if (page == m_textmap->m_state.page) {
      textmap = m_textmap;
    } else {
      printf("find in page %d\n", page);
      emit setBusy(false);
      emit setPageCount(page - getPageOffset());
      if (!getTextMap(page, &tmp_textmap)) {
	printf("failed to get textmap page %d\n", page);
	return(false);
      }
      textmap = &tmp_textmap;
    }
    if (page == first_page) {
      //      printf("first_page\n");
      if (*from_char < 0) {
	*from_char = textmap->numChars() - 1;
      }
      if (*to_char < 0) {
	*to_char = textmap->numChars() - 1;
      }
    } else {
      if (reverse) {
	*from_char = textmap->numChars() - 1;
	*to_char = 0;
      } else {
	*from_char = 0;
	*to_char = textmap->numChars() - 1;
      }
    }
    //    printf("find str %s from %d to %d\n", str.latin1(), *from_char, *to_char);
    if (textmap->findString(str, from_char, to_char)) {
      *found_page = page;
      return(true);
    }
    if (page == last_page) {
      break;
    }
    page += reverse? -1 : 1;
  }
  return(false);
}



QDocView::QDocView ( QWidget *parent, const char *name, int flags, QPDFconfig *config) :
  QGesture ( parent, name, WRepaintNoErase | WResizeNoErase | flags )
{
	
  setHScrollBarMode ( QScrollView::AlwaysOff );
  setVScrollBarMode ( QScrollView::AlwaysOff );
  setFocusPolicy ( WheelFocus );
	
  m_refresh_timer = new QTimer(this);
  connect(m_refresh_timer, SIGNAL(timeout()),
	  this, SLOT(refresh()));


  m_pixmap = NULL;
  m_rotated_pixmap = NULL;
  m_textmap = new QDocText;

  m_inside_read_page = false;
  m_abort = false;
  m_state.clear();
  m_target_state.clear();

  m_highlight_mode = true;

  setTextSelectionMode(false);
  m_mouse_debounce_time_window = false;
  m_mouse_click_time_window = false;
  m_mouse_down_in_selection = false;
  m_mouse_click_timer = new QTimer(this);
  connect(m_mouse_click_timer, SIGNAL(timeout()),
	  this, SLOT(mouseClickTimeout()));

  m_config = config;

  m_selection_drag = false;

  clearSelection();

}


void 
QDocView::updateRotate()
{

  // during repaint, have both m_pixmap (for drawing) and m_rotated_pixmap (for showing)
  // on exit from repaint, can get ride of m_pixmap
  // on entrance to repaint, or on anti-rotate, make m_pixmap from m_rotated_pixmap
  // on rotate, and not in repaint, make m_rotated_pixmap, delete m_pixmap

  if (m_target_state.isRotated) {
    if (m_inside_read_page || m_pixmap != NULL) {
      // currently drawing on m_pixmap, so need both pixmaps to display and redraw
      // rotated pixmap should be out of date
      if (m_rotated_pixmap != NULL) {
	delete m_rotated_pixmap;
	m_rotated_pixmap = NULL;
      }
    }
    if (m_rotated_pixmap == NULL && m_pixmap != NULL) {
      m_rotated_pixmap = rotatePixmap(m_pixmap, false);
      if (!m_inside_read_page) {
	delete m_pixmap;
	m_pixmap = NULL;
      }
    }
    if (m_rotated_pixmap != NULL && ! m_state.isRotated) {
      m_state.isRotated = true;
      rotateView(m_rotated_pixmap->width(), m_rotated_pixmap->height(), false);
    }
  } else {    
    if (m_pixmap == NULL && m_rotated_pixmap != NULL) {
      //      assert(m_state.isRotated);
      m_pixmap = rotatePixmap(m_rotated_pixmap, true);
      delete m_rotated_pixmap;
      m_rotated_pixmap = NULL;
    }
    if (! m_inside_read_page && m_rotated_pixmap != NULL) {
      delete m_rotated_pixmap;
      m_rotated_pixmap = NULL;
    }
    if (m_pixmap != NULL && m_state.isRotated) {
      m_state.isRotated = false;
      rotateView(m_pixmap->width(), m_pixmap->height(), true);
    }
  }
}

QPoint
QDocView::antirotate(QPoint p) 
{
  if (m_state.isRotated) {
    int tmp = p.x();
    p.setX(contentsHeight() - p.y());
    p.setY(tmp);
    //    p.setX(contentsWidth() - tmp - visibleHeight());
  }
  return(p);
}

QPoint
QDocView::rotate(QPoint p) 
{
  if (m_state.isRotated) {
    int tmp = p.y();
    p.setY(contentsHeight() - p.x());
    p.setX(tmp);
    //    p.setX(contentsWidth() - tmp - visibleHeight());
  }
  return(p);
}

QRect
QDocView::rotate(QRect r)
{
  if (!m_state.isRotated) {
    return(r);
  }
  QRect rr;
  QPoint p1,p2;
  p1 = rotate(r.topLeft());
  p2 = rotate(r.bottomRight());
  rr.setCoords(p1.x(), p1.y(), p2.x(), p2.y());
  rr.normalize();
  return(rr);
}


QRegion
QDocView::rotate(QRegion r)
{
  if (!m_state.isRotated) {
    return(r);
  }
  QRegion rr;
  QArray<QRect> rects = r.rects();
  int i;
  for(i=0; i<rects.count(); i++) {
    rr |= rotate(rects.at(i));
  }
  return(rr);
}

QPoint
QDocView::mapScreen(QPoint p)
{
  if (m_state.isRotated) {
    return(QPoint(contentsWidth() - p.y() - visibleHeight(), p.x()));
  }
  return(p);
}

int
QDocView::mapContentsX()
{
  if (m_state.isRotated) {
    return(contentsWidth() - contentsY() - visibleHeight());
  }
  return(contentsX());
}

int
QDocView::mapContentsY()
{
  if (m_state.isRotated) {   // m_target_state?
    return(contentsX());
  } 
  return(contentsY());
}

QPoint
QDocView::currentUpperLeftPix()
{
  return(QPoint(mapContentsX(), mapContentsY()));
}

QPoint
QDocView::currentCenterPix()
{
  return(mapScreen(QPoint(contentsX() + visibleWidth()/2,
			  contentsY() + visibleHeight()/2)));
}

void
QDocView::rotateView(int width, int height, bool unrotate)
{
  resizeContents(width, height);
  emit setPageCount(getPageNum() - getPageOffset(),1000);
  if (unrotate) {
    // use state.x, state.y and have updateState rotate the coord as needed
    //  then we could always work in non-rotated coord
    //  problem with this is then we have to rotate all event coord
    setContentsPos(contentsWidth() - contentsY() - visibleHeight(), contentsX());
  } else {
    setContentsPos(contentsY(), contentsHeight() - contentsX() - visibleHeight());
  }
  m_highlight_region = rotate(m_notes.highlights(m_state.page));
  setTextSelection();
}

QPixmap *
QDocView::rotatePixmap(QPixmap *old_pixmap, bool unrotate)
{
  int height = old_pixmap->height();
  int width = old_pixmap->width();
  int depth = old_pixmap->depth();
  int in_bpl = old_pixmap->bytesPerLine();
  QPixmap *new_pixmap = new QPixmap(height, width, depth);
  int out_bpl = new_pixmap->bytesPerLine();
  typedef unsigned char *pixptr;
  pixptr in_base, out_base;
  pixptr in_ptr, out_ptr;
  pixptr out_line;
  int i, ii, j, k;
  int reverse = 0;
  
#if 0
  printf("height=%d width=%d depth=%d in_bpl=%d out_bpl=%d\n", 
	 height, width, depth, in_bpl, out_bpl);
#endif

  depth /= 8;

  if (qt_screen->isTransformed()) {
    int tmp = width;
    width = height;
    height = tmp;
    reverse = 1;
  }

  in_base = old_pixmap->scanLine(0);
  out_base = new_pixmap->scanLine(0);

  if (unrotate) {
    out_line = out_base;
  } else {
    out_line = out_base + (width - 1) * out_bpl;
  }
  for(i=ii=0; i<width; i++, ii+=depth) {
    if (unrotate) {
      in_ptr = in_base + ii + (height - 1) * in_bpl;
      out_ptr = out_line;
      for(j=0; j<height; j++) {
	for(k=0; k<depth; k++) {
	  *out_ptr++ = in_ptr[k];
	}
	in_ptr -= in_bpl;
      }
      out_line += out_bpl;
    } else {
      in_ptr = in_base + ii;
      out_ptr = out_line;
      for(j=0; j<height; j++) {
	for(k=0; k<depth; k++) {
	  *out_ptr++ = in_ptr[k];
	}
	in_ptr += in_bpl;
      }
      out_line -= out_bpl;
    }
  }

  return(new_pixmap);
}

void
QDocView::gotoPage(int page)
{
  printf("gotoPage(%d)\n", page);
  if (page <= 0 || page > lastPageNum()) {
    return;
  }
  m_target_state.page = page;
  m_target_state.home();
  changeState();
  autoSenseRenderText();
}

void
QDocView::autoSenseRenderText()
{
  /* This is a hack to turn off font rendering
     if there are no embedded fonts.
     The reason is that the Adjust Spacing hack
     doesn't always get along well with embedded fonts. */

#if 0
  if (m_state.auto_sense_render_text
      && m_textmap->numWords() > 100
      && m_state.render_text) {
    if (! foundRenderedText()) {
      m_state.render_text = false;
      m_target_state.render_text = false;
      // okay, this looks bad
      // changeState();
    }
    m_state.auto_sense_render_text = false;
    m_target_state.auto_sense_render_text = false;
  }
#endif
}

// This is used in Text mode to keep track of the current page,
// but it is not painted
void
QDocView::setPage(int page)
{
  //  if (getTextMap(page)) {
    m_target_state.page = page;
    //  }
}

void
QDocView::gotoPage(int page, QPoint pos)
{
  if (page < 0 || page > lastPageNum()) {
    return;
  }
  m_target_state.page = page;
  m_target_state.pos(pos);
  printf("gotoPage(%d, %d,%d)\n", page, pos.x(), pos.y());
  changeState();
  autoSenseRenderText();
}

void
QDocView::setReProportionText(bool repro)
{
  m_target_state.reproportion_text = repro;
}

bool
QDocView::getReProportionText()
{
  return(m_target_state.reproportion_text);
}

void
QDocView::setReProportionAndRender(bool repro)
{
  m_target_state.reproportion_and_render = repro;
}

bool
QDocView::getReProportionAndRender()
{
  return(m_target_state.reproportion_and_render);
}

#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
void
QDocView::setRenderText(bool repro)
{
  m_target_state.render_text = repro;
}

bool
QDocView::getRenderText()
{
  return(m_target_state.render_text);
}
#endif

void
QDocView::setRotation(bool rotate)
{
  m_target_state.isRotated = rotate;
}

bool
QDocView::isRotated()
{
  return(m_target_state.isRotated);
}

bool
QDocView::highlightMode()
{
  return(m_highlight_mode);
}

void
QDocView::setHighlightMode(bool highlight)
{
  m_highlight_mode = highlight;
  refresh();
}



void
QDocView::setZoom(float zoom)
{
  m_target_state.zoom = zoom;
  if (m_state.zoom > 0.0) {
    float ratio = zoom / m_state.zoom;
    m_target_state.x = (int)(m_state.x * ratio);
    m_target_state.y = (int)(m_state.y * ratio);
    m_target_state.doMove = true;
  }
}

float
QDocView::getZoom()
{
  return(m_target_state.zoom);
}

int
QDocView::getPageNum()
{
  return(m_target_state.page);
}

QDocState *
QDocView::currentState()
{
  return(m_target_state.mode == PDF_VIEW? &m_state : &m_textmap->m_state);
}

int
QDocView::getPageOffset()
{
  return(m_target_state.page_offset);
}

void
QDocView::setPageOffset(int page)
{
  m_target_state.page_offset = page;
}

int
QDocView::lastPageNum()
{
  return(m_target_state.numPages);
}

QString 
QDocView::getFilename()
{
  return(m_target_state.filename());
}

QString 
QDocView::getPath()
{
  return(m_target_state.path);
}

void
QDocView::setPath(QString path)
{
  m_target_state.path = path;
  //  m_notes.setPath(notesPath(path));
}

void
QDocView::setMode(ViewMode mode)
{
  m_target_state.mode = mode;
}


void
QDocView::changeState()
{
  changeState(state());
}

bool
QDocView::isCurrentPixmap(QDocState &state)
{
  if (state.mode == PDF_VIEW) {
    return(m_state.isSamePixmap(state)
	   && m_textmap->m_state.isSamePixmap(state));
  } else {
    return(m_textmap->m_state.isSamePixmap(state));
  }
}

void 
QDocView::changeState(QDocState state, int refresh_rate)
{

  // one day this could handle drawing a part of the page
  // if the whole page would not fit in memory at the current zoom
  // (but, by then we will have more memory...)


  printf("changeState(path %s page %d/%d zoom %f rotate %d)\n",
	 state.path.ascii(), state.page, state.numPages, (float)state.zoom, state.isRotated);

  limitState(state);
  if (state.path.isEmpty()) {
    return;
  }
  //  assert(state.mode != 0);

  // would need a real semiphore here if we were multithreaded
  m_target_state = state;
  emit setPageCount(getPageNum() - getPageOffset(), 1000);

  if (m_inside_read_page) {
    if (! isCurrentPixmap(m_target_state)) {
      m_abort = true;
      abortReadPage();
      printf("abortReadPage()\n");
    }
    if (m_target_state.mode == PDF_VIEW) {
      refresh();
    }
    return;
  }

  // don't process events here, because they may cause more calls
  // into change state (ie. autorepeated down arrow at bottom of page will just
  // run through many pages per second)
  //  qApp-> processEvents ( );
  m_inside_read_page = true;
  emit setBusy(true);

  // get scroll out of the way if possible
  m_state.doMove = m_target_state.doMove;
  if (m_target_state.mode == PDF_VIEW) {
    refresh();

    if (refresh_rate < 0) {
      refresh_rate = refreshRate(m_target_state.isRotated);
    }
    if (refresh_rate > 0) {
      m_refresh_timer->start(refresh_rate,FALSE);
    }
  }

  // m_state should only change in this loop which is protected by m_inside_read_page
  // only one stack frame should exist for this loop
  // however, m_target_state can change during this loop due to events from the user, 
  //   it is a moving target!
  limitState(m_target_state);


  while(! isCurrentPixmap(m_target_state)
	|| m_abort) {

    m_abort = false;

#if 1
    printf("m_target_state(path %s page %d/%d zoom %f textpage %d zoom %f rotate %d x %d y %d move %d mode %d)\n",
	   m_target_state.path.ascii(), m_target_state.page, m_target_state.numPages,
	   (float)m_target_state.zoom,
	   m_textmap->m_state.page, (float)m_textmap->m_state.zoom,
    	   m_target_state.isRotated, 
    	   m_target_state.x, m_target_state.y, m_target_state.doMove, m_target_state.mode);
    printf("state(path %s page %d/%d zoom %f textpage %d zoom %f rotate %d mode %d)\n",
	   m_state.path.ascii(), m_state.page, m_state.numPages, (float)m_state.zoom, 
	   m_textmap->m_state.page, (float)m_textmap->m_state.zoom,
	   m_state.isRotated, m_state.mode);
#endif
    
    qApp-> processEvents ( );  
    emit updateMenuItems();
    qApp-> processEvents ( );  

    // originally the design had 3 states:
    //   the current state (only changed in this function)
    //   the target state (changed anywhere at any time)
    //   the copy of the target we are trying to match (only copied by this function)
    // this was a better idea than what I ended up doing, which seemed simpler
    // the trouble is that the target really does change at ANY point!

    if (m_state.path != m_target_state.path && ! m_abort) {
      clearSelection();
      emit flush();
      m_textmap->m_state.path = m_target_state.path;
      m_state.path = m_target_state.path;
      m_textmap->m_state.page = m_state.page = 0;
      m_textmap->m_state.numPages = m_target_state.numPages = 0;

      printf("switchDoc path %s target %s\n", m_state.path.latin1(),
	     m_target_state.path.latin1());

      if (! switchDocument(&m_state)) { 
	m_state.clear();
	m_target_state.clear();
	printf("********* switchDocument failed\n");
	goto exit_repaint;
      }
      //      assert(m_notes.isClean());
      m_notes.setPath(notesPath(m_state.path));
      // FIXME: during load we can not allow flush()!!
      if (m_notes.load(NULL)) {
	printf("loaded notes\n");
      } else {
	printf("**** failed to load notes\n");
      }
      m_textmap->m_state.numPages = m_target_state.numPages = m_state.numPages;
      limitState(m_target_state);
      continue;
    }

    if (! isCurrentPixmap(m_target_state)
	|| m_abort) {

      if (! currentState()->isSamePage(m_target_state)) {
	clearSelection();
	switchPage(m_target_state);
      }

      m_highlight_region = QRegion();
      setSelection(QRegion(), false);
      printf("readPage start mode %d\n", m_target_state.mode);
      if ((m_target_state.mode == PDF_VIEW && ! readPage(m_target_state, &m_pixmap))
	  || (m_target_state.mode != PDF_VIEW && ! readTextPage(m_target_state, &m_textmap))
	  || m_abort) {
	m_state.page = 0;
	printf("readPage abort\n");
	if (! m_abort) {
	  m_state.clear();
	  m_target_state.clear();
	  goto exit_repaint;
	}
	continue;
      }
      printf("readPage success\n");
      if (m_target_state.mode == PDF_VIEW) {
	m_state.page = m_target_state.page;
	m_state.zoom = m_target_state.zoom;
	m_state.reproportion_text = m_target_state.reproportion_text;
	m_state.reproportion_and_render = m_target_state.reproportion_and_render;
#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
	m_state.render_text = m_target_state.render_text;
#endif
	m_target_state.width = m_state.width = m_pixmap->width();
	m_target_state.height = m_state.height = m_pixmap->height();
	m_textmap->setState(&m_state);
      } else {
	m_textmap->m_state.page = m_target_state.page;
	m_textmap->m_state.zoom = m_target_state.zoom;
	m_textmap->m_state.reproportion_text = m_target_state.reproportion_text;
	m_textmap->m_state.reproportion_and_render = m_target_state.reproportion_and_render;
#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
	m_textmap->m_state.render_text = m_target_state.render_text;
#endif
	m_target_state.width = m_state.width = m_textmap->m_state.width;
	m_target_state.height = m_state.height = m_textmap->m_state.height;
      }
      m_state.mode = m_target_state.mode;
      continue;
    }
  }

  //  limitState(m_target_state);
  m_notes.setupPage(m_textmap);
  printf("repaint done\n");

 exit_repaint:
  m_inside_read_page = false;
  emit setBusy(false);
  m_refresh_timer->stop();
  updateHighlights();
  if (m_state.mode == PDF_VIEW) {
    setSelection(QRegion(), false);
    refresh();
    setTextSelection(-1, -1, !m_state.doMove);
  }
  m_state.doMove = false;
  if (!m_notes.isClean()) {
    emit flush();
  }
  emit changeStateComplete();
  printf("repaint exit\n");
}


void
QDocView::open(QString path)
{
  printf("QDocView::open(%s)\n", path.latin1());
  if (state().path == path) {
    return;
  }
  state().clear();
  state().path = path;
  state().page = 1;
  changeState(); 
}

bool
QDocView::getTextMap(int page, QDocText *textmap_ptr)
{
  // since most of the textPage stuff is UI driven (e.g. underlining, etc.)
  // we want it immediately, and since there is no way to switch back to a stack
  //   frame under us, and come back (ie. real threads)...
  // it is not a big deal if we don't get it, since we are changing pages anyway.
  QDocState state;
  QDocText *textmap;

  if (m_inside_read_page) {
    return(false);
  }

  textmap = textmap_ptr;
  if (textmap == NULL) {
    textmap = m_textmap;
  }
  if (textmap->m_state.page == page) {
    printf("getTextMap already okay\n");
    return(true);
  }

  m_inside_read_page = true;
  m_abort = false;
  state = m_target_state;  // get correct zoom for pdf_view even if in text_view
  state.page = page;
  if (! readTextPage(state, &textmap)) {
    m_inside_read_page = false;
    textmap->m_state.page = 0;
    return(false);
  }
  m_inside_read_page = false;
  textmap->m_state = state;
  if (m_abort) {
    // we need to do a full change state
    printf("getTextMap aborted page %d\n", page);
    changeState(m_target_state);
    return(false);
  }
  printf("getTextMap success\n");
  return(true);
}
  

void 
QDocView::refresh()
{
  QPixmap *pixmap;
  updateRotate();
  if ( m_state.isRotated ) {
    pixmap = m_rotated_pixmap;
  } else {
    pixmap = m_pixmap;
  }
  if (pixmap && 
      (contentsWidth() != pixmap->width() 
       || contentsHeight() != pixmap->height())) {
    printf("change size %d %d\n", pixmap->width(), pixmap->height());
    resizeContents(pixmap->width(), pixmap->height());
    //    emit layoutPageCounter();
    viewport()->repaint();
  }
  if ( m_target_state.doMove ) {
    if (m_target_state.x == -1 && m_target_state.y == -1) {
      if (m_target_state.isRotated) {
	m_target_state.y = contentsHeight() - visibleHeight();
      } else {
	m_target_state.y = 0;
      }
      m_target_state.x = 0;
    }
    // keep trying to move to correct place
     if ((m_target_state.x + visibleWidth() <= contentsWidth() 
	 && m_target_state.y + visibleHeight() <= contentsHeight())) {
       printf("*** doing move to %d %d\n", m_target_state.x, m_target_state.y);
       setContentsPos( m_target_state.x, m_target_state.y );
       m_target_state.doMove = false;
    } else {
      printf("***  not doing move contents %d %d, move %d %d\n",
      	     contentsWidth(), contentsHeight(), m_target_state.x, m_target_state.y);
    }
  }
  // would be nice to know what we really have to redraw,
  // but pdf can draw anywhere at any time...
  updateContents(contentsX(), contentsY(),
  		 visibleWidth(),visibleHeight());
}


void 
QDocView::drawContents(QPainter *p, int clipx, int clipy, int clipw, int cliph )
{
  QPixmap *pixmap;

  if ( m_state.isRotated ) {
    pixmap = m_rotated_pixmap;
  } else {
    pixmap = m_pixmap;
  }
  if ( pixmap != NULL ) {
    p-> drawPixmap ( clipx, clipy, *pixmap, clipx, clipy, clipw, cliph );
  } else {
    p-> fillRect ( clipx, clipy, clipw, cliph, white );
  }

  QRect clip ( clipx, clipy, clipw, cliph );

  if ( ! m_selection.isEmpty() ) {
    if (m_selection.boundingRect().intersects(clip)) {
      RasterOp rop = p-> rasterOp ( );
      QArray<QRect> urects = ( m_selection & clip ). rects ( );
      p-> setRasterOp ( XorROP );
      for ( uint i = 0; i < urects. count ( ); i++ ) {
	p-> fillRect ( urects[i], Qt::white );
      }
      p-> setRasterOp ( rop );
    }
  } 

  if (m_highlight_mode && !m_highlight_region.isEmpty()) {
    if (m_highlight_region.boundingRect().intersects(clip)) {
      RasterOp rop = p-> rasterOp ( );
      QArray<QRect> urects = ( m_highlight_region & clip ). rects ( );
      p-> setRasterOp ( XorROP );
      for ( uint i = 0; i < urects. count ( ); i++ ) {
	p-> fillRect ( urects[i], Qt::blue );
      }
      p-> setRasterOp ( rop );
    }
  } 

}

#if 0
void 
QDocView::keyPressEvent ( QKeyEvent *e )
{
  //  printf("QDocView: %x ascii %x\n", e->key(), e->ascii());

  switch ( e-> key ( )) {
    // Wheel Center (select)
  case Key_Space: 
    gestureStart();
    break;
    
    // Cancel
  case Qt::Key_Escape:
    gestureAbort();
    m_mouse_click_time_window = false;
    m_mouse_debounce_time_window = false;
    setTextSelectionMode(false);
    emit topLevelKeyPressEvent(e);
    break;

  default:
    emit topLevelKeyPressEvent(e);
    break;
  }
}
#endif



QRegion
QDocView::selection ( ) const
{
  return m_selection;
}


void
QDocView::setSelection ( const QRect &new_selection, bool scrollto )
{
  setSelection(QRegion(new_selection), scrollto);
}

void 
QDocView::setSelection ( const QRegion &new_selection, bool scrollto )
{
  QRegion oldsel = m_selection;
  m_selection = new_selection;
  
  QArray<QRect> urects = ( oldsel ^  m_selection ). rects ( );
  
  // this can be slow for big regions, but since region clip doesn't work...
  for ( uint i = 0; i < urects. count ( ); i++ )
    repaintContents ( urects [i] );

  if ( scrollto ) {
    QRect bb = new_selection.boundingRect();
    QPoint center = bb.center ( );
    
    ensureVisible ( center.x(), center.y(), bb.width ( ) / 2 + 5, bb.height ( ) / 2 + 5 );

#if 0
    if (new_selection.contains(center)) {
      printf("scroll contains center\n");
      ensureVisible ( center.x(), center.y(), bb.width ( ) / 2 + 5, bb.height ( ) / 2 + 5 );
    } else {
      printf("scroll does not contains center\n");
      QArray<QRect> rects = new_selection.rects();
      QRect first;
      for(uint i = 0; i < rects.count(); i++) {
	QRect rect = rects[i];
	if (first.isEmpty()
	    || (rect.left() < first.left() && rect.top() < first.top())) {
	  first = rect;
	}
      }
      ensureVisible ( first.top(), first.left(), first.width ( ) / 2 + 5, first.height ( ) / 2 + 5 );
    }
#endif
  }
}

void
QDocView::updateHighlights()
{
  m_highlight_region = rotate(m_notes.highlights(m_state.page));
}

void
QDocView::resizeEvent( QResizeEvent *e )
{
  //  m_state.visibleWidth = visibleWidth();
  //  m_state.visibleHeight = visibleHeight();
  //  printf("resize %d x %d\n", visibleWidth(), visibleHeight());
  QScrollView::resizeEvent(e);
  QGesture::resizeEvent(e);
  //  layoutPageCounter();
}


void
QDocView::setContentsPos(QPoint p)
{
  setContentsPos(p.x(), p.y());
}

void
QDocView::setContentsPos( int x, int y )
{
  m_state.x = x;
  m_state.y = y;
  m_target_state.doMove = false;
  QScrollView::setContentsPos(x,y);
}


#if 0
// called by newPage of client subclass
// not used, but maybe should go back to this...
QPixmap *
QDocView::newPixmap(int width, int height)
{
  //  assert(! m_inside_read_page);
  delete m_pixmap;
  clearSelection();
	
  m_pixmap = new QPixmap ( width, height);

  QPDFDBG( printf ( "NEW PIXMAP (%ld x %ld)\n", width, height) );
	
  resizeContents ( m_pixmap-> width ( ), m_pixmap-> height ( ));

  m_pixmap-> fill ( white ); // clear window
  return(m_pixmap);
}
#endif


  
void QDocView::mouseClickTimeout()
{
#if 0
  QPoint contentPos( m_mouse_start + QPoint(contentsX(), contentsY()) );
  if (m_doing_text_selection) {
    if (isTextMapValid()) {
      m_text_selection_start = m_text_selection_end = m_textmap->findWord(antirotate(contentPos), true);
      setTextSelection( m_text_selection_start, m_text_selection_end, false );
      m_selection_cursor = contentPos; //??
    }
  }
#endif

  // this timer is used for two purposes:
  //   m_mouse_click_time_window
  //     the time window after a mouse press for a click has run out
  //   m_mouse_debounce_time_window
  //     and for the time when the mouse must be up after the click

  if (m_mouse_click_time_window) {
    m_mouse_click_time_window = false;
    m_mouse_debounce_time_window = false;
  } else {
    if (m_mouse_debounce_time_window) {
      m_mouse_debounce_time_window = false;
      mouseClick();
    }
  }
}


void QDocView::viewportMousePressEvent ( QMouseEvent *e )
{
  if ( e-> button ( ) == LeftButton ) {
    QPoint contentPos( e->pos() + QPoint(contentsX(), contentsY()) );
    if (gestureMousePress(e)) {

      // TODO: move this into gesture code, so it can make a zoom arrows
      //   if it is a box, or arrows from side if it is a zoom out
      m_selection_start = contentPos;
      m_selection_cursor = m_selection_start;
      setSelection(QRegion(), false); // not clearSelection, keep word selection around 

    } else {

      m_mouse_click_time_window = true;
      m_mouse_debounce_time_window = false;
      m_mouse_click_timer->start(250, true);
      
      m_mouse_start = e->pos();
      m_mouse_prev = m_mouse_start;
      m_mouse_prev_delta = QPoint(0,0);
      m_mouse_prev_dist = 0;
      
      m_mouse_down_in_selection = false;
      if (hasSelection()
	  && m_selection.contains(contentPos)) {
	m_mouse_down_in_selection = true;
	getTextSelection(&m_mouse_down_selection_start, &m_mouse_down_selection_end);
      } 
      if (m_doing_text_selection) {
	clearSelection();
	setTextSelectionMode(true);
      }
    }
    m_selection_drag = true;
  }
}

void QDocView::viewportMouseMoveEvent ( QMouseEvent *e )
{
  int dist;

  if ( e-> state ( ) & LeftButton ) {
    if (! m_selection_drag) {  // needed???
      return;
    }
    if (gestureMouseMove(e)) {
      // we are doing zoom/rotate gesture
      // move this to QGesture, and expand the visual feedback part of interface
      // would be better to show a zoom arrow across diagonal of zoom in gesture,
      //   an arrow from size for zoom out
      //   a circle for rotate
      // for now just show a rectangle
      QPoint to ( e-> pos ( ) + QPoint ( contentsX ( ), contentsY ( )));
      
      if ( to != m_selection_cursor ) {
	setSelection ( QRect ( m_selection_start, to ). normalize ( ), true );
	m_selection_cursor = to;
      }
      return;
    }
    
    if (m_doing_text_selection) {
      if ( isTextMapValid()) {
	QPoint to ( e-> pos ( ) + QPoint ( contentsX ( ), contentsY ( )));
	
	// extend word selection?
	if ( to != m_selection_cursor ) {
	  int word_cursor = m_textmap->findWord(antirotate(to), true);
	  m_selection_cursor = to;
	  if (word_cursor != -1 && word_cursor != m_text_selection_end) {
	    m_text_selection_end = word_cursor;
	    if (m_text_selection_start < 0) {
	      m_text_selection_start = word_cursor;
	    }
	    setTextSelection ( m_text_selection_start, m_text_selection_end, true );
	  } else {
	    ensureVisible ( m_selection_cursor. x ( ), m_selection_cursor. y ( ), 10, 10 );
	  }
	}
      }

    } else {
      
      QPoint delta;
      delta = m_mouse_prev - e->pos();
      dist = delta.manhattanLength();
      
      if (m_mouse_click_time_window) {
	if (dist < screenRes(10)) {
	  return;
	}
	m_mouse_click_time_window = false;
      }
      
      if (dist <= screenRes(2) && m_mouse_prev_dist <= screenRes(2)) {
	QPoint delta2 = delta - m_mouse_prev_delta;
	if (delta2.manhattanLength() <= screenRes(2)) {
	  m_mouse_prev_delta = delta;
	  m_mouse_prev_dist = dist;
	  return;
	}
      }
      m_mouse_prev_delta = delta;
      m_mouse_prev_dist = dist;
      m_mouse_prev = e->pos();
      
      if (dist > screenRes(12)) {
	delta *= 4;
      } else if (dist > screenRes(9)) {
	delta *= 2;
      } else if (dist > screenRes(5)) {
	delta += delta/2;
      }
      scrollBy(delta.x(), delta.y());
      m_target_state.doMove = false;
    }
  }
}


void 
QDocView::viewportMouseReleaseEvent ( QMouseEvent *e )
{
  if (gestureMouseRelease(e)) {
    return;
  } 

  QPoint contentPos( e->pos() + QPoint(contentsX(), contentsY()) );

  if (! isTextMapValid()) {
#if 0
    printf("QDocView::viewportMouseReleaseEvent text map invalid\n");
    printf("m_inside_read_page=%d  page=%d zoom=%f  textmap page=%d zoom=%f\n",
	   m_inside_read_page, m_state.page, (float)m_state.zoom, 
	   m_textmap->m_state.page, (float)m_textmap->m_state.zoom);
#endif
    clearSelection();
    setTextSelectionMode(false);
  }

#if 0
  if (! m_selection_drag) {  // needed??
    setSelection( QRegion(), false);
    return;
  }
  m_selection_drag = false;
#endif

  if (m_mouse_click_time_window) {
    QPoint delta = m_mouse_start - e->pos();
    if (delta.manhattanLength() < screenRes(10)) {
      // was a mouse click, now be sure it wasn't a bounce over a bump on the screen protector
      m_mouse_click_time_window = false;
      m_mouse_click_timer->stop();
      m_mouse_debounce_time_window = true;
      m_mouse_click_pos = contentPos;
      m_mouse_click_abs_pos = e->pos();
      // mouse must be down 100ms, then up for 100ms before we take it as a click
      m_mouse_click_timer->start(100, true);
      return;
    }
  }

  if (hasSelection()) {
    if (m_text_selection_start > m_text_selection_end) {
      int tmp = m_text_selection_start;
      m_text_selection_start = m_text_selection_end;
      m_text_selection_end = tmp;
    }
    
    if (m_doing_text_selection) {
      emit selectionChanged(getText(m_text_selection_start, m_text_selection_end));
    }
  }
  //  setTextSelectionMode(false);
}


void
QDocView::mouseClick()
{
  // a LITTLE paranoia never hurt anyone
  m_mouse_click_time_window = false;
  m_mouse_debounce_time_window = false;
  m_mouse_click_timer->stop();

  if (isTextMapValid()) {
    QDocNote *note;
    note = m_notes.find(m_state.page, 
			mapPixToWord(antirotate(m_mouse_click_pos), false), false);
    if (note) {
      // click inside an old note
      setTextSelectionMode(false);
      setTextSelection(note->firstWord(), note->lastWord());
      setContentsPos(rotate(m_textmap->wordBox(note->firstWord()).topLeft()));
      emit setNoteMode();
      return;
    }
    
    if (m_mouse_down_in_selection
	&& m_mouse_down_selection_start >= 0
	&& m_mouse_down_selection_end >= 0
	&& ! isBusy()) {
      QString quote = m_textmap->getText(m_mouse_down_selection_start,
					 m_mouse_down_selection_end);
      while (quote.left(1) == "\n") {
	quote.remove(0,1);
      }
      while (quote.right(1) == "\n") {
	quote.remove(quote.length()-1,1);
      }

      // click inside current text selection => make it a note
      setTextSelectionMode(false);
      m_notes.add(m_state.page, 
		  m_textmap->wordBox(m_mouse_down_selection_start).center(), 
		  m_textmap->wordBox(m_mouse_down_selection_end).center(), 
		  "[\"" + quote + "\"]\n",
		  m_textmap);
      updateHighlights();
      refresh();
      clearSelection();
      return;
    }
  }
  
  int corner = detectCorners(m_mouse_click_abs_pos);
  if (detectCorners(m_mouse_start) == corner) {
    switch(corner) {
    case CORNER_UPPER_LEFT:
      gotoPage(getPageNum()-1);
      return;
    case CORNER_UPPER_RIGHT:
      gotoPage(getPageNum()+1);
      return;
    case CORNER_LOWER_LEFT:
      emit setPageCount(getPageNum() - getPageOffset());
      return;
    default:
      int edge = detectEdges(m_mouse_click_abs_pos);
      if (detectEdges(m_mouse_start) == edge) {
	if (edge == EDGE_TOP) {
	  emit toggleFullscreen();
	  return;
	}
      }
    }
  }
  
  int prev = m_doing_text_selection;
  clearSelection();
  setTextSelectionMode(! prev);
  return;
}

#if 0
void QDocView::viewportMouseMoveEvent ( QMouseEvent *e )
{

  if ( e-> state ( ) & LeftButton ) {
    if ( m_selection_drag ) {
      QPoint to ( e-> pos ( ) + QPoint ( contentsX ( ), contentsY ( )));
      
      if ( to != m_selection_cursor ) {
	setSelection ( QRect ( m_selectionstart, to ). normalize ( ), false );
	m_selection_cursor = to;
      }
      ensureVisible ( m_selection_cursor. x ( ), m_selection_cursor. y ( ), 5, 5 );
    }
  }
}

void QDocView::viewportMouseReleaseEvent ( QMouseEvent *e )
{
  if ( e-> button ( ) == LeftButton ) {
    if ( m_selection_drag ) {
      m_selection_drag = false;
      
      setSelection ( selection ( ), false ); // emit signal
    } 
    else {
      setSelection ( QRect ( 0, 0, 0, 0 ), false );
    }		
  }
}

#endif


// 0 = not edge, 1 = top, 2 = left, 3 = right, 4 = bottom
int
QDocView::detectEdges( QPoint pos)
{
  const int edge_size = screenRes(10);

  // for other platforms (like new zaurus C700!) don't ignore qt_screen->transformed

  if (pos.y() < edge_size) {
    return(EDGE_TOP);
  }
  if (pos.y() > visibleHeight() - edge_size) {
    return(EDGE_BOTTOM);
  }
  if (pos.x() < edge_size) {
    return(EDGE_LEFT);
  }
  if (pos.x() > visibleWidth() - edge_size) {
    return(EDGE_RIGHT);
  }
  return(0);
}



// 0 = not corner, 1 = upper left, 2 = upper right, 3 = lower left, 4 = lower right
int
QDocView::detectCorners( QPoint pos ) 
{
  const int corner_size = screenRes(18);
  QPoint delta;
  QPoint upper_right(visibleWidth(),0);
  QPoint upper_left(0,0);
  QPoint lower_right(visibleWidth(),visibleHeight());

  if (m_state.isRotated) {
    upper_left.setY(visibleHeight());
    upper_right.setX(0);
    lower_right.setY(0);
  }

  delta = pos - upper_left;
  if (delta.manhattanLength() < corner_size) {
    return(CORNER_UPPER_LEFT);
  }

  delta = pos - upper_right;
  if (delta.manhattanLength() < corner_size) {
    return(CORNER_UPPER_RIGHT);
  }

  delta = pos - lower_right;
  if (delta.manhattanLength() < corner_size) {
    return(CORNER_LOWER_LEFT);
  }
  return(CORNER_LOWER_RIGHT);
}


void 
QDocView::gestureDiagonal()
{
  float zoom_factor;
  // diagonal move = zoom in

  emit saveState();

  zoom_factor = (float)visibleHeight() / (float)m_gesture_box.height();
  if ((float)visibleWidth() / (float)m_gesture_box.width() < zoom_factor) {
    zoom_factor = (float)visibleWidth() / (float)m_gesture_box.width();
  }
  // center = m_gesture_box.center();
  //  center(delta.x() + contentsX(), delta.y() + contentsY());
  state().x = (int)((contentsX() + m_gesture_box.left()) * zoom_factor);
  state().y = (int)((contentsY() + m_gesture_box.top()) * zoom_factor);
  printf("zoom_factor = %f   %d=(%d + %d) * %f, %d=(%d + %d) * %f\n",
	 zoom_factor, 
	 state().x, contentsX(), m_gesture_box.left(), zoom_factor,
	 state().y, contentsY(), m_gesture_box.top(), zoom_factor);

  state().doMove = true;
  state().zoom *= zoom_factor;
  changeState();
}

void 
QDocView::gestureHorizontal()
{
  float zoom_factor;
  int length;

  if (! state().isRotated && 
      (detectEdges(m_gesture_start) != EDGE_LEFT && detectEdges(m_gesture_start) != EDGE_RIGHT)) {
    // should move this to qgesture
    if (m_gesture_delta.x() > 0) {
      //      printf("===> forward gesture\n");
      emit forwardState();
    } else {
      //      printf("===> backward gesture\n");
      emit backwardState();
    }
    return;
  }

  emit saveState();
  length = m_gesture_delta.x();

  if (length > visibleHeight() - visibleHeight()/8) {
    zoom_factor = (float)contentsWidth() / (float)visibleWidth();
  } else {
    zoom_factor = (float)visibleWidth() / (float)(visibleWidth() + QABS(length));
  }

//  printf("zoom out horz %f (length %d)\n", zoom_factor, length);
  if (state().isRotated) {
    if (length > 0) {
      state().x = (int)(zoom_factor * (contentsX() + visibleWidth() - 1)) - visibleWidth() + 1;
    } else {
      state().x = (int)(zoom_factor * contentsX());
    }
    state().y = (int)(zoom_factor * (contentsY() + visibleHeight() - 1)) - visibleHeight() + 1;
  } else {
    if (length > 0) {
      state().x = (int)(zoom_factor * (contentsX() + visibleWidth() - 1)) - visibleWidth() + 1;
    } else {
      state().x = (int)(zoom_factor * contentsX());
    }
    state().y = (int)(zoom_factor * contentsY());
  }
  state().doMove = true;
  state().zoom *= zoom_factor;
  changeState();
}

void
QDocView::gestureVertical()
{
  float zoom_factor;
  int length;

  if (state().isRotated && 
      (detectEdges(m_gesture_start) != EDGE_TOP && detectEdges(m_gesture_start) != EDGE_BOTTOM)) {
    if (m_gesture_delta.y() > 0) {
      emit forwardState();
    } else {
      emit backwardState();
    }
    return;
  }

  emit saveState();
  length = m_gesture_delta.y();

  if (length > visibleHeight() - visibleHeight()/8) {
    zoom_factor = (float)contentsHeight() / (float)visibleHeight();
    printf("zoom out to contents height\n");
 } else {
    zoom_factor = (float)visibleHeight() / (float)(visibleHeight() + QABS(length));
  }

  if (m_state.isRotated) {
    if (length > 0) {
      state().y = (int)(zoom_factor * (contentsY() + visibleHeight() - 1)) - visibleHeight() + 1;
    } else {
      state().y = (int)(zoom_factor * contentsY());
    }
    state().x = (int)(zoom_factor * contentsX());
  } else {  
    if (length > 0) {
      state().y = (int)(zoom_factor * (contentsY() + visibleHeight() - 1)) - visibleHeight() + 1;
    } else {
      state().y = (int)(zoom_factor * contentsY());
    }
    state().x = (int)(zoom_factor * contentsX());
  }  
  state().doMove = true;
  state().zoom *= zoom_factor;
  changeState();
}

void
QDocView::gestureTouch()
{
  emit backwardState();
}

void
QDocView::gestureCheck()
{
  printf("check gesture\n");
}

void
QDocView::gestureCircle()
{
  state().isRotated = 1 - state().isRotated;
  changeState();
}





// used by TEXT_VIEW selection of note
void
QDocView::addNote(int from_char, int to_char)
{
  if (isBusy()) {
    return;
  }
  int from_word = mapCharToWord(from_char);
  int to_word = mapCharToWord(to_char);
  if (from_word >= 0 && to_word >= 0 && to_word >= from_word) {
    QString quote = m_textmap->getText(from_word, to_word);
    printf("quote = %s\n", quote.latin1());
    while (quote.left(1) == "\n") {
      printf("addNote trim starting return\n");
      quote.remove(0,1);
    }
    while (quote.right(1) == "\n") {
      printf("addNote trim ending return\n");
      quote.remove(quote.length()-1,1);
    }
    m_notes.add(m_target_state.page, 
		m_textmap->wordBoxPDF(from_word).center(), 
		m_textmap->wordBoxPDF(to_word).center(), 
		"[\"" + quote + "\"]\n",
		m_textmap);
    emit setNoteMode();
  }
}

// static function, maybe should be in qdocstate ?
QString
QDocView::notesPath(const QString &path)
{
  QString notePath;
  notePath = path;
  int i = notePath.findRev(".");
  if (i < 0 || notePath.find(".pdf",i) != i) {
    printf("Warning: filename %s does not have .pdf at end\n", notePath.latin1());
    if (i >= 0) {
      notePath = notePath.left(i-1) + ".pdf";
    }
  }
  notePath = notePath.left(notePath.length()-4) + ".xml";
  return(notePath);
}

bool
QDocView::saveNotes(QString text, int noteCharPos, 
		    QDocNote **note)
{
  bool changed;
  changed = m_notes.fromText(text, noteCharPos, note);
   // might only save on changed, but this is safer 
  //  (e.g. ran out of "disk" on previous save)
  if (!saveNotes() && note) {
    // issue here is that during saveNotes() (esp. during "fail" dialog)
    // many things could happen async that would invalidate the note
    // real fix might be to copy the note and have *note not **note
    *note = NULL;
  }
  return(changed);
}

bool
QDocView::saveNotes()
{
  // could make this so we can assert that notes are valid for this file
  if (!m_state.path.isEmpty() && m_state.numPages > 0) {
    m_state.x = contentsX();
    m_state.y = contentsY();
    printf("save state pos %d, %d\n", m_state.x, m_state.y);
    return(m_notes.save(&m_state));
  } else {
    printf("******* save notes page %d path %s\n", m_state.page,
	   m_state.path.latin1());
    //    assert(m_state.page > 0 && !m_state.path.isEmpty())
  }
  return(false);
}

QString
QDocView::getNotes(QPoint pixPos, int *noteCharPos)
{
  QDocNote *note;
  note = m_notes.find(m_target_state.page, mapPixToWord(pixPos, true), true);
  if (! note) {
    *noteCharPos = -1;
    printf("getNotes: can not find note\n");
    //    return(m_notes.toText(m_target_state.page));
    return(m_notes.toText(-1));
  }
  // could return just the notes on the current page....
  //  return(m_notes.toText(m_target_state.page, note, noteCharPos));
  return(m_notes.toText(-1, note, noteCharPos));
}

QString
QDocView::getNotes(int word, int *noteCharPos)
{
  QDocNote *note;
  note = m_notes.find(m_target_state.page, word, true);
  if (! note) {
    printf("getNotes: can not find note\n");
    //    return(m_notes.toText(m_target_state.page));
    return(m_notes.toText(-1));
  }
  // could return just the notes on the current page....
  //  return(m_notes.toText(m_target_state.page, note, noteCharPos));
  return(m_notes.toText(-1, note, noteCharPos));
}

QString
QDocView::getNotes(int firstWord, int lastWord, int *noteCharPos)
{
  QDocNote *note;
  note = m_notes.find(m_target_state.page, firstWord, lastWord);
  if (! note) {
    printf("getNotes: can not find note\n");
    //    return(m_notes.toText(m_target_state.page));
    return QString();
  }
  // could return just the notes on the current page....
  //  return(m_notes.toText(m_target_state.page, note, noteCharPos));
  return(m_notes.toText(-1, note, noteCharPos));
}
