#include "aconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <iostream>
#include <assert.h>

#include <qgfx_qws.h>
#include <qpixmap.h>
#include <qimage.h>
#include <qpainter.h>
#include <qdict.h>
#include <qtimer.h>
#include <qlabel.h>
#include <qscrollview.h>
#include <qmessagebox.h>
#include <qapplication.h>
#include <qdatetime.h>
#include <qimage.h>

#include "qbusybar.h"
#include "qdocstate.h"
#include "qdoctext.h"
#include "qgesture.h"
#include "qdocnotes.h"
#include "qpdfconfig.h"


const int MAX_STATE_HISTORY = 20;
  

class QDocView : public QGesture {
  Q_OBJECT

public:
  QDocView(QWidget *paren, const char *name, int flags, QPDFconfig *config);
  virtual ~QDocView();
  
  QDocState *currentState();

  void setPage(int page);
  int getPageNum();
  int lastPageNum();

  int getPageOffset();
  void setPageOffset(int);

  void setRotation(bool rotated);
  bool isRotated();

  void setZoom(float zoom);
  float getZoom();

  void setReProportionText(bool mode);
  bool getReProportionText();
  void setReProportionAndRender(bool mode);
  bool getReProportionAndRender();

#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
  void setRenderText(bool mode);
  bool getRenderText();
#endif

  QString getFilename();
  QString getPath();
  void setPath(QString path);
  void setMode(ViewMode mode);
  void open(QString path);

  static QString notesPath(const QString &pdfFilename);

  void setPixPosition(QPoint);
  void setWordPosition(int);
  void setCharPosition(int);
  void setCharIndex(int);

  // current X,Y of upper left corner in coord of QScrollView content
  // (takes into account screen rotation if needed)
  int mapContentsX();  
  int mapContentsY();
  QPoint mapScreen(QPoint p);
  QPoint currentUpperLeftPix();
  QPoint currentCenterPix();

  QPoint rotate(QPoint p);
  QPoint antirotate(QPoint p);
  QRect rotate(QRect r);
  QRegion rotate(QRegion r);

  void clearSelection();
  bool hasSelection();
  QRegion selection ( ) const;
  void setSelection ( const QRect &r, bool scrollto = false );
  void setSelection ( const QRegion &r, bool scrollto = false );
  void setTextSelection(int first_word = -1, int last_word = -1, bool scroll = true);
  void getTextSelection(int *first_word, int *last_word);
  QString selectedText();

  void highlight(int from_word, int to_word, int thickness = 1, QColor color = Qt::white);
  void updateHighlights();

  bool findText(const QString &string, int *from_char, int *to_char, int *page,
		bool reverse, bool multipage);

  bool findTextInternal(QString text, int *from_char, int *to_char, 
			int first_page = 0, int last_page = -1,
			int *page = NULL);

  int mapPixToWord(QPoint pix, bool closest = true, bool lower_right = false); // return -1 if textmap not valid
  QPoint mapWordToPix(int word, bool lastChar = false);
  int mapWordToChar(int, bool lastChar = false);
  int mapCharToWord(int);
  QPoint absolutePosition(QPoint p);
  
  int numWords();
  QString getText(int from_word = 0, int to_word = -1, int page = -1);
  QRegion getTextRegion(int from_word = 0, int to_word = -1);

  //  QDocNote *getNote(int page, int note);
  //  void addNote(QDocNote note);

  void changeState();
  void changeState(QDocState new_state, int refresh_rate = -1);
  bool isValid();
  bool isBusy() { return(m_inside_read_page); }
  void setBusyFlag(bool busy) { m_inside_read_page = busy; }

  QDocState &state() { return(m_target_state); }

  // various screen size related variables
  inline int screenRes(int x) { return(m_config->c7xx ? (x<<1) : x); }

  // measured with a ruler... no doubt not exact
  int dpi() { return(m_config->c7xx ? 213 : 115); }

  int refreshRate(bool rotated) {
    if (rotated) {
      return(m_config->c7xx? 2000 : 1100);
    }
    return(m_config->c7xx? 800 : 600);
  }

  QPixmap scaleIcon(QPixmap pixmap) {
    if (m_config->c7xx) {
      QImage image(pixmap.convertToImage());
      QPixmap bigpixmap;
      image = image.smoothScale(pixmap.width() + pixmap.width()/3, 
				pixmap.height() + pixmap.height()/3);
      bigpixmap.convertFromImage(image);
      return(bigpixmap);
    }
    return(pixmap);
  }
    

  bool saveNotes(QString text,
		 int noteCharPos, 
		 QDocNote **note);
  bool saveNotes();
  QString getNotes(QPoint pixPos, int *noteCharPos);
  QString getNotes(int word, int *noteCharPos);  // not used
  QString getNotes(int fromWord, int toWord, int *noteCharPos);

  // should be less public
  bool getTextMap(int page, QDocText *textmap_ptr = NULL);

  QDocNotes m_notes;
  QDocText *m_textmap;
  QDocState m_target_state;

signals:
  void selectionChanged ( const QString &str );
  void toggleFullscreen();
  void setNoteMode();
  void updateMenuItems();
  void setBusy(bool);
  void setPageCount(int page, int time = 4000);
  void saveState();
  void forwardState();
  void backwardState();
  void changeStateComplete();
  void flush();
  
public slots:
  void gotoPage(int page);
  void gotoPage(int page, QPoint pos);

  void setContentsPos(QPoint p);
  virtual void setContentsPos( int x, int y );

  void addNote(int from_char, int to_char); 

  void setHighlightMode(bool mode);
  bool highlightMode();

  void setTextSelectionMode(bool mode);
  bool textSelectionMode();
  bool doingSelection() { return(m_doing_text_selection); }


private slots:
  virtual void refresh();
  void mouseClickTimeout();

protected:
  // this is the interface that needs to be implemented
  // by subclasses of QDocView
  virtual bool switchDocument(QDocState* target) = 0;
  virtual void switchPage(QDocState target) = 0;
  virtual bool readPage(QDocState target, QPixmap **pixmap_ptr) = 0;
  virtual bool readTextPage(QDocState target, QDocText **textmap_ptr) = 0;
  virtual void abortReadPage() {};
  virtual void limitState(QDocState &state) = 0;
  virtual bool foundRenderedText() = 0;

  // Qt user interface
  //  virtual void keyPressEvent ( QKeyEvent *e );
  virtual void viewportMousePressEvent ( QMouseEvent *e );
  virtual void viewportMouseMoveEvent ( QMouseEvent *e ); 
  virtual void viewportMouseReleaseEvent ( QMouseEvent *e );
  virtual void drawContents ( QPainter *p, int, int, int, int );
  virtual void resizeEvent( QResizeEvent *e );

  
  // Gesture user interfrace
  virtual void gestureCircle();
  virtual void gestureDiagonal();
  virtual void gestureHorizontal();
  virtual void gestureVertical();
  virtual void gestureCheck();
  virtual void gestureTouch();
  //  virtual void gestureBackForth();


  // state (display rotation, position, zoom, etc.) change
  void autoSenseRenderText();

  // could try to push these back into private by rearranging functions
  // from QPdfView => QDocView
  bool m_inside_read_page;

  QDocState m_state;  // should only change in changeState

  QPDFconfig *m_config;

private:

  // changeState variables
  bool m_abort;
  QTimer *m_refresh_timer;

  // selections
  QRegion m_selection;
  QPoint  m_selection_start;
  QPoint  m_selection_cursor;
  bool    m_selection_drag;

  bool m_doing_text_selection;
  int m_text_selection_start;
  int m_text_selection_end;

  // for HAND_MODE, i.e. scrolling around
  QPoint  m_mouse_start;
  QPoint  m_mouse_prev;
  QPoint  m_mouse_prev_delta;
  int     m_mouse_prev_dist;

  bool m_mouse_click_time_window;
  bool m_mouse_debounce_time_window;
  QPoint m_mouse_click_pos;
  QPoint m_mouse_click_abs_pos;
  QTimer *m_mouse_click_timer;


  bool m_mouse_down_in_selection;
  int m_mouse_down_selection_start;
  int m_mouse_down_selection_end;
  void mouseClick();

  // for highlighting of notes
  bool m_highlight_mode;
  QRegion m_highlight_region;

  int detectEdges(QPoint pos);
  int detectCorners(QPoint pos);

  void updateRotate();
  void rotateView(int width, int height, bool anti_rotate);
  QPixmap *rotatePixmap(QPixmap *pixmap, bool anti_rotate);

  QPixmap *m_pixmap;
  QPixmap *m_rotated_pixmap;

  bool isCurrentPixmap(QDocState &s);

  bool isTextValid() { return(!m_inside_read_page 
			      && m_textmap->isSamePage(m_state)  ); }
  bool isTextValid(QDocState &state) { return(!m_inside_read_page
					      && m_textmap->isSamePage(state)); }
  bool isTextMapValid() { return(!m_inside_read_page
				 && m_textmap->isSamePixmap(m_state)); }
};
