#ifdef __GNUC__
#pragma implementation
#endif

#if 0
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <iostream>

#include <qgfx_qws.h>
#include <qpixmap.h>
#include <qimage.h>
#include <qpainter.h>
#include <qdict.h>
#include <qtimer.h>
#include <qlabel.h>
#include <qscrollview.h>
#include <qapplication.h>

#include "qbusybar.h"
#endif

#include "qgesture.h"


/*
TODO:

add left/right, right/left, up/down, and down/up movement
check

*/


QGesture::QGesture( QWidget *parent, const char *name, int flags) 
  : QScrollView ( parent, name, WRepaintNoErase | WResizeNoErase | flags )
{
  m_doing_gesture = false;
  m_gesture_started = false;

  m_gesture_timer = new QTimer(this);
  connect(m_gesture_timer, SIGNAL(timeout()),
	  this, SLOT(gestureTimerEvent()));
  
  float dpi = 115;
  if (qApp->desktop()->width() > 600 || qApp->desktop()->height() > 600) {
    dpi = 213;
  }
  const float box_size = 0.25;  // 1/4 inch
  m_min_box_size = (int)(box_size * dpi);
}

bool
QGesture::doingGesture()
{
  return(m_doing_gesture);
}

void QGesture::gestureStart()
{
  m_doing_gesture = true;
  m_gesture_started = false;
}

void QGesture::gestureAbort()
{
  m_doing_gesture = false;
  m_gesture_timer->stop();
}

void QGesture::gestureTimerEvent()
{
  if (m_doing_gesture) {
    gestureEnd();
  }
}

bool
QGesture::gestureMousePress(QMouseEvent *e)
{
  if (! m_doing_gesture) {
    return(false);
  }
  if (! m_gesture_started) {
    m_gesture_started = true;
    m_gesture_start = e->pos();
    m_gesture_box = QRect(m_gesture_start, m_gesture_start);
  }
  m_gesture_timer->stop();
  return(true);
}


bool
QGesture::gestureMouseMove(QMouseEvent *e)
{
  if (! m_doing_gesture) {
    return(false);
  }
  m_gesture_box.normalize();
  m_gesture_box |= QRect(e->pos(), e->pos());
  return(true);
}


bool
QGesture::gestureMouseRelease(QMouseEvent *e)
{
  if (! m_doing_gesture) {
    return(false);
  }
  m_gesture_end = e->pos();
  m_gesture_box.normalize();
  m_gesture_box |= QRect(m_gesture_end, m_gesture_end);
  m_gesture_timer->start(150,TRUE);
  return(true);
}

void QGesture::gestureEnd()
{
  m_doing_gesture = false;
  m_gesture_timer->stop();

  m_gesture_box.normalize();

  m_gesture_delta = m_gesture_end - m_gesture_start;
  m_gesture_length = m_gesture_delta.manhattanLength();

#if 1
  printf("length %d delta (%d,%d)  box (%d,%d) (%d,%d) = %d x %d\n",
	 m_gesture_length, m_gesture_delta.x(), m_gesture_delta.y(), 
	 m_gesture_box.left(), m_gesture_box.top(),
	 m_gesture_box.bottom(), m_gesture_box.right(),
	 m_gesture_box.width(), m_gesture_box.height());
#endif

  if (m_gesture_box.width() > m_min_box_size 
      && m_gesture_box.height() > m_min_box_size) {

    // vaguely circular = rotate
    if (m_gesture_length < QMIN(m_gesture_box.width(), m_gesture_box.height())) {
      printf("circle\n");
      gestureCircle();
      return;
    } 

    printf("diagonal\n");
    gestureDiagonal();
    return;
  }

  // horizontal or vert move = zoom out
  if (m_gesture_box.width() > m_min_box_size) {
    printf("horzontal\n");
    gestureHorizontal();
    return;
  }
  if (m_gesture_box.height() > m_min_box_size) {
    printf("vertical\n");
    gestureVertical();
    return;
  }

  printf("touch\n");
  gestureTouch();
}
