#define ASYNC
#define QPDFFILESELECTOR

#define QPDF_VERSION  "QPDF2 2.2.1 (02/17/04) pkohn@mail.nih.gov"

//========================================================================
//
// qpdf.cc
//
// Copyright 2003-4, Philip Kohn
// Copyright 2001 Robert Griebl
//
//========================================================================


#include "qpdf.h"
#include <qpe/qpemenubar.h>
#include <qdir.h>

#ifdef AUTOTEST
#define MAX_ZOOM 1.2
#else
#define assert(X)
#endif

void sig_handler(int x)
{
  printf("got signal %d\n",x);
}

QPDFWindow *main_window;

void finish(int)
{
  main_window->flush();
  exit(-1);
}

int qpdf_debug_level = 0;

int main ( int argc, char **argv ) 
{
	QPEApplication app ( argc, argv );

	if (argc >= 1 
	    && strcmp(argv[1],"-d") == 0) {
	  qpdf_debug_level = atoi(argv[2]);
	}

	// read config file
	globalParams = new GlobalParams ( "" );
	globalParams-> setErrQuiet ( true );


	printf(QPDF_VERSION);
	printf("\nsend bug reports to pkohn@mail.nih.gov\n");

	main_window = new QPDFWindow();
	app. showMainDocumentWidget ( main_window );

	QObject::connect( &app, SIGNAL( flush() ), main_window, SLOT( flush() ) );
	//	QObject::connect( &app, SIGNAL( reload() ), &main_window, SLOT( reload() ) );

#if 0
	for(int i=1; i<=31; i++) {
	  if (i != SIGPIPE && i != SIGPROF && i != SIGSEGV 
	      && i != SIGINT && i != SIGILL && i != SIGTERM
	      && i != SIGBUS) 
	    signal(i,sig_handler);
	}
#endif
	signal(SIGTERM, finish);
	
	signal(SIGSTOP, sig_handler);
	signal(SIGCONT, sig_handler);
 	signal(SIGTSTP, sig_handler);

	if (( app. argc ( ) == 3 ) && ( app. argv ( ) [1] == QCString ( "-f" ))) {
	  main_window->openFile ( app. argv ( ) [2] );
	} else {
	  main_window->startup();
	}
	return app. exec ( );
}


QPDFWindow::QPDFWindow ( ) : QMainWindow (0, "qpdf", 0)
{
	setCaption ( QPDF_VERSION );
	setIcon ( Resource::loadPixmap ( "qpdf/qpdf_icon" ));
	m_fullscreen = false;
		
	setToolBarsMovable ( false );

	m_stack = new QWidgetStack ( this );
	setCentralWidget ( m_stack );
	//	m_stack->setActiveWindow();
	m_stack-> setSizePolicy ( QSizePolicy ( QSizePolicy::Expanding, QSizePolicy::Expanding ));

	m_config = new QPDFconfig(m_stack);
	m_config->read();
	m_stack->addWidget(m_config, PREFERENCES_PAGE);

	m_pdfview = new QPDFView ( m_stack, "qpdfview", 0, m_config );

#if 0
	QFont font;
	font.setPointSize(m_config->default_font_size);
	font.setStyleHint(QFont::SansSerif, QFont::PreferQuality);
	qApp->setFont(font, true);
	//	m_pdfview->setFont(font);
	//	m_stack->setFont(font);
	//	setFont(font);
#endif

	QFont font(qApp->font());
	font.setPointSize(m_config->default_font_size);
	if (m_config->default_font_name != "") {
	  font.setFamily(m_config->default_font_name);
	}
	qApp->setFont(font, true);

	m_stack->addWidget( m_pdfview, PDF_VIEW);

	m_pdfview->setHScrollBarMode ( QScrollView::AlwaysOff );
	m_pdfview->setVScrollBarMode ( QScrollView::AlwaysOff );

	m_doc_browser = NULL;
	m_dir_browser = NULL;

	m_textedit = new TextEdit(m_stack, "textedit", 0);
	m_textedit->setReadOnly(true);
	m_textedit->setEdited(false);
	m_stack->addWidget( m_textedit, TEXT_VIEW);

#if 0
	m_textedit->setFontSize(m_config->default_font_size);
	if (m_config->default_font_name != "") {
	  m_textedit->setFontFamily(m_config->default_font_name);
	}
#endif

	//	m_tb_menu = new QToolBar ( this );
	m_tb_menu = new QToolBar ( this );
	m_tb_menu-> setHorizontalStretchable ( true );

	QMenuBar *mb = new QMenuBar ( m_tb_menu );

	
	// file menu
	m_pm_file = new QPopupMenu ( mb, "file" );

	// recent file list
	m_pm_recent_files = new QPopupMenu(m_pm_file, "recent");
	int id;
	for(int i = 0; i<m_config->recent_files.count(); i++) {
	  QString filename;
	  int slash;
	  filename = m_config->recent_files[i];
	  slash = filename.findRev('/');
	  if (slash > 0) {
	    filename = filename.mid(slash+1);
	  }
	  id = m_pm_recent_files->insertItem( filename );
	  m_pm_recent_files->setWhatsThis(id, m_config->recent_files[i] );
	  m_pm_recent_files->connectItem(id, this, SLOT(openRecentFile(int)));
	  m_pm_recent_files->setItemParameter(id, id);
	}

	m_pm_priority = new QPopupMenu( m_pm_file, "priority" );
	for(int i = 0; i <= 10; i++) {
	  m_pm_priority->insertItem( QString::number(i), i );
	  m_pm_priority->connectItem(i, this, SLOT(changePriority(int)));
	}

	m_pm_categories = new QPopupMenu( m_pm_file, "category" );
	m_pm_categories->insertItem( tr("New Category"), 10000 );
	m_pm_categories->connectItem(10000, this, SLOT(newCategory()));
	for(int i = 0; i < m_config->category_list.count(); i++) {
	  m_pm_categories->insertItem( m_config->category_list[i], i);
	  m_pm_categories->connectItem(i, this, SLOT(changeCategory(int)));
	}


	mb-> insertItem ( tr("File"), m_pm_file );
	m_pm_file-> insertItem ( Resource::loadPixmap ( "qpdf/qpdf_icon" ), tr("Document Browser"), this, 
				 SLOT(openDocumentBrowser()), CTRL+Key_O);
	m_pm_file-> insertItem ( Resource::loadIconSet ( "folder_open" ), tr("Directory Browser"), this, 
				 SLOT(openDirectoryBrowser()), CTRL+Key_D);
	m_pm_file-> insertItem ( tr("Recent"), m_pm_recent_files);
	m_pm_load_directory_notes = m_pm_file-> insertItem ( tr("Load Directory Notes"), this, SLOT(loadDirectoryNotes()));
	m_pm_file-> insertItem ( tr("Rename"), this, SLOT(openRenameBar()), CTRL+Key_R);
	//	m_pm_file-> insertSeparator();
	m_pm_file-> insertItem ( tr("Priority"), m_pm_priority);
	m_pm_file-> insertItem ( tr("Categories"), m_pm_categories);
	//	m_pm_file-> insertSeparator();
	m_pm_file-> insertItem ( tr("Help"), this,  SLOT(openHelp()) );

	// edit menu

	m_pm_edit = new QPopupMenu ( mb, "edit" );
	mb-> insertItem ( tr( "Edit" ), m_pm_edit );

	m_pm_edit-> insertItem ( tr( "Copy" ),  m_textedit, 
				 SLOT(editCopy()), CTRL+Key_C, PM_EDIT_COPY);
	m_pm_edit-> insertItem ( tr( "Cut" ),  m_textedit, 
				 SLOT(editCut()), CTRL+Key_X, PM_EDIT_CUT);
	m_pm_edit-> insertItem ( tr( "Paste" ),  m_textedit, 
				 SLOT(editPaste()), CTRL+Key_V, PM_EDIT_PASTE);

	m_pm_edit-> insertSeparator();
	m_pm_view_find = m_pm_edit-> insertItem ( tr( "Find" ),  this, 
						  SLOT(toggleFindBar()));


	m_pm_edit-> insertSeparator();

	m_pm_edit-> insertItem ( tr("Preferences"), this, 
				 SLOT(openPreferences()) );

	// view menu
	m_pm_view = new QPopupMenu ( mb, "view" );
	m_pm_view-> setCheckable ( true );
	mb-> insertItem ( tr( "View" ), m_pm_view );

#if 0
	m_pm_view-> insertItem ( tr( "PDF" ),  this, 
				 SLOT(setPDFMode()), CTRL+Key_P, PM_VIEW_PDF);
	m_pm_view-> insertItem ( tr( "Text" ),  this, 
				 SLOT(setTextMode()), CTRL+Key_T, PM_VIEW_TEXT);
	m_pm_view-> insertItem ( tr( "Notes" ),  this, 
				 SLOT(setNoteMode()), CTRL+Key_N, PM_VIEW_NOTES);
	m_pm_view-> insertSeparator ( );
#endif

	m_pm_view_rotate = m_pm_view-> insertItem ( Resource::loadIconSet ( "repeat" ), tr( "Rotate" ),  this, 
						    SLOT(toggleRotate()), CTRL+Key_R);

	m_pm_view_fullscreen = m_pm_view-> insertItem ( Resource::loadIconSet ( "fullscreen" ), tr( "Full Screen" ),  this, 
							SLOT(toggleFullscreen()), CTRL+Key_F);


	m_pm_view_highlight = m_pm_view-> insertItem ( tr( "Highlight" ),  this, 
						       SLOT(toggleHighlight()));
	m_pm_view-> setItemChecked(m_pm_view_highlight, true);


	m_pm_view_reproportion_text = m_pm_view-> insertItem ( tr( "Adjust Spacing" ),  
							       this, 
							       SLOT(toggleReProportionText()));
	m_pm_view-> setItemChecked ( m_pm_view_reproportion_text, 
				     m_pdfview->getReProportionText());


#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
#if 0
	m_pm_view_reproportion_and_render = m_pm_view-> insertItem ( tr( "More Adjusted" ),  
							       this, 
							       SLOT(toggleReProportionAndRender()));
	m_pm_view-> setItemChecked ( m_pm_view_reproportion_and_render, 
				     m_pdfview->getReProportionAndRender());
#endif

	m_pm_view_render_text = m_pm_view-> insertItem ( tr( "Embedded Fonts" ),  
							       this, 
							       SLOT(toggleRenderText()));
	m_pm_view-> setItemChecked ( m_pm_view_render_text, 
				     m_pdfview->getRenderText());
#endif


	m_pm_view-> insertSeparator ( );

	m_pm_view_zoomin = m_pm_view-> insertItem ( tr( "Zoom In" ), 
						    this, SLOT(setZoom(int)), Key_Plus);
	m_pm_view-> setItemParameter(m_pm_view_zoomin, 1);

	m_pm_view_zoomout = m_pm_view-> insertItem ( tr( "Zoom Out" ), 
						     this, SLOT(setZoom(int)), Key_Minus);
	m_pm_view-> setItemParameter(m_pm_view_zoomout, -1);

	id = m_pm_view-> insertItem ( tr( "Fit to width" ), this, SLOT(setZoom(int)));
	m_pm_view-> setItemParameter(id, PM_VIEW_FIT_TO_WIDTH);

	id = m_pm_view-> insertItem ( tr( "Fit to page" ), this, SLOT(setZoom(int)));
	m_pm_view-> setItemParameter(id, PM_VIEW_FIT_TO_PAGE);

	// 12 per menu
	m_pm_zoom1 = new QPopupMenu(m_pm_view, "zoom1");
	for(int zooms = 10; zooms < 100; zooms += 10) {
	  id = m_pm_zoom1-> insertItem ( tr( "%1%" ).arg(zooms), this, SLOT(setZoom(int)));
	  m_pm_zoom1-> setItemParameter(id, zooms);
	}
	m_pm_zoom1_id = m_pm_view-> insertItem ( tr( "Zoom < 100%" ), m_pm_zoom1);
	m_pm_zoom2 = new QPopupMenu(m_pm_view, "zoom2");
	for(int zooms = 100; zooms <= 200; zooms += 10) {
	  id = m_pm_zoom2-> insertItem ( tr( "%1%" ).arg(zooms), this, SLOT(setZoom(int)));
	  m_pm_zoom2-> setItemParameter(id, zooms);
	}
	m_pm_zoom2_id = m_pm_view-> insertItem ( tr( "Zoom >= 100%" ), m_pm_zoom2);

	// go menu

	m_pm_page = new QPopupMenu ( mb, "page menu" );
	mb-> insertItem ( tr( "Page" ), m_pm_page );

#if 0
	m_pm_page-> insertItem ( Resource::loadIconSet ( "forward" ), tr( "next" ),  this, 
				 SLOT(nextPage()),
				 m_config->c7xx? Key_Period : Key_Apostrophe, PM_PAGE_NEXT);
	m_pm_page-> insertItem ( Resource::loadIconSet ( "back" ), tr( "previous" ),  this, 
				 SLOT(prevPage()),
				 m_config->c7xx? Key_Comma : Key_Slash, PM_PAGE_PREV);
#endif

	m_pm_page-> insertItem ( Resource::loadIconSet ( "fastforward" ), tr( "last" ),  this, 
				 SLOT(lastPage()),
				 Key_Period, PM_PAGE_LAST);
	m_pm_page-> insertItem ( Resource::loadIconSet ( "fastback" ), tr( "first" ),  this, 
				 SLOT(firstPage()),
				 Key_Tab, PM_PAGE_FIRST);

	m_pm_page-> insertSeparator ( );

	m_pm_page_forward = m_pm_page-> insertItem ( tr( "Forward" ),  this,
						     SLOT(forwardState()), Key_Greater);
	m_pm_page_backward = m_pm_page-> insertItem ( tr( "Backward" ), this,
						      SLOT(backwardState()), Key_Less);

	m_pm_page-> setItemEnabled(m_pm_page_forward, false);
	m_pm_page-> setItemEnabled(m_pm_page_backward, false);


	m_pm_page-> insertSeparator ( );

	m_pm_page-> insertItem ( tr( "Set Page 1" ), this, 
				 SLOT(setPageOffset()));

	m_pm_page-> insertSeparator ( );


	m_pm_page_start_pagenum_index = 
	  m_pm_page->indexOf(m_pm_page-> insertSeparator ( ));



	m_tb_menu-> addSeparator();

	// change mode buttons


	m_pdf_move_pixmap = m_pdfview->scaleIcon(Resource::loadPixmap("qpdf/pdf_move_icon"));
	m_pdf_text_select_pixmap = m_pdfview->scaleIcon(Resource::loadPixmap("qpdf/pdf_text_icon"));

	m_but_pdf = new QToolButton ( m_pdfview->scaleIcon(Resource::loadPixmap("qpdf/pdf_icon")), 
				      tr( "PDF" ),  
				      tr( "PDF" ),  this,
				      SLOT(setPDFMode()), m_tb_menu, tr( "PDF" ));
	// this one doesn't work!!
	m_but_pdf->setAccel(CTRL+Key_P);
	
	m_but_text = new QToolButton ( m_pdfview->scaleIcon(Resource::loadPixmap("qpdf/text_icon")), 
				       tr( "Text" ),  
				       tr( "Text" ),  this,
				       SLOT(setTextMode()), m_tb_menu, tr( "Text" ));
	m_but_text->setAccel(CTRL+Key_T);

	m_but_note = new QToolButton ( m_pdfview->scaleIcon(Resource::loadPixmap("qpdf/note_icon")), 
				       tr( "Notes" ),  
				       tr( "Notes" ),  this,
				       SLOT(setNoteMode()), m_tb_menu, tr( "Notes" ));
	m_but_note->setAccel(CTRL+Key_N);


	// arrows
	
	m_but_prev_page = new QToolButton ( LeftArrow, m_tb_menu, "prev page");
	connect(m_but_prev_page, SIGNAL( pressed() ), this, SLOT( prevPage() ) );

	m_but_next_page = new QToolButton ( RightArrow, m_tb_menu, "next page");
	connect(m_but_next_page, SIGNAL( pressed() ), this, SLOT( nextPage() ) );


#if 0
	new QToolButton ( Resource::loadIconSet ( "back" ), 
			  tr( "Previous page" ), 
			  QString::null, this, 
			  SLOT( prevPage ( )),  m_tb_menu, "last" );

	new QToolButton ( Resource::loadIconSet ( "forward" ),     
			  tr( "Next page" ), 
			  QString::null, this, 
			  SLOT( nextPage ( )),  m_tb_menu, "next" );
#endif

	
	// Search toolbar
	m_tb_find = new QToolBar ( this );
	addToolBar ( m_tb_find, "Search", QMainWindow::Top, true );
	m_tb_find-> setHorizontalStretchable ( true );
	m_tb_find-> hide ( );

	m_find_line_edit = new QLineEdit ( m_tb_find, "findedit" );
	m_tb_find-> setStretchableWidget ( m_find_line_edit );
	connect ( m_find_line_edit, SIGNAL( textChanged ( const QString & )), this, 
		  SLOT( findTextInPage ( const QString & )));
	//	connect ( m_find_line_edit, SIGNAL( returnPressed() ), this, 
	//		  SLOT( search_again() ) );

#if 0
	new QToolButton ( Resource::loadIconSet ( "next" ),        
			  tr( "Next" ), 
			  QString::null, this, 
			  SLOT( findTextForward ( )),  m_tb_find, "findnext" );
	new QToolButton ( Resource::loadIconSet ( "back" ),        
			  tr( "Back" ), 
			  QString::null, this, 
			  SLOT( findTextBackward ( )),  m_tb_find, "findprev" );
#endif
	new QToolButton ( Resource::loadIconSet ( "down" ),        
			  tr( "PrevPage" ), 
			  QString::null, this, 
			  SLOT( findTextMultipage ( )),  m_tb_find, "findpagenext" );
	new QToolButton ( Resource::loadIconSet ( "up" ),        
			  tr( "NextPage" ), 
			  QString::null, this, 
			  SLOT( findTextBackwardMultipage ( )),  m_tb_find, "findpageprev" );
	new QToolButton ( Resource::loadIconSet ( "close" ),        
			  tr( "Close" ), 
			  QString::null, this, 
			  SLOT( closeFindBar ()),  m_tb_find, "findclose" );


	// Rename toolbar
	m_tb_rename = new QToolBar ( this );
	addToolBar ( m_tb_rename, tr("Rename"), QMainWindow::Top, true );
	m_tb_rename-> setHorizontalStretchable ( true );
	m_tb_rename-> hide ( );
	m_rename_line_edit = new QLineEdit ( m_tb_rename, "rename" );
	m_tb_rename-> setStretchableWidget ( m_rename_line_edit );
	new QToolButton ( Resource::loadIconSet ( "done" ),        
			  tr( "Rename" ), 
			  QString::null, this, 
			  SLOT( closeRenameBar ()),  m_tb_rename, "rename" );
	new QToolButton ( Resource::loadIconSet ( "close" ),        
			  tr( "Rename" ), 
			  QString::null, this, 
			  SLOT( abortRenameBar ()),  m_tb_rename, "rename" );


	m_please_wait = new QLabel(this);
	m_please_wait-> setAlignment ( AlignCenter | SingleLine );
	m_please_wait-> hide();
	m_please_wait-> setSizePolicy ( QSizePolicy ( QSizePolicy::Expanding, QSizePolicy::Expanding ));
	m_please_wait-> setAutoResize(true);
	m_please_wait-> setFrameStyle(QFrame::PopupPanel | QFrame::Raised);
	m_please_wait-> setText(tr("Reading directories, please wait..."));
	
	connect ( m_pdfview, SIGNAL( selectionChanged ( const QString & )), this, 
		  SLOT( selectionChanged ( const QString & )));

	connect( m_pdfview, SIGNAL( toggleFullscreen()), 
		 this, SLOT(toggleFullscreen()));

	//	connect (m_pdfview, SIGNAL( newDocument(int) ),
	//		 this, SLOT(newDocument(int)));

	connect (m_pdfview, SIGNAL( setPageCount(int,int) ),
		 this, SLOT(setPageCount(int,int)));

	connect (m_pdfview, SIGNAL( flush() ),
		 this, SLOT( flush() ));

	connect (m_pdfview, SIGNAL( setBusy(bool) ),
		 this, SLOT(setBusy(bool)));

	connect (m_pdfview, SIGNAL( updateMenuItems() ),
		 this, SLOT(updateMenuItems()));
	connect (m_pdfview, SIGNAL( changeStateComplete() ),
		 this, SLOT(changeStateComplete()));

	connect (m_pdfview, SIGNAL( setNoteMode() ),
		 this, SLOT(setNoteMode()));

	connect (m_pdfview, SIGNAL( saveState() ),
		 this, SLOT(saveState()));
	connect (m_pdfview, SIGNAL( forwardState() ),
		 this, SLOT(forwardState()));
	connect (m_pdfview, SIGNAL( backwardState() ),
		 this, SLOT(backwardState()));

	connect (m_textedit, SIGNAL( addNote(int,int) ),
		 m_pdfview, SLOT(addNote(int,int)));

	connect (m_textedit, SIGNAL( nextPage() ),
		 this, SLOT(nextPage()));

	connect (m_textedit, SIGNAL( prevPage() ),
		 this, SLOT(prevPage()));

	connect (m_textedit, SIGNAL( toggleFullscreen() ),
		 this, SLOT(toggleFullscreen()));

	m_flash_timer = new QTimer(this);
	connect(m_flash_timer, SIGNAL(timeout()),
		this, SLOT(flashFindBarTimeout()));

	m_cancel_timer = new QTimer(this);
	connect(m_cancel_timer, SIGNAL(timeout()),
		this, SLOT(cancelTimeout()));

	m_cancel_message = new QLabel(this);
	m_cancel_message-> setAlignment ( AlignCenter | SingleLine );
	m_cancel_message-> hide();
	m_cancel_message-> setSizePolicy ( QSizePolicy ( QSizePolicy::Expanding, QSizePolicy::Expanding ));
	m_cancel_message-> setAutoResize(true);
	m_cancel_message-> setFrameStyle(QFrame::PopupPanel | QFrame::Raised);
	
	m_cancel_message-> setText(tr("Press CANCEL again to exit"));

	m_counter = new QLabel ( this );
	m_counter-> setAlignment ( AlignCenter | SingleLine );
	m_counter-> hide();
	
	m_isbusy = false;
	m_busybar = new QBusyBar ( this );
	m_busybar-> setParameters ( 12, 8, 200 );
	m_busybar-> hide ( );

	m_mode = PDF_VIEW;
	m_pdfview->setMode(m_mode);

	m_pageCounterTimer = new QTimer(this);
	connect(m_pageCounterTimer, SIGNAL(timeout()),
		this, SLOT(pageCounterTimeout()));
  

	for(int i = 0; i < MAX_STATE_HISTORY; i++) {
	  m_state_history[i].clear();
	}
	m_current_history_state = 0;
	m_last_history_state = -1;

#ifdef AUTOTEST
	autotester();
#endif
	/* 
	   This line (and the one like it in the fullscreen code) allow us
	   to use Escape/Cancel without having it close the app
	*/
	/* Maybe it is just me, but I don't like on/off and cancel on same button */
	reparent ( 0, WStyle_Customize, QPoint ( 0, 0 ));
}
	
void
QPDFWindow::startup()
{
  FILE *file;
  printf("startup last viewed %d filename %s page %d zoom %f\n",
	 m_config->startup_last_viewed_document,
	 m_config->last_state.filename().latin1(),
	 m_config->last_state.page,
	 (float)m_config->last_state.zoom);
  if (m_config->startup_last_viewed_document
      && m_config->last_state.path != "" 
      && m_config->last_state.page > 0
      && (file = fopen(m_config->last_state.path.latin1(),"r")) != NULL) {
    //	    && QFileInfo(m_config->last_state.filename()).exists()) {
    fclose(file);
    
    m_mode = m_config->last_state.mode;
    printf("restoring mode = %d  path = %s\n", m_mode, m_config->last_state.path.latin1());

    // do this in case we are restarting on a document that crashes
    m_config->startup_last_viewed_document = false;
    m_config->startup_directory_browser = true;
    m_config->clean = false;
    writeConfig();

    openFile(m_config->last_state.path);
    
    // open succeeded
    m_config->startup_last_viewed_document = true;
    m_config->startup_directory_browser = false;
    m_config->clean = false;
    writeConfig();

  } else if (m_config->startup_document_browser) {
    openDocumentBrowser();
  } else if (m_config->startup_directory_browser) {
    openDirectoryBrowser();
  } else {
    openDocumentBrowser();
  }
}

void
QPDFWindow::cancelTimeout()
{
  m_cancel_message->hide();
}

void 
QPDFWindow::keyPressEvent ( QKeyEvent *e )
{
  static int pageNum = 0;
  static int pageSign = 1;

  // use keys for find()

  // left = prev_page (like an open book)
  // right = next_page
  // up = backward state
  // down = forward state

  // rotated view:
  //  real up => rotated right = next_page
  //  real left => rotated up = backward state
  //  real down => rotated left = prev_page
  //  real right => rotated down = forward state

  //  printf("QPDFWindow keypress 0x%x ascii 0x%x state 0x%x text '%s'\n", 
  //	 e->key(), e->ascii(), e->state(), e->text().latin1());

  if (m_stack->visibleWidget() != m_pdfview
      && m_stack->visibleWidget() != m_textedit) {
    return;
  }

  int cy = m_pdfview->contentsY();
  int cx = m_pdfview->contentsX();
  const int scroll_fraction = 20;

  int key = e->key();
  if (e->ascii() == 010) { // control-H
    //    printf("control-h == backspace\n");
    key = Qt::Key_Backspace;
  }

  switch ( key ) {

    // cancel
  case Qt::Key_Escape:
    if (m_cancel_timer->isActive()) {
      m_cancel_timer->stop();
      flush();
      qApp->quit();
    }
    if (m_fullscreen || m_pdfview->doingGesture() || m_pdfview->doingSelection()) {
      setFullscreen(false);
      if (m_mode == PDF_VIEW) {
	m_pdfview->gestureAbort();
	m_pdfview->setTextSelectionMode(false);
      }
    } else {
      m_cancel_message-> setGeometry ( m_stack->x() + m_stack->width()/2 - m_cancel_message->width()/2, 
				       m_stack->y() + m_stack->height()/2, 
				       m_cancel_message->width(),
				       m_cancel_message->height());
      m_cancel_message->show();
      m_cancel_timer->start(2000,true);
    }
    break;

  case Qt::Key_Up:
#ifdef MYWHEEL
    if (!(m_mode == PDF_VIEW && m_pdfview->isRotated())) {
      setZoom(-1);
    } else {
      nextPage();
    }
#else
    m_pdfview->scrollBy(0,-m_pdfview->visibleHeight()/scroll_fraction);
    if (!(m_mode == PDF_VIEW && m_pdfview->isRotated())) {
      if (cy == m_pdfview->contentsY() && !m_pdfview->isBusy()) {
	gotoPagePos(m_pdfview->getPageNum()-1, cx, 
		    m_pdfview->contentsHeight() - m_pdfview->visibleHeight() - 1);
      }
    }
#endif
    break;

  case Qt::Key_Right:
#ifdef MYWHEEL
    if (!(m_mode == PDF_VIEW && m_pdfview->isRotated())) {
      nextPage();
    } else {
      setZoom(1);
    }
#else
    m_pdfview->scrollBy(m_pdfview->visibleWidth()/scroll_fraction,0);
    if (m_mode == PDF_VIEW && m_pdfview->isRotated()) {
      if (cx == m_pdfview->contentsX() && !m_pdfview->isBusy()) {
	gotoPagePos(m_pdfview->getPageNum()+1, 0, cy);
      }
    }
#endif    
    break;

  case Qt::Key_Left:
#ifdef MYWHEEL
    if (!(m_mode == PDF_VIEW && m_pdfview->isRotated())) {
      prevPage();
    } else {
      setZoom(-1);
    }
#else
    m_pdfview->scrollBy(-m_pdfview->visibleWidth()/scroll_fraction,0);
    if (m_mode == PDF_VIEW && m_pdfview->isRotated()) {
      if (cx == m_pdfview->contentsX() && !m_pdfview->isBusy()) {
	gotoPagePos(m_pdfview->getPageNum()-1, m_pdfview->contentsWidth(), cy);
      }
    }
#endif
    break;

  case Qt::Key_Down:
#ifdef MYWHEEL
    if (!(m_mode == PDF_VIEW && m_pdfview->isRotated())) {
      setZoom(1);
    } else {
      prevPage();
    }
#else
    m_pdfview->scrollBy(0,m_pdfview->visibleHeight()/scroll_fraction);
    if (!(m_mode == PDF_VIEW && m_pdfview->isRotated())) {
      if (m_pdfview->contentsY() == cy && !m_pdfview->isBusy()) {
	gotoPagePos(m_pdfview->getPageNum()+1, cx, 0);
      }
    }
#endif
    break;

  case Key_Minus:
    pageSign = -1;
    break;

  case Key_Plus:
    pageSign = 1;
    break;

  case Key_0: case Key_1: case Key_2: case Key_3: case Key_4: case Key_5:
  case Key_6: case Key_7: case Key_8: case Key_9:
    pageNum = pageNum * 10 + e->key() - Key_0;
    setPageCount(pageNum * pageSign);
    break;

    // Okay 
  case Key_Return:
  case Key_Enter:
    if (m_config->c7xx && qt_screen->transformOrientation() == 0) {
      // if we have c7xx and it is folded so keyboard is hidden, then
      // use the okay button to start a gesture
      m_pdfview->gestureStart();
    } else {
      if (pageNum > 0) {
	saveState();
	gotoPage(pageNum * pageSign + m_pdfview->getPageOffset());
	pageNum = 0;
	pageSign = 1;
      } else {
	search_again();
      }
    }
    break;

  case Key_Backspace:
    if (pageNum > 0) {
      pageNum /= 10;
      setPageCount(pageNum * pageSign);
      if (pageNum == 0) {
	pageSign = 1;
      }
    } else {
      search(key);
    }
    break;

  case Key_Space:
    if (m_mode == PDF_VIEW) {
      m_pdfview->gestureStart();
    }
    break;
    
  case 0x200f:  // c760 zoom -
    setZoom(-1);
    break;

  case 0x2010:  // c760 zoom +
    setZoom(1);
    break;

  case 0x2019: // c760 fn-n
    nextPage();
    break;

  case 0x2016: // c760 fn-p
    prevPage();
    break;

  case 0x201b:  // fn-5
    toggleFullscreen();
    break;

  case 0x103a:  // menu button
  case Key_Menu:
    printf("menu button %d\n",
	   QPEMenuToolFocusManager::manager()->isActive());
    QPEMenuToolFocusManager::manager()->setActive(!QPEMenuToolFocusManager::manager()->isActive());
    break;

  default:
    if (((e->key() & 0xff00) == 0) && isalpha(e->ascii())) {
      //      printf("search: %x ascii %x\n", e->key(), e->ascii());
      search(e->ascii());
    } else {
      // e.g. menu button (0x103a)
      //      QMainWindow::keyPressEvent(e);
    }
    break;
  }
}      


void
QPDFWindow::selectionChanged(const QString &str)
{
  qApp-> clipboard ( )-> setText ( str );
}



bool
QPDFWindow::makePageMenu()
{
  int n = m_pdfview->lastPageNum();
  int offset = m_pdfview->getPageOffset();
  int i, j;
  int id;

  static int num_pages = -1;
  static int page_offset = -1;

  if (n == num_pages && page_offset == offset) {
    return(false);
  }
  num_pages = n;
  page_offset = offset;

  printf("make page menu %d\n", n);
  for(i=m_pm_page->count()-1; i>m_pm_page_start_pagenum_index; i--) {
    // leaks popup menus if you don't delete them manually!
    if (m_pm_page->findItem(m_pm_page->idAt(i))->popup()) {
      delete m_pm_page->findItem(m_pm_page->idAt(i))->popup();
    } else {
      m_pm_page->removeItemAt(i);
    }
  }
  if (n > MAX_PAGES_ON_PAGE_MENU) {
    for(i=1; i<=n; i+=m_submenu_size) {
      QPopupMenu *popup = new QPopupMenu(m_pm_page, "subpage popup");
      int m = i + m_submenu_size - 1;
      if (m > n) {
	m = n;
      }
      for(j=i; j<=m; j++) {
	id = popup->insertItem(tr("%1").arg(j - offset), j);
	popup->setItemParameter(id, j);
	popup->connectItem(id, this, SLOT(gotoPageMenu(int)));
      }
      id = m_pm_page->insertItem(tr("%1 - %2").arg(i - offset).arg(m - offset), popup, i/m_submenu_size + 1000);
    }
  } else {
    for(i=1; i<=n; i++) {
      id = m_pm_page->insertItem(tr("%1").arg(i), i);
      //      assert(id == i);
      m_pm_page->setItemParameter(id, i);
      m_pm_page->connectItem(id, this, SLOT(gotoPageMenu(int)));
    }
  }
  return(true);
}
    

void QPDFWindow::updateMenuItems()
{
  int page = m_pdfview->getPageNum();

  m_submenu_size = (m_pdfview->lastPageNum()/100 + 1) * 10;

  static int last_page = -1;
  if (makePageMenu()) {
    last_page = -1;
  }

#if 0
  if (m_pdfview->lastPageNum() <= 0 
      || m_pdfview->getPageNum() <= 0 
      || m_pdfview->getPageNum() > m_pdfview->lastPageNum()) {
    printf("@@@@@@ funky page %d numPages %d\n", m_pdfview->getPageNum(),
	   m_pdfview->lastPageNum());
  }
#endif

  if (last_page > 0) {
    m_pm_page->setItemChecked(last_page, false);
    if (m_pdfview->lastPageNum() > MAX_PAGES_ON_PAGE_MENU) {
      m_pm_page->setItemChecked((last_page-1)/m_submenu_size + 1000, false);
    }
  }
  if (m_mode == NOTES_VIEW) {
    last_page = -1;
  } else {
    last_page = page;
    m_pm_page->setItemChecked(last_page, true);
    if (m_pdfview->lastPageNum() > MAX_PAGES_ON_PAGE_MENU) {
      m_pm_page->setItemChecked((last_page-1)/m_submenu_size + 1000, true);
    }
  }
    
#if 0
  for(int i = 1; i<=m_pdfview->lastPageNum(); i++) {
    m_pm_page->setItemChecked(i, i == page);
    m_pm_page->setItemEnabled(i, m_mode != NOTES_VIEW);
  }
  if (m_pdfview->lastPageNum() > MAX_PAGES_ON_PAGE_MENU) {
    int n = m_pdfview->lastPageNum()/m_submenu_size;
    int m = (page-1)/m_submenu_size;
    for(int i = 0; i<=n; i++) {
      m_pm_page->setItemChecked(i + 1000, i == m);
      m_pm_page->setItemEnabled(i + 1000, m_mode != NOTES_VIEW);
    }
  }
#endif

  updateCaption();

  static QString last_filename;

  if (m_pdfview->getFilename() != last_filename
      && !m_pdfview->getFilename().isNull()) {
    for(int i=m_pm_recent_files->count()-1; i>=0; i--) {
      QMenuItem *item = m_pm_recent_files->findItem( m_pm_recent_files->idAt(i) );
      if (item->whatsThis() == m_pdfview->getPath() || i > m_config->max_recent_files) {
	m_pm_recent_files->removeItemAt(i);
      }
    }
    // need to reverse order here
    int id = m_pm_recent_files->insertItem( m_pdfview->getFilename(), -1, 0);
    m_pm_recent_files->setWhatsThis(id, m_pdfview->getPath() );
    m_pm_recent_files->connectItem(id, this, SLOT(openRecentFile(int)));
    m_pm_recent_files->setItemParameter(id, id);
#if 0
    for(int i=0; i < m_pm_recent_files->count(); i++) {
      m_pm_recent_files->setItemParameter(m_pm_recent_files->idAt(i), m_pm_recent_files->idAt(i));
    }
#endif
    last_filename = m_pdfview->getFilename();
    writeConfig();  // in case we are killed by fuser
  }
    
  m_pm_page->setItemEnabled(m_pm_page_forward, forwardStateAllowed());
  m_pm_page->setItemEnabled(m_pm_page_backward, backwardStateAllowed());

  m_pm_page->setItemEnabled(PM_PAGE_FIRST, m_mode != NOTES_VIEW 
			    && m_pdfview->getPageNum() != 1);
  m_pm_page->setItemEnabled(PM_PAGE_LAST, m_mode != NOTES_VIEW
			    && m_pdfview->getPageNum() != m_pdfview->lastPageNum());

  m_pm_page->setItemEnabled(PM_PAGE_PREV, m_mode != NOTES_VIEW 
			    && m_pdfview->getPageNum() > 1);
  m_pm_page->setItemEnabled(PM_PAGE_NEXT, m_mode != NOTES_VIEW
			    && m_pdfview->getPageNum() < m_pdfview->lastPageNum());

  m_but_next_page->setEnabled(m_mode != NOTES_VIEW 
			      && m_pdfview->getPageNum() < m_pdfview->lastPageNum());
  m_but_prev_page->setEnabled(m_mode != NOTES_VIEW 
			      && m_pdfview->getPageNum() > 1);


  m_pm_view->setItemEnabled(PM_VIEW_FIT_TO_WIDTH, m_mode == PDF_VIEW);
  m_pm_view->setItemEnabled(PM_VIEW_FIT_TO_PAGE, m_mode == PDF_VIEW);
  m_pm_view->setItemEnabled(m_pm_view_highlight, m_mode == PDF_VIEW);
  m_pm_view->setItemChecked(m_pm_view_highlight, m_mode == PDF_VIEW
			    && m_pdfview->highlightMode());
  m_pm_view->setItemEnabled(m_pm_view_rotate, m_mode == PDF_VIEW);

  m_pm_edit->setItemEnabled(PM_EDIT_PASTE, m_mode != PDF_VIEW);
  m_pm_edit->setItemEnabled(PM_EDIT_CUT, m_mode != PDF_VIEW);

  // mode buttons
  m_but_pdf->setDown(m_mode == PDF_VIEW);
  m_but_text->setDown(m_mode == TEXT_VIEW);
  m_but_note->setDown(m_mode == NOTES_VIEW);

  if (m_mode == PDF_VIEW) {
    if (m_pdfview->textSelectionMode()) {
      m_but_pdf->setPixmap(m_pdf_text_select_pixmap);
    } else {
      m_but_pdf->setPixmap(m_pdf_move_pixmap);
    }
  } else {
    m_but_pdf->setPixmap(m_pdf_move_pixmap);
  }
      
  int zoom = -1;
  switch(m_mode) {
  case PDF_VIEW:
    zoom = (int)(m_pdfview->getZoom() * 100);
    break;
  case TEXT_VIEW:
  case NOTES_VIEW:
    zoom = (int)(m_textedit->getFontSize() * (100.0 / 12.0));
    break;
  }

  bool checked = false;
  for( QPopupMenu *zoom_menu = m_pm_zoom1; 
       zoom_menu != NULL; 
       zoom_menu = zoom_menu==m_pm_zoom1? m_pm_zoom2 : NULL) {
    for ( uint i = 0; i < zoom_menu-> count ( ); i++ ) {
      int id = zoom_menu->idAt(i);
      int z = zoom_menu-> itemParameter ( id );
      if (z > 1) {
	if (z >= zoom && !checked) {
	  zoom_menu-> setItemChecked( id, true );
	  m_pm_view-> setItemChecked( zoom_menu == m_pm_zoom1? m_pm_zoom1_id : m_pm_zoom2_id, true);
	  m_pm_view-> setItemChecked( zoom_menu == m_pm_zoom1? m_pm_zoom2_id : m_pm_zoom1_id, false);
	  checked = true;
	} else {
	  zoom_menu-> setItemChecked( id, false );
	}
      }
    }
  }

  m_pm_view-> setItemChecked ( m_pm_view_find, 
			       m_tb_find-> isVisible ( ));
  m_pm_view-> setItemChecked ( m_pm_view_reproportion_text, m_pdfview->getReProportionText() );
#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
  //  m_pm_view-> setItemChecked ( m_pm_view_reproportion_and_render, m_pdfview->getReProportionAndRender() );
  m_pm_view-> setItemChecked ( m_pm_view_render_text, m_pdfview->getRenderText() );
#endif

  for(int i = 0; i <= 10; i++) {
    m_pm_priority->setItemChecked(i, m_pdfview->m_notes.m_info.m_priority == i);
  }

  for(int i = 0; i < m_config->category_list.count(); i++) {
    m_pm_categories->setItemChecked(i, 
				    m_pdfview->m_notes.m_info.m_categories.contains(m_pm_categories->text(i)));
  }

  m_but_text->setEnabled(! m_pdfview->getText().isEmpty());
}


    

QPDFWindow::~QPDFWindow ( )
{
  if (m_doc_browser) {
    delete m_doc_browser;
  }
  if (m_dir_browser) {
    delete m_dir_browser;
  }
  delete m_pdfview;
  delete m_textedit;
  delete m_busybar;
  delete m_find_line_edit;
  for(int i=m_pm_page->count()-1; i>m_pm_page_start_pagenum_index; i--) {
    m_pm_page->removeItemAt(i);
  }
}


void
QPDFWindow::setTextMode()
{
  setMode(TEXT_VIEW);
}

void
QPDFWindow::setPDFMode()
{
  if (m_mode == PDF_VIEW) {
    // toggle text selection mode
    m_pdfview->setTextSelectionMode(! m_pdfview->textSelectionMode());
  } else {
    setMode(PDF_VIEW);
  }
}

void
QPDFWindow::setNoteMode()
{
  setMode(NOTES_VIEW);
}

/*
Problem: when we save notes (switch out of note view), we save to the xml file, all notes
that are for the currently open pdf file.  This is good because we can add to the notes
for that file.  It is bad because we don't save changes in notes for other files.
*/
  
void
QPDFWindow::loadDirectoryNotes()
{
  if (m_pdfview->isBusy()
      || !m_pdfview->isValid()
      || m_pdfview->m_notes.isLocked()) {

    // maybe over protective, but we don't want to be between states here
    printf("@@@ pdfview busy %d pdfview valid %d textmap valid %d\n",
	   m_pdfview->isBusy(), m_pdfview->isValid(), m_pdfview->m_textmap->m_state.isValid());
    printf("@@@ textmap notes path %s notes path %s target state %s current state %s\n",
	   m_pdfview->notesPath(m_pdfview->m_textmap->m_state.path).latin1(),
	   m_pdfview->m_notes.getPath().latin1(),
	   m_pdfview->getPath().latin1(),
	   m_pdfview->m_textmap->m_state.path.latin1(),
	   m_pdfview->currentState()->path.latin1());

    return;
  }


  m_pdfview->m_notes.checkdups();

  assert(m_pdfview->notesPath(m_pdfview->m_textmap->m_state.path)
	 == m_pdfview->m_notes.getPath());

  QString assert_filepath = m_pdfview->m_notes.getPath();


  if (!m_pdfview->m_notes.getLoadDirectory().isNull()) {
    flush();
    QString notes_path = m_pdfview->m_notes.getPath();
    // this is all sort of stupid, should just clear out the notes and load them
    // all the more reason this function should be in qdocnotes!
    m_pdfview->m_notes.clear();
    m_pdfview->m_notes.setPath(notes_path);
    m_pdfview->m_notes.setupPage(m_pdfview->m_textmap);
    m_pdfview->m_notes.load(notes_path, NULL);
    m_pm_file->setItemChecked(m_pm_load_directory_notes, false);
    printf("<<< clearing load_directory_notes >>>\n");

  } else {

    // this code should be in qdocview
    m_pdfview->setBusyFlag(true);
    setBusy(true);
    m_pdfview->m_notes.setLocked(true);
    
    QFileInfo fi(m_pdfview->m_notes.getPath());
    QString current_dir =  fi.dirPath(true);
    QDir dir( current_dir );
    bool found_one = false;
    
    printf("load directory notes: %s\n", current_dir.latin1());
    if( dir.exists() ){
      dir.setFilter( QDir::Files | QDir::Readable );
      const QFileInfoList *list = dir.entryInfoList();
      QFileInfoListIterator it( *list );
      QFileInfo *fi;
      bool ok;
      
      while( (fi=it.current() ) ) {
	if( fi->extension(false) == QString::fromLatin1("xml") ){
	  m_pdfview->m_notes.setLocked(false);
	  if (! found_one) {
	    flush();  // needed to dump out old notes if changing directory
	    m_pdfview->m_notes.setLoadDirectory(current_dir);
	    m_pm_file->setItemChecked(m_pm_load_directory_notes, true);
	    found_one = true;
	  }
	  //	printf("loading notes: %s\n", fi->absFilePath().latin1());
	  ok = m_pdfview->m_notes.load(fi->absFilePath(), NULL);
	  m_pdfview->m_notes.setLocked(true);
	  //	if (!ok) printf("error\n");
	}
	++it;
      }
      if (found_one) {

	m_pdfview->m_notes.checkdups();

	// reload current files notes to set qdocinfo and state stuff
	assert(m_pdfview->m_notes.getPath() == assert_filepath);
	assert(m_pdfview->notesPath(m_pdfview->m_textmap->m_state.path) ==
	       m_pdfview->m_notes.getPath());
	
	printf("reloading current notes: %s (%d)\n", 
	       m_pdfview->m_notes.getPath().latin1(),
	       m_pdfview->m_notes.m_info.m_num_notes);
	m_pdfview->m_notes.setLocked(false);
	m_pdfview->m_notes.load(m_pdfview->m_notes.getPath(), NULL);
	m_pdfview->m_notes.setLocked(true);
	printf("reloading done %d\n", m_pdfview->m_notes.m_info.m_num_notes);
	//      m_pdfview->m_notes.setPath(m_pdfview->notesPath(m_pdfview->m_textmap->m_state.path));
	//      m_pdfview->m_notes.setupPage(m_pdfview->m_textmap);
	
      }
    }
  }

  m_pdfview->m_notes.checkdups();

  m_new_state.clear();
  m_pdfview->setMode(NOTES_VIEW);
  m_new_state.set(m_pdfview->state());
  setBusy(false);
  m_pdfview->setBusyFlag(false);
  m_pdfview->m_notes.setLocked(false);
  flush();  // mostly to reset clean flag
  changeStateComplete();
}


void
QPDFWindow::openHelp()
{
  setMode(HELP_VIEW);
}

void
QPDFWindow::changePriority(int priority)
{
  //  printf("change priority = %d\n", priority);
  m_pdfview->m_notes.m_info.m_priority = priority;
  flush();
  updateMenuItems();
}

void
QPDFWindow::changeCategory(int cat)
{
  if (m_pdfview->m_notes.m_info.m_categories.contains(m_pm_categories->text(cat))) {
    m_pdfview->m_notes.m_info.m_categories.remove(m_pm_categories->text(cat));
  } else {
    m_pdfview->m_notes.m_info.m_categories.append(m_pm_categories->text(cat));
  }
  flush();
  updateMenuItems();
}

void
QPDFWindow::newCategory()
{
  QDialog *dialog = new QDialog ( this, "new_cat_dlg", true );
  dialog-> setCaption ( tr( "New Category" ));
	
  QLineEdit *edit = new QLineEdit ( dialog, "cat");
  edit->setMinimumWidth(qApp->desktop()->size().width()-30);
  connect(edit, SIGNAL(returnPressed()),
	  dialog, SLOT(accept()));
  if ( dialog-> exec ( ) == QDialog::Accepted ) {
    int newcat = m_config->category_list.count();
    m_config->category_list.append(edit->text());
    if (m_dir_browser) {
      m_dir_browser->m_all_categories = m_config->category_list;
    }
    m_pm_categories->insertItem( m_config->category_list.last(), newcat);
    m_pm_categories->connectItem(newcat, this, SLOT(changeCategory(int)));
    m_config->clean = false;
    writeConfig();
    changeCategory(newcat);
  }    
  delete dialog;
}

void
dump_font_name(QFont f)
{
  QFontInfo finfo(f);
  printf ("Font %s%d/%d%s%s%s%s\n",
	  finfo.family().ascii()? finfo.family().ascii() : "", 
	  finfo.pointSize(), 
	  finfo.weight(), 
	  finfo.bold()?" bold ":"", 
	  finfo.italic()? " ital ":"", 
	  finfo.underline()? " under ":"",
	  finfo.strikeOut()? " strikeout ":"");
}

QPDFstate::QPDFstate() 
{
  clear();
}

void 
QPDFstate::clear() 
{
  mode = PDF_VIEW;
  wordPosition0 = wordPosition1 = -1;
  charPosition0 = charPosition1 = -1;
  pixPosition0 = QPoint();
  pixPosition1 = QPoint();
  absPixPosition0 = QPoint();
  absPixPosition1 = QPoint();
  range = false;
  QDocState::clear();
}

void
QPDFWindow::save(QPDFstate *state)
{
  // This is made more complex because many of the
  // mapXXX functions need the pagemap to already be in place for that page.
  // Switching out of note view, it may be a new page.
  // So, have to delay them until gotoPage is done in restore()

  QDocNote *note;

  //  if (isBusy()) {
  //    return;
  //  }

  state->clear();

  state->path = m_pdfview->getPath();
  state->page = m_pdfview->getPageNum();
  state->zoom = m_pdfview->getZoom();
  state->mode = m_mode;

  // close out old mode
  switch(m_mode) {

  case PDF_VIEW:
    m_pdfview->gestureAbort();
    if (m_pdfview->hasSelection()) {
      m_pdfview->getTextSelection(&state->wordPosition0, &state->wordPosition1);
      state->charPosition0 = m_pdfview->mapWordToChar(state->wordPosition0, false);
      state->charPosition1 = m_pdfview->mapWordToChar(state->wordPosition1, true);
      state->pixPosition0 = m_pdfview->mapWordToPix(state->wordPosition0, false);
      state->pixPosition1 = m_pdfview->mapWordToPix(state->wordPosition1, true);
      state->range = true;
    } else {
      // maybe better to go to closest note...
      state->pixPosition0 = m_pdfview->currentUpperLeftPix();   // was currentCenterPix()
      state->wordPosition0 = m_pdfview->mapPixToWord(state->pixPosition0, true, true);
      if (state->wordPosition0 < 0) {
	state->wordPosition0 = 0;
      }
      state->charPosition0 = m_pdfview->mapWordToChar(state->wordPosition0);
      if (state->charPosition0 < 0) {
	state->charPosition0 = 0;
      }
    }
    break;

  case TEXT_VIEW:

    int line0, col0, line1, col1;
    
    if (m_textedit->getSelection(&line0, &col0, &line1, &col1)) {
      state->charPosition0 = m_textedit->positionToOffset(line0, col0);
      state->charPosition1 = m_textedit->positionToOffset(line1, col1);
      state->range = true;
    } else {
      state->charPosition0 = m_textedit->getTopCharPosition();
    }
    break;

  case NOTES_VIEW:
    bool changed;
    changed = m_pdfview->saveNotes(m_textedit->text(), 
				   m_textedit->getCharPosition(), 
				   &note);
    m_textedit->setReadOnly(true);
    m_textedit->setEdited(false);
    // note: state->pixPosition0, 1  are relative to full image size
    // there may not be any valid wordStart, wordEnd since page may
    // not have been loaded yet
    if (note == NULL) {
      state->page = m_pdfview->getPageNum();
      //      state->wordPosition0 = state->wordPosition1 = 0;
      state->range = false;
    } else {
      //      printf("note %s\n", note->m_filename.latin1());
      //      printf("state path %s\n", state->path.latin1());
      if (!note->isSameFile(state->path)) {
	QString path = m_pdfview->m_notes.findPath(*note);
	if (! path.isEmpty() ) {
	  state->path = path;
	}
	printf("*** find notes file new path = %s\n", state->path.latin1());
      }
      state->page = note->m_page;
      state->absPixPosition0 = note->m_start;
      state->absPixPosition1 = note->m_end;
      state->range = true;
    }
    break;

  }

#if 1
  printf("SAVE mode %d char %d word %d pix %d,%d abs %d,%d (%s)\n", m_mode,
	 state->charPosition0, state->wordPosition0, state->pixPosition0.x(), state->pixPosition0.y(),
	 state->absPixPosition0.x(), state->absPixPosition0.y(),
	 (state->wordPosition0 < 0)? "" :
	 m_pdfview->getText(state->wordPosition0,state->wordPosition0).latin1());
  if (state->range) {
    printf("  => char %d word %d pix %d,%d abs %d,%d (%s)\n",
	   state->charPosition1, state->wordPosition1, state->pixPosition1.x(), state->pixPosition1.y(),
	   state->absPixPosition1.x(), state->absPixPosition1.y(),
	   (state->wordPosition0 <0 || state->wordPosition1 < 0)? "" :
	   m_pdfview->getText(state->wordPosition0, state->wordPosition1).latin1());
  }
#endif


}


// to understand this you need to know that a lot of the work
// is done by the callback changeStateComplete

// the m_textview readonly flag indicates when there are not valid "notes" present

void
QPDFWindow::restore(QPDFstate *state)
{

  //  if (isBusy()) {
  //    return;
  //  }

  flush();

  m_new_state = *state;

  m_pdfview->setZoom(state->zoom);
  m_pdfview->setPath(state->path);

  m_pdfview->setMode(m_new_state.mode);
  printf("restore: mode %d\n", m_new_state.mode);

  // start new mode
  switch(m_new_state.mode) {

  case PDF_VIEW:
    m_stack->raiseWidget(m_pdfview);
    m_pdfview->setPage(state->page);
    m_pdfview->changeState();
    if (!m_new_state.pixPosition0.isNull()) {
      m_pdfview->state().doMove = true;
      m_pdfview->state().pos(m_new_state.pixPosition0);
    }
    break;

  case TEXT_VIEW:
    assert(!m_textedit->edited() && m_pdfview->m_notes.isClean());
    m_textedit->setReadOnly(true);
    m_textedit->setEdited(false);
    m_stack->raiseWidget(m_textedit);
    m_pdfview->setPage(state->page);
    m_pdfview->changeState();
    break;

  case NOTES_VIEW:
    m_textedit->setReadOnly(true);
    m_textedit->setEdited(false);
    m_stack->raiseWidget(m_textedit);
    m_pdfview->changeState();
    break;

  case HELP_VIEW:
    m_stack->raiseWidget(m_textedit);
    assert(!m_textedit->edited() && m_pdfview->m_notes.isClean());
    if (! m_textedit->openFile("/home/QtPalmtop/help/qpdf2.txt")) {
      printf("can not open help file /home/QtPalmtop/help/qpdf2.txt\n");
    }
    m_textedit->setReadOnly(true);
    m_textedit->setEdited(false);
    break;
  }

  updateMenuItems();
}

void
QPDFWindow::changeStateComplete()
{
  QString noteText;
  QDocNote *note;

  m_pdfview->limitState(m_new_state);

  printf("changeStateComplete(path %s(%s) page %d(%d)/%d zoom %f(%f) rotate %d x %d y %d move %d mode %d(%d))\n",
	 m_pdfview->state().path.ascii(), m_new_state.path.latin1(),
	 m_pdfview->getPageNum(), m_new_state.page,
	 m_pdfview->state().numPages, 
	 (float)m_pdfview->state().zoom, (float)m_new_state.zoom,
	 m_pdfview->state().isRotated, 
	 m_pdfview->state().x, m_pdfview->state().y, m_pdfview->state().doMove, 
	 m_pdfview->state().mode, m_new_state.mode);


  printf("m_new_state.word1 = %d to %d range %d\n",
	 m_new_state.wordPosition0, 
	 m_new_state.wordPosition1, 
	 m_new_state.range);

  if (m_new_state.path == m_pdfview->getPath()
      && m_new_state.path != ""
      && m_pdfview->getPageNum() > 0
      && m_new_state.mode == m_pdfview->state().mode
      && (m_mode == NOTES_VIEW
	  || (m_new_state.page == m_pdfview->getPageNum()
	      && m_new_state.zoom == m_pdfview->getZoom()))) {
    printf("changeStateComplete success\n");

    m_mode = m_new_state.mode;

    if (m_mode == TEXT_VIEW 
	&& m_pdfview->getText().isEmpty()
	&& ! isBusy()) {
      m_new_state.mode = PDF_VIEW;
      restore(&m_new_state);
    }

    // start new mode
    switch(m_mode) {

    case PDF_VIEW:
      m_stack->raiseWidget(m_pdfview);
      m_textedit->setReadOnly(true);
      
      if (!m_new_state.absPixPosition0.isNull()) {
	m_new_state.pixPosition0 = m_pdfview->absolutePosition(m_new_state.absPixPosition0);
      }
      if (!m_new_state.absPixPosition1.isNull()) {
	m_new_state.pixPosition1 = m_pdfview->absolutePosition(m_new_state.absPixPosition1);
      }
      if (m_new_state.charPosition0 != -1) {
	m_new_state.wordPosition0 = m_pdfview->mapCharToWord(m_new_state.charPosition0);
      }
      if (m_new_state.charPosition1 != -1) {
	m_new_state.wordPosition1 = m_pdfview->mapCharToWord(m_new_state.charPosition1);
      }
      if (!m_new_state.pixPosition0.isNull()) {
	m_new_state.wordPosition0 = m_pdfview->mapPixToWord(m_new_state.pixPosition0, true);
      }
      if (!m_new_state.pixPosition1.isNull()) {
	m_new_state.wordPosition1 = m_pdfview->mapPixToWord(m_new_state.pixPosition1, true);
      }
#if 0
      if (m_new_state.mode == NOTES_VIEW && m_new_state.note != NULL) {
	if (m_new_state.wordPosition0 != m_new_state.note->m_firstWord
	    || m_new_state.wordPosition1 != m_new_state.note->m_lastWord) {
	  printf("*** BUG *** wordPosition %d %d note %d %d\n",
		 m_new_state.wordPosition0, m_new_state.wordPosition1, m_new_state.note->m_firstWord, m_new_state.note->m_lastWord);
	}
      }
#endif
      if (m_new_state.range && m_new_state.mode == NOTES_VIEW
	  && (note = m_pdfview->m_notes.find(m_new_state.page, m_new_state.pixPosition0)) != NULL) {
	m_pdfview->setTextSelection(note->m_firstWord, note->m_lastWord, false);
	      printf("range note words\n");
      } else if (m_new_state.range && m_new_state.wordPosition0 != -1 && m_new_state.wordPosition1 != -1) {
	      printf("range words\n");
	m_pdfview->setTextSelection(m_new_state.wordPosition0, m_new_state.wordPosition1, false);
	//      } else if (m_pdfview->state().doMove) {
      } else {
	if (!m_new_state.pixPosition0.isNull()) {
	  printf("set pix position\n");
	  m_pdfview->setPixPosition(m_new_state.pixPosition0);
	} else if (m_new_state.wordPosition0 != -1) {
	  printf("set word postition\n");
	  m_pdfview->setWordPosition(m_new_state.wordPosition0);
	}
      }
      if (m_pdfview->hasSelection()) {
	int word0, word1;
	m_pdfview->getTextSelection(&word0, &word1);
	//      printf("selection %d to %d\n", word0, word1);
	QPoint pt = m_pdfview->mapWordToPix(word0);
	//      printf("center %d,%d\n", pt.x(), pt.y());
	m_pdfview->center(pt.x(), pt.y());
      }
      break;

    case TEXT_VIEW:
      assert(!m_textedit->edited() && m_pdfview->m_notes.isClean());
      m_textedit->setReadOnly(true);
      m_stack->raiseWidget(m_textedit);
      if (m_textedit->text() != m_pdfview->getText()) {
	m_textedit->setText(m_pdfview->getText());
      }
      m_textedit->setEdited(false);
      if (!m_new_state.absPixPosition0.isNull()) {
	m_new_state.pixPosition0 = m_pdfview->absolutePosition(m_new_state.absPixPosition0);
      }
      if (!m_new_state.absPixPosition1.isNull()) {
	m_new_state.pixPosition1 = m_pdfview->absolutePosition(m_new_state.absPixPosition1);
      }
      if (!m_new_state.pixPosition0.isNull() && m_new_state.wordPosition0 == -1) {
	m_new_state.wordPosition0 = m_pdfview->mapPixToWord(m_new_state.pixPosition0, true);
      }
      if (m_new_state.charPosition0 == -1 && m_new_state.wordPosition0 != -1) {
	m_new_state.charPosition0 = m_pdfview->mapWordToChar(m_new_state.wordPosition0, false);
	printf("word to char\n");
      }
      if (m_new_state.range) {
	if (!m_new_state.pixPosition1.isNull() && m_new_state.wordPosition1 == -1) {
	  m_new_state.wordPosition1 = m_pdfview->mapPixToWord(m_new_state.pixPosition1, true);
	  printf("== mapPixToWord(%d,%d) => %d\n", m_new_state.pixPosition1.x(),
		 m_new_state.pixPosition1.y(), m_new_state.wordPosition1);
	}
	if (m_new_state.charPosition1 == -1 && m_new_state.wordPosition1 != -1) {
	  m_new_state.charPosition1 = m_pdfview->mapWordToChar(m_new_state.wordPosition1, true);
	}
      }
      if (m_new_state.charPosition0 != -1) {
	printf("set char pos %d\n", m_new_state.charPosition0);
	m_textedit->setTopCharPosition(m_new_state.charPosition0);
	m_textedit->setCharPosition(m_new_state.charPosition0, false);
	if (m_new_state.range && m_new_state.charPosition1 != -1) {
	  m_textedit->setCharPosition(m_new_state.charPosition1, true);
	}
      } else {
	m_textedit->setTopCharPosition(0);
	m_textedit->setCharPosition(0);
      }
      //    m_textedit->setFocusPolicy(QWidget::StrongFocus);
      m_textedit->setFocus();
      break;
      
    case NOTES_VIEW:
      m_stack->raiseWidget(m_textedit);
      
      if (m_new_state.charPosition0 != -1 && m_new_state.wordPosition0 == -1) {
	m_new_state.wordPosition0 = m_pdfview->mapCharToWord(m_new_state.charPosition0);
      }
      if (m_new_state.charPosition1 != -1 && m_new_state.wordPosition1 == -1) {
      m_new_state.wordPosition1 = m_pdfview->mapCharToWord(m_new_state.charPosition1);
      }
      if (m_new_state.range && m_new_state.wordPosition0 != -1 && m_new_state.wordPosition1 != -1) {
	noteText = m_pdfview->getNotes(m_new_state.wordPosition0, m_new_state.wordPosition1, &m_new_state.charPosition0);
      }
      if (noteText.isNull() && !m_new_state.pixPosition0.isNull()) {
	noteText = m_pdfview->getNotes(m_new_state.pixPosition0, &m_new_state.charPosition0);
      }
      if (noteText.isNull()) {
	noteText = m_pdfview->getNotes(-1,NULL);
      }
      assert(!m_textedit->edited() && m_pdfview->m_notes.isClean());
      m_textedit->setText(noteText);
      m_textedit->setReadOnly(false);
      m_textedit->setEdited(false);
      //    m_textedit->setFocusPolicy(QWidget::StrongFocus);
      m_textedit->setFocus();
      if (m_new_state.charPosition0 >= 0) {
	//      printf("set pos %d\n", m_new_state.charPosition0);
	m_textedit->setCharPosition(m_new_state.charPosition0);
	m_textedit->setTopCharPosition(m_new_state.charPosition0);
      } else {
	m_textedit->setTopCharPosition(0);
      }
      break;
      
    case HELP_VIEW:
      m_stack->raiseWidget(m_textedit);
      if (! m_textedit->openFile("/home/QtPalmtop/help/qpdf2.txt")) {
	printf("can not open help file /home/QtPalmtop/help/qpdf2.txt\n");
      }
      m_textedit->setReadOnly(true);
      m_textedit->setEdited(false);
      break;
    }
  } else {
    printf("<<<<<<<< state complete failed >>>>>>>>>\n");
  }

#if 1
  printf("RESTORE mode %d char %d word %d pix %d,%d abs %d,%d (%s)\n", m_mode,
	 m_new_state.charPosition0, m_new_state.wordPosition0, m_new_state.pixPosition0.x(), m_new_state.pixPosition0.y(),
	 m_new_state.absPixPosition0.x(), m_new_state.absPixPosition0.y(),
	 (m_new_state.wordPosition0 <0)? "" :
	 m_pdfview->getText(m_new_state.wordPosition0,m_new_state.wordPosition0).latin1());
  if (m_new_state.range) {
    printf("  => char %d word %d pix %d,%d abs %d,%d (%s)\n",
	   m_new_state.charPosition1, m_new_state.wordPosition1, m_new_state.pixPosition1.x(), m_new_state.pixPosition1.y(),
	   m_new_state.absPixPosition1.x(), m_new_state.absPixPosition1.y(),
	   (m_new_state.wordPosition0 <0 || m_new_state.wordPosition1 < 0)? "" :
	   m_pdfview->getText(m_new_state.wordPosition0, m_new_state.wordPosition1).latin1());
  }
#endif

  setPageCount(m_pdfview->getPageNum() - m_pdfview->getPageOffset());
  updateMenuItems();
  //  flush();  // needed?  does flush() twice to restore()
  m_new_state.clear();
}    




void
QPDFWindow::setMode(ViewMode new_mode)
{

  QPDFstate state;

  if (! m_pdfview || !m_pdfview-> isOk ( )) {
    return;
  }
  if (m_mode == new_mode) {
    return;
  }

  //  saveState();
  save(&state);
  state.mode = new_mode;
  m_pdfview->setMode(new_mode);
  restore(&state);
}


void
QPDFWindow::saveState()
{
  //  static bool in_save_state = false;

  if (m_pdfview->getPageNum() <= 0 || isBusy()) {
      // || in_save_state) {
    return;
  }
  if (m_current_history_state > 0
      && m_pdfview->state().path == m_state_history[m_current_history_state-1].path
      && m_pdfview->state().page == m_state_history[m_current_history_state-1].page
      && m_pdfview->state().zoom == m_state_history[m_current_history_state-1].zoom) {
    return;
  }

  //  in_save_state = true;
  if (m_current_history_state >= MAX_STATE_HISTORY) {
    int i;
    for(i=0; i<MAX_STATE_HISTORY-1; i++) {
      m_state_history[i] = m_state_history[i+1];
    }
    m_current_history_state = MAX_STATE_HISTORY-1;
  }
#if 0
  printf("** before save current=%d last=%d page %d zoom %f\n", m_current_history_state,
	 m_last_history_state,
	 m_pdfview->getPageNum(), m_pdfview->getZoom());
  assert(m_current_history_state >= 0
	 && m_current_history_state < MAX_STATE_HISTORY);
#endif
  m_last_history_state = m_current_history_state;
  m_current_history_state++;
  save(&m_state_history[m_current_history_state-1]);
#if 0
  printf("** save current=%d last=%d page %d zoom %f\n", m_current_history_state,
	 m_last_history_state,
	 m_pdfview->getPageNum(), m_pdfview->getZoom());
  assert(m_current_history_state >= 0
	 && m_current_history_state <= MAX_STATE_HISTORY);
#endif
  //  in_save_state = false;
}

void
QPDFWindow::backwardState()
{

  if (m_last_history_state == m_current_history_state-1) {
    // when we start going backward, keep current state
    // so we can return to before historical exploration
    //    printf("first backwardState current=%d last=%d\n", m_current_history_state, m_last_history_state);
    saveState();
    m_current_history_state--;
  }

  if (m_current_history_state > 0) {
    m_current_history_state--;
    assert(m_current_history_state >= 0
	   && m_current_history_state < MAX_STATE_HISTORY);
    restore(&m_state_history[m_current_history_state]);
#if 0
    printf("** backward current=%d last=%d page %d zoom %f (%f)\n", m_current_history_state,
	   m_last_history_state,
	   m_pdfview->getPageNum(),
	   m_pdfview->getZoom(), m_state_history[m_current_history_state].zoom);
#endif
  }
}

bool 
QPDFWindow::backwardStateAllowed()
{
  return(m_current_history_state > 0);
}

void
QPDFWindow::forwardState()
{
  if (m_current_history_state < m_last_history_state) {
    assert(m_current_history_state >= 0
	   && m_current_history_state < MAX_STATE_HISTORY);
    m_current_history_state++; 
    restore(&m_state_history[m_current_history_state]);
#if 0
    printf("** forward current=%d last=%d page %d zoom %f (%f)\n", m_current_history_state,
	   m_last_history_state,
	   m_pdfview->getPageNum(), 
	   m_pdfview->getZoom(), m_state_history[m_current_history_state].zoom);
#endif
  }
}

bool 
QPDFWindow::forwardStateAllowed()
{
  return(m_current_history_state < m_last_history_state);
}



#if 0
QString
QPDFWindow::changedRegion(QString t1, QString t2)
{
  if (t1.isEmpty()) {
    return(t2);
  }
  if (t2.isEmpty()) {
    return(t1);
  }
  QString result;
  const QChar *u1 = t1.unicode();
  const QChar *u2 = t2.unicode();
  const QChar *e1 = u1 + t1.length() - 1;
  const QChar *e2 = u2 + t2.length() - 1;

  while(*e2 == *e1) {
    e2--;
    e1--;
  }
  while(*u1 == *u2) {
    u1++;
    u2++;
  }
  int n = 0;
  while(u1 <= e1) {
    result[n++] = *u1++;
  }
  return(result);
}
#endif

  

// vv Fullscreen handling (for broken QT-lib) [David Hedbor, www.eongames.com]

void QPDFWindow::resizeEvent ( QResizeEvent * )
{
  
  if ( m_fullscreen && ( size ( ) != qApp -> desktop() -> size ( )) ) {
    setFullscreen ( true );
  }
}

void QPDFWindow::focusInEvent ( QFocusEvent * )
{
  //  printf("QPDFWindow::focusInEvent\n");
  if ( m_fullscreen ) {
        setFullscreen ( true );
  }
}

void QPDFWindow::toggleFullscreen ( )
{
  setFullscreen ( !m_fullscreen );
}


void QPDFWindow::setFullscreen ( bool b )
{
	static QSize normalsize;


	m_fullscreen = b;

	// doesn't do much good in full screen, you can't see it anyway
	m_pm_view-> setItemChecked ( m_pm_view_fullscreen, b );

	if ( b ) {
	  if ( !normalsize. isValid ( )) {
	    normalsize = size ( );
	  }    
	  
	  setFixedSize ( qApp-> desktop ( )-> size ( ));
	  showNormal ( );
	  reparent ( 0, WStyle_Customize | WStyle_NoBorder, 
		     QPoint ( 0, 0 ));
	  showFullScreen ( );

	  m_tb_menu-> hide ( );
	  m_tb_find-> hide ( );
	  m_tb_rename-> hide ( );

	} else {
	  showNormal ( );
	  reparent ( 0, WStyle_Customize, QPoint ( 0, 0 ));
	  resize ( normalsize );
	  showMaximized ( );
	  normalsize = QSize ( );
	  
	  m_tb_menu-> show ( );
	  m_tb_find-> hide ( );
	  m_tb_rename-> hide ( );
	}	
	layoutPageCounter();
}
	
// ^^ Fullscreen handling (for broken QT-lib)




void QPDFWindow::toggleRotate()
{
  setRotate(1 - m_pdfview->isRotated());
}

void QPDFWindow::setRotate(bool rotated)
{
  m_pm_view-> setItemChecked ( m_pm_view_rotate, rotated );
  m_pdfview-> setRotation( rotated );
  m_pdfview-> changeState();
}

void QPDFWindow::toggleHighlight()
{
  setHighlight(1 - m_pdfview->highlightMode());
}

void QPDFWindow::toggleReProportionText()
{
  m_pdfview->setReProportionText(! m_pdfview->getReProportionText() );
  m_pm_view-> setItemChecked ( m_pm_view_reproportion_text, m_pdfview->getReProportionText() );
  m_pdfview-> changeState();
}

void QPDFWindow::toggleReProportionAndRender()
{
  m_pdfview->setReProportionAndRender(! m_pdfview->getReProportionAndRender() );
  m_pdfview-> changeState();
  //  m_pm_view-> setItemChecked ( m_pm_view_reproportion_and_render, m_pdfview->getReProportionAndRender() );
}


void QPDFWindow::toggleRenderText()
{
  // would get rid of the whole function, but then moc does not
  // understand conditional compilation in slots of qpdf.h
#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
  m_pdfview->setRenderText(! m_pdfview->getRenderText() );
  m_pm_view-> setItemChecked ( m_pm_view_render_text, m_pdfview->getRenderText() );
  m_pdfview-> changeState();
#endif
}


void QPDFWindow::setHighlight(bool highlight)
{
  m_pm_view-> setItemChecked ( m_pm_view_highlight, highlight );
  m_pdfview-> setHighlightMode( highlight );
}


void QPDFWindow::setZoom ( int z )
{
  float zoom = -1.0;

  printf("QPDFWindow::setZoom(%d)\n", z);
  switch(m_mode) {
  case PDF_VIEW:
    if (m_pdfview->visibleWidth() <= 0 || m_pdfview->visibleHeight() <= 0
	|| m_pdfview->contentsWidth() <= 0 || m_pdfview->contentsHeight() <= 0 ) {
      return;
    }
    
    switch ( z ) {
      
    case PM_VIEW_FIT_TO_WIDTH:
      zoom = (float)m_pdfview-> visibleWidth ( ) /  (float)m_pdfview-> contentsWidth();
      zoom *= m_pdfview->getZoom();
      //      printf("%f = %f/%f\n", (float)zoom,
      //	     (float)m_pdfview-> visibleWidth ( ), (float)m_pdfview-> contentsWidth());
      break;
      
      
    case PM_VIEW_FIT_TO_PAGE:
      zoom = QMIN( (float)m_pdfview-> visibleWidth() / (float)m_pdfview-> contentsWidth(),
		   (float)m_pdfview-> visibleHeight() / (float)m_pdfview-> contentsHeight());
      zoom *= m_pdfview->getZoom();
      //      printf("%f = %f/%f %f/%f\n", (float)zoom,
      //	     (float)m_pdfview-> visibleWidth ( ), (float)m_pdfview-> contentsWidth(),
      //	     (float)m_pdfview-> visibleHeight(), (float)m_pdfview-> contentsHeight());
      break;
      

    case 1:
      zoom = m_pdfview->getZoom() * 1.25;
      break;

    case -1:
      zoom = m_pdfview->getZoom() / 1.25;
      break;

    default:
      if (z > 0 && z < 200) {
	zoom = (float)z  / 100.0;
      }
      break;
      
    }
    if (zoom > MAX_ZOOM) {
      zoom = MAX_ZOOM;
      printf("hit max zoom limit\n");
    }
    if (zoom > 0.0) {
      saveState();
      m_pdfview->setZoom(zoom);
      m_pdfview->changeState();
    }
    break;

  case TEXT_VIEW:
  case NOTES_VIEW:
    switch(z) {
    case 1:
      m_textedit->zoomIn();
      break;
    case -1:
      m_textedit->zoomOut();
      break;
    default:
      m_textedit->setFontSize((int)((float)z * 12.0 / 100.0));
      break;
    }
  }

  updateMenuItems();
}


void QPDFWindow::updateCaption ( )
{
  QString cap;
	
  cap = QString ( "[%1] %2 " )
    . arg ( m_pdfview->getPageNum() - m_pdfview->getPageOffset() )
    . arg ( m_pdfview->getFilename() );
  setCaption ( cap );
}

#if 0
void QPDFWindow::gotoPageDialog ( )
{
	QDialog *d = new QDialog ( this, "gotodlg", true );
	d-> setCaption ( tr( "Goto page" ));
	
	QBoxLayout *lay = new QVBoxLayout ( d, 4, 4  );	
	
	QLabel *l = new QLabel ( tr( "Select from 1 .. %1:" ). arg ( m_pages ), d );
	lay-> addWidget ( l );
	
	QSpinBox *spin = new QSpinBox ( 1, m_pages, 1, d );
	spin-> setValue ( m_currentpage );
	spin-> setWrapping ( true );
	spin-> setButtonSymbols ( QSpinBox::PlusMinus );
	lay-> addWidget ( spin );
	
	if ( d-> exec ( ) == QDialog::Accepted ) {
		gotoPage ( spin-> value ( ));
	}
	
	delete d;
}
#endif

void QPDFWindow::toggleFindBar ( )
{
#if 0
	if ( m_to_find-> isOn ( ) == m_tb_find-> isVisible ( ))
		m_to_find-> setOn ( !m_tb_find-> isVisible ( ));
#endif
	setFindBar(! m_tb_find-> isVisible ( ));
}

void QPDFWindow::closeFindBar( )
{
  setFindBar(false);
}

void QPDFWindow::setFindBar( bool active )
{
  if ( m_tb_find-> isVisible ( ) != active ) {
    if ( m_tb_find-> isVisible ( ) ) {
      m_tb_find-> hide ( );
      if (m_mode == PDF_VIEW) {
	m_pdfview->setFocus();
      } else {
	m_textedit->setFocus();
      }
    } else {
      m_tb_find-> show ( );
      m_find_line_edit-> setFocus ( );
    }
  }
  m_pm_view-> setItemChecked ( m_pm_view_find, active );

}

void QPDFWindow::search(const int search_char)
{
  switch(search_char) {
  case Key_Backspace:
    m_find_line_edit->backspace();
    break;
  case Qt::Key_Escape:
    m_find_line_edit->clear();
    break;
  case Key_Return:
  case Key_Enter:
    // already get it from keyPressEvent
    //  search_again();
    break;
  default:
    //    printf("search(0x%x)\n", search_char);
    m_find_line_edit->insert(QChar((char)search_char));
    break;
  }
}

void
QPDFWindow::search_again()
{
  // this is a hack, should move forward a char to get next match
  // unfortunely, we remember in units of words, so we'd have to
  // have char indices off on the side, which would be ugly too.
   int from, to;
   switch(m_mode) {
   case PDF_VIEW:
     if (m_pdfview->hasSelection()) {
       m_pdfview->getTextSelection(&from,&to);
       //       printf("from=%d to=%d\n", from, to);
       m_pdfview->setTextSelection(from+1,from+1);
     }
     break;
   case TEXT_VIEW:
     m_textedit->setCharPosition(m_textedit->getCharPosition()+1);
     break;
   }
   findTextInPage(m_find_line_edit->text());
}
 
bool QPDFWindow::findText ( const QString &str, bool reverse, bool multipage )
{
  int from, to;
  int page;
  bool found;
  int start_page = m_pdfview->getPageNum();

  to = -1;
  from = -1;
  switch(m_mode) {
  case PDF_VIEW:
    if (m_pdfview->hasSelection()) {
      m_pdfview->getTextSelection(&from, &to);
      //      printf("start word from=%d, to=%d\n", from,to);
      from = m_pdfview->mapWordToChar(from);
      to = m_pdfview->mapWordToChar(to);
      if (from == to) {
	to = -1;
      }
      //      printf("start char from=%d, to=%d\n", from,to);
    } else {
      //    from = m_pdfview->mapPixToWord(m_pdfview->currentUpperLeftPix());
      from = 0;
    }
    break;
  case TEXT_VIEW:
    int line0, col0, line1, col1;
    if (m_textedit->getSelection(&line0, &col0, &line1, &col1)) {
      from = m_textedit->positionToOffset(line0, col0);
    } else {
      from = m_textedit->getCharPosition();
    }
    //    printf("start search %d\n", from);
    break;
  case NOTES_VIEW:
    return(m_textedit->find(str, false, reverse));
  }

  found = m_pdfview->findText ( str, &from, &to, &page, reverse, multipage );

  if (!found && !multipage && !reverse) {
    // wrap around
    //    printf("search wrap found %d multi %d reverse %d\n", found, multipage, reverse);
    from = 0;
    to = -1;
    found = m_pdfview->findText ( str, &from, &to, &page, reverse, multipage );
  }
  
  if (found) {
    saveState();

    if (page != start_page) {
      m_new_state.clear();
      m_pdfview->setPage(page);
      m_new_state.set(m_pdfview->state());

      printf("found char pos %d to %d\n", from, to);
      m_new_state.charPosition0 = from;
      m_new_state.charPosition1 = to+1;
      m_new_state.range = true;
      m_pdfview->changeState();
    } else {
      if (m_mode == TEXT_VIEW) {
      //      printf("found char pos %d to %d\n", from, to);
	m_textedit->setCharPosition(from, false);
	m_textedit->setCharPosition(to+1, true);
      }
      if (m_mode == PDF_VIEW) {
	m_pdfview->setTextSelection(m_pdfview->mapCharToWord(from), 
				    m_pdfview->mapCharToWord(to), true);
	m_pdfview->setCharPosition(from);
      }
    }
  }
    
  return(found);
}

void QPDFWindow::findNoMatch(bool flag)
{
  if (flag) {
    m_tb_find->setPalette(QPalette(Qt::red));
  } else {
    m_tb_find->setPalette(QPalette(Qt::white));
  }
}

void QPDFWindow::findTextForward ( )
{
  findNoMatch(false);
  if (!findText( m_find_line_edit-> text ( ), false, false )) {
    findNoMatch(true);
  }
}


void QPDFWindow::findTextMultipage ( )
{
  findNoMatch(false);
  if (!findText ( m_find_line_edit-> text ( ), false, true )) {
    findNoMatch(true);
  }
}

void QPDFWindow::findTextBackward ( )
{
  findNoMatch(false);
  if (!findText ( m_find_line_edit-> text ( ), true, false )) {
    findNoMatch(true);
  }
}

void QPDFWindow::findTextBackwardMultipage ( )
{
  findNoMatch(false);
  if (!findText ( m_find_line_edit-> text ( ), true, true )) {
    findNoMatch(true);
  }
}

void QPDFWindow::findTextInPage(const QString &str)
{
  //  printf("QPDFWindow::findTextInPage\n");
  flashFindBar();
  findNoMatch(false);
  if (!findText( str, false, false )) {
    // kill last char since it didn't match
    // m_find_line_edit->setText( m_find_line_edit->text().left(m_find_line_edit->text().length() - 1) );
    findNoMatch(true);
  }
}


void QPDFWindow::editCut ( )
{
  printf("editCut mode %d readonly %d\n",
	 m_mode, m_textedit->isReadOnly());
  if (m_mode == NOTES_VIEW
    && !m_textedit->isReadOnly()) {
    m_textedit->cut();
  }
}

void QPDFWindow::editPaste ( )
{
  printf("editPaste mode %d readonly %d\n",
	 m_mode, m_textedit->isReadOnly());
  if (m_mode == NOTES_VIEW 
      && !m_textedit->isReadOnly()) {
    m_textedit->paste();
  }
}

void QPDFWindow::editCopy ( )
{
  switch(m_mode) {
  case PDF_VIEW:
    qApp-> clipboard ( )-> setText ( m_pdfview->selectedText() );
    break;
  case TEXT_VIEW:
  case NOTES_VIEW:
    qApp-> clipboard ( )-> setText ( m_textedit->selectedText() );
    break;
  }
}

void QPDFWindow::gotoPageMenu( int page )
{
  saveState();
  gotoPage(page);
}

void QPDFWindow::gotoPage( int page ) 
{
  gotoPagePos(page, -1, -1);
}

void QPDFWindow::gotoPagePos ( int page, int x, int y)
{
  printf("QPDFWindow::gotoPage(%d)\n", page);
  if (m_mode != PDF_VIEW && m_mode != TEXT_VIEW) {
    return;
  }
  if (!m_pdfview->isValid()) {
    printf("**** not valid state\n");
    return;
  }
  if (page > m_pdfview->lastPageNum()) {
    page = m_pdfview->lastPageNum();
  }
  if (page <= 0) {
    page = 1;
  }

  m_new_state.clear();
  m_new_state.set(m_pdfview->state());
  m_new_state.page = page;

  switch(m_mode) {
  case PDF_VIEW:
    if (x == -1 && y == -1) {
      m_pdfview->gotoPage(page);
    } else {
      m_pdfview->gotoPage(page, QPoint(x, y));
    }
    break;
  case TEXT_VIEW:
    assert(!m_textedit->edited() && m_pdfview->m_notes.isClean());
    m_textedit->setReadOnly(true);
    m_textedit->setEdited(false);
    m_pdfview->gotoPage(page);
    //    m_textedit->setText( m_pdfview->getText(0,-1,page) );
    break;
    // TODO: note mode, select page of notes
  }
  updateMenuItems();
}

void QPDFWindow::setPageOffset()
{
  m_pdfview->setPageOffset(m_pdfview->getPageNum() - 1);
  updateMenuItems();
}

void QPDFWindow::firstPage ( )
{
  saveState();
  gotoPage(1);
}

void QPDFWindow::prevPage ( )
{
  gotoPage(m_pdfview->getPageNum() - 1);
}

void QPDFWindow::nextPage ( )
{
  gotoPage(m_pdfview->getPageNum() + 1);
}

void QPDFWindow::lastPage ( )
{
  saveState();
  gotoPage(m_pdfview->lastPageNum());
}


void QPDFWindow::closeFileSelector ( )
{
  if (m_tb_menu) {
    m_tb_menu-> show ( );
  }
  if (m_doc_browser) {
    m_doc_browser->hide();
  }
  if (m_dir_browser) {
    m_dir_browser->hide();
  }
  if (m_mode == PDF_VIEW) {
    m_stack->raiseWidget(m_pdfview);
  } else {
    m_stack->raiseWidget(m_textedit);
  }
}

void QPDFWindow::openPreferences ( )
{
  QWidget *prev = m_stack->visibleWidget();
  writeConfig();
  m_stack->raiseWidget(PREFERENCES_PAGE);
  m_config->run();
  m_stack->raiseWidget(prev);
  QFont font;
  font.setPointSize(m_config->default_font_size);
  if (m_config->default_font_name != "") {
    font.setFamily(m_config->default_font_name);
  }
  qApp->setFont(font, true);
  if (m_dir_browser) {
    m_dir_browser->listView()->setFont(font);
    m_dir_browser->initPics();
  }
  layoutPageCounter();
  m_config->clean = false;
  writeConfig();
}

void QPDFWindow::openDocumentBrowser ( )
{
  if (m_doc_browser == NULL) {
    m_please_wait-> setGeometry ( m_stack->x() + m_stack->width()/2 - m_please_wait->width()/2, 
				     m_stack->y() + m_stack->height()/2, 
				     m_please_wait->width(),
				     m_please_wait->height());
    m_please_wait->show();
    qApp-> processEvents ( );
    m_doc_browser = new HackedFileSelector ( "application/pdf", m_stack, "fs", false, true );
    m_please_wait->hide();
    m_stack->addWidget( m_doc_browser, PDF_DOC_BROWSER);
    connect ( m_doc_browser, SIGNAL( closeMe ( )), this, 
	      SLOT( closeFileSelector ( )));
    connect ( m_doc_browser, SIGNAL( fileSelected ( const DocLnk & )), this, 
	      SLOT( openFile ( const DocLnk & )));
    //    connect ( m_doc_browser, SIGNAL( returnPressed ( const DocLnk & )), this, 
    //	      SLOT( openFile ( const DocLnk & )));
  } else {
    m_please_wait->show();
    m_doc_browser->reopen();
    m_please_wait->hide();
  }
  //  QPEApplication::setStylusOperation( m_doc_browser->view->viewport(), QPEApplication::RightOnHold );
  m_tb_menu-> hide ( );
  m_tb_find-> hide ( );
  m_tb_rename-> hide ( );
  m_stack-> raiseWidget ( PDF_DOC_BROWSER );
  m_doc_browser->setFocus();
}

void QPDFWindow::openDirectoryBrowser ( )
{
  if (m_dir_browser == NULL) {
    m_dir_browser = new QpdfFileSelector ( "application/pdf", m_stack, "fs", 
					   m_config->recent_dirs, m_config->category_list,
					   false, true);
    m_dir_browser->setChooserVisible(false);
    m_dir_browser->setSelector(QpdfFileSelector::EXTENDED_ALL);

    m_stack->addWidget( m_dir_browser, PDF_DIR_BROWSER);
    connect ( m_dir_browser, SIGNAL( closeMe ( )), this, 
	      SLOT( closeFileSelector ( )));
    connect ( m_dir_browser, SIGNAL( fileSelected ( const DocLnk & )), this, 
	      SLOT( openFile ( const DocLnk & )));
  }
  m_tb_menu-> hide ( );
  m_tb_find-> hide ( );
  m_tb_rename-> hide ( );
  m_stack-> raiseWidget ( PDF_DIR_BROWSER );
  m_dir_browser->listView()->setFocus();
}

void QPDFWindow::openRecentFile( int id )
{
  QMenuItem *item = m_pm_recent_files->findItem(id);
  printf(">>>>>>*** open recent file %d\n", id);
  if (item) {
    printf(">>>>> open recent %s\n", item->whatsThis().latin1());
    openFile( item->whatsThis() );
  }
}

void QPDFWindow::openFile ( const QString &f )
{
  //  printf("enter openFile\n");
	DocLnk nf;
	nf. setType ( "application/pdf" );
	nf. setFile ( f );
	QFileInfo fi ( f );
	nf. setName ( fi. baseName ( ));
	openFile ( nf );
}

void QPDFWindow::openFile ( const DocLnk &f )
{
	QString fn = f. file ( );
	QFileInfo fi ( fn );
	
	if ( fi. exists ( )) {
	  closeFileSelector ( );
	  qApp-> processEvents ( );
	  saveState();
	  flush();
	  //	  m_stack->raiseWidget( m_pdfview );
	  assert(!m_textedit->edited() && m_pdfview->m_notes.isClean());
	  m_textedit->setReadOnly(true);

	  // should push all this into qdocview
	  // make a setDefaults function
	  m_pdfview->setReProportionText(m_config->default_adjust_spacing);
#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
	  m_pdfview->setRenderText(m_config->default_render_text);
#endif
	  m_pdfview->setPageOffset(0);
	  m_pdfview->setPath(fn);
	  m_pdfview->setZoom(m_config->default_zoom);  
	  m_pdfview->state().page = 1;
	  m_pdfview->state().home();
	  // allow open in current viewing mode for quick access to text or notes views
	  //	  m_mode = PDF_VIEW;

	  if (QDocNotes::loadState(m_pdfview->notesPath(fn.latin1()), m_pdfview->state())) {
	    printf("loaded xml state\n");
	    if (!m_config->open_last_viewed_page && !m_config->open_notes_view) {
	      m_pdfview->state().page = 1;
	      m_pdfview->state().home();
	    }
	    //	    if (m_config->open_last_viewed_page) {
	    //	      m_mode = m_pdfview->state().mode;
	    //	      printf("last viewed mode %d\n", m_mode);
	    //      }
	  }

	  printf("open zoom = %f\n", m_pdfview->getZoom());

	  if (m_config->open_notes_view) {
	    QDocInfo info;
	    info.quickRead(m_pdfview->getPath());
	    if (info.m_num_notes > 0) {
	      printf("going to note mode\n");
	      m_mode = NOTES_VIEW;
	    }
	  }

	  m_pdfview->setMode(m_mode);
	  printf("openFile set mode %d\n", m_mode);

	  m_new_state.clear();
	  m_new_state.set(m_pdfview->state());
	  m_pdfview->state().doMove = true;
	  m_new_state.pixPosition0.setX(m_pdfview->state().x);
	  m_new_state.pixPosition0.setY(m_pdfview->state().y);
	  m_new_state.charPosition0 = m_pdfview->state().char_index;
	  
	  m_pdfview->changeState();

	} else {
	  printf("can not open file %s\n", fn.latin1());
#ifndef AUTOTEST
	  QMessageBox::warning ( this, tr( "Error" ), tr( "File does not exist !" ));
#endif
	}		
}

void QPDFWindow::setDocument ( const QString &f )
{
	if ( f. find ( ".desktop", 0, true ) == -1 ) 
		openFile ( f );
	else 
		openFile ( DocLnk ( f ));
		
}

void QPDFWindow::openRenameBar ( )
{
  QString path = m_pdfview-> getPath ( );
  int pos = path. findRev ( '/' );
  if ( pos < 0 ) {
    pos = path. length ( );
  }
  m_rename_line_edit-> setText( path );
  m_rename_line_edit-> setCursorPosition( pos );
  setRenameBar( true );
}

void QPDFWindow::setRenameBar ( bool active )
{
  if (active != m_tb_rename->isVisible()) {
    if (active) {
      m_tb_rename-> show ( );
      m_tb_rename-> setFocus ( );
    } else {
      m_tb_rename-> hide ( );
      if (m_mode == PDF_VIEW) {
	m_pdfview->setFocus();
      } else {
	m_textedit->setFocus();
      }
    }
  }
}

void QPDFWindow::abortRenameBar ( )
{
  setRenameBar( false );
}


#ifdef ARM
extern int errno;
#else
extern "C" {
  int errno;
};
#endif

void
QPDFWindow::writeConfig()
{
  //  if (m_config->clean) {
  //    return;
  //  }
  m_config->recent_files.clear();
  int n;
  n = QMIN(m_pm_recent_files->count(), m_config->max_recent_files);
  for(int i=0; i < n; i++) {
    m_config->recent_files += m_pm_recent_files->findItem(m_pm_recent_files->idAt(i))->whatsThis();
  }
  if (m_dir_browser) {
    m_config->recent_dirs = m_dir_browser->getRecentDirs();
    if (m_config->recent_dirs.count() > m_config->max_recent_dirs) {
      for(int i=m_config->recent_dirs.count()-1; i>= m_config->max_recent_dirs; i--) {
	m_config->recent_dirs.remove(m_config->recent_dirs[i]);
      }
    }
    m_config->category_list = m_dir_browser->m_all_categories;
  }
  m_pdfview->state().x = m_pdfview->contentsX();
  m_pdfview->state().y = m_pdfview->contentsY();
  m_config->write(m_pdfview->state());
  m_config->clean = true;
}


void
QPDFWindow::flush()
{
  printf("<<<<flush>>>>\n");
  if (m_pdfview->isValid()) {
    writeConfig();
    // save text edit position in bookmark -
    // more indications that state should be saved/maintained in this class, not qdocview/qpdfview
    m_pdfview->setCharIndex(m_textedit->getTopCharPosition());
    if (m_textedit->edited()) {
      assert(!m_textedit->isReadOnly());
      m_pdfview->saveNotes(m_textedit->text(), -1, NULL);
      m_textedit->setEdited(false);
    } else {
      // really should be saveXmlState...
      m_pdfview->saveNotes();
    }
  } else {
    printf(">>>> no notes flush, invalid state <<<<<\n");
  }
}

void
QPDFWindow::closeEvent( QCloseEvent *e )
{
  flush();
  QMainWindow::closeEvent(e);
}

void QPDFWindow::closeRenameBar ( )
{
  QString newPath = m_rename_line_edit-> text( );
  QString oldPath = m_pdfview->getPath();

  if (isBusy()) {
    // don't change files if you are using them!!
    return;
  }
  if (newPath != oldPath ) {
    flush();
#ifdef AUTOTEST
    printf("RENAME TURNED OFF FOR DEBUGGING!!\n");
#else
    
    if ( ::rename( oldPath, newPath ) < 0 ) {
      qWarning( "problem renaming file %s to %s errno %d",
		oldPath.latin1(), newPath.latin1(), errno  );
      newPath = oldPath;
    } else {
      if ( ::rename( m_pdfview->notesPath(oldPath), m_pdfview->notesPath(newPath) ) < 0 ) {
	qWarning( "problem renaming file %s to %s errno %d",
		  m_pdfview->notesPath(oldPath).latin1(), 
		  m_pdfview->notesPath(newPath).latin1(), errno  );
      } 
      printf("renamed %s => %s\n", oldPath.latin1(), newPath.latin1());
      //      m_pdfview->setPath(newPath);
      //      updateCaption();
    }
    m_pdfview->setPath(newPath);
#endif
    // problem: will attempt to close out old non-existant file
    m_pdfview->changeState();
  }
  setRenameBar( false );
}

void
QPDFWindow::flashFindBar(int time)
{
  if (! (m_tb_find->isVisible() && !m_flash_timer->isActive()) ) {
    // don't mess with it if we have find bar on manually (visible and timer not active)
    setFindBar(true);
    if (m_flash_timer->isActive()) {
      m_flash_timer->changeInterval(time);
    } else {
      m_flash_timer->start(time, true);
    }
  }
}

void
QPDFWindow::flashFindBarTimeout()
{
  setFindBar(false);
}



void QPDFWindow::layoutPageCounter()
{
  QFont font = qApp->font();
  QFontMetrics fm(font);
  int w = fm.width("5") * 10;
  int h = fm.height() + fm.height()/2;
  //  int w = m_pdfview->screenRes(50);
  //  int h = m_pdfview->screenRes(12);

  if (m_mode == PDF_VIEW && m_pdfview->isRotated()) {
    m_counter-> setGeometry ( m_stack->x() + m_stack->width() - w - 2, m_stack->y(), w, h);
    m_busybar-> setGeometry ( m_stack->x() + m_stack->width() - w - 2, m_stack->y(), w, h);
  } else {
    m_counter-> setGeometry ( m_stack->x() + m_stack->width() - w + 2, 
			      m_stack->y() + m_stack->height() - h + 2, w, h);
    m_busybar-> setGeometry ( m_stack->x() + m_stack->width() - w + 2, 
			      m_stack->y() + m_stack->height() - h + 2, w, h);
  }
}

void QPDFWindow::setPageCount ( int actp, int time )
{
  //  printf("setPageCount(%d,%d)\n", actp, time);
  m_counter-> setText ( QString ( "%1 / %2" ). arg ( actp ). arg ( m_pdfview->lastPageNum() ));
  showPageCount(time);
}

void QPDFWindow::showPageCount(int time)
{
  m_pageCounterTimer->stop();
  m_pageCounterTimer->start(time, TRUE);
  layoutPageCounter();
  m_busybar->hide();
  m_counter->show();
}

void QPDFWindow::pageCounterTimeout()
{
  m_counter->hide();
  if (m_isbusy) {
    m_busybar->show();
  }
}

void QPDFWindow::setBusy ( bool b )
{
  if ( b != m_isbusy ) {
    if ( b ) {
      m_busybar-> beginBusy ( );
      layoutPageCounter();
      m_busybar-> show ( );
      m_counter-> hide ( );
    }
    else {
      m_counter-> show ( );	
      m_busybar-> hide ( );
      m_busybar-> endBusy ( );
    }		
    m_isbusy = b;
  }
}

/*
  m_isbusy is for display purposes, while this one indicates
  that it is actually reading the document and may process events
  in the middle of it, so don't screw with the file!!
*/
bool 
QPDFWindow::isBusy ( ) const
{
  return m_pdfview->isBusy();
}

