#ifndef __QPDF_H__
#define __QPDF_H__


//#define AUTOTEST

#define INCLUDE_MENUITEM_DEF 

#include "aconf.h"

#include "qpdfview.h"

#include "signal.h"

#include <qpe/qpeapplication.h>
#include <qpe/resource.h>
#include <qpe/applnk.h>
#include <qpe/qcopenvelope_qws.h>
#include <qpe/fileselector.h>
#undef CONFIG_H  // we need this because xpdf has a config.h too!
#include <qpe/config.h>
#include <qmainwindow.h>

#include <qclipboard.h>
#include <qpe/qpetoolbar.h>
#include <qtoolbutton.h>
#include <qpe/qpemenubar.h>

#include <qpopupmenu.h>
#include <qmenudata.h>
#include <qwidgetstack.h>
#include <qtimer.h>
#include <qfileinfo.h>
#include <qstring.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qlayout.h>
#include <qdialog.h>
#include <qlabel.h>
#include <qmessagebox.h>

#include "qpdffileselector.h"
#include <qpe/fileselector.h>
#include "textedit.h"

#include <qobjectlist.h>
#include <qsignalslotimp.h>
#include <qpe/lnkproperties.h>

static void removeObjFromList( QObjectList *objList, const QObject *obj,
			       bool single=FALSE )
{
    if ( !objList )
	return;
    int index = objList->findRef( obj );
    while ( index >= 0 ) {
	objList->remove();
	if ( single )
	    return;
	index = objList->findNextRef( obj );
    }
}


// goal here is to be able to turn Z on and use it right away.
// without this hack, it will stay frozen until all files have been scanned again
// (a pretty long delay with a large CF card!)
// since TrollTech has not allowed us any nice virtual functions or other controls,
// [hey, if anyone from Troll reads this, making reread() virtual would be nice for many things!]
// this hacks into the guts of the class, and the QObject message system

class HackedFileSelector : public FileSelector {
  Q_OBJECT
 public:
  HackedFileSelector( const QString &mimefilter, QWidget *parent, const char *name, 
		      bool newVisible = TRUE, bool closeVisible = TRUE ) 
    : FileSelector(mimefilter, parent, name, newVisible, closeVisible) {
    // don't refresh files whenever external cards are remounted (which happens when you
    // turn the Zaurus on).  
    // I like to be able to start reading as soon as the power is on.
    QObject *obj;
    int err;
    if ( ((QObject*)view)->senderObjects ) {			// disconnect from senders
	QObjectList *tmp = ((QObject*)view)->senderObjects;
	senderObjects = 0;
	obj = tmp->first();
	while ( obj ) {				// for all senders...
	  if (strcmp(obj->className(), "QCopChannel") == 0) {
	    err = obj->disconnect( (QObject*)view );
	    //	    printf("disconnect sender class %s, err = %d\n", obj->className(), err);
	    err = connect(obj, SIGNAL(received(const QCString &, const QByteArray &)),
			  this, SLOT(hackedCardMessage( const QCString &, const QByteArray &)) );
	    //	    printf("reconnect err = %d\n", err);
	  }
	  obj = tmp->next();
	}
    }
#ifdef AUTOTEST
    // during AUTOTEST don't get into right-mouse (press-hold mouse) menus that allow
    // files to be deleted!! (only happens on x86 qtopia)
    err = disconnect((QObject*)view, SIGNAL( mouseButtonPressed( int, QListViewItem *, const QPoint &, int ) ),
		     this, SLOT( filePressed( int, QListViewItem *, const QPoint &, int ) ) );
    printf("disconnect mouseButtonPressed err=%d\n", err);
    connect( (QObject*)view, SIGNAL( mouseButtonPressed( int, QListViewItem *, const QPoint &, int ) ),
	     this, SLOT( hackedFilePressed( int, QListViewItem *, const QPoint &, int ) ) );
    printf("connect mouseButtonPressed err=%d\n", err);
#endif
    do_reread_on_open = false;
  }


  public slots:

    // card inserted, or power on
    // in either case, just set a flag rather than freezing
    // the machine until we have reread all the directories!
    // it will reread when the browser is used.
  void hackedCardMessage( const QCString &msg, const QByteArray &) {
    if ( msg == "mtabChanged()" ) {
      printf("got card change, set flag\n");
      do_reread_on_open = true;
    }
  }  
  

  void hackedFilePressed( int button, QListViewItem *i, const QPoint &, int )  {
      printf("inside hackedFilePressed i=%d\n", i);
  }
  //    if ( !i )
  //        return;
  //    if ( button == Qt::RightButton ) {
  //      DocLnk l = ((FileSelectorItem *)i)->file();
  //      LnkProperties prop( &l );
  //      prop.showMaximized();
  //      prop.exec();
  //      d->cm->reload();
  //      do_reread_on_open = true;
  //    }
  //  }

  void reopen() {
    if (do_reread_on_open) {
      reread();
    }
    do_reread_on_open = false;
  }

 private:
  bool do_reread_on_open;

};


//#define QPDF_QPE_ONLY 1

//class QPEOutputDev;
//class PDFDoc;

class DocLnk;
class QWidgetStack;
class QLineEdit;

const int MAX_PAGES_ON_PAGE_MENU = 10;

enum {
  PM_VIEW_FIT_TO_WIDTH = 2, PM_VIEW_FIT_TO_PAGE = 3
};

enum {
  PM_EDIT_CUT, PM_EDIT_COPY, PM_EDIT_PASTE,
};

enum {
  PM_PAGE_MARK = 20001, 
  PM_PAGE_NEXT, PM_PAGE_PREV, PM_PAGE_FIRST, PM_PAGE_LAST
};




struct QPDFstate : public QDocState {
  int wordPosition0, wordPosition1;
  int charPosition0, charPosition1;
  QPoint pixPosition0, pixPosition1;
  QPoint absPixPosition0, absPixPosition1;
  bool range;

  void clear();
  QPDFstate();

#if 0
  void set(QDocState &s) {
    QDocState::set(s);
    if (mode == PDF_VIEW) {
      pixPosition0.setX(x);
      pixPosition0.setY(y);
      printf("QPDFstate::set x=%d y=%d\n", x, y);
    } else {
      charPosition0 = char_index;
      printf("QPDFstate::set char_index=%d\n", char_index);
    }
  }
#endif
};

class QPDFWindow : public QMainWindow {
	Q_OBJECT

public:
	QPDFWindow ( );
	virtual ~QPDFWindow ( );
	void startup();

	void save(QPDFstate *state);
	void restore(QPDFstate *state);

public slots:

	void openDirectoryBrowser ( );
	void openDocumentBrowser ( );
	void openPreferences();
	
	void openFile ( const QString & );
	void openFile ( const DocLnk & );	
	void openRecentFile( int );

	void loadDirectoryNotes();

	void firstPage ( );
	void prevPage ( );
	void nextPage ( );
	void lastPage ( );
	
	void setPageOffset();

	void gotoPage(int n);
	void gotoPagePos ( int n, int x, int y);
	void gotoPageMenu(int n);

	void setZoom ( int z );

	void setFullscreen ( bool b = true );
	void toggleFullscreen ( );

	void setRotate( bool b );
	void toggleRotate ( );	

	void setHighlight( bool b );
	void toggleHighlight ( );
	void toggleReProportionText();
	void toggleReProportionAndRender();
	void toggleRenderText();

	void setNoteMode ( );
	void setTextMode ( );
	void setPDFMode ( );
	void setMode( ViewMode );

	void saveState();
	void forwardState();
	void backwardState();

	void updateMenuItems();
	void changeStateComplete();

	void toggleFindBar ( );	
	void setFindBar ( bool active );
	void closeFindBar ( );
	void search(const int search_char);
	void search_again();

	void selectionChanged( const QString &str );

	bool findText( const QString &str, bool reverse, bool multipage );

	void findTextInPage ( const QString & );	
	void findTextForward ( );
	void findTextBackward ( );
	void findTextMultipage ( );
	void findTextBackwardMultipage ( );
	
	void openRenameBar ( );
	void setRenameBar ( bool active );
	void abortRenameBar ( );
	void closeRenameBar ( );

	void openHelp();

	void setDocument ( const QString & );
	
	void editCopy();
	void editCut();
	void editPaste();

	//	void newDocument(int);

	void findNoMatch(bool);
	void flashFindBar(int time = 2000);

	void showPageCount(int time = 4000);
	void setPageCount(int page, int time = 4000);

	bool isBusy ( ) const;
	void setBusy ( bool b = true );

	void flush();

private slots:
	void keyPressEvent(QKeyEvent*);
	void closeFileSelector ( );
	void updateCaption ( );
	void pageCounterTimeout();
	void flashFindBarTimeout();
	void cancelTimeout();

	void changePriority(int);
	void changeCategory(int);
	void newCategory();

#ifdef AUTOTEST
	void autotester();
	void autotester_timer_event();
#else
	void autotester() {};
	void autotester_timer_event() {};
#endif

protected:

	virtual void resizeEvent ( QResizeEvent *e );
	virtual void focusInEvent ( QFocusEvent *e );
	virtual void closeEvent( QCloseEvent * );
	
	//	QString changedRegion(QString s, QString org_s);

private:
	bool forwardStateAllowed();
	bool backwardStateAllowed();

	bool makePageMenu();
	void layoutPageCounter();
	void writeConfig();

	QWidgetStack *m_stack;
	QPDFView *m_pdfview;
	TextEdit *m_textedit;

	QpdfFileSelector *m_dir_browser;
	HackedFileSelector *m_doc_browser;
	
	QToolBar *m_tb_menu;
	QToolBar *m_tb_find;
	QToolBar *m_tb_rename;

	QLineEdit *m_find_line_edit;
	QLineEdit *m_rename_line_edit;

	QPopupMenu *m_pm_priority;
	QPopupMenu *m_pm_categories;
	QPopupMenu *m_pm_file;
	QPopupMenu *m_pm_recent_files;
	QPopupMenu *m_pm_edit;
	QPopupMenu *m_pm_view;
	QPopupMenu *m_pm_zoom1;
	int m_pm_zoom1_id;
	QPopupMenu *m_pm_zoom2;
	int m_pm_zoom2_id;
	QPopupMenu *m_pm_page;
	int m_pm_page_forward;
	int m_pm_page_backward;
	int m_pm_page_start_pagenum_index;
	int m_submenu_size;

	int m_pm_load_directory_notes;

	int m_pm_view_rotate;
	int m_pm_view_fullscreen;
	int m_pm_view_highlight;
	int m_pm_view_reproportion_text;
	int m_pm_view_reproportion_and_render;
#ifdef RENDER_EMBEDDED_SYMBOL_FONTS
	int m_pm_view_render_text;
#endif
	int m_pm_view_find;
	int m_pm_view_zoomin;
	int m_pm_view_zoomout;

	QButton *m_but_next_page;
	QButton *m_but_prev_page;
	QButton *m_but_pdf;
	QButton *m_but_note;
	QButton *m_but_text;
	QPixmap m_pdf_move_pixmap;
	QPixmap m_pdf_text_select_pixmap;

	QToolButton *m_to_find, *m_to_full;

	bool m_fullscreen;
	bool m_highlights;
	bool m_rotated;

	ViewMode m_mode;  // kept in sync with qdoview's state.mode (not good design)

	QTimer *m_flash_timer;

	QTimer *m_pageCounterTimer;
	QLabel *m_counter;


	QBusyBar *m_busybar;
	bool m_isbusy;

	QTimer *m_cancel_timer;
	QLabel *m_cancel_message;

	QPDFconfig *m_config;

	int m_current_history_state; // index to backward+1 or foward-1
	int m_last_history_state;    // index to write new states
	QPDFstate m_state_history[MAX_STATE_HISTORY];

	QPDFstate m_new_state;

	QLabel *m_please_wait;

	
};


#endif
