#include "qpdf.h"
#include <qpushbutton.h>
#include <qradiobutton.h>
#include <qcheckbox.h>
#include <qdialog.h>

QPDFconfig::QPDFconfig(QWidget *parent) : qpdf_configuration(parent, "QPDFconfig", true)
{
  c7xx = false;
  if (qApp->desktop()->width() > 600 || qApp->desktop()->height() > 600) {
    c7xx = true;
  }
  connect(okay, SIGNAL(pressed()), this,  SLOT(okaySlot()));
  connect(cancel, SIGNAL(pressed()), this,  SLOT(cancelSlot()));
  clean = true;
}

void
QPDFconfig::run()
{
  document_browser->setChecked(startup_document_browser);
  directory_browser->setChecked(startup_directory_browser);
  last_document->setChecked(startup_last_viewed_document);
  last_view->setChecked(open_last_viewed_page);
  notes_view->setChecked(open_notes_view);
  page_one->setChecked(!open_notes_view && !open_last_viewed_page);
  render_text->setChecked(default_render_text);
  adjust_spacing->setChecked(default_adjust_spacing);
  zoom_percent->setValue((int)(default_zoom*100.0));
  font_size->setValue(default_font_size);
  font_name->setText(default_font_name);
  exec();
}

void
QPDFconfig::okaySlot()
{
  accept();
  startup_document_browser = document_browser->isChecked();
  startup_directory_browser = directory_browser->isChecked();
  startup_last_viewed_document = last_document->isChecked();
  open_last_viewed_page = last_view->isChecked();
  open_notes_view = notes_view->isChecked();
  default_render_text = render_text->isChecked();
  default_adjust_spacing = adjust_spacing->isChecked();
  default_zoom = (float)zoom_percent->value() / 100.0;
  if (default_zoom <= 0.0) {
    default_zoom = 0.5;
  }
  default_font_size = font_size->value();
  if (default_font_size > 24) {
    default_font_size = 24;
  }
  if (default_font_size < 4) {
    default_font_size = 4;
  }
  default_font_name = font_name->text();
}

void
QPDFconfig::cancelSlot()
{
  reject();
}
  
void
QPDFconfig::read()
{
  Config cfg("qpdf");
  QString s;

  cfg.setGroup("startup");
  startup_document_browser = cfg.readBoolEntry("startup_document_browser",false);
  startup_directory_browser = cfg.readBoolEntry("startup_directory_browser",false);
  startup_last_viewed_document = cfg.readBoolEntry("startup_last_viewed_document",true);

  open_notes_view = cfg.readBoolEntry("open_notes_view",false);
  open_last_viewed_page = cfg.readBoolEntry("open_last_viewed_page",true);

  cfg.setGroup("defaults");
  default_render_text = cfg.readBoolEntry("default_render_text",false);
  default_adjust_spacing = cfg.readBoolEntry("default_adjust_spacing",false);

  default_zoom = cfg.readEntry("default_zoom","-1.0").toFloat();

  if (default_zoom < 0.0) {
    default_zoom = c7xx? 0.7 : 0.9;
  }

  default_font_size = cfg.readNumEntry("default_font_size",-1);
  if (default_font_size == -1) {
    // 16/10 a little small, 18/10 view menu too long, 24/12 was too big
    default_font_size = c7xx? 16 : 10;
  }

  default_font_name = cfg.readEntry("default_font_name");

  cfg.setGroup("browser");
  category_list = cfg.readListEntry("category_list",'|');
  recent_dirs = cfg.readListEntry("recent_dirs",'|');
  recent_files = cfg.readListEntry("recent_files",'|');
  max_recent_dirs = cfg.readNumEntry("max_recent_dirs",12);

  cfg.setGroup("menus");
  max_recent_files = cfg.readNumEntry("max_recent_files",12);

  cfg.setGroup("bookmark");
  last_state.clear();
  last_state.load(cfg.readEntry("last_state"),0,true);
  printf("reading last state = %s\n", cfg.readEntry("last_state").latin1());
  printf("loaded last state:  ");
  last_state.dump();
}

void
QPDFconfig::write(QDocState &state)
  
{
  Config cfg("qpdf");

  cfg.setGroup("startup");
  cfg.writeEntry("startup_document_browser", startup_document_browser);
  cfg.writeEntry("startup_directory_browser", startup_directory_browser);
  cfg.writeEntry("startup_last_viewed_document", startup_last_viewed_document);
  cfg.writeEntry("open_notes_view", open_notes_view);
  cfg.writeEntry("open_last_viewed_page", open_last_viewed_page);

  cfg.setGroup("defaults");
  cfg.writeEntry("default_render_text", default_render_text);
  cfg.writeEntry("default_adjust_spacing", default_adjust_spacing);
  cfg.writeEntry("default_zoom", QString::fromLatin1("%1").arg(default_zoom));
  cfg.writeEntry("default_font_size", default_font_size);
  cfg.writeEntry("default_font_name", default_font_name);
  
  cfg.setGroup("browser");
  cfg.writeEntry("category_list", category_list, '|');
  cfg.writeEntry("recent_dirs", recent_dirs, '|');
  cfg.writeEntry("recent_files", recent_files, '|');
  cfg.writeEntry("max_recent_dirs", max_recent_dirs);

  cfg.setGroup("menus");
  cfg.writeEntry("max_recent_files", max_recent_files);

  if (state.isValid()) {
    cfg.setGroup("bookmark");
    QString s;
    state.save(s);
    cfg.writeEntry("last_state", s);
  }
}


