/*
keep favorite directories cfg list (or keep count of usage automatically)


*/

#include "qpdf.h"

#include <qwmatrix.h>
#include <qcheckbox.h>
#include <qcombobox.h>
#include <qheader.h>
#include <qlabel.h>
#include <qabstractlayout.h>
#include <qlayout.h>
#include <qlineedit.h>
#include <qlistview.h>
#include <qmessagebox.h>
#include <qpainter.h>
#include <qpushbutton.h>
#include <qwidgetstack.h>
#include <qpopupmenu.h>
#include <qdir.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qtimer.h>

#include <qpe/qpeapplication.h>
#include <qpe/applnk.h>
#include <qpe/fileselector.h>
#include <qpe/global.h>
//#include <qpe/mimetype.h>
#include <qpe/resource.h>
#include <qpe/storage.h>
#include <qpe/timeconversion.h>
#include <qpe/stringutil.h>
#include <qpe/config.h>

#include <unistd.h>
#include <stdlib.h>
#include <sys/stat.h>


//QMap<QString,QPixmap> *QpdfFileSelector::m_pixmaps = 0;

namespace {
  int indexByString( const QComboBox *box, const QString &str ){
    int index= 0;
    for(int i= 0; i < box->count(); i++ ){
      if( str == box->text(i ) ){
	index= i;
	break;
      }
    }
    return index;
  }

  class QpdfFileSelectorItem : public QListViewItem {
  public:
    QpdfFileSelectorItem(QpdfFileSelector *selector,
			 QListView *view, const QPixmap &icon, const QString &path, 
			 const int priority, 
			 const QDate &pub_date, 
			 const QDateTime &last_open, 
			 const QDateTime &last_note, 
			 const QDateTime &creation_date,
			 const int num_notes,
			 const QStringList &categories,
			 const QString &mDir, bool isLocked=false, bool isDir=false)
			 : QListViewItem(view) {
      setPixmap(0, icon);
      setText(1, path );
      setText(2, priority >= 0? QString::number(priority) : QString("") );
      setText(3, last_open.isValid()? last_open.toString() : QString("") );
      setText(4, last_note.isValid()? last_note.toString() : QString("") );
      setText(5, isDir? QString("") : QString::number(num_notes) );
      setText(6, pub_date.isValid()? pub_date.toString() : QString("") );
      setText(7, creation_date.isValid()? creation_date.toString() : QString("") );
      setText(8, categories.join(" ") );

      m_selector = selector;
      m_dir = mDir;
      dir = isDir;
      mLocked = isLocked;
      m_priority = priority;
      m_pub_date = pub_date;
      m_last_open = last_open;
      m_last_note = last_note;
      m_num_notes = num_notes;
      m_creation_date = creation_date;
      m_categories = categories;
    }

    bool isLocked() const{
      return mLocked;
    }

    QString directory()const{
      return m_dir;
    }

    bool isDir()const{
      return dir;
    }

    QString path()const{
      return text(1 );
    }

    QString key(int id, bool )const {
      QString ke;
      QDateTime epoch( QDate( 2002, 1, 1 ) );

      switch(id) {
      case 0:
      case 1:  // path
	if( dir ){
	  ke.append("0" );
	  ke.append( text(1) );
	} else{
	  ke.append("1" );
	  ke.append( text(1) );
	}
	break;
      case 2:
	ke.sprintf( "%09d", m_priority );
	break;
      case 3:
	ke.sprintf( "%09d", epoch.secsTo( m_last_open ) );
	break;
      case 4:
	ke.sprintf( "%09d", epoch.secsTo( m_last_note ) );
	break;
      case 5:
	ke.sprintf( "%09d", m_num_notes);
	break;
      case 6:
	ke.sprintf( "%09d", epoch.secsTo( m_pub_date ) );
	break;
      case 7:
	ke.sprintf( "%09d", epoch.secsTo( m_creation_date ) );
	break;
#ifdef CATEGORY_SELECTOR
      case 8:
	for(QStringList::ConstIterator  it = m_selector->m_selected_categories.begin();
	    it != m_selector->m_selected_categories.end(); ++it) {
	  if (m_categories.contains(*it)) {
	    ke += "1";
	  } else {
	    ke += "0";
	  }
	}
	break;
#endif
      default:
	ke = text(id);
	break;
      }
      return ke;
    };

  private:
    bool mLocked:1;
    bool dir:1;
    QString m_dir;
    int m_priority, m_num_notes;
    QDateTime m_last_open, m_last_note;
    QDate m_pub_date;
    QDateTime m_creation_date;
    QpdfFileSelector *m_selector;
    QStringList m_categories;
  };
};


#if 0
QpdfFileSelector::QpdfFileSelector( QWidget *wid, int mode, int selector,
				    const QString &dirName,
				    const QStringList &recent_dirs,
				    const QStringList &categories,
				    const QString &fileName,
				    const QMap<QString,QStringList>& mimeTypes)
  : QWidget( wid, "QpdfFileSelector")
{


  m_mimetypes = mimeTypes;
  if (mode == SAVE )
    m_name = fileName;
  initVars();
  m_mode = mode;
  m_selector = selector;
  m_currentDir =  dirName;
  init();
  m_location->insertStringList(recent_dirs, 0);
  m_location->setCurrentItem( 0 );
  m_all_categories = categories;

  //QTimer::singleShot(6*1000, this, SLOT( slotTest() ) );
}
#endif

QpdfFileSelector::QpdfFileSelector(const QString &mimeFilter, QWidget *parent,
				   const char *name, 
				   const QStringList &recent_dirs,
				   const QStringList &categories,
				   bool newVisible,
				   bool closeVisible)
  : QWidget( parent, name )
{
  if (!mimeFilter.isEmpty() ) {
    QStringList list = QStringList::split(";", mimeFilter );
    m_mimetypes.insert(mimeFilter, list );
  }
  initVars();
  m_currentDir = QPEApplication::documentDir();
  m_mode = FILESELECTOR;
  m_selector = NORMAL;
  m_shClose = closeVisible;
  m_shNew = newVisible;
  m_shLne = false;
  m_shPerm = false;
  m_shYesNo = false;
  init();
  setToolbarVisible(true);
  if (recent_dirs.count() > 0) {
    m_location->insertStringList(recent_dirs, 0);  // put recent dirs at the top
    m_currentDir = recent_dirs[0];
    m_location->setCurrentItem( 0 );
  }
  m_all_categories = categories;

  m_please_wait = new QLabel(this);
  m_please_wait-> setAlignment ( AlignCenter | SingleLine );
  m_please_wait-> hide();
  m_please_wait-> setSizePolicy ( QSizePolicy ( QSizePolicy::Expanding, QSizePolicy::Expanding ));
  m_please_wait-> setAutoResize(true);
  m_please_wait-> setFrameStyle(QFrame::PopupPanel | QFrame::Raised);
  m_please_wait-> setText(tr("Reading directory, please wait..."));

  m_pixmap_size = -1;
}

void
QpdfFileSelector::close()
{
  emit closeMe();
}

QStringList
QpdfFileSelector::getRecentDirs()
{
  QStringList list;
  int i;

  if (m_location) {
    for(i=0; i<m_location->count(); i++) {
      if (m_location->text(i).find("<-") < 0) {
	list.append(m_location->text(i));
      }
    }
  }
  return(list);
}

QpdfFileSelector::~QpdfFileSelector()
{
}

void QpdfFileSelector::setNewVisible( bool visible )
{
  m_shNew = visible;
  if( m_selector == NORMAL ){
#if 0
    delete m_select;
    // we need to initialize but keep the selected mimetype
    QString mime = currentMimeType();
    m_select = new FileSelector( mime ,
				 m_stack, "fileselector",
				 m_shNew, m_shClose);
    connect(m_select, SIGNAL(fileSelected( const DocLnk & ) ),
	    this, SLOT( slotFileBridgeSelected(const DocLnk & ) ) );
    connect(m_select, SIGNAL(closeMe() ),
	    this, SIGNAL(closeMe() ) );
    //connect to close me and other signals as well
    m_stack->addWidget( m_select, NORMAL );
#endif
  }else{
    if (m_shNew) {
      m_new->show();
    } else {
      m_new->hide();
    }
  }
}
void QpdfFileSelector::setCloseVisible( bool visible )
{
  m_shClose = visible;
  if( m_selector == NORMAL ){
    setNewVisible( m_shNew ); // yeah baby
  }else{
    if (visible) {
      m_close->show();
    } else {
      m_close->hide();
    }
  }
}
void QpdfFileSelector::reread()
{
  if( m_selector == NORMAL ){
    setNewVisible( m_shNew ); // make it a initializeSelector
  }else if ( m_selector == EXTENDED || m_selector == EXTENDED_ALL ){
    reparse();
  }
}

const DocLnk *QpdfFileSelector::selected()
{
  DocLnk *lnk = new DocLnk(selectedDocument() );
  return lnk;
}

void QpdfFileSelector::setYesCancelVisible( bool show )
{
  initializeYes(); // FIXME if YesCancel is not shown we will initialize it to hide it :(
  m_shYesNo = show;
  if( m_shYesNo )
    m_boxOk->show();
  else
    m_boxOk->hide();
  
}
void QpdfFileSelector::setToolbarVisible( bool show )
{
  m_shTool = show;
  initializeListView(); // FIXME see above waste of memory
  if(!m_shTool ){
    m_location->hide();
    m_up->hide();
    m_homeButton->hide();
    m_docButton->hide();
  }else{
    m_location->show();
    m_up->show();
    //    m_homeButton->show();
    //    m_docButton->show();
  }
}
void QpdfFileSelector::setPermissionBarVisible( bool show )
{
  m_shPerm = show;
  initializePerm();
  if( m_shPerm )
    m_checkPerm->show();
  else
    m_checkPerm->hide();
}
void QpdfFileSelector::setLineEditVisible( bool show )
{
  if( show ){
    initializeName();
    m_boxName->show();
  }else{
    if( m_shLne && m_boxName != 0 ){ // check if we showed before this is the way to go
      m_boxName->hide();
    }
  }
  m_shLne = show;
}

void QpdfFileSelector::setChooserVisible( bool show )
{
  m_shChooser = show;
  initializeChooser();
  if( m_shChooser ){
    m_boxView->show();
  }else{
    m_boxView->hide();
  }
}

QCheckBox* QpdfFileSelector::permissionCheckbox()
{
  if( m_selector == NORMAL )
    return 0l;
  else
    return m_checkPerm;
}
bool QpdfFileSelector::setPermission()const
{
  return m_checkPerm == 0 ? false : m_checkPerm->isChecked();
}
void QpdfFileSelector::setPermissionChecked( bool check )
{
  if( m_checkPerm )
    m_checkPerm->setChecked( check );
}

void QpdfFileSelector::setMode(int mode) // FIXME do direct raising
{
  m_mode = mode;
  if( m_selector == NORMAL )
    return;
}
void QpdfFileSelector::setShowDirs(bool )
{
  m_dir = true;
  reparse();
}
void QpdfFileSelector::setCaseSensetive(bool caSe )
{
  m_case = caSe;
  reparse();
}
void QpdfFileSelector::setShowFiles(bool show )
{
  m_files = show;
  reparse();
}
///
bool QpdfFileSelector::cd(const QString &path )
{
  m_currentDir = path;
#ifdef AUTOTEST
  if (m_currentDir.find(QPEApplication::documentDir()) != 0) {
    m_currentDir = QPEApplication::documentDir();
  }
#endif
  reparse();
  return true;
}
void QpdfFileSelector::setSelector(int mode )
{
  QString text;
  switch( mode ){
  case NORMAL:
    text = tr("Documents");
    break;
  case EXTENDED:
    text = tr("Files");
    break;
  case EXTENDED_ALL:
    text = tr("All Files");
    break;
  }
  slotViewCheck( text );
}

void QpdfFileSelector::setPopupMenu(QPopupMenu *popup )
{
  m_custom = popup;
  m_showPopup = true;
}

//void QpdfFileSelector::updateL

QString QpdfFileSelector::selectedName() const
{
  QString name;
  if( m_selector == NORMAL ){
    //    DocLnk lnk = m_select->selectedDocument();
    DocLnk lnk = *m_select->selected();
    name = lnk.file();
  }else if( m_selector == EXTENDED || m_selector == EXTENDED_ALL ){
    if ( m_shLne ) {
      name = m_currentDir + "/" +m_edit->text();
    }else{
      QListViewItem *item  = m_View->currentItem();
      if( item != 0 )
	name = m_currentDir + "/" + item->text( 1 );
    }
  }else { // installed view
    ;
  }
  return name;
}
QStringList QpdfFileSelector::selectedNames()const
{
  QStringList list;
  if( m_selector == NORMAL ){
    list << selectedName();
  }else if ( m_selector == EXTENDED || m_selector == EXTENDED_ALL ) {
    list << selectedName(); // FIXME implement multiple Selections
  }
  return list;
}
/** If mode is set to the Dir selection this will return the selected path.
 *
 *
 */
QString QpdfFileSelector::selectedPath()const
{
  QString path;
  if( m_selector == NORMAL ){
    path = QPEApplication::documentDir();
  }else if( m_selector == EXTENDED || m_selector == EXTENDED_ALL ){
    ;
  }
  return path;
}
QStringList QpdfFileSelector::selectedPaths() const
{
  QStringList  list;
  list << selectedPath();
  return list;
}
QString QpdfFileSelector::directory()const
{
  if( m_selector == NORMAL )
    return QPEApplication::documentDir();
  
  return QDir(m_currentDir).absPath();
}

int QpdfFileSelector::fileCount() const
{
  int count;
  switch( m_selector ){
  case NORMAL:
    count = m_select->fileCount();
    break;
    //case CUSTOM:
  case EXTENDED:
  case EXTENDED_ALL:
  default:
    count = m_View->childCount();
    break;
  }
  return count;
}
DocLnk QpdfFileSelector::selectedDocument() const
{
  DocLnk lnk;
  switch( m_selector ){
  case NORMAL:{
    //    lnk = m_select->selectedDocument();
    lnk = *m_select->selected();
    break;
  }
  case EXTENDED:
  case EXTENDED_ALL:
  default:
    lnk = DocLnk( selectedName() ); // new DocLnk
    break;
  }
  return lnk;
}
QValueList<DocLnk> QpdfFileSelector::selectedDocuments() const
{
  QValueList<DocLnk> docs;
  docs.append( selectedDocument() );
  return docs;
}


// slots internal

void QpdfFileSelector::slotOk()
{
  emit ok();
}
void QpdfFileSelector::slotCancel()
{
  emit cancel();
}
void QpdfFileSelector::slotViewCheck(const QString &sel)
{
  if( sel == tr("Documents" ) ){
#if 0
    if( m_select == 0 ){
      // autMime? fix cause now we use All and not the current
      // yes currentMime fixes that for us
      QString mime = currentMimeType();
      m_select = new FileSelector(mime,
				  m_stack, "fileselector",
				  m_shNew, m_shClose);
      connect(m_select, SIGNAL(fileSelected( const DocLnk & ) ),
	      this, SLOT( slotFileBridgeSelected(const DocLnk & ) ) );
      connect(m_select, SIGNAL(closeMe() ),
              this, SIGNAL(closeMe() ) );
      //connect to close me and other signals as well
      
      m_stack->addWidget( m_select, NORMAL );
    }
    m_stack->raiseWidget( NORMAL );
    m_selector = NORMAL;
#endif
  }else if( sel == tr("Files") ){
    m_selector = EXTENDED;
    initializeListView();
    reparse();
    m_stack->raiseWidget( EXTENDED );
  }else if( sel == tr("All Files") ){
    m_selector = EXTENDED_ALL;
    initializeListView();
    reparse();
    m_stack->raiseWidget( EXTENDED ); // same widget other QFileFilter
  }
}

// not yet finished.....
QString QpdfFileSelector::currentMimeType() const
{

  return("application/pdf");
#if 0
  QString mime;
  QString currentText;
  currentText = m_mimeCheck->currentText();
  
  if (tr("All") == currentText ) return QString::null;
  else if (currentText.isEmpty() ) {
    ;
  }else {
    QMap<QString, QStringList>::ConstIterator it;
    it = m_mimetypes.find( currentText );
    if ( it == m_mimetypes.end() ) {
      mime = it.data().join(";");
    }else{
      mime = currentText;
    }
  }
  return mime;
#endif
}

void QpdfFileSelector::slotMimeCheck(const QString &mime)
{
  if( m_selector == NORMAL ){
#if 0
    //if( m_autoMime ){
    QString newMimeType;
    if (mime != tr("All") ) {
      QMap<QString, QStringList>::Iterator it;
      it = m_mimetypes.find(mime);
      if ( it != m_mimetypes.end() ) {
	newMimeType = it.data().join(";");
      }else{
	newMimeType = mime;
      }
    }
    delete m_select;
    m_select = new FileSelector( newMimeType,
				 m_stack, "fileselector",
				 m_shNew, m_shClose);
    
    connect(m_select, SIGNAL(fileSelected( const DocLnk & ) ),
	    this, SLOT( slotFileBridgeSelected(const DocLnk & ) ) );
    connect(m_select, SIGNAL(closeMe() ),
            this, SIGNAL(closeMe() ) );
    //connect to close me and other signals as well
    m_stack->addWidget( m_select, NORMAL );
    m_stack->raiseWidget( NORMAL );
    updateMimes();
    updateMimeCheck();
    m_mimeCheck->setCurrentItem(indexByString( m_mimeCheck, mime) );
#endif
    //}
  }else{ // others
    qWarning("Mime %s", mime.latin1() );
    if(m_shChooser ){
      qWarning("Current Text %s", m_mimeCheck->currentText().latin1() );
      //m_mimeCheck->setCurrentItem(indexByString( m_mimeCheck, mime)  );
    }
    reparse();
  }
  
}
/*
 * Ok if a non dir gets inserted into this combobox
 * we need to change it
 * QFileInfo and dirPath will give us the right Dir
 */
void QpdfFileSelector::slotLocationActivated(const QString &file)
{
  QString name = file.left( file.find("<-", 0, TRUE ) );
  QFileInfo info( name );
  if ( info.isFile() )
    cd(info.dirPath( TRUE ) ); //absolute
  else
    cd(name );
  reparse();
}

void QpdfFileSelector::slotInsertLocationPath(const QString &currentPath, int index)
{
  int n = m_location->count();
  QString path = currentPath;
  if (path.right(1) == "/") {
    path = path.left(path.length()-1);
  }
  while(path.find("//") >= 0) {
    path.remove(path.find("//"),1);
  }
  if (path == "") {
    path = "/";
  }
  for(int i = n-1; i >= 0; i--) {
    if (m_location->text(i) == path) {
      m_location->removeItem(i);
    }
  }
  m_location->insertItem( path, index );
  m_location->setCurrentItem(index);
}

/*
 * Do not crash anymore
 * don't try to change dir to a file
 */
void QpdfFileSelector::locationComboChanged()
{
  QFileInfo info( m_location->lineEdit()->text() );
  if (info.isFile() )
    cd(info.dirPath(TRUE) ); //absolute path
  else
    cd( m_location->lineEdit()->text() );
  reparse();
}

void QpdfFileSelector::init()
{
  m_lay = new QVBoxLayout( this );
  m_lay->setSpacing(0 );
  
  m_stack = new QWidgetStack( this );
  if( m_selector == NORMAL ){
#if 0
    QString mime;
    if (!m_autoMime) {
      if (!m_mimetypes.isEmpty() ) {
	QMap<QString, QStringList>::Iterator it;
	it = m_mimetypes.begin(); // cause we're in the init
	mime = it.data().join(";");
      }
    }
    m_select = new FileSelector(mime,
				m_stack, "fileselector",
				m_shNew, m_shClose);
    
    connect(m_select, SIGNAL(fileSelected( const DocLnk & ) ),
	    this, SLOT( slotFileBridgeSelected(const DocLnk & ) ) );
    connect(m_select, SIGNAL(closeMe() ),
	    this, SIGNAL( closeMe() ) );
    //connect to close me and other signals as well
    
    m_stack->addWidget( m_select, NORMAL );
    m_stack->raiseWidget( NORMAL );
#endif
  }else{ // we're in init so it will be EXTENDED or EXTENDED_ALL
    // and initializeListview will take care of those
    // toolbar get's generade in initializeListView
    initializeListView( ); // will raise the widget as well
    m_stack->raiseWidget( EXTENDED );
  }
  m_lay->addWidget( m_stack, 100 ); // add to the layout 10 = stretch
  
  if( m_shLne ) // the LineEdit with the current FileName
    initializeName();
  
  if( m_shPerm ) // the Permission QCheckBox
    initializePerm();
  
  if( m_shChooser ) // the Chooser for the view and Mimetypes
    initializeChooser();
  
  if( m_shYesNo ) // the Yes No button row
    initializeYes( );
  
  if (m_selector != NORMAL )
    reparse();
}

void QpdfFileSelector::updateMimes()
{
#if 0
  if( m_autoMime ){
    m_mimetypes.clear();
    m_mimetypes.insert( tr("All"), QString::null );
    if( m_selector == NORMAL ){
      DocLnkSet  set;
      Global::findDocuments(&set, QString::null );
      QListIterator<DocLnk> dit( set.children() );
      for( ; dit.current(); ++dit ){
	if( !m_mimetypes.contains( (*dit)->type() ) )
	  m_mimetypes.insert( (*dit)->type(), (*dit)->type() );
      }
    }// else done in reparse
  }
#endif
}

void QpdfFileSelector::initVars()
{
#if 0
  if( m_mimetypes.isEmpty() )
    m_autoMime = true;
  else
    m_autoMime = false;
#endif
  m_autoMime = false;


  m_shClose = false;
  m_shNew = false;
  m_shTool = true;
  m_shPerm = false;
  m_shLne = true;
  m_shChooser = true;
  m_shYesNo = false;
  m_case = false;
  m_dir = true;
  m_files = true;
  m_showPopup = false;

  //  if(m_pixmaps == 0 ) // init the pixmaps
  //    initPics();
  
  // pointers
  m_location = 0;
  m_mimeCheck = 0;
  m_viewCheck = 0;
  m_homeButton = 0;
  m_docButton = 0;
  m_hideButton = 0;
  m_ok = 0;
  m_cancel = 0;
  m_reread = 0;
  m_up = 0;
  m_View = 0;
  m_checkPerm = 0;
  m_pseudo = 0;
  m_pseudoLayout = 0;
  m_select = 0;
  m_stack = 0;
  m_lay = 0;
  m_Oselector = 0;
  m_boxToolbar = 0;
  m_boxOk = 0;
  m_boxName = 0;
  m_boxView = 0;
  m_custom = 0;
  m_edit = 0;
  m_fnLabel = 0;
  m_new = 0;
  m_close = 0;
}

void QpdfFileSelector::addFile(const QString &mime, QFileInfo *info, bool symlink)
{
  QPixmap pix;
  QString dir;
  QString name;
  QDocInfo docinfo;
  bool locked;
  bool selected = false;


  if(!m_files)
    return;
#if 0
  MimeType type( info->absFilePath() );
  //  printf("%s complies(%s) %d  id=%s %d\n", info->absFilePath().latin1(),
  //	 mime.latin1(),
  //	 compliesMime(info->absFilePath(), mime ),
  //	 type.id().latin1(),
  //	 compliesMime( type.id() ) );

  if( !compliesMime(info->absFilePath(), mime ) )
    return;
  if (!compliesMime( type.id() ) )
    return;
  
  if( pix.isNull() )
    pix = Resource::loadPixmap( "UnknownDocument-14");  // FIXME
#endif
  int dot = info->absFilePath().findRev('.');
  QString ext = info->absFilePath().mid(dot>=0? dot+1 : info->absFilePath().length()-3);

  if (ext != "pdf") {
    return;
  }

  pix = m_pixmap_qpdf;

  dir = info->dirPath( true );
  
  if( symlink ) {
    name = info->fileName() + " -> " +info->dirPath() + "/" + info->readLink();
    selected = true;
  } else {
    name = info->fileName();
    if( ( m_mode == OPEN && !info->isReadable() )||
	( m_mode == SAVE && !info->isWritable() ) ){
      
      locked = true; pix = m_pixmap_locked;
    }
    
    docinfo.quickRead(QDocView::notesPath(info->absFilePath()));
  
    int n = docinfo.m_categories.count();
    // be sure category list is up to date (this is n**2, could use hash table)
    for( QStringList::Iterator it = docinfo.m_categories.begin(); 
	 it != docinfo.m_categories.end(); 
	 it++) {
      QString cat;
      cat = (*it);
      if (! m_all_categories.contains(cat)) {
	m_all_categories.append(cat);
      }
      if (m_selected_categories.contains(cat)) {
	selected = true;
      }
    }
  }

  if (m_selected_categories.count() == 0
      || selected) {

    new QpdfFileSelectorItem( this, 
			      m_View, pix, name, 
			      docinfo.m_priority, 
			      docinfo.m_pub_date,
			      docinfo.m_last_open,
			      docinfo.m_last_note,
			      info->lastModified(),
			      docinfo.m_num_notes,
			      docinfo.m_categories,
			      info->dirPath());
  }
}

void QpdfFileSelector::addDir(const QString &/*mime*/, QFileInfo *info, bool symlink )
{
  if(!m_dir)
    return;
  if( m_selector == EXTENDED_ALL || m_selector == EXTENDED ){
    bool locked = false;
    QString name;

    QPixmap pix;
    if( ( m_mode == OPEN && !info->isReadable() ) || ( m_mode == SAVE && !info->isWritable() ) ){
      locked = true;
      if( symlink )
	pix = m_pixmap_locked_link;  
      else
	pix = m_pixmap_locked_dir;
    }else { // readable
      pix = symlink ?  m_pixmap_link : m_pixmap_dir;
    }

    name = symlink ? info->fileName() + "->" + info->dirPath(true) + "/" +info->readLink() : info->fileName() ;

    new QpdfFileSelectorItem( this, m_View, pix, name,
			      -1, QDate(), QDateTime(), QDateTime(), QDateTime(), -1, QStringList(),
			      info->dirPath(), false, true);
    
  }// else CUSTOM View
}
void QpdfFileSelector::delItems()
{
  
}
void QpdfFileSelector::initializeName()
{
  /**  Name Layout Line
   *   This is the Layout line arranged in
   *   horizontal way each components
   *   are next to each other
   *   but we will only do this if
   *   we didn't initialize a while ago.
   */
  if( m_boxName == 0 ){
    m_boxName = new QHBox( this ); // remove this this? or use a QHBox
    m_fnLabel = new QLabel( m_boxName );
    m_fnLabel->setText( tr("Name:") );
    m_edit = new QLineEdit( m_boxName );
    m_edit->setText( m_name );
    //m_boxName->addWidget( m_fnLabel );
    m_boxName->setMargin( 5 );
    m_boxName->setSpacing( 8 );
    //m_boxName->setStretchFactor(m_edit, 100 ); // 100 is stretch factor
    
    m_lay->addWidget( m_boxName, 0 ); // add it to the topLevel layout
  }// else we already initialized
  // maybe show the components?
  //
}
void QpdfFileSelector::initializeYes()
{
  /** The Save Cancel bar
   *
   */
  if( m_boxOk == 0 ){
    m_boxOk = new QHBox( this );
    m_ok = new QPushButton( tr("&Save"),m_boxOk , "save" );
    m_cancel = new QPushButton( tr("C&ancel"), m_boxOk, "cancel" );
    
    //m_boxOk->addWidget( m_ok );
    //m_boxOk->addWidget( m_cancel );
    m_boxOk->setMargin( 5 );
    m_boxOk->setSpacing( 10 );
    m_lay->addWidget( m_boxOk, 0 );
    
    connect( m_ok, SIGNAL( clicked() ),
	     this, SLOT(slotOk() ) );
    connect( m_cancel, SIGNAL( clicked() ),
	     this, SLOT( slotCancel() ) );
  }
}
/*
 * OK m_mimeCheck is a QComboBox we now want to fill
 * out that combobox
 * if automime we need to update the mimetypes
 */
void QpdfFileSelector::updateMimeCheck() {
#if 0
  m_mimeCheck->clear();
  if (m_autoMime ) {
    //m_mimeCheck->insertItem( tr("All") );
    updateMimes();
  }
  
  QMap<QString, QStringList>::Iterator it;
  for (it = m_mimetypes.begin(); it != m_mimetypes.end(); ++it ) {
    m_mimeCheck->insertItem( it.key() );
  }
#endif
}

void QpdfFileSelector::initializeChooser()
{
  if( m_boxView == 0 ){
    m_boxView = new QHBox( this );
    m_viewCheck = new QComboBox( m_boxView, "view check");
    m_mimeCheck = new QComboBox( m_boxView, "mime check");
    m_boxView->setSpacing( 8 );
    m_lay->addWidget(m_boxView, 0 );
    
    m_viewCheck->insertItem( tr("Documents") );
    m_viewCheck->insertItem( tr("Files") );
    m_viewCheck->insertItem( tr("All Files") );
    updateMimeCheck();
    
    connect( m_viewCheck, SIGNAL( activated(const QString & ) ),
	     this, SLOT( slotViewCheck(const QString & ) ) );
    connect( m_mimeCheck, SIGNAL( activated(const QString & ) ),
	     this, SLOT( slotMimeCheck( const QString & ) ) );
  }
}
void QpdfFileSelector::initializeListView()
{
  if( m_pseudo == 0 ){
    m_pseudo = new QWidget( m_stack, "Pseudo Widget");
    m_pseudoLayout = new QVBoxLayout( m_pseudo );
    // toolbar
    m_boxToolbar = new QHBox( m_pseudo );
    m_boxToolbar->setSpacing(0 ); // next to each other please
    
    // toolbar members
    {
      // location QComboBox
      m_location = new QComboBox( m_boxToolbar );
      m_location->setEditable( TRUE );
      m_location->setInsertionPolicy(QComboBox::AtTop);
      m_location->setDuplicatesEnabled( FALSE );
      connect( m_location, SIGNAL(activated(const QString &) ),
	       this, SLOT( slotLocationActivated(const QString &) ) );
      connect( m_location->lineEdit(), SIGNAL(returnPressed() ),
	       this, SLOT(locationComboChanged() ) );

      // Hey Trolltech, I thought this was what widgets were supposed to do for you!
      m_location->setMaximumWidth( qApp->desktop()->width()-70 );

      // UP Button
      m_up = new QPushButton(Resource::loadIconSet("up"),"",
			     m_boxToolbar,"cdUpButton");
      m_up->setFixedWidth( 30 );
      //      m_up->setFixedSize( QSize( 20, 20 ) );
      connect(m_up ,SIGNAL(clicked()),this,SLOT(cdUP()  ) );
      m_up->setFlat(TRUE);
      
      // Home Button
      m_homeButton = new QPushButton(Resource::loadIconSet("home") ,
				     "", m_boxToolbar);
      //      m_homeButton->setFixedSize( QSize( 20, 20 ) );
      m_homeButton->setFixedWidth( 30 );
      connect(m_homeButton,SIGNAL(clicked()),this,SLOT(slotHome() ) );
      m_homeButton->setFlat(TRUE);
      // Documents Button
      m_docButton = new QPushButton(Resource::loadIconSet("DocsIcon"),"",
				    m_boxToolbar,"docsButton");
      //      m_docButton->setFixedSize( QSize( 20, 20 ) );
      m_docButton->setFixedWidth( 30 );
      connect(m_docButton,SIGNAL(clicked()),this,SLOT(slotDoc() ) );
      m_docButton->setFlat(TRUE);

      m_homeButton->hide();
      m_docButton->hide();
      
      // Close button
      m_close = new QPushButton( Resource::loadIconSet( "close"), "",
				 m_boxToolbar );
      connect( m_close, SIGNAL(clicked() ), this, SIGNAL(closeMe()  ) );
      m_close->setFixedWidth( 30 );
      //      m_close->setFixedSize( 20, 20 );
      
      //      m_boxToolbar->setFixedHeight( 20 );
      m_pseudoLayout->addWidget(m_boxToolbar );
      
      // let;s fill the Location ComboBox
      StorageInfo storage;
      const QList<FileSystem> &fs = storage.fileSystems();
      QListIterator<FileSystem> it ( fs );
      for( ; it.current(); ++it ){
	const QString disk = (*it)->name();
	const QString path = (*it)->path();
	m_location->insertItem(path+ "<-"+disk);
      }

      QString currentPath = m_currentDir;
      if (currentPath.right(1) == "/") {
	currentPath = currentPath.left(currentPath.length()-1);
      }
      m_location->insertItem( currentPath );

      // due to the New and Close button we can not simply hide m_boxToolBar to not show it
      if( !m_shTool ){
	m_location->hide(  );
	m_up->hide(  );
	m_homeButton->hide(  );
	m_docButton->hide( );
      }
      if(!m_shClose )
	m_close->hide();
      //if(!m_shNew)
      //m_close->hide();
      
    } // off toolbar
    // the Main ListView
    // make a QWidgetStack first so Views can share the Toolbar
    m_View = new QListView( m_pseudo, "Extended view");
    QPEApplication::setStylusOperation( m_View->viewport(),
					QPEApplication::RightOnHold);
    m_View->addColumn(" " );
    m_View->addColumn(tr("Name"), -1 );
    m_View->addColumn(tr("Priority"), -1  );
    m_View->addColumn(tr("Last Open"), -1 );
    m_View->addColumn(tr("Last Note"), -1 );
    m_View->addColumn(tr("#Notes"), -1 );
    m_View->addColumn(tr("Published"), -1  );
    m_View->addColumn(tr("Created"), -1  );
    m_View->addColumn(tr("Categories"), -1 );
    QHeader *header = m_View->header();
    //    header->hide();

    m_View->setSorting( 1 );
    m_View->setAllColumnsShowFocus( TRUE );
    
    connect(m_View, SIGNAL(selectionChanged() ),
	    this, SLOT(slotSelectionChanged() ) );
    
    connect(m_View, SIGNAL(returnPressed( QListViewItem * )),
	    this, SLOT(slotReturnPressed(QListViewItem*)));

    connect(m_View, SIGNAL(mouseButtonClicked(int, QListViewItem*, const QPoint &, int) ),
	    this, SLOT(slotClicked( int, QListViewItem *, const QPoint &, int) ) );
    
#if 0
    connect(m_View, SIGNAL(clicked(QListViewItem *) ),
	    //    connect(m_View, SIGNAL(currentChanged(QListViewItem *) ),
	    this, SLOT(slotCurrentChanged(QListViewItem * ) ) );

    connect(m_View, SIGNAL(mouseButtonPressed(int, QListViewItem *, const QPoint &, int )),
	    this, SLOT(slotRightButton(int, QListViewItem *, const QPoint &, int  ) ) );
#endif    

    m_pseudoLayout->addWidget( m_View, 200 );  // was 288, where did that come from?
    m_stack->addWidget( m_pseudo, EXTENDED );
  }
}
void QpdfFileSelector::initializePerm()
{
  if( m_checkPerm == 0 ){
    m_checkPerm = new QCheckBox(tr("Set Permission"), this, "perm");
    m_checkPerm->setChecked( false );
    m_lay->addWidget( m_checkPerm );
    
  }
}

QPixmap
QpdfFileSelector::scaleIcon(QPixmap pixmap, int size)
{
  QImage image(pixmap.convertToImage());
  QPixmap scaled_pixmap;
  size += size/4;
  image = image.smoothScale(size, size);
  scaled_pixmap.convertFromImage(image);
  return(scaled_pixmap);
}

void QpdfFileSelector::initPics()
{
  //  m_pixmaps = new QMap<QString,QPixmap>;
  //  printf("resize pics size %d\n", m_View->font().pixelSize());
  m_pixmap_size = m_View->font().pixelSize();

  m_pixmap_qpdf = scaleIcon(Resource::loadPixmap("qpdf/qpdf_icon"), m_pixmap_size);
  m_pixmap_dir = scaleIcon(Resource::loadIconSet("folder").pixmap(), m_pixmap_size);
  m_pixmap_locked = scaleIcon(Resource::loadIconSet("locked").pixmap(), m_pixmap_size);
  // get a better link icon
  m_pixmap_link = scaleIcon(Resource::loadIconSet( "next").pixmap(), m_pixmap_size);
  m_pixmap_locked_dir = scaleIcon(Resource::loadIconSet( "lockedfolder").pixmap(), m_pixmap_size);
  m_pixmap_locked_link = m_pixmap_link;

#if 0
  QPainter painter( &m_pixmap_dir_link );
  painter.drawPixmap( m_pixmap_dir.width()-m_pixmap_link.width(), m_pixmap_dir.height()-m_pixmap_link.height(), lnk );
  m_pixmap_dir.setMask( m_pixmap_dir.createHeuristicMask( FALSE ) );
  m_pixmap_dir_link = pm;
  
  QPixmap pm2 = Resource::loadIconSet( "lockedfolder" );
  QPainter pen(&pm2 );
  pen.drawPixmap(pm2.width()-m_pixmap_link.width(), pm2.height()-m_pixmap_link.height(), lnk );
  pm2.setMask( pm2.createHeuristicMask( FALSE ) );
  m_pixmap_locked_link = pm;
  m_pixmaps->insert("symlinkedlocked", pm2 );
#endif
}

// if a mime complies with the m_mimeCheck->currentItem
bool QpdfFileSelector::compliesMime( const QString &path, const QString &mime )
{
#if 0
  if( mime == "All" )
    return true;
  MimeType type( path );
  if( type.id() == mime )
    return true;
  return false;
#endif
  return true;
}

/*  check if the mimetype in mime
 *  complies with the  one which is current
 */
/*
 * We've the mimetype of the file
 * We need to get the stringlist of the current mimetype
 *
 * mime = image/jpeg
 * QStringList = 'image/*'
 * or QStringList = image/jpeg;image/png;application/x-ogg
 * or QStringList = application/x-ogg;image/*;
 * with all these mime filters it should get acceptes
 * to do so we need to look if mime is contained inside
 * the stringlist
 * if it's contained return true
 * if not ( I'm no RegExp expert at all ) we'll look if a '/*'
 * is contained in the mimefilter and then we will
 * look if both are equal until the '/'
 */
bool QpdfFileSelector::compliesMime( const QString& mime ) {
#if 0
  //  qWarning("mimetype is %s", mime.latin1() );
  QString currentText;

  currentText = m_mimeCheck->currentText();
  
  //  qWarning("current text is %s", currentText.latin1() );
  QMap<QString, QStringList>::Iterator it;
  QStringList list;
  if ( currentText == tr("All") ) return true;
  else if ( currentText.isEmpty() && !m_mimetypes.isEmpty() ) {
    it = m_mimetypes.begin();
    list = it.data();
  }else if ( currentText.isEmpty() ) return true;
  else{
    it = m_mimetypes.find(currentText );
    if ( it == m_mimetypes.end() ) list << currentText;
    else list = it.data();
  }
  // dump it now
  //for ( QStringList::Iterator it = list.begin(); it != list.end(); ++it ) {
  //    qWarning( "%s", (*it).latin1() );
  //}
  
  if ( list.contains(mime) ) return true;
  //  qWarning("list doesn't contain it ");
  QStringList::Iterator it2;
  int pos;
  for ( it2 = list.begin(); it2 != list.end(); ++it2 ) {
    pos = (*it2).findRev("/*");
    if ( pos >= 0 ) {
      if ( mime.contains( (*it2).left(pos) ) ) return true;
    }
  }
  return false;
#endif
  return true;
}
void QpdfFileSelector::slotFileSelected( const QString &string )
{
  if( m_shLne )
    m_edit->setText( string );
  emit fileSelected( string );
}
void QpdfFileSelector::slotFileBridgeSelected( const DocLnk &lnk )
{
  slotFileSelected( lnk.name() );
  emit fileSelected( lnk );
}
void QpdfFileSelector::slotSelectionChanged()
{
  
}
void QpdfFileSelector::slotCurrentChanged(QListViewItem* item )
{
  if( item == 0 )
    return;
  if( m_selector == EXTENDED || m_selector == EXTENDED_ALL ) {
    QpdfFileSelectorItem *sel = (QpdfFileSelectorItem*) item; // start to use the C++ casts ;)
    qWarning("current changed");
    if(!sel->isDir() ){
      if( m_shLne )
	m_edit->setText( sel->text(1) );
      
      if (m_mode == FILESELECTOR ) {
	QStringList str = QStringList::split("->", sel->text(1) );
	QString path =sel->directory() + "/" + str[0].stripWhiteSpace();
	//	emit fileSelected(path  );
	DocLnk lnk( path );
	emit fileSelected(lnk );
	//	printf("slotCurrentChanged\n");
      }
    }
  }
}

void QpdfFileSelector::slotReturnPressed(QListViewItem *item)
{
  slotClicked(Qt::LeftButton, item, QPoint(), 0);
}

void QpdfFileSelector::slotClicked( int button, QListViewItem *item, 
				    const QPoint &, int)
{
  if ( item == 0 )
    return;
  
  if( button != Qt::LeftButton )
    return;
  
  switch( m_selector ){
  default:
    break;
  case EXTENDED: // fall through
  case EXTENDED_ALL:{
    QpdfFileSelectorItem *sel = (QpdfFileSelectorItem*)item;
    if(!sel->isLocked() ){
      QStringList str = QStringList::split("->", sel->text(1) );
      if( sel->isDir() ){
	QString newdir = sel->directory() + "/" + str[0].stripWhiteSpace() ;
	//	printf("change dir %s\n", newdir.latin1());
	cd( newdir );
	// if MODE Dir m_shLne set the Text
      }else{
	if( m_shLne )
	  m_edit->setText(  str[0].stripWhiteSpace() );
        qWarning("selected here in slot clicked");
	QString filepath = sel->directory() + "/" + str[0].stripWhiteSpace() ;
	emit fileSelected( filepath );
	DocLnk lnk( filepath );
        qWarning("file selected %s\n", filepath.latin1());
	emit fileSelected( lnk );
      }
    }
    break;
  }
  }
}
void QpdfFileSelector::slotRightButton(int button, QListViewItem *item, const QPoint &, int )
{
  if( item == 0 )
    return;
  
  if( button != Qt::RightButton )
    return;
  slotContextMenu( item );
}

void QpdfFileSelector::slotContextMenu( QListViewItem * /*item*/)
{
}

void QpdfFileSelector::slotChangedDir()
{
  QpdfFileSelectorItem *sel = (QpdfFileSelectorItem*)m_View->currentItem();
  if(sel->isDir() ){
    QStringList str = QStringList::split("->", sel->text(1) );
    cd( sel->directory() + "/" + str[0].stripWhiteSpace() );
  }
}

void QpdfFileSelector::slotOpen()
{
  QpdfFileSelectorItem *sel = (QpdfFileSelectorItem*)m_View->currentItem();
  if(!sel->isDir() ){
    QStringList str = QStringList::split("->", sel->text(1) );
    slotFileSelected( sel->directory() +"/" +str[0].stripWhiteSpace() );
    qWarning("slot open");
    // DocLnk lnk( sel->directory() + "/" + str[0].stripWhiteSpace() );
    //emit fileSelected( lnk );
  }
}
void QpdfFileSelector::slotRescan()
{
  
}
void QpdfFileSelector::slotRename()
{
  reparse();
}
void QpdfFileSelector::slotDelete()
{
  QpdfFileSelectorItem *sel = (QpdfFileSelectorItem*)m_View->currentItem();
  QStringList list = QStringList::split("->", sel->text(1) );
  if( sel->isDir() ){
    QString str = QString::fromLatin1("rm -rf ") + sel->directory() +"/" + list[0]; //better safe than sorry
    switch ( QMessageBox::warning(this,tr("Delete"),tr("Do you really want to delete\n")+list[0],
                                  tr("Yes"),tr("No"),0,1,1) ) {
    case 0:
      ::system(str.utf8().data() );
    break;
    }
  } else {
    QFile::remove( list[0] );
  }
  m_View->takeItem( sel );
  delete sel;
}

void QpdfFileSelector::cdUP()
{
  QDir dir( m_currentDir );
  dir.cdUp();
  if(dir.exists() ){
    m_currentDir = dir.absPath();
    reparse();
    slotInsertLocationPath( m_currentDir, 0 );
  }
}

void QpdfFileSelector::slotHome()
{
  cd(QDir::homeDirPath() );
}
void QpdfFileSelector::slotDoc()
{
  cd(QPEApplication::documentDir() );
}
void QpdfFileSelector::slotNavigate( )
{
  
}

// fill the View with life
void QpdfFileSelector::reparse()
{
  if( m_selector == NORMAL )
    return;
  if( m_selector == EXTENDED || m_selector == EXTENDED_ALL )
    m_View->clear();
  else // custom view
    ; // currentView()->clear();
  if( m_shChooser)
    qWarning("reparse %s", m_mimeCheck->currentText().latin1() );
  
  QString currentMimeType = "pdf";
#if 0
  // let's update the mimetype

  currentMimeType = m_mimeCheck->currentText();

  if( m_autoMime ){
    m_mimetypes.clear();
    // ok we can change mimetype so we need to be able to give a selection
    if( m_shChooser ) {
      m_mimeCheck->clear();
      
      // let's find possible mimetypes
      QDir dir( m_currentDir );
      if( dir.exists() ){
        dir.setFilter( QDir::Files | QDir::Readable );
        dir.setSorting( QDir::Size );
        const QFileInfoList *list = dir.entryInfoList();
        QFileInfoListIterator it( *list );
        QFileInfo *fi;
        while( (fi=it.current() ) ) {
	  if( fi->extension() == QString::fromLatin1("desktop") ){
	    ++it;
	    continue;
	  }
	  MimeType type( fi->absFilePath() );
	  if( !m_mimetypes.contains( type.id() ) ){
	    //qWarning("Type %s", type.id().latin1() );
	    m_mimetypes.insert( type.id(), type.id() );
	  }
	  
	  ++it;
        }
        // add them to the chooser
        updateMimeCheck();
        m_mimeCheck->setCurrentItem( indexByString( m_mimeCheck, currentMimeType ) );
        currentMimeType = m_mimeCheck->currentText();
      }
    }
  } else { // no autoMime
    // let the mimetype be set from out side the m_mimeCheck FEATURE
    
    if( m_shChooser ){
      updateMimeCheck();
    }
  }
  // now we got our mimetypes we can add the files
  //  printf("**mimetype = %s\n", currentMimeType.latin1());
#endif
  
  QDir dir( m_currentDir );
  
  if( dir.exists() ){
    int sort;

    // now go through all files
    m_please_wait-> setGeometry ( qApp->desktop()->width()/2 - m_please_wait->width()/2, 
				  qApp->desktop()->height()/2,
				  m_please_wait->width(),
				  m_please_wait->height());

    m_please_wait->show();
    m_please_wait->raise();

    printf("raise/show please wait %d x %d + %d,%d\n",
	   m_please_wait->width(), m_please_wait->height(),
	   m_please_wait->x(), m_please_wait->y());


    if ( m_case )
      sort = (QDir::IgnoreCase | QDir::Name | QDir::DirsFirst | QDir::Reversed);
    else
      sort = (QDir::Name | QDir::DirsFirst | QDir::Reversed);

    dir.setSorting( sort );
    
    int filter;
    if( m_selector == EXTENDED_ALL /*|| m_selector ==CUSTOM_ALL */ ){
      filter = QDir::Files | QDir::Dirs | QDir::Hidden | QDir::All;
    }else
      filter = QDir::Files | QDir::Dirs | QDir::All;

    dir.setFilter( filter );
    
    qApp-> processEvents ( );

    const QFileInfoList *list = dir.entryInfoList();

    qApp-> processEvents ( );

    if (list == NULL) {
      goto done;
    }

    qApp-> processEvents ( );

    if (m_pixmap_size != m_View->font().pixelSize()) {
      initPics();
    }

    QFileInfoListIterator it( *list );
    QFileInfo *fi;
    while( (fi=it.current() ) ){

      if( fi->fileName() == QString::fromLatin1("..") || fi->fileName() == QString::fromLatin1(".") ){
        ++it;
        continue;
      }

      if( fi->isSymLink() ){
#ifdef FOLLOW_LINKS
	//        QString file = fi->dirPath( true ) + "/" + fi->readLink();
        QString file = fi->readLink();
	//	printf("readLink(%s) = %s\n", fi->fileName().latin1(), file.latin1());
	if (file.left(1) != "/") {
	  file = fi->dirPath(true) + "/" + file;
	}
        for( int i = 0; i<=8; i++) { // 9 tries to prevent dos
	  QFileInfo info( file );
	  if( !info.exists() ){
	    printf("%s doesn't exist\n", file.latin1());
	    addSymlink( currentMimeType, fi, TRUE );
	    break;
	  }else if( info.isDir() ){
	    addDir( currentMimeType, fi, TRUE );
	    break;
	  }else if( info.isFile() ){

	    // I'm sure some people will not agree, but I think it is nice
	    // to see all the info about the file it links to, rather than the link itself
	    addFile( currentMimeType, &info, FALSE ); 
	    //	    addFile( currentMimeType, fi, TRUE );

	    break;
	  }else if( info.isSymLink() ){
	    QString follow = info.readLink() ;
	    if (follow.left(1) != "/") {
	      follow = fi->dirPath(true) + "/" + follow;
	    }
	    //	    printf("follow link %s -> %s\n", file.latin1(), follow.latin1());
	    file = follow;
	    break;
	  }else if( i == 4){
	    addSymlink( currentMimeType, fi );
	  }
        } // off for loop
#else	
        addFile( currentMimeType, fi );
#endif
      }else if( fi->isDir() ){
        addDir( currentMimeType, fi );
      }else if( fi->isFile() ){
        addFile( currentMimeType, fi );
      }
      //qWarning( "%s", fi->fileName().latin1() );
      ++it;
    } // of while loop
  } // if ( dir.exists() )
  m_View->sort();
  m_View->setCurrentItem(m_View->firstChild());

  //  if( m_shTool ){
  slotInsertLocationPath(m_currentDir, 0);
  //  }

 done:

  m_please_wait->hide();
  m_View->setFocus();
  // reenable painting and updates
}
