

#include <aconf.h>

#ifdef USE_GCC_PRAGMAS
#pragma implementation
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <math.h>
#include <ctype.h>
#include "gmem.h"
#include "GString.h"
#include "GList.h"
#include "config.h"
#include "Error.h"
#include "GlobalParams.h"
#include "UnicodeMap.h"
#include "GfxState.h"


#include "qdocview.h"
#include "TextOutputDev.h"

/*
 Here is the xpdf interface for making a textmap.
 All information about words, text and the
 mapping to screen positions are kept in the textmap.
*/


void
TextPage::makeTextMap(QDocText *textmap )
{
  QString ss;
  QString eol;
  int eolLen;
  TextWord *word0;
  TextLine *line0;
  TextFlow *flow0;
  TextBlock *blk0;
  int i;
  int wordStart;
  int charIndex = 0;

  textmap->clear();

  eolLen = 0; // make gcc happy
  switch (globalParams->getTextEOL()) {
  case eolUnix:
    eolLen=1; eol += "\n";
    break;
  case eolDOS:
    eolLen=2; eol += "\r\n";
    break;
  case eolMac:
    eolLen=1; eol += "\r";
    break;
  }

  fouble  next_line_spacing, line_spacing, prev_line_base;

  wordStart = charIndex;
  for (flow0 = flows; flow0; flow0 = flow0->next) {
    line_spacing = -1;
    prev_line_base = -1;
    for (blk0 = flow0->blocks; blk0; blk0 = blk0->next) {
      //      QString debug_line;
      for (line0 = blk0->lines; line0; line0 = line0->next) {
#if 0
	debug_line = "";
	printf("%f > %f + %f * 2 ===",
	       (float)line0->words->xMin, (float)blk0->xMin,
	       (float)line0->words->fontSize);
#endif
	// if line spacing increases over previous line pair -> line break
	if (prev_line_base > 0 && line0->base != prev_line_base) {
	  next_line_spacing = line0->base - prev_line_base;
	  if (next_line_spacing < 0) {
	    line_spacing = prev_line_base = -1;
	  }
	}
	if ((line0->words->xMin > blk0->xMin + line0->words->fontSize * (fouble)0.8
	     && prev_line_base != line0->base)
	     //	     && line0->words->xMin < flow0->xMin + line0->words->fontSize * (fouble)2.0)
	    || (line_spacing > 0 && next_line_spacing > line_spacing + line0->words->fontSize * (fouble)0.5)) {
#if 0
	  printf("line_spacing %f next_line_spacing %f prev_line_base %f base %f\n",
		 (float)line_spacing, (float)next_line_spacing,
		 (float)prev_line_base, (float)line0->base);
#endif
	  ss += eol;
	  charIndex += eolLen;
	  ss += "     ";
	  charIndex += 5;
	}
	prev_line_base = line0->base;
	line_spacing = next_line_spacing;

	for (word0 = line0->words; word0; word0 = word0->next) {
	  for(i=0; i<word0->len; i++) {
	    ss += QChar(word0->text[i]);
	    //	    debug_line += QChar(word0->text[i]);
	    charIndex++;
	  }
	  if (word0->next 
	      || !((line0->next || blk0->next)
		   && (ss.right(1) == "-" || line0->hyphenated))) {
	    ss += " ";
	    //	    debug_line += " ";
	    charIndex++;
	  } else {
	    if (ss.right(1) == "-") {
	      ss.truncate(ss.length()-1);
	      charIndex--;
	    }
	  }
	  fouble xmax = word0->xMax;
	  fouble xScreenMax = word0->xScreenMax;
	  if (word0->spaceAfter && word0->next) {
	    xmax = word0->next->xMin;
	    xScreenMax = word0->next->xScreenMin;
	  }
	  textmap->addWord(lrint(word0->xMin),lrint(word0->yMin),
			   lrint(xmax),lrint(word0->yMax),
			   lrint(word0->xScreenMin), lrint(xScreenMax),
			   wordStart, charIndex-1);
	  // this will include eol's at start of word, but at least there are no holes
	  wordStart = charIndex;
	}
	//	printf("%s\n", debug_line.latin1());
#if 0
	if (!line0->hyphenated) {
	  if (line0->next) {
	    ss += " ";
	    charIndex++;
	  } else if (blk0->next) {
	    //~ this is a bit of a kludge - we should really do a more
	    //~ intelligent determination of paragraphs
	    if (blk0->next->lines->words->fontSize == blk0->lines->words->fontSize) {
	      ss += " ";
	      charIndex++;
	    } else {
	      ss += eol;
	      charIndex += eolLen;
	    }
	  }
	}
#endif
      }
      if (blk0->next
	  && blk0->next->lines->words->fontSize
	  != blk0->lines->words->fontSize) {
	ss += eol;
	charIndex += eolLen;
	ss += eol;
	charIndex += eolLen;
      }
      //      ss += eol;
      //      charIndex += eolLen;
    }
    // sometimes we start a new flow in next column, and it shouldn't
    // be a break
    ss += eol;
    charIndex += eolLen;
    ss += eol;
    charIndex += eolLen;
  }
  textmap->setText(ss);
  clear(); 
}
