// Copyright 2003, Philip Kohn


#ifdef __GNUC__
#pragma implementation
#endif

#include "aconf.h"

#include "qpdf.h"
//#include "qpdfview.h"
#include "QOutputDev.h"


QPDFView::~QPDFView()
{
  closeDocument();
  if (m_pdfdoc)
    delete m_pdfdoc;
  if (m_outdev)
    delete m_outdev;
}

QPDFView::QPDFView( QWidget *parent, const char *name, int flags, QPDFconfig *config) 
  : QDocView( parent, name, flags, config )
{
  m_pdfdoc = NULL;
  m_outdev = new QOutputDev(parent, name, flags);
}

bool
QPDFView::isOk()
{
  if (!m_pdfdoc) {
    return(false);
  }
  return(m_pdfdoc->isOk());
}

void
QPDFView::switchPage(QDocState target)
{
}

void
QPDFView::closeDocument()
{
  if (m_pdfdoc != NULL) {
    //    printf("closeDocument path = %s\n", m_state.path.latin1());
    delete m_pdfdoc;
  }
  m_pdfdoc = NULL;
  //  m_notes.clear();
  m_state.page = 0;
  m_state.numPages = 0;
}

bool
QPDFView::switchDocument(QDocState *target)
{
  closeDocument();
  return(openDocument(target));
}

bool
QPDFView::openDocument(QDocState *target)
{
  m_pdfdoc = new PDFDoc ( new GString ( target->path.local8Bit ( )), 0, 0 );
  if (!m_pdfdoc || !m_pdfdoc->isOk()) {
    printf("@@@@@@@@ NOT READABLE: %s\n", target->path.latin1());
#ifndef AUTOTEST
    QMessageBox::warning ( this, tr( "Error" ), 
			   tr( "File was not readable PDF" ));
#endif
    if (m_pdfdoc) delete m_pdfdoc;
    m_pdfdoc = NULL;
    return(false);
  }

  m_outdev->startDoc(m_pdfdoc->getXRef());
  target->numPages = m_pdfdoc->getNumPages();

  printf("===> opened %s (%d pages)\n", target->path.latin1(),
  	 target->numPages);

  return(true);
}

    
static bool abort_paint_page_flag;

static GBool abort_pdf_repaint_check(void *dev) {
  static int processEventCount = 0;
  // another attempt to make things lively without the help
  // of threads or time slicing

  // could most likely remove many of the other processQtEvents
  // now that this one is here
  // (only exceptions are very time consuming things like image drawing
  if (processEventCount++ > 40) {
    ((QOutputDev*)dev)->processQtEvents();
    processEventCount = 0;
  }
  return(abort_paint_page_flag);
}

/*
here is the problem:
outdev needs m_pixmap to write on during repaint
outdev is in best positon to manage creation of m_pixmap
docview needs complete control of m_pixmap when not repainting (for rotate)
  and read access all the time (for repaintContents)


 m_pixmap = new QPixmap ( lrint ( state-> getPageWidth ( )), lrint ( state-> getPageHeight ( )));

  m_outdev->setPixmap(newPixmap(m_pdfdoc->getPageWidth(target.page),
				m_pdfdoc->getPageHeight(target.page)));
*/

bool
QPDFView::readPage(QDocState target, QPixmap **pixmap_ptr)
{
  fouble zoom_dpi;

  abort_paint_page_flag = 0;
  if (!m_pdfdoc || !m_pdfdoc->isOk()) {
    return(false);
  }
  limitState(target);
  // everything here needs some help, maybe outdev should be deleted and created?
  m_outdev->setPixmapPtr(pixmap_ptr); // still don't like this
  m_outdev->m_reproportion_text = target.reproportion_text;
  m_outdev->m_always_render_font = target.render_text;
  m_outdev->m_reproportion_and_render = target.reproportion_and_render;
  printf("readpage zoom = %f dpi=%d\n", target.zoom, dpi());
  zoom_dpi = target.zoom * dpi();
  m_pdfdoc->displayPage ( m_outdev, target.page, zoom_dpi, zoom_dpi,
			  0, gFalse,
			  &abort_pdf_repaint_check, (void*)m_outdev);
  m_outdev->setPixmapPtr(NULL);  // take control back

  m_outdev->makeTextMap(m_textmap);
  m_textmap->m_state.page = target.page;
  m_textmap->m_state.width = (*pixmap_ptr)->width();
  m_textmap->m_state.height = (*pixmap_ptr)->height();
  setFocus();
  return(true);
}

bool
QPDFView::readTextPage(QDocState target, QDocText **textmap_ptr)
{
  QTextOutputDev *textOut = 0;
  fouble zoom_dpi;

  if (!m_pdfdoc || !m_pdfdoc->isOk()) {
    return(false);
  }
  limitState(target);

  if (! *textmap_ptr) {
    *textmap_ptr = new QDocText;
  }

  textOut = new QTextOutputDev ( 0, gFalse, gFalse, gFalse);

  if ( !textOut-> isOk ( )) {
    delete textOut;
    return(false);
  }

  abort_paint_page_flag = 0;

  //  textOut->m_reproportion_text = target.reproportion_text && !target.render_text;
  textOut->m_reproportion_text = target.reproportion_text;
  textOut->m_always_render_font = target.render_text;

  zoom_dpi = target.zoom * dpi();
  printf("readTextPage %d zoom %f\n", target.page, (float)target.zoom);
  m_pdfdoc->displayPage(textOut, target.page, zoom_dpi, zoom_dpi, 0,
			gFalse, &abort_pdf_repaint_check, (void*)m_outdev);
  if (! abort_paint_page_flag) {
    printf("readTextPage success\n");
    textOut->makeTextMap(*textmap_ptr);
    // needed?  most likely not, since we are either in sync with m_target_state
    // (e.g. in PDF page mode on a page), or we aren't (e.g. text mode)
    (*textmap_ptr)->m_state.page = target.page;
    (*textmap_ptr)->m_state.zoom = target.zoom;
    (*textmap_ptr)->m_state.width = textOut->m_width;
    (*textmap_ptr)->m_state.height = textOut->m_height;
    printf("set textmap size %d x %d\n", textOut->m_width, textOut->m_height);
    delete textOut;
    return(true);
  }
  delete textOut;
  printf("readTextPage abort\n");
  return(false);
}

void 
QPDFView::abortReadPage()
{
  abort_paint_page_flag = 1;
}


void
QPDFView::limitState(QDocState &state)
{
  // check zoom for reasonable level
  if (state.zoom < MIN_ZOOM)
    state.zoom = MIN_ZOOM;
  // need to limit zoom in by pixmap size that will fit in
  // limited memory of zaurus
  // could query kernel here...
  if (state.zoom > MAX_ZOOM)
    state.zoom = MAX_ZOOM;
  
  // could be moved into DocState since this is pretty generic
  if (state.page <= 0) 
    state.page = 1;
  if (state.numPages > 0 && state.page > state.numPages)
    state.page = state.numPages;
}



#if 0
QString
QPDFView::wrapText(QString text)
{
  if ( text.isEmpty() )
	return text;
  QString result;
  const QChar *from = text.unicode();
  const QChar *fromend = from+text.length();
  int outc=0;
  for (;;) {
    while ( from!=fromend && from->isSpace() )
      from++;
    while ( from!=fromend && !from->isSpace() ) {
      result[outc] = *from++;
      if (result[outc] == '.'
	  && (
	      (outc >= 2  
	       && result[outc-1].category() == QChar::Letter_Lowercase
	       && result[outc-2].category() == QChar::Letter_Lowercase)
	      || 
	      (outc >= 1 && result[outc-1].isPunct()))) {
	result[++outc] = '\n';
      }
      outc++;
    }
    if ( from!=fromend )
      result[outc++] = ' ';
    else
      break;
  }
  if ( outc > 0 && result[outc-1] == ' ' )
    outc--;
  result.truncate( outc );
  return result;
}
#endif



