
/*
  This class provides very basic mutual exclusion (assuming no real interrupts or threads!)
  and a little caching.
  
  To implement a new type of document, all you need is getPage and abortGetPage.
*/
  
#include "qdocbase.h"

QDocBase::QDocBase()
{
  m_busy = false;
  m_zoom = 0.0;
  m_page = -1;
  m_pixmap_ptr = NULL;
  m_pixmap_textmap = new QDocTextmap();

  m_aux_textmap_page = 0;
  m_aux_textmap = new QDocTextmap();
}

bool
QDocBase::getPixmap(QString path, int page, fouble zoom, QPixmap **pixmap_ptr, QDocTextmap **textmap_ptr)
{
  bool success;

  if (page <= 0) {
    printf("QDocBase::getTextmap page = %d\n", page);
    return(false);
  }

  if (m_path != path) {
    m_aux_textmap_page = -1;
  }
  m_path = path;
  m_page = page;
  m_zoom = zoom;
  m_pixmap_ptr = pixmap_ptr;

  if (m_busy) {
    m_abort = true;
    abortGetPage();
    return;
  }

  m_busy = true;

  do {
    m_abort = false;
    success = getPage(m_path, m_page, m_zoom, m_pixmap_ptr, m_pixmap_textmap, &m_numPages);
    if (! success) {
      m_page = -1;
      return(false);
    }
  } while(m_abort);
  m_busy = false;
  if (m_textmap_ptr) {
    *m_textmap_ptr = m_pixmap_textmap;
  }
  emit gotPage;
  return(true);
}

bool 
QDocBase::getTextmap(QString path, int page, QDocTextmap **textmap_ptr)
{
  bool success;

  if (page <= 0) {
    printf("QDocBase::getTextmap page = %d\n", page);
    return(false);
  }

  if (m_busy) {
    return(false);
  }
  if (path == m_path) {
    if (page == m_page) {
      *textmap_ptr = m_pixmap_textmap_ptr;
      return(true);
    }
    if (page == m_textmapPage) {
      *textmap_ptr = m_aux_textmap_ptr;
      return(true);
    }
  }
  m_textmapPage = page;
  m_busy = true;
  m_abort = false;
  success = getPage(m_path, m_textmapPage, 0.0, NULL, m_textmap_ptr);
  m_busy = false;
  if (m_abort) {
    getPage(m_path, m_page, m_zoom, m_pixmap_ptr, m_pixmap_textmap, &m_num_pages);
    return(false);
  }
  if (!success) {
    m_aux_textmap_page = -1;
  }
  return(success);
}



  
