#include "xmlparser.h"


class QDocNote {
 public:
  friend class QDocNotes;
  QDocNote();
  ~QDocNote();

  void save(QString &str);
  int load(const QString &str, int index);
  int firstWord() { return(m_firstWord); }
  int lastWord() { return(m_lastWord); }

  // we remember notes as a range of words, but since the
  // PDF => word algorithm is always being improved, we can't count
  // on word indices to be conserved across versions of the xpdf code.
  // Instead, we remember the start and end coordinates as a ratio of the
  // full size of the page image.  From these and the textmap, we fill in
  // the firstWord, lastWord and regionWords fields.
  int m_page;
  QPoint m_start, m_end;
  int m_firstWord;
  int m_lastWord;

 protected:
  QString notesPath(const QString &pdfFilename);

  QDateTime m_date;
  QString m_note;
  QRegion m_regionWords;
};

class QDocNotes : public QDocState, public QDocTextEdit {
 public:   

  QDocNotes();
  ~QDocNotes();

  void setupPage(QDocText *text);

  virtual void clear();
  virtual void open();
  virtual void close();
  virtual void changeState();
  virtual void abortChangeState();

  QDocNote *find(QDocState &state);
  QDocNote *find(int page, QPoint point);
  QDocNote *find(int page, int word, bool closest);
  QDocNote *find(int page, int firstWord, int lastWord);

  QString toText(int page = -1, QDocNote *selected_note = NULL, int *selected_note_position = NULL);
  bool fromText(const QString &text, int noteCharPos, 
		QDocNote **note);

  QRegion highlights(int page);

  QDocNote *add(int page, const QPoint start, const QPoint end, const QString &note, QDocText *text);

  bool load(const QString &filename, QDocState *state);
  bool load(QDocState *state);
  bool save(const QString &filename, QDocState *state);
  bool save(QDocState *state);

 protected:
  void add(QDocNote *);
  QString notesPath();

 private:
  QList<QDocNote> m_notes;
  QString m_noteSeperator;
  QString m_filename;
  bool m_clean;

};

