#include "aconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <iostream>
#include <assert.h>

#include <qgfx_qws.h>
#include <qpixmap.h>
#include <qimage.h>
#include <qpainter.h>
#include <qdict.h>
#include <qtimer.h>
#include <qlabel.h>
#include <qscrollview.h>
#include <qmessagebox.h>
#include <qapplication.h>
#include <qdatetime.h>

#include "qbusybar.h"
#include "qdocstate.h"
#include "qdoctext.h"
#include "qgesture.h"
#include "qdocnotes.h"


class QDocPix : public QDocView {
  Q_OBJECT

public:
  QDocPix(QWidget *parent = 0, const char *name = 0, int flags = 0, QDocBase *doc);
  virtual ~QDocPix();

  QPoint rotate(QPoint p);
  QPoint antirotate(QPoint p);
  QRect rotate(QRect r);
  QRegion rotate(QRegion r);

  virtual void setSelection(int from_word, int to_word, QColor color = Qt::white);
  virtual void showSelection();

  virtual void showHighlights();
  virtual void hideHighlights();

  virtual void changeState();

  QRegion selection ( ) const;
  void setSelection ( const QRect &r, bool scrollto = false );
  void setSelection ( const QRegion &r, bool scrollto = false );
  void setHighlights ( const QRegion &r );

  bool hasSelection();
  QRegion selection ( ) const;
  void setSelection ( const QRect &r, bool scrollto = false );
  void setSelection ( const QRegion &r, bool scrollto = false );
  void setTextSelection(int first_word = -1, int last_word = -1, bool scroll = true);
  void getTextSelection(int *first_word, int *last_word);
  QString selectedText();

  void highlight(int from_word, int to_word, int thickness = 1, QColor color = Qt::white);
  void updateHighlights();

  bool findText(const QString &string, int *from_char, int *to_char, int *page,
		bool reverse, bool multipage);

  bool findTextInternal(QString text, int *from_char, int *to_char, 
			int first_page = 0, int last_page = -1,
			int *page = NULL);

  QString getText(int from_word = 0, int to_word = -1, int page = -1);
  QRegion getTextRegion(int from_word = 0, int to_word = -1);

  //  QDocNote *getNote(int page, int note);
  //  void addNote(QDocNote note);

  void changeState(bool saveHistory = true);
  bool isValid();
  bool isBusy() { return(m_inside_read_page); }

signals:
  void selectionChanged ( const QString &str );
  void toggleFullscreen();
  void setNoteMode();
  void updateMenuItems();
  void setBusy(bool);
  void setPageCount(int page, int time = 4000);
  
public slots:
  void gotoPage(int page);
  void gotoPage(int page, QPoint pos);

  void setContentsPos(QPoint p);
  virtual void setContentsPos( int x, int y );

  void addNote(int from_char, int to_char); 

  void setHighlightMode(bool mode);
  bool highlightMode();

  void setTextSelectionMode(bool mode);
  bool textSelectionMode();


protected:

  // Qt user interface
  //  virtual void keyPressEvent ( QKeyEvent *e );
  virtual void viewportMousePressEvent ( QMouseEvent *e );
  virtual void viewportMouseMoveEvent ( QMouseEvent *e ); 
  virtual void viewportMouseReleaseEvent ( QMouseEvent *e );
  virtual void drawContents ( QPainter *p, int, int, int, int );
  virtual void resizeEvent( QResizeEvent *e );
  
  // Gesture user interfrace
  virtual void gestureCircle();
  virtual void gestureDiagonal();
  virtual void gestureHorizontal();
  virtual void gestureVertical();
  virtual void gestureCheck();
  virtual void gestureTouch();
  //  virtual void gestureBackForth();

  // state (display rotation, position, zoom, etc.) change
  virtual void changeState();

  QDocBase *m_doc;

private:
  // changeState variables
  QTimer *m_refresh_timer;

  // selections
  QRegion m_selection;
  QPoint  m_selection_start;
  QPoint  m_selection_cursor;
  bool    m_selection_drag;

  bool m_doing_text_selection;
  int m_text_selection_start;
  int m_text_selection_end;

  // for HAND_MODE, i.e. scrolling around
  QPoint  m_mouse_start;
  QPoint  m_mouse_prev;
  QPoint  m_mouse_prev_delta;
  int     m_mouse_prev_dist;

  bool m_mouse_click_time_window;
  bool m_mouse_debounce_time_window;
  QPoint m_mouse_click_pos;
  QPoint m_mouse_click_abs_pos;
  QTimer *m_mouse_click_timer;


  bool m_mouse_down_in_selection;
  int m_mouse_down_selection_start;
  int m_mouse_down_selection_end;
  void mouseClick();

  // for highlighting of notes
  bool m_highlight_mode;
  QRegion m_highlight_region;

  int detectEdges(QPoint pos);
  int detectCorners(QPoint pos);

  void updateRotate();
  void rotateView(int width, int height, bool anti_rotate);
  QPixmap *rotatePixmap(QPixmap *pixmap, bool anti_rotate);

  QPixmap *m_pixmap;
  QPixmap *m_rotated_pixmap;

  bool isTextValid() { return(!m_inside_read_page 
			      && m_textmap->isSamePage(m_state)  ); }
  bool isTextValid(QDocState &state) { return(!m_inside_read_page
					      && m_textmap->isSamePage(state)); }
  bool isTextMapValid() { return(!m_inside_read_page
				 && m_textmap->isSamePixmap(m_state)); }

};
