// Copyright 2003, Philip Kohn

#include "aconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <iostream>
#include <assert.h>

#include <qgfx_qws.h>
#include <qpixmap.h>
#include <qimage.h>
#include <qpainter.h>
#include <qdict.h>
#include <qtimer.h>
#include <qlabel.h>
#include <qscrollview.h>
#include <qmessagebox.h>
#include <qapplication.h>

#include "qdocview.h"

QDocState::QDocState(bool top)
{
  m_actual = NULL;
  if (top) {
    m_actual = new QDocState(false);
  }
  clear();
}

void
QDocState::clear()
{
  m_pos = QPoint();
  m_wordIndex = -1;
  m_charIndex = -1;
  m_move = false;

  m_range = false;
  m_endPos = QPoint();
  m_endWordIndex = -1;
  m_endCharIndex = -1;

  m_rotated = false;
  m_zoom = DEFAULT_ZOOM;

  m_path = QString();
  m_page = -1;
  m_num_pages = -1;
  m_width = -1;
  m_height = -1;

  m_mode = FILE_SELECTOR_MODE;
  if (m_actual) {
    m_actual->clear();
  }
}

void
QDocState::setPath(QString path)
{
  m_path = path;
}

QString
QDocState::path()
{
  return(m_path);
}

QString
QDocState::filename()
{
  int sep = m_path. findRev ( '/' );
  if ( sep > 0 )
    return(m_path.mid( sep + 1 ));
  return(m_path);
}

QString
QDocState::changeExtension(const QString &ext)
{
  int i = m_path.findRev(".");
  if (i > 0) {
    return(m_path.left(i) + ext);
  } else {
    printf("QDocState::changeExntension %s had no ext\n", ext.latin1());
    return(m_path + "." + ext);
  }
}

int
QDocState::numPages()
{
  return(m_num_pages);
}

int
QDocState::page()
{
  return(m_page);
}

void
QDocState::setPage(int page)
{
  m_page = page;
}

void
QDocState::setZoom(float zoom)
{
  m_zoom = zoom;
}

float
QDocState::zoom()
{
  return(m_zoom);
}

int
QDocState::x()
{
  return(m_x);
}

int
QDocState::y()
{
  return(m_y);
}

void
QDocState::setX(int x)
{
  m_x = x;
}

void
QDocState::setY(int y)
{
  m_y = y;
}

QPoint
QDocState::pos()
{
  return(QPoint(m_x, m_y));
}

void
QDocState::setPos(int x, int y)
{
  m_x = x;
  m_y = y;
}

void
QDocState::home()
{
  m_x = m_y = -1;
  m_move = true;
}

bool
QDocState::move()
{
  return(m_move);
}

void
QDocState::setMove()
{
  m_move = true;
}

void
QDocState::setPos(int x, int y)
{
  m_x = x;
  m_y = y;
  m_move = true;
}


void
QDocState::save(QString &str)
{
  str += QString::fromLatin1("<bookmark page=\"%1\" zoom=\"%2\" x=\"%3\" y=\"%4\" rotate=\"%5\" file=\"%6\" />\n")
    .arg(page).arg(zoom).arg(x).arg(y).arg(isRotated).arg(path);
}

int
QDocState::load(const QString &str, int index)
{
  int nextIndex;
  XmlParser parser;
  nextIndex = parser.parse(str, index);
  if (!parser.checkHead("bookmark")) {
    return(-1);
  }
  // path = parser.get("file");
  page = parser.get("page").toInt();
  zoom = parser.get("zoom").toFloat();
  x = parser.get("x").toInt();
  y = parser.get("y").toInt();
  isRotated = parser.get("rotate").toInt();
  doMove = true;
  printf("**loaded state page = %d\n", page);
  return(nextIndex);
}


bool
QDocState::isSamePixmap(QDocState &state)
{
  return(page == state.page
	 && zoom == state.zoom
	 && path == state.path);
}

bool
QDocState::isSamePage(QDocState &state)
{
  return(page == state.page
	 && path == state.path);
}

