


typedef enum { PDF_MODE = 1, TEXT_MODE, NOTES_MODE, FILE_SELECTOR_MODE, HELP_MODE } ViewMode;

class QDocState  {
public:
  QDocState(bool toplevel);

  virtual void clear();
  virtual void changeState() = 0;
  virtual void abortChangeState() {}
  virtual void open() = 0;
  virtual void close() = 0;

  virtual void mapPos() = 0;  // set word & char index (and range if there is one)
  virtual void mapCharIndex() = 0;
  virtual void mapWordIndex() = 0;

  virtual void showRange();

  void setPath(QString &path);
  QString path() const;
  QString filename() const;
  QString changeExtension(QString &ext) const;

  int numPages();
  int page();
  void setPage(int);

  void setZoom(float zoom);
  float zoom();

  int x();
  int y();
  void setX(int);
  void setY(int);
  QPoint pos();
  void setPos(int x, int y);
  void setPos(QPoint p) { setPos(p.x(),p.y()); }
  void home();
  bool move();
  void setMove();

  int percentX();
  int percentY();
  void setPercentX(int);
  void setPercentY(int);

  void setWordIndex(int);
  int wordIndex();
  void setCharIndex(int);
  int charIndex();
  
  bool range();
  void clearRange();

  int endX(), endY();
  int endWordIndex();
  int endCharIndex();

  void setEndX(int);
  void setEndY(int);
  void setEndWordIndex(int);
  void setEndCharIndex(int);

  int numWords();
  int numChars();

  void setRotated(bool);
  bool rotated();

  void setMode(ViewMode);
  ViewMode mode();

  void toXml(QString &str);
  int fromXml(const QString &str, int index);

  bool isSamePixmap(QDocState &);
  bool isSamePage(QDocState &);
  bool isValid() { return(page > 0 && numPages > 0 && !path.isEmpty()); }

  void sanitizeState();

  void dump() {
    printf("state x=%d y=%d move=%d rot=%d zoom=%f page=%d path=%s\n",
	   x,y,doMove, isRotated, zoom, page, path.latin1());
  }


 protected:


  void setSize(int w, int h);
  void setNumPages(int);

  QPoint m_pos;  // upper left corner
  int m_wordIndex;
  int m_charIndex;
  bool m_move;

  bool m_range;
  QPoint m_endPos;
  int m_endWordIndex;
  int m_endCharIndex;

  bool m_rotated;
  float m_zoom;
  QString m_path;
  int m_page;
  int m_num_pages;
  int m_width, m_height;

  ViewMode m_mode; 

  // actual state of this document
  // changeState() works to bring *m_actual == *this
  QDocState *m_actual;
};
