#include "aconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <iostream>
#include <assert.h>

#include <qgfx_qws.h>
#include <qpixmap.h>
#include <qimage.h>
#include <qpainter.h>
#include <qdict.h>
#include <qtimer.h>
#include <qlabel.h>
#include <qscrollview.h>
#include <qmessagebox.h>
#include <qapplication.h>
#include <qdatetime.h>

#include "qbusybar.h"
#include "qdocstate.h"
#include "qdoctext.h"
#include "qgesture.h"


class QDocView : public QDocState, public QGesture {
  Q_OBJECT

public:
  QDocView(QWidget *parent = 0, const char *name = 0, int flags = 0, QDocBase *doc);
  virtual ~QDocView();
  
  QString selectedText();

  virtual void setSelection(int from_word, int to_word, QColor color = Qt::white);
  virtual void clearSelection();
  virtual void showSelection();

  virtual void showHighlights();
  virtual void hideHighlights();

  virtual void changeState();
  bool isValid();
  bool isBusy() { return(m_inside_read_page); }

signals:
  void selectionChanged ( const QString &str );
  void toggleFullscreen();
  void setNoteMode();
  void updateMenuItems();
  void setBusy(bool);
  void setPageCount(int page, int time = 4000);
  void addNote(int from_char, int to_char); 
  
public slots:
  void setContentsPos(QPoint p);
  virtual void setContentsPos( int x, int y );

  void setHighlightMode(bool mode);
  bool highlightMode();

  void setTextSelectionMode(bool mode);
  bool textSelectionMode();

private slots:
  virtual void refresh();
  void mouseClickTimeout();

protected:

  // Qt user interface
  //  virtual void keyPressEvent ( QKeyEvent *e );
  virtual void viewportMousePressEvent ( QMouseEvent *e );
  virtual void viewportMouseMoveEvent ( QMouseEvent *e ); 
  virtual void viewportMouseReleaseEvent ( QMouseEvent *e );
  virtual void drawContents ( QPainter *p, int, int, int, int );
  virtual void resizeEvent( QResizeEvent *e );

  
  // Gesture user interfrace
  virtual void gestureCircle();
  virtual void gestureDiagonal();
  virtual void gestureHorizontal();
  virtual void gestureVertical();
  virtual void gestureCheck();
  virtual void gestureTouch();
  //  virtual void gestureBackForth();


  // state (display rotation, position, zoom, etc.) change
  virtual void changeState();

  QDocBase *m_doc;

private:
  // changeState variables
  QTimer *m_refresh_timer;

  // selections
  QRegion m_selection;
  QPoint  m_selection_start;
  QPoint  m_selection_cursor;
  bool    m_selection_drag;

  bool m_doing_text_selection;
  int m_text_selection_start;
  int m_text_selection_end;

  // for HAND_MODE, i.e. scrolling around
  QPoint  m_mouse_start;
  QPoint  m_mouse_prev;
  QPoint  m_mouse_prev_delta;
  int     m_mouse_prev_dist;

  bool m_mouse_click_time_window;
  bool m_mouse_debounce_time_window;
  QPoint m_mouse_click_pos;
  QPoint m_mouse_click_abs_pos;
  QTimer *m_mouse_click_timer;


  bool m_mouse_down_in_selection;
  int m_mouse_down_selection_start;
  int m_mouse_down_selection_end;
  void mouseClick();

  // for highlighting of notes
  bool m_highlight_mode;
  QRegion m_highlight_region;


  void updateRotate();
  void rotateView(int width, int height, bool anti_rotate);
  QPixmap *rotatePixmap(QPixmap *pixmap, bool anti_rotate);

  QPixmap *m_pixmap;
  QPixmap *m_rotated_pixmap;

  bool isTextValid() { return(!m_inside_read_page 
			      && m_textmap->isSamePage(m_state)  ); }
  bool isTextValid(QDocState &state) { return(!m_inside_read_page
					      && m_textmap->isSamePage(state)); }
  bool isTextMapValid() { return(!m_inside_read_page
				 && m_textmap->isSamePixmap(m_state)); }

};
