#define ASYNC
//========================================================================
//
// qpdf.cc
//
// Copyright 2003, Philip Kohn
// Copyright 2001 Robert Griebl
//
//========================================================================


#include "qpdf.h"

#ifdef AUTOTEST
#define MAX_ZOOM 1.2
#endif

void sig_handler(int x)
{
  printf("got signal %d\n",x);
}

int main ( int argc, char **argv ) 
{
	QPEApplication app ( argc, argv );

	// read config file
	globalParams = new GlobalParams ( "" );
	globalParams-> setErrQuiet ( true );

	QPDFWindow mw;

	QObject::connect( &app, SIGNAL( flush() ), &mw, SLOT( flush() ) );
	//	QObject::connect( &app, SIGNAL( reload() ), &mw, SLOT( reload() ) );

	printf("qpdf version 2.0.0\n");
	printf("please read BUGS file before sending a report!\nplease send pdf file,\n how to replicate it,\n and this output\n  to pkohn@codon.nih.gov\n");

	app. showMainDocumentWidget ( &mw );

#if 0
	for(int i=1; i<=31; i++) {
	  if (i != SIGPIPE && i != SIGPROF && i != SIGSEGV 
	      && i != SIGINT && i != SIGILL && i != SIGTERM
	      && i != SIGBUS && i != SIGTSTP)
	    signal(i,sig_handler);
	}
#endif

	if (( app. argc ( ) == 3 ) && ( app. argv ( ) [1] == QCString ( "-f" )))
		mw.openFile ( app. argv ( ) [2] );

	return app. exec ( );
}


QPDFWindow::QPDFWindow ( ) : QMainWindow ()
{
	setCaption ( tr( "QPdf" ));
	setIcon ( Resource::loadPixmap ( "qpdf/qpdf_icon" ));
	
	m_fullscreen = false;
		
	setToolBarsMovable ( false );

	m_stack = new QWidgetStack ( this );
	setCentralWidget ( m_stack );
	m_stack-> setSizePolicy ( QSizePolicy ( QSizePolicy::Expanding, QSizePolicy::Expanding ));

	m_pdfview = new QPDFView ( m_stack, "qpdfview", 0 );
	m_stack->addWidget( m_pdfview, PDF_VIEW);

	m_pdfview->setHScrollBarMode ( QScrollView::AlwaysOff );
	m_pdfview->setVScrollBarMode ( QScrollView::AlwaysOff );

#ifdef OZ
	m_filesel = new OFileSelector ( "application/pdf", m_stack, "fs", false, true );
#else
	m_filesel = new FileSelector ( "application/pdf", m_stack, "fs", false, true );
#endif
	m_stack->addWidget( m_filesel, FILESELECTOR_VIEW);


	m_textedit = new TextEdit(m_stack, "textedit", 0);
	m_stack->addWidget( m_textedit, TEXT_VIEW);

	connect ( m_filesel, SIGNAL( closeMe ( )), this, 
		  SLOT( closeFileSelector ( )));
	connect ( m_filesel, SIGNAL( fileSelected ( const DocLnk & )), this, 
		  SLOT( openFile ( const DocLnk & )));

	m_tb_menu = new QToolBar ( this );
	m_tb_menu-> setHorizontalStretchable ( true );

	QMenuBar *mb = new QMenuBar ( m_tb_menu );

	// file menu
	m_pm_file = new QPopupMenu ( mb );
	mb-> insertItem ( tr("File"), m_pm_file );
	m_pm_file-> insertItem ( Resource::loadIconSet ( "fileopen" ), tr("Open"), this, 
				 SLOT(openFile()), CTRL+Key_O);
	m_pm_file-> insertItem ( tr("Rename"), this, 
				 SLOT(openRenameBar()), CTRL+Key_R);

	// view menu
	m_pm_view = new QPopupMenu ( mb );
	m_pm_view-> setCheckable ( true );
	mb-> insertItem ( tr( "View" ), m_pm_view );

#if 0
	m_pm_view-> insertItem ( tr( "PDF" ),  this, 
				 SLOT(setPDFMode()), CTRL+Key_P, PM_VIEW_PDF);
	m_pm_view-> insertItem ( tr( "Text" ),  this, 
				 SLOT(setTextMode()), CTRL+Key_T, PM_VIEW_TEXT);
	m_pm_view-> insertItem ( tr( "Notes" ),  this, 
				 SLOT(setNoteMode()), CTRL+Key_N, PM_VIEW_NOTES);
	m_pm_view-> insertSeparator ( );
#endif

	m_pm_view_rotate = m_pm_view-> insertItem ( Resource::loadIconSet ( "repeat" ), tr( "Rotate" ),  this, 
						    SLOT(toggleRotate()), CTRL+Key_R);

	m_pm_view_fullscreen = m_pm_view-> insertItem ( Resource::loadIconSet ( "fullscreen" ), tr( "Full Screen" ),  this, 
							SLOT(toggleFullscreen()), CTRL+Key_F);

	m_pm_view_highlight = m_pm_view-> insertItem ( tr( "Highlight" ),  this, 
						       SLOT(toggleHighlight()), CTRL+Key_H);
	m_pm_view-> setItemChecked(m_pm_view_highlight, true);

	m_pm_view_find = m_pm_view-> insertItem ( tr( "Find" ),  this, 
						  SLOT(toggleFindBar()));

	m_pm_view-> insertSeparator ( );

	m_pm_view_zoomin = m_pm_view-> insertItem ( tr( "Zoom In" ), this, SLOT(setZoom(int)), Key_Plus);
	m_pm_view-> setItemParameter(m_pm_view_zoomin, 1);

	m_pm_view_zoomout = m_pm_view-> insertItem ( tr( "Zoom Out" ), this, SLOT(setZoom(int)), Key_Plus);
	m_pm_view-> setItemParameter(m_pm_view_zoomout, -1);

	for(int zooms = 70; zooms <= 160; zooms += 10) {
	  int id;
	  id = m_pm_view-> insertItem ( tr( "%1%" ).arg(zooms), this, SLOT(setZoom(int)));
	  m_pm_view-> setItemParameter(id, zooms);
	}

	//	m_pm_view-> insertItem ( tr( "Fit to width" ), PM_VIEW_FIT_TO_WIDTH);
	//	m_pm_view-> insertItem ( tr( "Fit to page" ),  PM_VIEW_FIT_TO_PAGE);
	

	// edit menu

	m_pm_edit = new QPopupMenu ( mb );
	mb-> insertItem ( tr( "Edit" ), m_pm_edit );

	m_pm_edit-> insertItem ( tr( "Copy" ),  m_textedit, 
				 SLOT(editCopy()), CTRL+Key_C, PM_EDIT_COPY);
	m_pm_edit-> insertItem ( tr( "Cut" ),  m_textedit, 
				 SLOT(editCut()), CTRL+Key_X, PM_EDIT_CUT);
	m_pm_edit-> insertItem ( tr( "Paste" ),  m_textedit, 
				 SLOT(editPaste()), CTRL+Key_V, PM_EDIT_PASTE);

	// go menu

	m_pm_page = new QPopupMenu ( mb );
	mb-> insertItem ( tr( "Go" ), m_pm_page );

	m_pm_page-> insertItem ( Resource::loadIconSet ( "forward" ), tr( "next" ),  this, 
				 SLOT(nextPage()),
				 Key_Apostrophe, PM_PAGE_NEXT);
	m_pm_page-> insertItem ( Resource::loadIconSet ( "back" ), tr( "previous" ),  this, 
				 SLOT(prevPage()),
				 Key_Slash, PM_PAGE_PREV);
	m_pm_page-> insertItem ( Resource::loadIconSet ( "fastforward" ), tr( "last" ),  this, 
				 SLOT(lastPage()),
				 Key_Period, PM_PAGE_LAST);
	m_pm_page-> insertItem ( Resource::loadIconSet ( "fastback" ), tr( "first" ),  this, 
				 SLOT(firstPage()),
				 Key_Tab, PM_PAGE_FIRST);

	m_pm_page-> insertSeparator ( );

#if 0
	m_pm_page-> insertItem ( tr( "Mark" ), this, 
				 SLOT(saveState()), CTRL+Key_S, PM_PAGE_MARK);
#endif
	m_pm_page-> insertItem ( tr( "Forward" ),  this,
				 SLOT(forwardState()), Key_Greater, PM_PAGE_FORW);
	m_pm_page-> insertItem ( tr( "Backward" ), this,
				 SLOT(backwardState()), Key_Less, PM_PAGE_BACK);

	m_pm_page-> setItemEnabled(PM_PAGE_FORW, false);
	m_pm_page-> setItemEnabled(PM_PAGE_BACK, false);

	m_pm_page-> insertSeparator ( );

#if 0
	m_pm_page-> insertItem ( Resource::loadIconSet ( "fastforward" ), tr( "first" ),  this, 
				 SLOT(firstPage()) );
	m_pm_page-> insertItem ( Resource::loadIconSet ( "fastback" ), tr( "last" ),  this, 
				 SLOT(lastPage()) );
#endif

	m_pm_page_start_pagenum_index = 
	  m_pm_page->indexOf(m_pm_page-> insertSeparator ( ));



	m_tb_menu-> addSeparator();

	// change mode buttons


	m_pdf_move_pixmap = Resource::loadPixmap("qpdf/pdf_move_icon");
	m_pdf_text_select_pixmap = Resource::loadPixmap("qpdf/pdf_text_icon");

	m_but_pdf = new QToolButton ( Resource::loadPixmap("qpdf/pdf_icon"), tr( "PDF" ),  
				      tr( "PDF" ),  this,
				      SLOT(setPDFMode()), m_tb_menu, tr( "PDF" ));
	// this one doesn't work!!
	m_but_pdf->setAccel(CTRL+Key_P);
	
	m_but_text = new QToolButton ( Resource::loadPixmap("qpdf/text_icon"), tr( "Text" ),  
				       tr( "Text" ),  this,
				       SLOT(setTextMode()), m_tb_menu, tr( "Text" ));
	m_but_text->setAccel(CTRL+Key_T);

	m_but_note = new QToolButton ( Resource::loadPixmap("qpdf/note_icon"), tr( "Notes" ),  
				       tr( "Notes" ),  this,
				       SLOT(setNoteMode()), m_tb_menu, tr( "Notes" ));
	m_but_note->setAccel(CTRL+Key_N);


	// arrows
	
	m_but_prev_page = new QToolButton ( LeftArrow, m_tb_menu, "prev page");
	connect(m_but_prev_page, SIGNAL( pressed() ), this, SLOT( prevPage() ) );

	m_but_next_page = new QToolButton ( RightArrow, m_tb_menu, "next page");
	connect(m_but_next_page, SIGNAL( pressed() ), this, SLOT( nextPage() ) );


#if 0
	new QToolButton ( Resource::loadIconSet ( "back" ), 
			  tr( "Previous page" ), 
			  QString::null, this, 
			  SLOT( prevPage ( )),  m_tb_menu, "last" );

	new QToolButton ( Resource::loadIconSet ( "forward" ),     
			  tr( "Next page" ), 
			  QString::null, this, 
			  SLOT( nextPage ( )),  m_tb_menu, "next" );
#endif

	
	// Search toolbar
	m_tb_find = new QToolBar ( this );
	addToolBar ( m_tb_find, "Search", QMainWindow::Top, true );
	m_tb_find-> setHorizontalStretchable ( true );
	m_tb_find-> hide ( );

	m_find_line_edit = new QLineEdit ( m_tb_find, "findedit" );
	m_tb_find-> setStretchableWidget ( m_find_line_edit );
	connect ( m_find_line_edit, SIGNAL( textChanged ( const QString & )), this, 
		  SLOT( findTextInPage ( const QString & )));
	//	connect ( m_find_line_edit, SIGNAL( returnPressed() ), this, 
	//		  SLOT( search_again() ) );

#if 0
	new QToolButton ( Resource::loadIconSet ( "next" ),        
			  tr( "Next" ), 
			  QString::null, this, 
			  SLOT( findTextForward ( )),  m_tb_find, "findnext" );
	new QToolButton ( Resource::loadIconSet ( "back" ),        
			  tr( "Back" ), 
			  QString::null, this, 
			  SLOT( findTextBackward ( )),  m_tb_find, "findprev" );
#endif
	new QToolButton ( Resource::loadIconSet ( "down" ),        
			  tr( "PrevPage" ), 
			  QString::null, this, 
			  SLOT( findTextMultipage ( )),  m_tb_find, "findpagenext" );
	new QToolButton ( Resource::loadIconSet ( "up" ),        
			  tr( "NextPage" ), 
			  QString::null, this, 
			  SLOT( findTextBackwardMultipage ( )),  m_tb_find, "findpageprev" );
	new QToolButton ( Resource::loadIconSet ( "close" ),        
			  tr( "Close" ), 
			  QString::null, this, 
			  SLOT( closeFindBar ()),  m_tb_find, "findclose" );


	// Rename toolbar
	m_tb_rename = new QToolBar ( this );
	addToolBar ( m_tb_rename, "Rename", QMainWindow::Top, true );
	m_tb_rename-> setHorizontalStretchable ( true );
	m_tb_rename-> hide ( );
	m_rename_line_edit = new QLineEdit ( m_tb_rename, "rename" );
	m_tb_rename-> setStretchableWidget ( m_rename_line_edit );
	new QToolButton ( Resource::loadIconSet ( "done" ),        
			  tr( "Rename" ), 
			  QString::null, this, 
			  SLOT( closeRenameBar ()),  m_tb_rename, "rename" );
	new QToolButton ( Resource::loadIconSet ( "close" ),        
			  tr( "Rename" ), 
			  QString::null, this, 
			  SLOT( abortRenameBar ()),  m_tb_rename, "rename" );


	connect ( m_pdfview, SIGNAL( selectionChanged ( const QString & )), this, 
		  SLOT( selectionChanged ( const QString & )));

	connect( m_pdfview, SIGNAL( toggleFullscreen()), 
		 this, SLOT(toggleFullscreen()));

	connect (m_pdfview, SIGNAL( newDocument() ),
		 this, SLOT(newDocument()));

	connect (m_pdfview, SIGNAL( setPageCount(int,int) ),
		 this, SLOT(setPageCount(int,int)));

	connect (m_pdfview, SIGNAL( setBusy(bool) ),
		 this, SLOT(setBusy(bool)));

	connect (m_pdfview, SIGNAL( updateMenuItems() ),
		 this, SLOT(updateMenuItems()));

	connect (m_pdfview, SIGNAL( setNoteMode() ),
		 this, SLOT(setNoteMode()));

	connect (m_textedit, SIGNAL( addNote(int,int) ),
		 m_pdfview, SLOT(addNote(int,int)));

	connect (m_textedit, SIGNAL( nextPage() ),
		 this, SLOT(nextPage()));

	connect (m_textedit, SIGNAL( prevPage() ),
		 this, SLOT(prevPage()));

	connect (m_textedit, SIGNAL( toggleFullscreen() ),
		 this, SLOT(toggleFullscreen()));

	m_flash_timer = new QTimer(this);
	connect(m_flash_timer, SIGNAL(timeout()),
		this, SLOT(flashFindBarTimeout()));


	m_counter = new QLabel ( this );
	m_counter-> setAlignment ( AlignCenter | SingleLine );
	m_counter-> hide();
	
	m_isbusy = false;
	m_busybar = new QBusyBar ( this );
	m_busybar-> setParameters ( 12, 8, 200 );
	m_busybar-> hide ( );

	m_mode = PDF_VIEW;

	m_pageCounterTimer = new QTimer(this);
	connect(m_pageCounterTimer, SIGNAL(timeout()),
		this, SLOT(pageCounterTimeout()));
  
#ifdef AUTOTEST
	autotester();
#endif
	/* 
	   This line (and the one like it in the fullscreen code) allow us
	   to use Escape/Cancel without having it close the app
	*/
	reparent ( 0, WStyle_Customize, QPoint ( 0, 0 ));
	
	openFile ( );
}


void 
QPDFWindow::keyPressEvent ( QKeyEvent *e )
{
  static int pageNum = 0;

  // use keys for find()

  // left = prev_page (like an open book)
  // right = next_page
  // up = backward state
  // down = forward state

  // rotated view:
  //  real up => rotated right = next_page
  //  real left => rotated up = backward state
  //  real down => rotated left = prev_page
  //  real right => rotated down = forward state

  printf("QPDFWindow keypress %x\n", e->key());

  switch ( e-> key ( )) {

  case Qt::Key_Escape:
    setFullscreen(false);
    if (m_mode == PDF_VIEW) {
      m_pdfview->gestureAbort();
      m_pdfview->setTextSelectionMode(false);
    }
    break;

  case Qt::Key_Up:
    if (!(m_mode == PDF_VIEW && m_pdfview->isRotated())) {
      backwardState();
    } else {
      nextPage();
    }
    break;

  case Qt::Key_Right:
    if (!(m_mode == PDF_VIEW && m_pdfview->isRotated())) {
      nextPage();
    } else {
      forwardState();
    }
    break;

  case Qt::Key_Left:
    if (!(m_mode == PDF_VIEW && m_pdfview->isRotated())) {
      prevPage();
    } else {
      backwardState();
    }
    break;

  case Qt::Key_Down:
    if (!(m_mode == PDF_VIEW && m_pdfview->isRotated())) {
      forwardState();
    } else {
      prevPage();
    }
    break;

  case Key_0: case Key_1: case Key_2: case Key_3: case Key_4: case Key_5:
  case Key_6: case Key_7: case Key_8: case Key_9:
    pageNum = pageNum * 10 + e->key() - Key_0;
    setPageCount(pageNum);
    break;

    // Okay 
  case Key_Return:
  case Key_Enter:
    if (pageNum > 0) {
      gotoPage(pageNum);
      pageNum = 0;
    } else {
      search_again();
    }
    break;

  case Key_Backspace:
    if (pageNum > 0) {
      pageNum /= 10;
      setPageCount(pageNum);
    } else {
      search(e->key());
    }
    break;

  case Key_Space:
    if (m_mode == PDF_VIEW) {
      m_pdfview->gestureStart();
      break;
    }

  default:
    if (((e->key() & 0xff00) == 0) && isalpha(e->ascii())) {
      printf("search: %x ascii %x\n", e->key(), e->ascii());
      search(e->ascii());
    } else {
      printf("QPDFWindow keypress %x\n", e->key());
    }
    break;
  }
}      


// should move from pdfview to here
void
QPDFWindow::saveState()
{
  m_pdfview->saveState();
}

void
QPDFWindow::forwardState()
{
  m_pdfview->forwardState();
  if (m_mode == TEXT_VIEW) {
    gotoPage(m_pdfview->getPageNum());
    // go to correct place on page
  }
  updateMenuItems();
}

void
QPDFWindow::backwardState()
{
  m_pdfview->backwardState();
  if (m_mode == TEXT_VIEW) {
    gotoPage(m_pdfview->getPageNum());
  }
  updateMenuItems();
}


void
QPDFWindow::selectionChanged(const QString &str)
{
  qApp-> clipboard ( )-> setText ( str );
}


void
QPDFWindow::newDocument()
{
  makePageMenu();
  updateMenuItems();
}


void
QPDFWindow::makePageMenu()
{
  int n = m_pdfview->lastPageNum();
  int i, j;
  int id;

  for(i=m_pm_page->count()-1; i>m_pm_page_start_pagenum_index; i--) {
    m_pm_page->removeItemAt(i);
  }
  if (n > MAX_PAGES_ON_PAGE_MENU) {
    for(i=1; i<=n; i+=10) {
      QPopupMenu *popup = new QPopupMenu();
      int m = i + 9;
      if (m > n) {
	m = n;
      }
      for(j=i; j<=m; j++) {
	id = popup->insertItem(tr("%1").arg(j), j);
	popup->setItemParameter(id, j);
	popup->connectItem(id, this, SLOT(gotoPage(int)));
      }
      id = m_pm_page->insertItem(tr("%1 - %2").arg(i).arg(m), popup, i/10 + 1000);
    }
  } else {
    for(i=1; i<=n; i++) {
      id = m_pm_page->insertItem(tr("%1").arg(i), i);
      assert(id == i);
      m_pm_page->setItemParameter(id, i);
      m_pm_page->connectItem(id, this, SLOT(gotoPage(int)));
    }
  }
}
    

void QPDFWindow::updateMenuItems()
{
  int page = m_pdfview->getPageNum();
  for(int i = 1; i<=m_pdfview->lastPageNum(); i++) {
    m_pm_page->setItemChecked(i, i == page);
    m_pm_page->setItemEnabled(i, m_mode != NOTES_VIEW);
  }
  if (m_pdfview->lastPageNum() > MAX_PAGES_ON_PAGE_MENU) {
    int n = m_pdfview->lastPageNum()/10;
    int m = (page-1)/10;
    for(int i = 0; i<=n; i++) {
      m_pm_page->setItemChecked(i + 1000, i == m);
      m_pm_page->setItemEnabled(i + 1000, m_mode != NOTES_VIEW);
    }
  }
  m_pm_page->setItemEnabled(PM_PAGE_FORW, m_pdfview->forwardStateAllowed());
  m_pm_page->setItemEnabled(PM_PAGE_BACK, m_pdfview->backwardStateAllowed());

  m_pm_page->setItemEnabled(PM_PAGE_FIRST, m_mode != NOTES_VIEW 
			    && m_pdfview->getPageNum() != 1);
  m_pm_page->setItemEnabled(PM_PAGE_PREV, m_mode != NOTES_VIEW 
			    && m_pdfview->getPageNum() != 1);
  m_pm_page->setItemEnabled(PM_PAGE_LAST, m_mode != NOTES_VIEW
			    && m_pdfview->getPageNum() != m_pdfview->lastPageNum());
  m_pm_page->setItemEnabled(PM_PAGE_NEXT, m_mode != NOTES_VIEW
			    && m_pdfview->getPageNum() != m_pdfview->lastPageNum());


  //  m_pm_view->setItemEnabled(m_VIEW_FIT_TO_WIDTH, m_mode == PDF_VIEW);
  //  m_pm_view->setItemEnabled(PM_VIEW_FIT_TO_PAGE, m_mode == PDF_VIEW);
  m_pm_view->setItemEnabled(m_pm_view_highlight, m_mode == PDF_VIEW);
  m_pm_view->setItemChecked(m_pm_view_highlight, m_mode == PDF_VIEW
			    && m_pdfview->highlightMode());
  m_pm_view->setItemEnabled(m_pm_view_rotate, m_mode == PDF_VIEW);

  m_pm_edit->setItemEnabled(PM_EDIT_PASTE, m_mode != PDF_VIEW);
  m_pm_edit->setItemEnabled(PM_EDIT_CUT, m_mode != PDF_VIEW);

  // mode buttons
  m_but_pdf->setDown(m_mode == PDF_VIEW);
  m_but_text->setDown(m_mode == TEXT_VIEW);
  m_but_note->setDown(m_mode == NOTES_VIEW);

  if (m_mode == PDF_VIEW) {
    if (m_pdfview->textSelectionMode()) {
      m_but_pdf->setPixmap(m_pdf_text_select_pixmap);
    } else {
      m_but_pdf->setPixmap(m_pdf_move_pixmap);
    }
  } else {
    m_but_pdf->setPixmap(m_pdf_move_pixmap);
  }
      
  bool checked = false;
  int zoom;
  switch(m_mode) {
  case PDF_VIEW:
    zoom = (int)(m_pdfview->getZoom() * 100);
    break;
  case TEXT_VIEW:
  case NOTES_VIEW:
    zoom = (int)(m_textedit->getFontSize() * (100 / 12));
    break;
  }

  for ( uint i = 0; i < m_pm_view-> count ( ); i++ ) {
    int id = m_pm_view->idAt(i);
    int z = m_pm_view-> itemParameter ( id );
    if (z > 0) {
      if (z >= zoom && !checked) {
	m_pm_view-> setItemChecked ( id, true );
	checked = true;
      } else {
	m_pm_view-> setItemChecked ( id, false );
      }
    }
  }

  updateCaption();
}

QPDFWindow::~QPDFWindow ( )
{
  delete m_pdfview;
  delete m_textedit;
}


void
QPDFWindow::setTextMode()
{
  setMode(TEXT_VIEW);
}

void
QPDFWindow::setPDFMode()
{
  if (m_mode == PDF_VIEW) {
    // toggle text selection mode
    m_pdfview->setTextSelectionMode(! m_pdfview->textSelectionMode());
  } else {
    setMode(PDF_VIEW);
  }
}

void
QPDFWindow::setNoteMode()
{
  setMode(NOTES_VIEW);
}


void
QPDFWindow::setMode(ViewMode new_mode)
{

  ViewMode old_mode;

  if (! m_pdfview || !m_pdfview-> isOk ( )) {
    return;
  }
  if (m_mode == new_mode) {
    return;
  }

  // a cop-out for now
  // should really allow this to go through, and
  // if needed flag it so that changeState will finish it up
  // when it exits
  if (isBusy() 
      && new_mode == TEXT_VIEW) {
    return;
  }

  int wordPosition0, wordPosition1;
  int charPosition0, charPosition1;
  QPoint pixPosition0, pixPosition1;
  bool range = false;
  int page;
  QDocNote *note;

  note = NULL;
  wordPosition0 = wordPosition1 = -1;
  charPosition0 = charPosition1 = -1;
  pixPosition0 = QPoint();
  pixPosition1 = QPoint();
  page = m_pdfview->getPageNum();

  // This is made more complex because many of the
  // mapXXX functions need the pagemap to already be in place for that page.
  // Switching out of note view, it may be a new page.
  // So, have to delay them until gotoPage is done in the second switch.

  // close out old mode
  switch(m_mode) {

  case PDF_VIEW:
    m_pdfview->gestureAbort();
    if (m_pdfview->hasSelection()) {
      m_pdfview->getTextSelection(&wordPosition0, &wordPosition1);
      charPosition0 = m_pdfview->mapWordToChar(wordPosition0, false);
      charPosition1 = m_pdfview->mapWordToChar(wordPosition1, true);
      pixPosition0 = m_pdfview->mapWordToPix(wordPosition0, false);
      pixPosition1 = m_pdfview->mapWordToPix(wordPosition1, true);
      range = true;
    } else {
      // maybe better to go to closest note...
      pixPosition0 = m_pdfview->currentUpperLeftPix();
      wordPosition0 = m_pdfview->mapPixToWord(pixPosition0, true);
      if (wordPosition0 < 0) {
	wordPosition0 = 0;
      }
      charPosition0 = m_pdfview->mapWordToChar(wordPosition0);
      if (charPosition0 < 0) {
	charPosition0 = 0;
      }
    }
    break;

  case TEXT_VIEW:

    int line0, col0, line1, col1;
    
    if (m_textedit->getSelection(&line0, &col0, &line1, &col1)) {
      charPosition0 = m_textedit->positionToOffset(line0, col0);
      charPosition1 = m_textedit->positionToOffset(line1, col1);
      range = true;
    } else {
      charPosition0 = m_textedit->getCharPosition();
    }
    break;

  case NOTES_VIEW:
    bool changed;
    changed = m_pdfview->saveNotes(m_textedit->text(), 
				   m_textedit->getCharPosition(), 
				   &note);
    // note: pixPosition0, 1  are relative to full image size
    // there may not be any valid wordStart, wordEnd since page may
    // not have been loaded yet
    if (note == NULL) {
      page = m_pdfview->getPageNum();
      wordPosition0 = wordPosition1 = 0;
      range = false;
    } else {
      page = note->m_page;
      pixPosition0 = note->m_start;
      pixPosition1 = note->m_end;
      range = true;
    }
    break;

  }


  printf("setMode %d char %d word %d pix %d,%d  (%s)\n", m_mode,
	 charPosition0, wordPosition0, pixPosition0.x(), pixPosition0.y(),
	 m_pdfview->getText(wordPosition0,wordPosition0).latin1());
  if (range) {
    printf("  => char %d word %d pix %d,%d  (%s)\n",
	   charPosition1, wordPosition1, pixPosition1.x(), pixPosition1.y(),
	   m_pdfview->getText(wordPosition0, wordPosition1).latin1());
  }

  old_mode = m_mode;
  m_mode = new_mode;

  // start new mode
  switch(new_mode) {

  case PDF_VIEW:
    m_stack->raiseWidget(m_pdfview);
    gotoPage(page);
    if (m_pdfview->getPageNum() != page) {
      return;
    }
    m_pdfview->updateHighlights();  // may have changed if note deleted
    if (old_mode == NOTES_VIEW) {
      if (!pixPosition0.isNull() && !pixPosition1.isNull()) {
	pixPosition0 = m_pdfview->absolutePosition(pixPosition0);
	pixPosition1 = m_pdfview->absolutePosition(pixPosition1);
      }
    }
    if (charPosition0 != -1) {
      wordPosition0 = m_pdfview->mapCharToWord(charPosition0);
    }
    if (charPosition1 != -1) {
      wordPosition1 = m_pdfview->mapCharToWord(charPosition1);
    }
    if (!pixPosition0.isNull()) {
      wordPosition0 = m_pdfview->mapPixToWord(pixPosition0, true);
    }
    if (!pixPosition1.isNull()) {
      wordPosition1 = m_pdfview->mapPixToWord(pixPosition1, true);
    }
    if (old_mode == NOTES_VIEW && note != NULL) {
      if (wordPosition0 != note->m_firstWord
	  || wordPosition1 != note->m_lastWord) {
	printf("*** BUG *** wordPosition %d %d note %d %d\n",
	       wordPosition0, wordPosition1, note->m_firstWord, note->m_lastWord);
      }
    }
    if (range && old_mode == NOTES_VIEW
	&& note != NULL && note->m_firstWord >= 0 && note->m_lastWord >=0) {
      m_pdfview->setTextSelection(note->m_firstWord, note->m_lastWord, true);
    } else if (range && wordPosition0 != -1 && wordPosition1 != -1) {
      m_pdfview->setTextSelection(wordPosition0, wordPosition1, true);
    } else if (wordPosition0 != -1) {
      m_pdfview->setWordPosition(wordPosition0);
    }
    break;

  case TEXT_VIEW:
    m_stack->raiseWidget(m_textedit);
    gotoPage(page);
    if (m_pdfview->getPageNum() != page) {
      return;
    }
    m_textedit->setText(m_pdfview->getText());
    m_textedit->setReadOnly(true);
    if (old_mode == NOTES_VIEW) {
      if (!pixPosition0.isNull() && !pixPosition1.isNull()) {
	pixPosition0 = m_pdfview->absolutePosition(pixPosition0);
	pixPosition1 = m_pdfview->absolutePosition(pixPosition1);
      }
    }
    if (!pixPosition0.isNull() && wordPosition0 == -1) {
      wordPosition0 = m_pdfview->mapPixToWord(pixPosition0, true);
    }
    if (charPosition0 == -1 && wordPosition0 != -1) {
      charPosition0 = m_pdfview->mapWordToChar(wordPosition0, false);
    }
    if (range) {
      if (!pixPosition1.isNull() && wordPosition1 == -1) {
	wordPosition1 = m_pdfview->mapPixToWord(pixPosition1, true);
      }
      if (charPosition1 == -1 && wordPosition1 != -1) {
	charPosition1 = m_pdfview->mapWordToChar(wordPosition1, true);
      }
    }
    if (charPosition0 != -1) {
      m_textedit->setCharPosition(charPosition0, false);
    }
    if (range && charPosition1 != -1) {
      m_textedit->setCharPosition(charPosition1, true);
    }
    break;

  case NOTES_VIEW:
    QString noteText;
    m_stack->raiseWidget(m_textedit);
    m_textedit->setReadOnly(false);
    if (charPosition0 != -1 && wordPosition0 == -1) {
      wordPosition0 = m_pdfview->mapCharToWord(charPosition0);
    }
    if (charPosition1 != -1 && wordPosition1 == -1) {
      wordPosition1 = m_pdfview->mapCharToWord(charPosition1);
    }
    if (range && wordPosition0 != -1 && wordPosition1 != -1) {
      noteText = m_pdfview->getNotes(wordPosition0, wordPosition1, &charPosition0);
    }
    if (noteText.isNull()) {
      noteText = m_pdfview->getNotes(pixPosition0, &charPosition0);
    }
    m_textedit->setText(noteText);
    if (charPosition0 >= 0) {
      m_textedit->setCharPosition(charPosition0);
    } 
    //    m_textedit->show();
    break;
  }

  printf("post setMode %d char %d word %d pix %d,%d  (%s)\n", m_mode,
	 charPosition0, wordPosition0, pixPosition0.x(), pixPosition0.y(),
	 m_pdfview->getText(wordPosition0,wordPosition0).latin1());
  if (range) {
    printf("  => char %d word %d pix %d,%d  (%s)\n",
	   charPosition1, wordPosition1, pixPosition1.x(), pixPosition1.y(),
	   m_pdfview->getText(wordPosition0, wordPosition1).latin1());
  }
  updateMenuItems();
}


#if 0
QString
QPDFWindow::changedRegion(QString t1, QString t2)
{
  if (t1.isEmpty()) {
    return(t2);
  }
  if (t2.isEmpty()) {
    return(t1);
  }
  QString result;
  const QChar *u1 = t1.unicode();
  const QChar *u2 = t2.unicode();
  const QChar *e1 = u1 + t1.length() - 1;
  const QChar *e2 = u2 + t2.length() - 1;

  while(*e2 == *e1) {
    e2--;
    e1--;
  }
  while(*u1 == *u2) {
    u1++;
    u2++;
  }
  int n = 0;
  while(u1 <= e1) {
    result[n++] = *u1++;
  }
  return(result);
}
#endif

  

// vv Fullscreen handling (for broken QT-lib) [David Hedbor, www.eongames.com]

void QPDFWindow::resizeEvent ( QResizeEvent * )
{
  
  if ( m_fullscreen && ( size ( ) != qApp -> desktop() -> size ( )) ) {
    setFullscreen ( true );
  }
}

void QPDFWindow::focusInEvent ( QFocusEvent * )
{
  if ( m_fullscreen ) {
        setFullscreen ( true );
  }
}

void QPDFWindow::toggleFullscreen ( )
{
  setFullscreen ( !m_fullscreen );
}


void QPDFWindow::setFullscreen ( bool b )
{
	static QSize normalsize;


	m_fullscreen = b;

	// doesn't do much good in full screen, you can't see it anyway
	m_pm_view-> setItemChecked ( m_pm_view_fullscreen, b );

	if ( b ) {
	  if ( !normalsize. isValid ( )) {
	    normalsize = size ( );
	  }    
	  
	  setFixedSize ( qApp-> desktop ( )-> size ( ));
	  showNormal ( );
	  reparent ( 0, WStyle_Customize | WStyle_NoBorder, 
		     QPoint ( 0, 0 ));
	  showFullScreen ( );

	  m_tb_menu-> hide ( );
	  m_tb_find-> hide ( );
	  m_tb_rename-> hide ( );

	} else {
	  showNormal ( );
	  reparent ( 0, WStyle_Customize, QPoint ( 0, 0 ));
	  resize ( normalsize );
	  showMaximized ( );
	  normalsize = QSize ( );
	  
	  m_tb_menu-> show ( );
	  m_tb_find-> hide ( );
	  m_tb_rename-> hide ( );
	}	
	layoutPageCounter();
}
	
// ^^ Fullscreen handling (for broken QT-lib)




void QPDFWindow::toggleRotate()
{
  setRotate(1 - m_pdfview->isRotated());
}

void QPDFWindow::setRotate(bool rotated)
{
  m_pm_view-> setItemChecked ( m_pm_view_rotate, rotated );
  m_pdfview-> setRotation( rotated );
}

void QPDFWindow::toggleHighlight()
{
  setHighlight(1 - m_pdfview->highlightMode());
}

void QPDFWindow::setHighlight(bool highlight)
{
  m_pm_view-> setItemChecked ( m_pm_view_highlight, highlight );
  m_pdfview-> setHighlightMode( highlight );
}


void QPDFWindow::setZoom ( int z )
{
  float zoom = -1.0;

  printf("setZoom(%d)\n", z);
  switch(m_mode) {
  case PDF_VIEW:
    if (m_pdfview->visibleWidth() <= 0 || m_pdfview->visibleHeight() <= 0
	|| m_pdfview->contentsWidth() <= 0 || m_pdfview->contentsHeight() <= 0 ) {
      return;
    }
    
    switch ( z ) {
      
#if 0
    case PM_VIEW_FIT_TO_WIDTH:
      zoom = (float)m_pdfview-> visibleWidth ( ) /  (float)m_pdfview-> contentsWidth();
      break;
      
      
    case PM_VIEW_FIT_TO_PAGE:
      zoom = QMIN( (float)m_pdfview-> visibleWidth() / (float)m_pdfview-> contentsWidth(),
		   (float)m_pdfview-> visibleHeight() / (float)m_pdfview-> contentsHeight());
      break;
      
#endif

    case 1:
      zoom = m_pdfview->getZoom() * 1.25;
      break;

    case -1:
      zoom = m_pdfview->getZoom() / 1.25;
      break;

    default:
      if (z > 0 && z < 200) {
	zoom = (float)z  / 100.0;
      }
      break;
      
    }
    if (zoom > MAX_ZOOM) {
      zoom = MAX_ZOOM;
      printf("hit max zoom limit\n");
    }
    if (zoom > 0.0) {
      m_pdfview->setZoom(zoom);
    }
    break;

  case TEXT_VIEW:
  case NOTES_VIEW:
    switch(z) {
    case 1:
      m_textedit->zoomIn();
      break;
    case -1:
      m_textedit->zoomOut();
      break;
    default:
      m_textedit->setFontSize((int)((float)z * 12.0 / 100.0));
      break;
    }
  }

  updateMenuItems();
}


void QPDFWindow::updateCaption ( )
{
  QString cap;
	
  cap = QString ( "[%1] %2 " )
    . arg ( m_pdfview->getPageNum() )
    . arg ( m_pdfview->getFilename() );
  setCaption ( cap );
}

#if 0
void QPDFWindow::gotoPageDialog ( )
{
	QDialog *d = new QDialog ( this, "gotodlg", true );
	d-> setCaption ( tr( "Goto page" ));
	
	QBoxLayout *lay = new QVBoxLayout ( d, 4, 4  );	
	
	QLabel *l = new QLabel ( tr( "Select from 1 .. %1:" ). arg ( m_pages ), d );
	lay-> addWidget ( l );
	
	QSpinBox *spin = new QSpinBox ( 1, m_pages, 1, d );
	spin-> setValue ( m_currentpage );
	spin-> setWrapping ( true );
	spin-> setButtonSymbols ( QSpinBox::PlusMinus );
	lay-> addWidget ( spin );
	
	if ( d-> exec ( ) == QDialog::Accepted ) {
		gotoPage ( spin-> value ( ));
	}
	
	delete d;
}
#endif

void QPDFWindow::toggleFindBar ( )
{
#if 0
	if ( m_to_find-> isOn ( ) == m_tb_find-> isVisible ( ))
		m_to_find-> setOn ( !m_tb_find-> isVisible ( ));
#endif
	setFindBar(! m_tb_find-> isVisible ( ));
}

void QPDFWindow::closeFindBar( )
{
  printf("closeFindBar\n");
  setFindBar(false);
}

void QPDFWindow::setFindBar( bool active )
{
  if ( m_tb_find-> isVisible ( ) != active ) {
    printf("setFindBar(%d)\n", active);
    if ( m_tb_find-> isVisible ( ) ) {
      m_tb_find-> hide ( );
      //      m_stack-> setFocus ( );
    } else {
      m_tb_find-> show ( );
      m_find_line_edit-> setFocus ( );
    }
    printf("done setFindBar(%d)\n", active);
  }
  m_pm_view-> setItemChecked ( m_pm_view_find, active );

}

void QPDFWindow::search(const int search_char)
{
  switch(search_char) {
  case Key_Backspace:
    m_find_line_edit->backspace();
    break;
  case Qt::Key_Escape:
    m_find_line_edit->clear();
    break;
  case Key_Return:
  case Key_Enter:
    // already get it from keyPressEvent
    //  search_again();
    break;
  default:
    printf("search(0x%x)\n", search_char);
    m_find_line_edit->insert(QChar((char)search_char));
    break;
  }
}

void
QPDFWindow::search_again()
{
  // this is a hack, should move forward a char to get next match
  // unfortunely, we remember we are are in units of words, so we'd have to
  // have char indices off on the side, which would be ugly too.
   int from, to;
   printf("search_again\n");
   switch(m_mode) {
   case PDF_VIEW:
     if (m_pdfview->hasSelection()) {
       m_pdfview->getTextSelection(&from,&to);
       m_pdfview->setTextSelection(from+1,to+1);
     }
     break;
   case TEXT_VIEW:
     m_textedit->setCharPosition(m_textedit->getCharPosition()+1);
     break;
   }
   findTextInPage(m_find_line_edit->text());
}
 
bool QPDFWindow::findText ( const QString &str, bool reverse, bool multipage )
{
  int from, to;
  int page;
  bool found;
  int start_page = m_pdfview->getPageNum();

  switch(m_mode) {
  case PDF_VIEW:
    if (m_pdfview->hasSelection()) {
      m_pdfview->getTextSelection(&from, &to);
      from = m_pdfview->mapWordToChar(from);
      to = m_pdfview->mapWordToChar(to);
    } else {
      //    from = m_pdfview->mapPixToWord(m_pdfview->currentUpperLeftPix());
      from = 0;
    }
    break;
  case TEXT_VIEW:
    int line0, col0, line1, col1;
    if (m_textedit->getSelection(&line0, &col0, &line1, &col1)) {
      from = m_textedit->positionToOffset(line0, col0);
    } else {
      from = m_textedit->getCharPosition();
    }
    printf("start search %d\n", from);
    break;
  case NOTES_VIEW:
    return(m_textedit->find(str, false, reverse));
  }

  to = -1;

  found = m_pdfview->findText ( str, &from, &to, &page, reverse, multipage );

  if (!found && !multipage && !reverse) {
    // wrap around
    printf("search wrap found %d multi %d reverse %d\n", found, multipage, reverse);
    from = 0;
    to = -1;
    found = m_pdfview->findText ( str, &from, &to, &page, reverse, multipage );
  }
  
  if (found) {
    if (page != start_page) {
      gotoPage(page);
    }
    if (m_mode == TEXT_VIEW) {
      printf("found char pos %d to %d\n", from, to);
      m_textedit->setCharPosition(from, false);
      m_textedit->setCharPosition(to+1, true);
    }
    if (m_mode == PDF_VIEW) {
      m_pdfview->setTextSelection(m_pdfview->mapCharToWord(from), 
				  m_pdfview->mapCharToWord(to), true);
      m_pdfview->setCharPosition(from);
    }
  }
    
  return(found);
}

void QPDFWindow::findNoMatch(bool flag)
{
  if (flag) {
    m_tb_find->setPalette(QPalette(Qt::red));
  } else {
    m_tb_find->setPalette(QPalette(Qt::white));
  }
}

void QPDFWindow::findTextForward ( )
{
  findNoMatch(false);
  if (!findText( m_find_line_edit-> text ( ), false, false )) {
    findNoMatch(true);
  }
}


void QPDFWindow::findTextMultipage ( )
{
  findNoMatch(false);
  if (!findText ( m_find_line_edit-> text ( ), false, true )) {
    findNoMatch(true);
  }
}

void QPDFWindow::findTextBackward ( )
{
  findNoMatch(false);
  if (!findText ( m_find_line_edit-> text ( ), true, false )) {
    findNoMatch(true);
  }
}

void QPDFWindow::findTextBackwardMultipage ( )
{
  findNoMatch(false);
  if (!findText ( m_find_line_edit-> text ( ), true, true )) {
    findNoMatch(true);
  }
}

void QPDFWindow::findTextInPage(const QString &str)
{
  printf("QPDFWindow::findTextInPage\n");
  flashFindBar();
  findNoMatch(false);
  if (!findText( str, false, false )) {
    // kill last char since it didn't match
    // m_find_line_edit->setText( m_find_line_edit->text().left(m_find_line_edit->text().length() - 1) );
    findNoMatch(true);
  }
}


void QPDFWindow::editCut ( )
{
  if (m_mode == NOTES_VIEW) {
    m_textedit->cut();
  }
}

void QPDFWindow::editPaste ( )
{
  if (m_mode == NOTES_VIEW) {
    m_textedit->paste();
  }
}

void QPDFWindow::editCopy ( )
{
  switch(m_mode) {
  case PDF_VIEW:
    qApp-> clipboard ( )-> setText ( m_pdfview->selectedText() );
    break;
  case TEXT_VIEW:
  case NOTES_VIEW:
    qApp-> clipboard ( )-> setText ( m_textedit->selectedText() );
    break;
  }
}

void QPDFWindow::gotoPage ( int page )
{
  printf("QPDFWindow::gotoPage(%d)\n", page);
  if (!m_pdfview->isValid()) {
    return;
  }
  if (page > m_pdfview->lastPageNum()) {
    page = m_pdfview->lastPageNum();
  }
  if (page <= 0) {
    page = 1;
  }
  switch(m_mode) {
  case PDF_VIEW:
    m_pdfview->gotoPage(page);
    break;
  case TEXT_VIEW:
    m_pdfview->setPage(page);
    m_textedit->setText( m_pdfview->getText(0,-1,page) );
    m_textedit->setCharPosition(0);
    break;
    // TODO: note mode, select page of notes
  }

  updateMenuItems();
}

void QPDFWindow::firstPage ( )
{
  gotoPage(1);
}

void QPDFWindow::prevPage ( )
{
  gotoPage(m_pdfview->getPageNum() - 1);
}

void QPDFWindow::nextPage ( )
{
  gotoPage(m_pdfview->getPageNum() + 1);
}

void QPDFWindow::lastPage ( )
{
  gotoPage(m_pdfview->lastPageNum());
}


void QPDFWindow::closeFileSelector ( )
{
	m_tb_menu-> show ( );
	m_stack-> raiseWidget ( m_pdfview );
}

void QPDFWindow::openFile ( )
{
  int num_file = m_filesel->fileCount();
  if (num_file == 0) {
    // FIXME: need to do something here? no pdf files, so exit??
  }
  printf("files found = %d\n", num_file);

	m_tb_menu-> hide ( );
	m_tb_find-> hide ( );
	m_tb_rename-> hide ( );
	m_stack-> raiseWidget ( m_filesel );
}

void QPDFWindow::openFile ( const QString &f )
{
	DocLnk nf;
	nf. setType ( "application/pdf" );
	nf. setFile ( f );
	QFileInfo fi ( f );
	nf. setName ( fi. baseName ( ));
	openFile ( nf );
}

void QPDFWindow::openFile ( const DocLnk &f )
{
	QString fn = f. file ( );
	QFileInfo fi ( fn );
	
	if ( fi. exists ( )) {
	  closeFileSelector ( );
	  m_mode = PDF_VIEW;
	  m_pdfview->open(fn);
	  m_stack-> setFocus ( );
	} else {
	  QMessageBox::warning ( this, tr( "Error" ), tr( "File does not exist !" ));
	}		
}

void QPDFWindow::setDocument ( const QString &f )
{
	if ( f. find ( ".desktop", 0, true ) == -1 ) 
		openFile ( f );
	else 
		openFile ( DocLnk ( f ));
		
	closeFileSelector ( );
}

void QPDFWindow::openRenameBar ( )
{
  QString path = m_pdfview-> getPath ( );
  int pos = path. findRev ( '/' );
  if ( pos < 0 ) {
    pos = path. length ( );
  }
  m_rename_line_edit-> setText( path );
  m_rename_line_edit-> setCursorPosition( pos );
  setRenameBar( true );
}

void QPDFWindow::setRenameBar ( bool active )
{
  if (active != m_tb_rename->isVisible()) {
    printf("setRenameBar(%d)\n", active);
    if (active) {
      m_tb_rename-> show ( );
      m_tb_rename-> setFocus ( );
    } else {
      m_tb_rename-> hide ( );
      //    m_stack->setFocus();
    }
    printf("exit setRenameBar(%d)\n", active);
  }
}

void QPDFWindow::abortRenameBar ( )
{
  setRenameBar( false );
}


#ifdef ARM
extern int errno;
#else
extern "C" {
  int errno;
};
#endif

void
QPDFWindow::flush()
{
  m_pdfview->closeDocument();
  m_pdfview->openDocument();
}



void QPDFWindow::closeRenameBar ( )
{
  QString newPath = m_rename_line_edit-> text( );
  QString oldPath = m_pdfview->getPath();

  if (isBusy()) {
    // don't change files if you are using them!!
    return;
  }
  if (newPath != oldPath ) {
    m_pdfview->closeDocument();
#ifdef AUTOTEST
    printf("RENAME TURNED OFF FOR DEBUGGING!!\n");
#else
    if ( ::rename( oldPath, newPath ) < 0 ) {
      qWarning( "problem renaming file %s to %s errno %d",
		oldPath.latin1(), newPath.latin1(), errno  );
      newPath = oldPath;
    } else {
      if ( ::rename( m_pdfview->notesPath(oldPath), m_pdfview->notesPath(newPath) ) < 0 ) {
	qWarning( "problem renaming file %s to %s errno %d",
		  m_pdfview->notesPath(oldPath).latin1(), 
		  m_pdfview->notesPath(newPath).latin1(), errno  );
      } 
      printf("renamed %s => %s\n", oldPath.latin1(), newPath.latin1());
      //      m_pdfview->setPath(newPath);
      //      updateCaption();
    }
    m_pdfview->setPath(newPath);
#endif
    m_pdfview->openDocument();
  }
  setRenameBar( false );
}

void
QPDFWindow::flashFindBar(int time)
{
  if (! (m_tb_find->isVisible() && !m_flash_timer->isActive()) ) {
    // don't mess with it if we have find bar on manually (visible and timer not active)
    setFindBar(true);
    if (m_flash_timer->isActive()) {
      m_flash_timer->changeInterval(time);
    } else {
      m_flash_timer->start(time, true);
    }
  }
}

void
QPDFWindow::flashFindBarTimeout()
{
  setFindBar(false);
}



void QPDFWindow::layoutPageCounter()
{
  int w = m_counter->width();
  int h = m_counter->height();
  w = 50;
  h = 15;

  if (m_mode == PDF_VIEW && m_pdfview->isRotated()) {
    m_counter-> setGeometry ( m_stack->x() + m_stack->width() - w - 2, m_stack->y(), w, h);
    m_busybar-> setGeometry ( m_stack->x() + m_stack->width() - w - 2, m_stack->y(), w, h);
  } else {
    m_counter-> setGeometry ( m_stack->x() + m_stack->width() - w + 2, 
			      m_stack->y() + m_stack->height() - h + 2, w, h);
    m_busybar-> setGeometry ( m_stack->x() + m_stack->width() - w + 2, 
			      m_stack->y() + m_stack->height() - h + 2, w, h);
  }
}

void QPDFWindow::setPageCount ( int actp, int time )
{
  printf("setPageCount(%d, %d)\n", actp, time);
  m_counter-> setText ( QString ( "%1 / %2" ). arg ( actp ). arg ( m_pdfview->lastPageNum() ));
  showPageCount(time);
}

void QPDFWindow::showPageCount(int time)
{
  m_pageCounterTimer->stop();
  m_pageCounterTimer->start(time, TRUE);
  layoutPageCounter();
  m_busybar->hide();
  m_counter->show();
}

void QPDFWindow::pageCounterTimeout()
{
  m_counter->hide();
  if (m_isbusy) {
    m_busybar->show();
  }
}

void QPDFWindow::setBusy ( bool b )
{
  if ( b != m_isbusy ) {
    if ( b ) {
      m_busybar-> beginBusy ( );
      layoutPageCounter();
      m_busybar-> show ( );
      m_counter-> hide ( );
    }
    else {
      m_counter-> show ( );	
      m_busybar-> hide ( );
      m_busybar-> endBusy ( );
    }		
    m_isbusy = b;
  }
}

/*
  m_isbusy is for display purposes, while this one indicates
  that it is actually reading the document and may process events
  in the middle of it, so don't screw with the file!!
*/
bool 
QPDFWindow::isBusy ( ) const
{
  return m_pdfview->isBusy();
}

