#ifndef __QPDF_H__
#define __QPDF_H__

//#define OZ

// NOTE: seems like you need to manually comment out
// the two lines inside the #ifdef below
//#define AUTOTEST

#include "aconf.h"

#include "qpdfview.h"

#include "signal.h"

#include <qpe/qpeapplication.h>
#include <qpe/resource.h>
#include <qpe/applnk.h>
#include <qpe/qcopenvelope_qws.h>
#include <qmainwindow.h>

#include <qclipboard.h>
#include <qpe/qpetoolbar.h>
#include <qtoolbutton.h>
#include <qpe/qpemenubar.h>
#include <qpopupmenu.h>
#include <qwidgetstack.h>
#include <qtimer.h>
#include <qfileinfo.h>
#include <qstring.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qlayout.h>
#include <qdialog.h>
#include <qlabel.h>
#include <qmessagebox.h>
#ifdef OZ
#include <opie/ofileselector.h>
#else
#include <qpe/fileselector.h>
#endif

#include "textedit.h"


const float DEFAULT_ZOOM = 0.9;
const int MAX_PAGES_ON_PAGE_MENU = 17;
const int MAX_STATE_HISTORY = 20;


class DocLnk;
class FileSelector;
class OFileSelector;
class QWidgetStack;
class QLineEdit;



enum {
  PM_EDIT_CUT, PM_EDIT_COPY, PM_EDIT_PASTE
};

enum {
  PM_PAGE_MARK = 20001, PM_PAGE_FORW, PM_PAGE_BACK, 
  PM_PAGE_NEXT, PM_PAGE_PREV, PM_PAGE_FIRST, PM_PAGE_LAST
};

class QPDFWindow : public QMainWindow {
	Q_OBJECT

public:
	QPDFWindow ( );
	virtual ~QPDFWindow ( );

public slots:

	void openFile ( );
	void openFile ( const QString & );
	void openFile ( const DocLnk & );	

	void firstPage ( );
	void prevPage ( );
	void nextPage ( );
	void lastPage ( );
	
	void gotoPage ( int n );
	
	void setZoom ( int z );

	void setFullscreen ( bool b = true );
	void toggleFullscreen ( );

	void setRotate( bool b );
	void toggleRotate ( );	

	void setHighlight( bool b );
	void toggleHighlight ( );

	void setNoteMode ( );
	void setTextMode ( );
	void setPDFMode ( );
	void setMode( ViewMode );


	void updateMenuItems();

	void toggleFindBar ( );	
	void setFindBar ( bool active );
	void closeFindBar ( );
	void search(const int search_char);
	void search_again();

	void selectionChanged( const QString &str );

	bool findText( const QString &str, bool reverse, bool multipage );

	void findTextInPage ( const QString & );	
	void findTextForward ( );
	void findTextBackward ( );
	void findTextMultipage ( );
	void findTextBackwardMultipage ( );
	
	void openRenameBar ( );
	void setRenameBar ( bool active );
	void abortRenameBar ( );
	void closeRenameBar ( );

	void setDocument ( const QString & );
	
	void editCopy();
	void editCut();
	void editPaste();

	void newDocument();

	void findNoMatch(bool);
	void flashFindBar(int time = 2000);

	void showPageCount(int time = 4000);
	void setPageCount(int page, int time = 4000);

	bool isBusy ( ) const;
	void setBusy ( bool b = true );

	void forwardState();
	void backwardState();
	bool forwardStateAllowed();
	bool backwardStateAllowed();
	void saveState();

	void flush();

private slots:
	void closeFileSelector ( );
	void updateCaption ( );
	void keyPressEvent(QKeyEvent*);
	void pageCounterTimeout();
	void flashFindBarTimeout();

	// for some reason moc seems to ignore this conditional compilation
	// resulting in moc_qpdf.cpp 'autotester' is not a memory of QPDFWindow...
	// So, here we comment it out as well.
	// (I know moc has to be smarter than that!)
#ifdef AUTOTEST
	//	void autotester();
	//	void autotester_timer_event();
#endif

protected:

	virtual void resizeEvent ( QResizeEvent *e );
	virtual void focusInEvent ( QFocusEvent *e );
	
	//	QString changedRegion(QString s, QString org_s);

private:
	void makePageMenu();
	void layoutPageCounter();

	QWidgetStack *m_stack;

	ViewMode m_mode;
	
	QDocView *m_current_viwe;

	QDocPixView *m_pdf_view;
	QDocTextView *m_text_view;
	QDocNoteView *m_note_view

#ifdef OZ
	OFileSelector *m_filesel;
#else
	FileSelector *m_filesel;
#endif
	
	QToolBar *m_tb_menu;
	QToolBar *m_tb_find;
	QToolBar *m_tb_rename;

	QLineEdit *m_find_line_edit;
	QLineEdit *m_rename_line_edit;

	QPopupMenu *m_pm_file;
	QPopupMenu *m_pm_edit;
	QPopupMenu *m_pm_view;
	QPopupMenu *m_pm_zoom;
	QPopupMenu *m_pm_page;
	int m_pm_page_start_pagenum_index;

	int m_pm_view_rotate;
	int m_pm_view_fullscreen;
	int m_pm_view_highlight;
	int m_pm_view_find;
	int m_pm_view_zoomin;
	int m_pm_view_zoomout;

	QButton *m_but_next_page;
	QButton *m_but_prev_page;
	QButton *m_but_pdf;
	QButton *m_but_note;
	QButton *m_but_text;
	QPixmap m_pdf_move_pixmap;
	QPixmap m_pdf_text_select_pixmap;

	QToolButton *m_to_find, *m_to_full;

	bool m_fullscreen;
	bool m_highlights;
	bool m_rotated;

	int m_current_history_state; // index to backward+1 or foward-1
	int m_last_history_state;    // index to write new states
	QDocState m_state_history[MAX_STATE_HISTORY];


	QTimer *m_flash_timer;

	QTimer *m_pageCounterTimer;
	QLabel *m_counter;


	QBusyBar *m_busybar;
	bool m_isbusy;

};


#endif
