/**********************************************************************
** Copyright (C) 2000-2002 Trolltech AS.  All rights reserved.
**
** This file is part of the Qtopia Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/

#ifndef TEXTEDIT_H
#define TEXTEDIT_H

#define QTEXTEDIT_OPEN_API

#include <qpe/filemanager.h>

#include <qmainwindow.h>
#include <qmultilineedit.h>
#include <qlist.h>
#include <qmap.h>
#include <qevent.h>

class QWidgetStack;
class QToolButton;
class QPopupMenu;
class QToolBar;
class QLineEdit;
class QAction;
class FileSelector;
class QpeEditor;



class QTextEdit : public QMultiLineEdit
{
    Q_OBJECT

public:
    friend class QMultiLineEdit;

    TextEdit( QWidget *parent = 0, const char *name = 0, WFlags f = 0);
    ~TextEdit();

    bool openFile( const QString & );
    QString getFilename();
    void setText(QString text);
    //   void show();
    
    void setCharPosition(int pos, bool mark = false);
    int getCharPosition();

    void mouseReleaseEvent( QMouseEvent *e );
    void mouseMoveEvent( QMouseEvent *e );
    void mousePressEvent( QMouseEvent *e );
    void resizeEvent( QResizeEvent *e );

    bool find( const QString &txt, bool caseSensitive,
	       bool backwards );

    // IF YOU ARE HAVING TROUBLE COMPILING THIS----
    // I have broken the "protected" status of qmultilineedit member functions.
    // To do this, I have put the following line into qmultilinedit.h:
    //   friend class TextEdit

    // (Hey Trolltech! There is no way to do serious customization of the behavior of
    //  qmultilineedit without having access to the screen mapping.)

    // (eventually, need to make a copy of qmultilineedit and change a lot of
    //  of things to allow same functionality as PDF_VIEW (e.g. multiple
    //  selections, rotation, gestures, etc.))
    void offsetToPosition( int position, int *row, int *col ) const;
    int positionToOffset( int row, int col ) const;

    bool getSelection(int *r1, int *c1, int *r2, int *c2) const {
      return(getMarkedRegion(r1,c1,r2,c2));
    }

    QString selectedText() {
      return(markedText());
    }

    int detectEdges(QPoint);
    int detectCorners(QPoint);

public slots:

    void editCut();
    void editCopy();
    void editPaste();

    void zoomIn();
    void zoomOut();
    int getFontSize();
    void setFontSize(int sz);
    void setFontSize(int sz, bool round_down_not_up);


signals:
    void setFullscreen(bool b);
    void toggleFullscreen();
    void addNote(int startingOffset, int endingOffset);
    void nextPage();
    void prevPage();

    //    void close();

protected:
    virtual void keyPressEvent ( QKeyEvent *e );

    //    void closeEvent( QCloseEvent *e );


private:
    void colorChanged( const QColor &c );
    int m_font_size;
    bool m_mouse_inside_selection;
    QPoint m_mouse_start;
};

#endif
