/**********************************************************************
** Copyright (C) 2000-2002 Trolltech AS.  All rights reserved.
**
** This file is part of the Qtopia Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/

#include <stdlib.h> //getenv

#include <qpe/global.h>
#include <qpe/fileselector.h>
#include <qpe/applnk.h>
#include <qpe/resource.h>
#include <qpe/config.h>
#include <qpe/qpeapplication.h>
#include <qpe/qpemenubar.h>
#include <qpe/qpetoolbar.h>
//#include <qpe/finddialog.h>

#include <qaction.h>
#include <qcolordialog.h>
#include <qfileinfo.h>
#include <qlineedit.h>
#include <qmessagebox.h>
#include <qobjectlist.h>
#include <qpopupmenu.h>
#include <qspinbox.h>
#include <qtoolbutton.h>
#include <qwidgetstack.h>
#include <qscrollview.h>

#include "qgesture.h"
#include "textedit.h"


bool TextEdit::openFile( const QString &f )
{
    DocLnk nf;
    nf.setType("text/plain");
    nf.setFile(f);
    return(openFile(nf));
}

bool TextEdit::openFile( const DocLnk &f )
{
    clear();
    FileManager fm;
    QString txt;
    if ( !fm.loadFile( f, txt ) ) {
      return(false);
    }
    setText(txt);
    setEdited(FALSE);
    return(true);
}


void 
TextEdit::offsetToPosition( int position, int *row, int *col ) const 
{
  QMultiLineEdit::offsetToPositionInternal(position, row, col);
}

int 
TextEdit::positionToOffset( int row, int col ) const
{
  return QMultiLineEdit::positionToOffsetInternal(row, col);
}

#ifdef FIXME // newline func
void 
TextEdit::offsetToPosition( int position, int *row, int *col ) const 
{
    if (position <= 0) {
	*row = 0;
	*col = 0;
	return;
    }
    else {
	int charsLeft = position;
	int i;

	for( i = 0; contents->at( i ); ++i ) {
	    if (lineLength( i ) < charsLeft)
		charsLeft -= lineLength( i );
	    else {
		*row = i;
		*col = charsLeft;
		return;
	    }
	    if ( contents->at( i )->newline )
		--charsLeft;
	}

	if (contents->at( i - 1) && !contents->at( i - 1 )->newline) {
	    *row = i - 1;
	    *col = lineLength( i - 1 );
	}
	else {
	    *row = i;
	    *col = 0;
	}
	return;
    }
    if (position <= 0) {
	*row = 0;
	*col = 0;
	return;
    }
    else {
	int charsLeft = position;
	int i;

	for( i = 0; contents->at( i ); ++i ) {
	    if (lineLength( i ) < charsLeft)
		charsLeft -= lineLength( i );
	    else {
		*row = i;
		*col = charsLeft;
		return;
	    }
	    if ( contents->at( i )->newline )
		--charsLeft;
	}

	if (contents->at( i - 1) && !contents->at( i - 1 )->newline) {
	    *row = i - 1;
	    *col = lineLength( i - 1 );
	}
	else {
	    *row = i;
	    *col = 0;
	}
	return;
    }
}


int 
TextEdit::positionToOffset( int row, int col ) const
{
    row = QMAX( QMIN( row, numLines() - 1), 0 ); // Sanity check
    col = QMAX( QMIN( col,  lineLength( row )), 0 ); // Sanity check
    if ( row == 0 )
	return QMIN( col, lineLength( 0 ));
    else {
	int lastI;
	lastI  = lineLength( row );
	int i, tmp = 0;

	for( i = 0; i < row ; i++ ) {
	    tmp += lineLength( i );
	    if ( contents->at( i )->newline )
		++tmp;
	}

	tmp += QMIN( lastI, col );

	return tmp;
    }
}
#endif

void
TextEdit::keyPressEvent ( QKeyEvent *e )
{
  //  printf("textedit: key 0x%x ascii 0x%x state 0x%x\n", e->key(), e->ascii(), e->state());

  //  if ( isAlphaNum(e->key()) ) {
  //    emit search(e->key());
  //  }

  //  printf("TextEdit keypress %x\n", e->key());

  if (e->key() == 0x48 && e->state() == ControlButton) {  // control-H
    QKeyEvent bs( QEvent::KeyPress, Qt::Key_Backspace, 0, 0 );
    QMultiLineEdit::keyPressEvent(&bs);
    printf("backspace\n");
    return;
  }
    
#if 0
  switch( e-> key ( )) {

  case Qt::Key_Right:
    emit nextPage();
    e->accept();
    break;
  case Qt::Key_Left:
    emit prevPage();
    e->accept();
    break;

  default:
      //      if (isReadOnly()) {
      //	emit topLevelKeyPressEvent(e);
      //      printf("QPDFWindow keypress %x\n", e->key());
  }
#endif

  QMultiLineEdit::keyPressEvent(e);
}

void 
TextEdit::mousePressEvent( QMouseEvent *e )
{
  int x, y;

  if (e->button() == LeftButton) {
    m_mouse_start = e->pos();
    pixelPosToCursorPos(e->pos(), &x, &y);
    if (inMark(x,y)) {
      // without this, the release will move the cursor in the note view
      m_mouse_inside_selection = true;
    }
  }
  if (! m_mouse_inside_selection) {
    QMultiLineEdit::mousePressEvent(e);
  }
}

/*
  All this is to keep the release event from moving the cursor
  in textedit after it goes into note mode

  It might be nicer to wait for the mouse release, then go into note
  mode, storing away the region during the mouse press...
*/

void 
TextEdit::mouseMoveEvent( QMouseEvent *e )
{
  if (! m_mouse_inside_selection) {
    QMultiLineEdit::mouseMoveEvent(e);
  }
}

void 
TextEdit::mouseReleaseEvent( QMouseEvent *e )
{

  QPoint delta;
  delta = e->pos() - m_mouse_start;
  if (delta.manhattanLength() > 10) {
    QMultiLineEdit::mouseReleaseEvent(e);
    m_mouse_inside_selection = false;
    return;
  }
  if (! m_mouse_inside_selection) {
    int corner = detectCorners(e->pos());
    switch(corner) {
    case CORNER_UPPER_LEFT:
      emit prevPage();
      return;
    case CORNER_UPPER_RIGHT:
      emit nextPage();
      return;
    default:
      int edge = detectEdges(e->pos());
      if (edge == EDGE_TOP) {
	emit toggleFullscreen();
	return;
      }
    }
  } else {
    int x1, y1, x2, y2;
    getMarkedRegion(&y1,&x1,&y2,&x2);
    //      printf("inmark region  %d %d - %d %d offset %d,%d\n",x1,x2,y1,y2,
    //	     positionToOffset(y1,x1), positionToOffset(y2,x2));
    emit addNote(positionToOffset(y1,x1), positionToOffset(y2,x2));
  }
  m_mouse_inside_selection = false;
}

  
// FIXME: same as ones in qdocview!!
// 0 = not edge, 1 = top, 2 = left, 3 = right, 4 = bottom
int
TextEdit::detectEdges( QPoint pos)
{
  const int edge_size = 16;

  // for other platforms (like new zaurus C700!) don't ignore qt_screen->transformed

  if (pos.y() < minViewY() + edge_size) {
    return(EDGE_TOP);
  }
  if (pos.y() > maxViewY() - edge_size) {
    return(EDGE_BOTTOM);
  }
  if (pos.x() < minViewX() + edge_size) {
    return(EDGE_LEFT);
  }
  if (pos.x() > maxViewX() - edge_size) {
    return(EDGE_RIGHT);
  }
  return(0);
}



// 0 = not corner, 1 = upper left, 2 = upper right, 3 = lower left, 4 = lower right
int
TextEdit::detectCorners( QPoint pos ) 
{
  const int corner_size = 18;
  QPoint delta;
  QPoint upper_right(maxViewX(),minViewY());
  QPoint upper_left(0,0);
  QPoint lower_right(maxViewX(), maxViewY());

  delta = pos - upper_left;
  if (delta.manhattanLength() < corner_size) {
    return(CORNER_UPPER_LEFT);
  }

  delta = pos - upper_right;
  if (delta.manhattanLength() < corner_size) {
    return(CORNER_UPPER_RIGHT);
  }

  delta = pos - lower_right;
  if (delta.manhattanLength() < corner_size) {
    return(CORNER_LOWER_LEFT);
  }
  return(CORNER_LOWER_RIGHT);
}


bool
TextEdit::find ( const QString &txt, bool caseSensitive,
		 bool reverse )
{
    bool wrap = FALSE;
    int line, col;

    if (txt.isNull()) {
      return(false);
    }

    getCursorPosition( &line, &col );

    for(;;) {
      int findCol;
      if (line < 0 || line >= numLines()) {
	return(false);
      }
      if (reverse) {
	findCol = getString( line )->findRev( txt, col, caseSensitive );
      } else {
	findCol = getString( line )->find( txt, col, caseSensitive );
      }
      if ( findCol >= 0 ) {
	//	printf("findCol %d line %d col %d\n", findCol, line, col);
	setCursorPosition( line, findCol, FALSE );
	col = findCol+ txt.length();
	setCursorPosition( line, col, TRUE );
	return(true);
      }
      //      printf("reverse %d line %d col %d numLines %d\n", reverse, line, col,
      //	     numLines());
      if (reverse) {
	line--;
	if (line < 0) {
	  if (wrap) {
	    return(false);
	  }
	  line = numLines() - 1;
	  wrap = true;
	}
	col = lineLength(line);
      } else {
	line++;
	if (line >= numLines()) {
	  if (wrap) {
	    return(false);
	  }
	  line = 0;
	  wrap = true;
	}
	col = 0;
      }
    }
}

static const int nfontsizes = 7;
//static const int fontsize[nfontsizes] = {8,10,12,14,16,18,24};
static const int fontsize[nfontsizes] = {8,10,12,14,18,24};

TextEdit::TextEdit( QWidget *parent, const char *name, WFlags f)
  : QMultiLineEdit( parent, name )
{
  clearTableFlags();
  setTableFlags( Tbl_vScrollBar );  // | Tbl_autoHScrollBar );
  setSizePolicy ( QSizePolicy ( QSizePolicy::Expanding, QSizePolicy::Expanding ));
  clearTableFlags( Tbl_hScrollBar );
  setAutoUpdate( true );
				
  setFrameStyle( QFrame::Panel | QFrame::Sunken );
  
  resize( 240, 300 );  //???
  
  //  Config cfg("qpdf");
  //  cfg.setGroup("View");
  //  setFontSize(cfg.readNumEntry("TextEditFontSize",10),TRUE);
  setWordWrap(QMultiLineEdit::WidgetWidth);
  m_mouse_inside_selection = false;
  dragScrolling = true;
  m_font_size = 12;
}

TextEdit::~TextEdit()
{
    Config cfg("qpdf");
    cfg.setGroup("View");
    QFont f = font();
    cfg.writeEntry("TextEditFontSize",f.pointSize());
}

void TextEdit::zoomIn()
{
    setFontSize(font().pointSize()+1,FALSE);
}

void TextEdit::zoomOut()
{
    setFontSize(font().pointSize()-1,TRUE);
}

void TextEdit::setFontSize(int sz, bool round_down_not_up)
{
    int s=8;
    for (int i=0; i<nfontsizes; i++) {
	if ( fontsize[i] == sz ) {
	    s = sz;
	    break;
	} else if ( round_down_not_up ) {
	    if ( fontsize[i] < sz )
		s = fontsize[i];
	} else {
	    if ( fontsize[i] > sz ) {
		s = fontsize[i];
		break;
	    }
	}
    }
    //    printf("setFontSize %d (was %d) sz %d round %d\n", s, font().pointSize(), sz, round_down_not_up);
    setFontSize(s);
}

void
TextEdit::setFontSize(int s)
{
    QFont f = font();
    f.setPointSize(s);
    printf("setFontSize %d\n", s);
    m_font_size = s;
    setFont(f);
}

int
TextEdit::getFontSize()
{
  return(m_font_size);
}


#if 0
void TextEdit::slotFind()
{
    FindDialog frmFind( "Text Editor", this );
    connect( &frmFind, SIGNAL(signalFindClicked(const QString &, bool, bool, int)), 
	     editor, SLOT(slotDoFind( const QString&,bool,bool)));

    //case sensitive, backwards, [category] 


    connect( editor, SIGNAL(notFound()), 
	     &frmFind, SLOT(slotNotFound()) );
    connect( editor, SIGNAL(searchWrapped()), 
	     &frmFind, SLOT(slotWrapAround()) );

    frmFind.exec();


}
#endif

void TextEdit::editCut()
{
  printf("TextEdit::editCut readonly %d\n", isReadOnly());
#ifndef QT_NO_CLIPBOARD
  if (!isReadOnly()) {
    cut();
  }
#endif
}

void TextEdit::editCopy()
{
#ifndef QT_NO_CLIPBOARD
    copy();
#endif
}

void TextEdit::editPaste()
{
  printf("TextEdit::editPaste readonly %d\n", isReadOnly());
#ifndef QT_NO_CLIPBOARD
  if (!isReadOnly()) {
    paste();
  }
#endif
}


#if 0
bool TextEdit::save()
{
    // case of nothing to save...
    if ( !doc )
	return true;
    if ( !edited() ) {
	delete doc;
	doc = 0;
	return true;
    }

    QString rt = text();

    if ( doc->name().isEmpty() ) {
	QString pt = rt.simplifyWhiteSpace();
	int i = pt.find( ' ' );
	QString docname = pt;
	if ( i > 0 )
	    docname = pt.left( i );
	// remove "." at the beginning
	while( docname.startsWith( "." ) )
	    docname = docname.mid( 1 );
	docname.replace( QRegExp("/"), "_" );
	// cut the length. filenames longer than that don't make sense and something goes wrong when they get too long.
	if ( docname.length() > 40 )
	    docname = docname.left(40);
	if ( docname.isEmpty() )
	    docname = "Empty Text";
	doc->setName(docname);
    }

	
    FileManager fm;
    if ( !fm.saveFile( *doc, rt ) ) {
	return false;
    }
    delete doc;
    doc = 0;
    setEdited( false );
    return true;
}
#endif

void 
TextEdit::setCharPosition(int charIndex, bool mark)
{
  int row, col;
  offsetToPosition( charIndex, &row, &col );
  setAutoUpdate(true); // needed??
  setCursorPosition( row, col, mark );
}

void 
TextEdit::setTopCharPosition(int charIndex)
{
  int row, col;
  offsetToPosition( charIndex, &row, &col );
  setTopCell(row);
}

int
TextEdit::getCharPosition()
{
  int row, col, r2, c2;
  if (!getSelection(&row, &col, &r2, &c2)) {
    cursorPosition(&row, &col);
  }
  return(positionToOffset(row, col));
}

int
TextEdit::getTopCharPosition()
{
  return(positionToOffset(topCell(), 0));
}

void
TextEdit::setText(QString s)
{
  QMultiLineEdit::setText(s);
}


void
TextEdit::resizeEvent( QResizeEvent *e )
{
// This is here because QMultiLineEdit clears the selection
// when it's size changes (e.g. if a tool bar comes or goes).
// Look in QMultiLineEdit::resize and you will see at the end it
// calls deselect if isVisible.  I can't imagine why you would 
// want to have your selections disappear just because you changed
// the screen size!!
  int line0, col0, line1, col1;
  if (!getSelection(&line0, &col0, &line1, &col1)) {
    line0 = -1;
  }
  QMultiLineEdit::resizeEvent(e);
  if (line0 >= 0) {
    setCursorPosition( line0, col0, false );
    setCursorPosition( line1, col1, true );
  }
}

  
