// Copyright, Philip Kohn, 2003

#include <stdio.h>
#include "qstring.h"
#include "qdict.h"
#include "xmlparser.h"
#include "qpe/stringutil.h"

XmlParser::XmlParser()
{
  m_dict.setAutoDelete(true);
}

int
XmlParser::parse(const QString &str, int index)
{
  int i, j;
  int len = str.length();
  i = str.find("<", index);
  if (i < 0) {
    return(-1);
  }
  j = str.find(" ", i);
  if (j < 0) {
    return(-1);
  }
  m_headName = str.mid(i+1, j-i-1);
  i = j + 1;

  m_dict.clear();

  for(;;) {
    while ( i < len && (str[i] == ' ' || str[i] == '\n' || str[i] == '\r') )
      ++i;
    if (i >= len || str[i] == '>') {
      return(i+1);
    }
#if 0
    if ( i >= len-2 || (str[i] == '/' && str[i+1] == '>') ) {
      return(i+2);
    }
#endif
    // we have another attribute, read it.
    int j = i;
    while ( j < len && str[j] != '=' )
      ++j;
    if (j >= len) {
      return(-1);
    }
    QString name = str.mid(i, j-i);
    i = ++j; // skip =
    while ( i < len && str[i] != '"' ) 
      ++i;
    i++;
    if (i >= len) {
      return(-1);
    }
    j = i;

    j = str.find('"', j);
    if (j < 0) {
      return(-1);
    }
    if ( i == j ) {
      // leave out empty attributes
      i = j + 1;
      continue;
    }
    if (j > i) {
      //      printf("set %s = %s =>", name.latin1(), str.mid(i,j-i).latin1());
      m_dict.insert(name, new QString(unescape(str.mid(i, j-i))));
      //      printf("%s\n", m_dict.find(name)->latin1());
    }
    i = j + 1;
  }
}

QString
XmlParser::unescape(const QString &str)
{
  QString un;
  bool haveAmp = FALSE;
  bool haveUtf = FALSE;
  int i;
  int n = str.length();
  for(int i=0; i<n; i++) {
    if ( str[i] == '&' )
      haveAmp = TRUE;
    if ( ((unsigned char)str[i]) > 0x7f )
      haveUtf = TRUE;
  }
  QString value = QString::fromUtf8(str);
#if 0
  if (haveUtf) {
    value = value.fromUtf8();
  }
#endif
  if (haveAmp) {
    value = Qtopia::plainString(value);
  }
  return(value);
}

QString
XmlParser::get(const QString &s)
{
  QString *str = m_dict.find(s);
  if (str == NULL || str->isNull() || *str == "") {
    //        printf("get(%s) = **EMPTY**\n", s.latin1());
    return(QString(""));
  }  
  //  printf("get(%s) = %s\n", s.latin1(), str->latin1());
  return(*str);
}

bool
XmlParser::checkHead(const QString &s)
{
  if (s != m_headName) {
    printf("bad header '%s' != '%s'\n", s.latin1(), m_headName.latin1());
    return(false);
  }
  return(true);
}
