//========================================================================
//
// FTFont.cc
//
// Copyright 2001-2003 Glyph & Cog, LLC
//
//========================================================================

#include <aconf.h>
#include <stdio.h>
#include <assert.h>

#if FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)

#include "freetype/freetype.h"

#ifdef USE_GCC_PRAGMAS
#pragma implementation
#endif

#include <math.h>
#include <string.h>
#include "gmem.h"
#include "freetype/ftoutln.h"
#include "freetype/internal/ftobjs.h"
#if 1 //~ cff cid->gid map
#include "freetype/internal/cfftypes.h"
#include "freetype/internal/tttypes.h"
#endif
#include "GlobalParams.h"
#include "GfxState.h"
#include "FTFont.h"

//------------------------------------------------------------------------

FTFontEngine::FTFontEngine(GBool aaA):
  SFontEngine() {

  ok = gFalse;
  if (FT_Init_FreeType(&lib)) {
    return;
  }
  aa = aaA;
  ok = gTrue;
}

FTFontEngine::~FTFontEngine() {
  FT_Done_FreeType(lib);
}

//------------------------------------------------------------------------

FTFontFile::FTFontFile(FTFontEngine *engineA, char *fontFileName,
		       char **fontEnc, Gushort *codeToGID) {
  char *name;
  int i;

  ok = gFalse;
  engine = engineA;
  codeMap = NULL;
  cidToGID = NULL;
  cidToGIDLen = 0;

  if (FT_New_Face(engine->lib, fontFileName, 0, &face)) {
    return;
  }

  if (!strcmp(face->driver->root.clazz->module_name, "type1") ||
      !strcmp(face->driver->root.clazz->module_name, "cff")) {
    mode = ftFontModeCodeMapDirect;
    codeMap = (Guint *)gmalloc(256 * sizeof(Guint));
    for (i = 0; i < 256; ++i) {
      codeMap[i] = 0;
      if ((name = fontEnc[i])) {
	codeMap[i] = FT_Get_Name_Index(face, name);
      }
    }

  } else {
    mode = ftFontModeCodeMapDirect;
    codeMap = (Guint *)gmalloc(256 * sizeof(Guint));
    for (i = 0; i < 256; ++i) {
      codeMap[i] = (int)codeToGID[i];
    }
  }

  ok = gTrue;
}

FTFontFile::FTFontFile(FTFontEngine *engineA, char *fontFileName,
		       Gushort *cidToGIDA, int cidToGIDLenA, GBool embedded) {
  int i;

  ok = gFalse;
  engine = engineA;
  codeMap = NULL;
  cidToGID = NULL;
  cidToGIDLen = 0;

  if (FT_New_Face(engine->lib, fontFileName, 0, &face)) {
    return;
  }
  cidToGIDLen = cidToGIDLenA;
  cidToGID = (Gushort *)gmalloc(cidToGIDLen * sizeof(Gushort));
  memcpy(cidToGID, cidToGIDA, cidToGIDLen * sizeof(Gushort));
  if (!strcmp(face->driver->root.clazz->module_name, "t1cid")) {
    mode = ftFontModeCID;
  } else if (!strcmp(face->driver->root.clazz->module_name, "cff")) {
    mode = ftFontModeCFFCharset;
  } else if (embedded) {
    mode = ftFontModeCIDToGIDMap;
  } else {
    mode = ftFontModeUnicode;
    for (i = 0; i < face->num_charmaps; ++i) {
      if ((face->charmaps[i]->platform_id == 3 &&
	   face->charmaps[i]->encoding_id == 1) ||
	  face->charmaps[i]->platform_id == 0) {
	break;
      }
    }
    if (i == face->num_charmaps) {
      i = 0;
    }
    FT_Set_Charmap(face, face->charmaps[i]);
  }
  ok = gTrue;
}

FTFontFile::FTFontFile(FTFontEngine *engineA, char *fontFileName,
		       GBool embedded) {
  int i;

  ok = gFalse;
  engine = engineA;
  codeMap = NULL;
  cidToGID = NULL;
  cidToGIDLen = 0;

  if (FT_New_Face(engine->lib, fontFileName, 0, &face)) {
    return;
  }
#if 0
  if (!strcmp(face->driver->root.clazz->module_name, "t1cid")) {
    mode = ftFontModeCID;
  } else if (!strcmp(face->driver->root.clazz->module_name, "cff")) {
    mode = ftFontModeCFFCharset;
  } else if (embedded) {
    mode = ftFontModeCIDToGIDMap;
  } else {
    mode = ftFontModeUnicode;
#else
    printf("driver %s embedded %d\n", face->driver->root.clazz->module_name, embedded);
  if (!strcmp(face->driver->root.clazz->module_name, "t1cid")) {
    mode = ftFontModeCID;
  } else if (embedded) {
    mode = ftFontModeCFFCharset;
  } else {
    mode = ftFontModeUnicode;
#endif
    for (i = 0; i < face->num_charmaps; ++i) {
      if ((face->charmaps[i]->platform_id == 3 &&
	   face->charmaps[i]->encoding_id == 1) ||
	  face->charmaps[i]->platform_id == 0) {
	break;
      }
    }
    if (i == face->num_charmaps) {
      i = 0;
    }
    FT_Set_Charmap(face, face->charmaps[i]);
  }
  ok = gTrue;
}

FTFontFile::~FTFontFile() {
  if (face) {
    FT_Done_Face(face);
  }
  if (codeMap) {
    gfree(codeMap);
  }
  if (cidToGID) {
    gfree(cidToGID);
  }
}

//------------------------------------------------------------------------

FTFont::FTFont(FTFontFile *fontFileA, fouble *m) {
  FTFontEngine *engine;
  FT_Face face;
  fouble size, div;
  int x, xMin, xMax;
  int y, yMin, yMax;
  int i;

  ok = gFalse;
  fontFile = fontFileA;
  engine = fontFile->engine;
  face = fontFile->face;
  if (FT_New_Size(face, &sizeObj)) {
    return;
  }
  face->size = sizeObj;
  size = sqrt(m[2]*m[2] + m[3]*m[3]);
  if (size <= 0 || FT_Set_Pixel_Sizes(face, 0, (int)size)) {
    return;
  }

  div = face->bbox.xMax > 20000 ? 65536 : 1;

  // transform the four corners of the font bounding box -- the min
  // and max values form the bounding box of the transformed font
  x = (int)((m[0] * (face->bbox.xMin) + m[2] * face->bbox.yMin) /
	    (div * face->units_per_EM));
  xMin = xMax = x;
  y = (int)((m[1] * face->bbox.xMin + m[3] * face->bbox.yMin) /
	    (div * face->units_per_EM));
  yMin = yMax = y;
  x = (int)((m[0] * face->bbox.xMin + m[2] * face->bbox.yMax) /
	    (div * face->units_per_EM));
  if (x < xMin) {
    xMin = x;
  } else if (x > xMax) {
    xMax = x;
  }
  y = (int)((m[1] * face->bbox.xMin + m[3] * face->bbox.yMax) /
	    (div * face->units_per_EM));
  if (y < yMin) {
    yMin = y;
  } else if (y > yMax) {
    yMax = y;
  }
  x = (int)((m[0] * face->bbox.xMax + m[2] * face->bbox.yMin) /
	    (div * face->units_per_EM));
  if (x < xMin) {
    xMin = x;
  } else if (x > xMax) {
    xMax = x;
  }
  y = (int)((m[1] * face->bbox.xMax + m[3] * face->bbox.yMin) /
	    (div * face->units_per_EM));
  if (y < yMin) {
    yMin = y;
  } else if (y > yMax) {
    yMax = y;
  }
  x = (int)((m[0] * face->bbox.xMax + m[2] * face->bbox.yMax) /
	    (div * face->units_per_EM));
  if (x < xMin) {
    xMin = x;
  } else if (x > xMax) {
    xMax = x;
  }
  y = (int)((m[1] * face->bbox.xMax + m[3] * face->bbox.yMax) /
	    (div * face->units_per_EM));
  if (y < yMin) {
    yMin = y;
  } else if (y > yMax) {
    yMax = y;
  }
  // This is a kludge: some buggy PDF generators embed fonts with
  // zero bounding boxes.
  if (xMax == xMin) {
    xMin = 0;
    xMax = (int)size;
  }
  if (yMax == yMin) {
    yMin = 0;
    yMax = (int)(1.2 * size);
  }
  // this should be (max - min + 1), but we add some padding to
  // deal with rounding errors, bogus bboxes, etc.
  glyphW = xMax - xMin + 3;
  glyphW += glyphW >> 1;
  glyphH = yMax - yMin + 3;
  glyphH += glyphH >> 1;
  if (engine->aa) {
    glyphSize = glyphW * glyphH;
  } else {
    glyphSize = ((glyphW + 7) >> 3) * glyphH;
  }

  //  printf("size=%f div=%f m = %f %f %f %f\n",
  //	 (float)size, (float)div, 
  //	 (float)m[0], (float)m[1], (float)m[2], (float)m[3]);

  // set up the glyph pixmap cache
  cacheAssoc = 8;
  if (glyphSize <= 256) {
    cacheSets = 8;
  } else if (glyphSize <= 512) {
    cacheSets = 4;
  } else if (glyphSize <= 1024) {
    cacheSets = 2;
  } else {
    cacheSets = 1;
  }
  cache = (Guchar *)gmalloc(cacheSets * cacheAssoc * glyphSize);
  cacheTags = (FTFontCacheTag *)gmalloc(cacheSets * cacheAssoc *
					sizeof(FTFontCacheTag));
  for (i = 0; i < cacheSets * cacheAssoc; ++i) {
    cacheTags[i].mru = i & (cacheAssoc - 1);
  }

  // create the XImage
  //  printf("FTFont pixmap %d x %d \n", glyphH, glyphW);
//  if (!(image = new QPixmap(glyphH, glyphW, 32))) {
	//XCreateImage(engine->display, engine->visual, engine->depth,
	//			     ZPixmap, 0, NULL, glyphW, glyphH, 8, 0))) {
//    return;
//  }
  //  image->data = (char *)gmalloc(glyphH * image->bytes_per_line);

  // compute the transform matrix
  assert(size != (fouble)0.0);
  float isize = 65536 / (float)size;
  matrix.xx = (FT_Fixed)((float)m[0] * isize);
  matrix.yx = (FT_Fixed)((float)m[1] * isize);
  matrix.xy = (FT_Fixed)((float)m[2] * isize);
  matrix.yy = (FT_Fixed)((float)m[3] * isize);

  ok = gTrue;
}

FTFont::~FTFont() {
  gfree(cacheTags);
  gfree(cache);
//  delete image;
}

GBool FTFont::drawChar(QPixmap *pixmap,
		       int x, int y, int r, int g, int b,
		       CharCode c, Unicode u) {
  FTFontEngine *engine;
  int bgR, bgG, bgB;
  Gulong colors[5];
  Guchar *bitmap, *p;
  GBool tempBitmap;
//  QPixmap *img;
  int pix;
  int xOffset, yOffset, x0, y0, gw, gh;
  int xx, yy, xx1;


  engine = fontFile->engine;

  //  printf("FTFont x=%d y=%d rgb=%d %d %d engine=%x pixmap=%x\n",
  //	 x,y,r,g,b,engine, pixmap);

  unsigned char *pixmap_base = pixmap->scanLine(0);
  const int pixmap_bpl = pixmap->bytesPerLine();
//  const int pixmap_depth = 4;
  const int pixmap_depth = pixmap->depth()/8;
  const int w = pixmap->width();
  const int h = pixmap->height();
//  unsigned char *pp;

  //  QPixmap test(16, 1);
  //  bool isFBLandScape = test.bytesPerLine()<16;
  bool isFBLandScape = (qt_screen->transformOrientation() != 0);

  if (pixmap_depth<2)return gFalse;
//#define PIXEL_PTR(x,y) ( (pixmap_base) + ( ((x) * (pixmap_bpl)) + (y) * (pixmap_depth) ) )
// SL-5x00, SL-B500, iPaq have 320x240x16 FB
#define PIXEL_PTR_LANDSCAPE(x,y) ((pixmap_base) + ((x) * (pixmap_bpl)) + (((h-1)-(y)) * (pixmap_depth)) )
// SL-A300 have 240x320x16 FB, SL-C7x0 have 480x640x16 FB
#define PIXEL_PTR_PORTRAIT(x,y)  ((pixmap_base) + ((y) * (pixmap_bpl)) + ((x) * (pixmap_depth)) )

#define PIXEL_PTR(x,y) (isFBLandScape ? (PIXEL_PTR_LANDSCAPE(x, y)) : (PIXEL_PTR_PORTRAIT(x, y)))

  // no Unicode index for this char - don't draw anything
  if (fontFile->mode == ftFontModeUnicode && u == 0) {
    return gFalse;
  }

  // generate the glyph pixmap
  if (!(bitmap = getGlyphPixmap(c, u, &xOffset, &yOffset, &gw, &gh,
				&tempBitmap))) {
    return gFalse;
  }

  //  printf("got glyph offset %d %d w %d h %d tmp %x bitmap %x\n",
  //	 xOffset, yOffset, gw, gh, tempBitmap, bitmap);

  if (gw==0 || gh==0)return gFalse;
  
  // compute: (x0,y0) = position in destination drawable
  //          (x1,y1) = position in glyph image
  //          (w0,h0) = size of image transfer
  x0 = x - xOffset;
  y0 = y - yOffset;


  // this is not clipping, but in the rare case
  // that a letter is partly outside the page, just nudge it
  // back on the page.  a small price to pay for a faster inner loop.
  if (x0 < 0) {
    x0 = 0;
  }
  if (x0 + gw > w) {
    x0 = w - gw;
    if (x0 < 0) {
      goto done;
    }
  }
  if (y0 < 0) {
    y0 = 0;
  }
  if (y0 + gh > h) {
    y0 = h - gh;
    if (y0 < 0) {
      goto done;
    }
  }

/*
  // getGlyphPixmap may have returned a larger-than-cache-entry
  // bitmap, in which case we need to allocate a temporary XImage here
  if (tempBitmap) {
    //    if (!(img = XCreateImage(engine->display, engine->visual, engine->depth,
    //			     ZPixmap, 0, NULL, gw, gh, 8, 0))) {
//    printf("FTFont temp pixmap %d x %d\n", gh, gw);
    if (!(image = new QPixmap(gh, gw, 3))) {
      goto done;
    }
    //    img->data = (char *)gmalloc(gh * img->bytes_per_line);
  } else {
    img = image;
  }
*/
  
  // read the X image
  //  XGetSubImage(engine->display, d, x0, y0, w0, gh, (1 << engine->depth) - 1,
  //	       ZPixmap, img, x1, y1);

#if 0
  printf("pixmap=%x bpl=%d depth=%d w=%d h=%d\n", pixmap, pixmap_bpl, pixmap->depth(),
	 pixmap->width(),
	 pixmap->height());

  printf("x0=%d y0=%d w0=%d h0=%d x1=%d y1=%d\n", x0, y0, w0, h0, x1, y1);
#endif

  if (engine->aa) {

    // compute the colors
    //    xcolor.pixel = XGetPixel(img, x1 + w0/2, y1 + h0/2);
    //    XQueryColor(engine->display, engine->colormap, &xcolor);
    //    bgR = xcolor.red;
    //    bgG = xcolor.green;
    //    bgB = xcolor.blue;
    //    printf("get pix %d %d\n", x0+w0/2, y0+h0/2);
//    pp = PIXEL_PTR(x0 + w0/2, y0 + h0/2);
//    bgR = *pp++; bgG = *pp++; bgB = *pp++;
    if (pixmap_depth==2){
      int p565 = *(unsigned short*)PIXEL_PTR( (x0 + gw/2), (y0 + gh/2));
      bgB = (p565 & 0x1f) << 11;
      p565 >>= 5;
      bgG = (p565 & 0x3f) << 10;
      p565 >>= 6;
      bgR = (p565 & 0x1f) << 11;
    } else {
      unsigned char *pp = PIXEL_PTR(x0 + gw/2, y0 + gh/2);
      bgR = *pp++ << 8; bgG = *pp++ << 8; bgB = *pp++ << 8;
    }

    colors[1] = engine->findColor((r + 3*bgR) / 4,
				  (g + 3*bgG) / 4,
				  (b + 3*bgB) / 4);
    colors[2] = engine->findColor((r + bgR) / 2,
				  (g + bgG) / 2,
				  (b + bgB) / 2);
    colors[3] = engine->findColor((3*r + bgR) / 4,
				  (3*g + bgG) / 4,
				  (3*b + bgB) / 4);
    colors[4] = engine->findColor(r, g, b);

    // stuff the glyph pixmap into the X image
    p = bitmap;

    if (pixmap_depth==2){
      int dx, dy;
      unsigned short *pp;
      if (isFBLandScape) {
	pp = (unsigned short*) PIXEL_PTR_LANDSCAPE(x0, y0);
	dx = (pixmap_bpl/2);
	dy = -gw * dx - 1;
      } else {
	pp = (unsigned short*) PIXEL_PTR_PORTRAIT(x0, y0);
	dx = 1;
	dy = pixmap_bpl/2 - gw;
      }
      for (yy = 0; yy < gh; ++yy) {
//      pp = PIXEL_PTR(x0-x1, y0-y1+yy);
        for (xx = 0; xx < gw; ++xx) {
          pix = *p++ & 0xff;
	// this is a heuristic which seems to produce decent
	// results -- the linear mapping would be:
	// pix = (pix * 5) / 256;
          pix = ((pix + 10) * 5) / 256;
          if (pix > 4) {
            pix = 4;
          }
          if (pix > 0) {
            //FIXME use shift, mult
//	  pp[0] = colors[pix] & 0xff;
//	  pp[1] = (colors[pix] >> 8) & 0xff;
//	  pp[2] = (colors[pix] >> 16) & 0xff;
	  // XPutPixel(img, xx, yy, colors[pix]);
            *pp = (unsigned short)colors[pix];
	    //            if (qpdf_debug_level>6) printf("%d",pix);//for debug
          } else {
	    //            if (qpdf_debug_level>6) printf(" ");//for debug
          }
//	pp += pixmap_depth;
          pp += dx;
        }
	//        if (qpdf_debug_level>6)printf("\n");//for debug
        pp += dy;
      }
    } else {//32bpp
      const int dx = (isFBLandScape ? (pixmap_bpl/pixmap_depth) : 1);
      const int dy = (isFBLandScape ? (- gw * dx -1) : (pixmap_bpl/pixmap_depth - gw));
      unsigned long* pp = (unsigned long*)PIXEL_PTR(x0, y0);
      for (yy = 0; yy < gh; ++yy) {
        for (xx = 0; xx < gw; ++xx) {
          pix = *p++ & 0xff;
	// this is a heuristic which seems to produce decent
	// results -- the linear mapping would be:
	// pix = (pix * 5) / 256;
          pix = ((pix + 10) * 5) / 256;
          if (pix > 4) {
            pix = 4;
          }
          if (pix > 0) {
            //FIXME use shift, mult
            *pp = colors[pix];
	    //            if (qpdf_debug_level>6)printf("%d",pix);//for debug
          } else {
	    //            if (qpdf_debug_level>6)printf(" ");//for debug
          }
          pp += dx;
        }
	//        if (qpdf_debug_level>6)printf("\n");//for debug
        pp += dy;
      }
    }

  } else {

    // one color
    colors[1] = engine->findColor(r, g, b);

    // stuff the glyph bitmap into the X image
    p = bitmap;
    if (pixmap_depth==2){
      const int dx = (isFBLandScape ? (pixmap_bpl/2) : 1);
      const int dy = (isFBLandScape ? (- gw * dx -1) : (pixmap_bpl/2 - gw));
      unsigned short *pp = (unsigned short*)PIXEL_PTR(x0, y0);
      for (yy = 0; yy < gh; ++yy) {
//      pp = PIXEL_PTR(x0-x1, y0-y1+yy);
        for (xx = 0; xx < gw; xx += 8) {
          pix = *p++;
          for (xx1 = xx; xx1 < xx + 8 && xx1 < gw; ++xx1) {
            if (pix & 0x80) {
//	    *pp++ = (colors[pix] & 0xff);
//	    *pp++ = (colors[pix] >> 8) & 0xff;
//	    *pp++ = (colors[pix] >> 16) & 0xff;
            *pp = (unsigned short)colors[1];
	    //	    XPutPixel(img, xx1, yy, colors[1]);
//            } else {
//	    *pp += 3;
            }
            pix <<= 1;
//            pp+=pixmap_depth/8;
            pp += dx;
          }
        }
        pp += dy;
      }
    } else {//32bpp
      const int dx = (isFBLandScape ? (pixmap_bpl/pixmap_depth) : 1);
      const int dy = (isFBLandScape ? (- gw * dx -1) : (pixmap_bpl/pixmap_depth - gw));
      unsigned long *pp = (unsigned long*)PIXEL_PTR(x0, y0);
      for (yy = 0; yy < gh; ++yy) {
        for (xx = 0; xx < gw; xx += 8) {
          pix = *p++;
          for (xx1 = xx; xx1 < xx + 8 && xx1 < gw; ++xx1) {
            if (pix & 0x80) {
              *pp = colors[1];
            }
            pix <<= 1;
            pp += dx;
          }
        }
        pp += dy;
      }
      
    }
  }

  // draw the X image
  //  XPutImage(engine->display, d, gc, img, x1, y1, x0, y0, w0, h0);

//  if (tempBitmap) {
//    delete img;
//  }
 done:
  if (tempBitmap) {
    gfree(bitmap);
  }
  return gTrue;
}

Guchar *FTFont::getGlyphPixmap(CharCode c, Unicode u,
			       int *x, int *y, int *w, int *h,
			       GBool *tempBitmap) {
  FT_GlyphSlot slot;
  FT_UInt idx;
  int rowSize;
  int i, j, k;
  Guchar *ret, *p, *q;

  // check the cache
  i = (c & (cacheSets - 1)) * cacheAssoc;
  for (j = 0; j < cacheAssoc; ++j) {
    if ((cacheTags[i+j].mru & 0x8000) && cacheTags[i+j].code == c) {
      *x = cacheTags[i+j].x;
      *y = cacheTags[i+j].y;
      *w = cacheTags[i+j].w;
      *h = cacheTags[i+j].h;
      for (k = 0; k < cacheAssoc; ++k) {
	if (k != j &&
	    (cacheTags[i+k].mru & 0x7fff) < (cacheTags[i+j].mru & 0x7fff)) {
	  ++cacheTags[i+k].mru;
	}
      }
      cacheTags[i+j].mru = 0x8000;
      *tempBitmap = gFalse;
      return cache + (i+j) * glyphSize;
    }
  }

  // generate the glyph pixmap or bitmap
  fontFile->face->size = sizeObj;
  FT_Set_Transform(fontFile->face, &matrix, NULL);
  slot = fontFile->face->glyph;
  idx = getGlyphIndex(c, u);

  // if we have the FT2 bytecode interpreter, autohinting won't be used
#ifdef TT_CONFIG_OPTION_BYTECODE_INTERPRETER
  if (FT_Load_Glyph(fontFile->face, idx,
		    fontFile->engine->aa ? FT_LOAD_NO_BITMAP
		                         : FT_LOAD_DEFAULT)) {
    return gFalse;
  }
#else
  // FT2's autohinting doesn't always work very well (especially with
  // font subsets), so turn it off if anti-aliasing is enabled; if
  // anti-aliasing is disabled, this seems to be a tossup - some fonts
  // look better with hinting, some without, so leave hinting on
  if (FT_Load_Glyph(fontFile->face, idx,
		fontFile->engine->aa ? FT_LOAD_NO_HINTING | FT_LOAD_NO_BITMAP
		                     : FT_LOAD_DEFAULT)) {
    return gFalse;
  }
#endif
  if (FT_Render_Glyph(slot,
		      fontFile->engine->aa ? ft_render_mode_normal :
		                             ft_render_mode_mono)) {
    return gFalse;
  }

  // copy the glyph into the cache or a temporary bitmap
  *x = -slot->bitmap_left;
  *y = slot->bitmap_top;
  *w = slot->bitmap.width;
  *h = slot->bitmap.rows;
  if (fontFile->engine->aa) {
    rowSize = *w;
  } else {
    rowSize = (*w + 7) >> 3;
  }
  if (*w > glyphW || *h > glyphH) {
    // the glyph doesn't fit in the bounding box -- return a
    // temporary, uncached bitmap (this shouldn't happen but some
    // fonts have incorrect bboxes)
    ret = (Guchar *)gmalloc(*h * rowSize);
    *tempBitmap = gTrue;
  } else {
    // store glyph pixmap in cache
    ret = NULL; // make gcc happy
    for (j = 0; j < cacheAssoc; ++j) {
      if ((cacheTags[i+j].mru & 0x7fff) == cacheAssoc - 1) {
	cacheTags[i+j].mru = 0x8000;
	cacheTags[i+j].code = c;
	cacheTags[i+j].x = *x;
	cacheTags[i+j].y = *y;
	cacheTags[i+j].w = *w;
	cacheTags[i+j].h = *h;
	ret = cache + (i+j) * glyphSize;
      } else {
	++cacheTags[i+j].mru;
      }
    }
    *tempBitmap = gFalse;
  }
  for (k = 0, p = ret, q = slot->bitmap.buffer;
       k < slot->bitmap.rows;
       ++k, p += rowSize, q += slot->bitmap.pitch) {
    memcpy(p, q, rowSize);
  }
  return ret;
}

GBool FTFont::getCharPath(CharCode c, Unicode u, GfxState *state) {
  static FT_Outline_Funcs outlineFuncs = {
    &charPathMoveTo,
    &charPathLineTo,
    &charPathConicTo,
    &charPathCubicTo,
    0, 0
  };
  FT_GlyphSlot slot;
  FT_UInt idx;
  FT_Glyph glyph;

  fontFile->face->size = sizeObj;
  FT_Set_Transform(fontFile->face, &matrix, NULL);
  slot = fontFile->face->glyph;
  idx = getGlyphIndex(c, u);
#ifdef TT_CONFIG_OPTION_BYTECODE_INTERPRETER
  if (FT_Load_Glyph(fontFile->face, idx, FT_LOAD_DEFAULT)) {
    return gFalse;
  }
#else
  // FT2's autohinting doesn't always work very well (especially with
  // font subsets), so turn it off if anti-aliasing is enabled; if
  // anti-aliasing is disabled, this seems to be a tossup - some fonts
  // look better with hinting, some without, so leave hinting on
  if (FT_Load_Glyph(fontFile->face, idx,
		    fontFile->engine->aa ? FT_LOAD_NO_HINTING
		                         : FT_LOAD_DEFAULT)) {
    return gFalse;
  }
#endif
  if (FT_Get_Glyph(slot, &glyph)) {
    return gFalse;
  }
  FT_Outline_Decompose(&((FT_OutlineGlyph)glyph)->outline,
		       &outlineFuncs, state);
  return gTrue;
}

int FTFont::charPathMoveTo(FT_Vector *pt, void *state) {
  ((GfxState *)state)->moveTo(pt->x / 64.0, -pt->y / 64.0);
  return 0;
}

int FTFont::charPathLineTo(FT_Vector *pt, void *state) {
  ((GfxState *)state)->lineTo(pt->x / 64.0, -pt->y / 64.0);
  return 0;
}

int FTFont::charPathConicTo(FT_Vector *ctrl, FT_Vector *pt, void *state) {
  fouble x0, y0, x1, y1, x2, y2, x3, y3, xc, yc;

  x0 = ((GfxState *)state)->getCurX();
  y0 = ((GfxState *)state)->getCurY();
  xc = ctrl->x / 64.0;
  yc = -ctrl->y / 64.0;
  x3 = pt->x / 64.0;
  y3 = -pt->y / 64.0;

  // A second-order Bezier curve is defined by two endpoints, p0 and
  // p3, and one control point, pc:
  //
  //     p(t) = (1-t)^2*p0 + t*(1-t)*pc + t^2*p3
  //
  // A third-order Bezier curve is defined by the same two endpoints,
  // p0 and p3, and two control points, p1 and p2:
  //
  //     p(t) = (1-t)^3*p0 + 3t*(1-t)^2*p1 + 3t^2*(1-t)*p2 + t^3*p3
  //
  // Applying some algebra, we can convert a second-order curve to a
  // third-order curve:
  //
  //     p1 = (1/3) * (p0 + 2pc)
  //     p2 = (1/3) * (2pc + p3)

  x1 = (1.0 / 3.0) * (x0 + 2 * xc);
  y1 = (1.0 / 3.0) * (y0 + 2 * yc);
  x2 = (1.0 / 3.0) * (2 * xc + x3);
  y2 = (1.0 / 3.0) * (2 * yc + y3);

  ((GfxState *)state)->curveTo(x1, y1, x2, y2, x3, y3);
  return 0;
}

int FTFont::charPathCubicTo(FT_Vector *ctrl1, FT_Vector *ctrl2,
			    FT_Vector *pt, void *state) {
  ((GfxState *)state)->curveTo(ctrl1->x / 64.0, -ctrl1->y / 64.0,
			       ctrl2->x / 64.0, -ctrl2->y / 64.0,
			       pt->x / 64.0, -pt->y / 64.0);
  return 0;
}

FT_UInt FTFont::getGlyphIndex(CharCode c, Unicode u) {
  FT_UInt idx;
  int j;

#if FREETYPE_MAJOR == 2 && FREETYPE_MINOR == 0
  CFF_Font *cff;
#else
  CFF_Font cff;
#endif

  idx = 0; // make gcc happy
  switch (fontFile->mode) {
  case ftFontModeUnicode:
    idx = FT_Get_Char_Index(fontFile->face, (FT_ULong)u);
    break;
  case ftFontModeCodeMapDirect:
    if (c <= 0xff) {
      idx = (FT_UInt)fontFile->codeMap[c];
    } else {
      idx = 0;
    }
    break;
  case ftFontModeCIDToGIDMap:
    if (fontFile->cidToGIDLen) {
      if ((int)c < fontFile->cidToGIDLen) {
	idx = (FT_UInt)fontFile->cidToGID[c];
      } else {
	idx = (FT_UInt)0;
      }
    } else {
      idx = (FT_UInt)c;
    }
    break;
  case ftFontModeCFFCharset:
#if 1 //~ cff cid->gid map   // TALK TO SOMEONE ABOUT THIS!!
#if FREETYPE_MAJOR == 2 && FREETYPE_MINOR == 0
    cff = (CFF_Font *)((TT_Face)fontFile->face)->extra.data;
#else
    cff = (CFF_Font)((TT_Face)fontFile->face)->extra.data;
#endif
    idx = 0;
    for (j = 0; j < (int)cff->num_glyphs; ++j) {
      if (cff->charset.sids[j] == c) {
	idx = j;
	break;
      }
    }
#endif
    break;
  case ftFontModeCID:
    idx = c;
    break;
  }
  return idx;
}

#endif // FREETYPE2 && (HAVE_FREETYPE_FREETYPE_H || HAVE_FREETYPE_H)
